/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include/numpy/arrayobject.h",
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include/numpy/arrayscalars.h",
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include/numpy/ndarrayobject.h",
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include/numpy/ndarraytypes.h",
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include"
        ],
        "name": "glycresoft._c.tandem.core_search",
        "sources": [
            "src/glycresoft/_c/tandem/core_search.pyx"
        ]
    },
    "module_name": "glycresoft._c.tandem.core_search"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__glycresoft___c__tandem__core_search
#define __PYX_HAVE_API__glycresoft___c__tandem__core_search
/* Early includes */
#include <string.h>
#include <stdio.h>
#include <math.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include <stddef.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/glycresoft/_c/tandem/core_search.pyx",
  "__init__.cython-30.pxd",
  "contextvars.pxd",
  "<stringsource>",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "peak_set.pxd",
  "peak_index.pxd",
  "averagine.pxd",
  "deconvoluter_base.pxd",
  "scoring.pxd",
  "peak_set.pxd",
  "ccomposition.pxd",
  "src/glycresoft/_c/chromatogram_tree/mass_shift.pxd",
  "glycan_composition.pxd",
  "count_table.pxd",
  "base.pxd",
  "glycan.pxd",
  "fragment.pxd",
  "sequence_methods.pxd",
  "base.pxd",
  "glycan.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":770
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":771
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":772
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":773
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":777
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":778
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":779
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":780
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":784
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":785
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":792
 * ctypedef double complex complex128_t
 * 
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * ctypedef npy_longlong   longlong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":795
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":796
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":798
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":799
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":800
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef float complex       cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< long double > __pyx_t_long_double_complex;
  #else
    typedef long double _Complex __pyx_t_long_double_complex;
  #endif
#else
    typedef struct { long double real, imag; } __pyx_t_long_double_complex;
#endif
static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double, long double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet;
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition;
struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase;
struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift;
struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift;
struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase;
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator;
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1096
 * 
 * # Iterator API added in v1.6
 * ctypedef int (*NpyIter_IterNextFunc)(NpyIter* it) noexcept nogil             # <<<<<<<<<<<<<<
 * ctypedef void (*NpyIter_GetMultiIndexFunc)(NpyIter* it, npy_intp* outcoords) noexcept nogil
 * 
 */
typedef int (*__pyx_t_5numpy_NpyIter_IterNextFunc)(NpyIter *);

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1097
 * # Iterator API added in v1.6
 * ctypedef int (*NpyIter_IterNextFunc)(NpyIter* it) noexcept nogil
 * ctypedef void (*NpyIter_GetMultiIndexFunc)(NpyIter* it, npy_intp* outcoords) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "numpy/arrayobject.h":
 */
typedef void (*__pyx_t_5numpy_NpyIter_GetMultiIndexFunc)(NpyIter *, npy_intp *);
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;
struct __pyx_t_7brainpy_2_c_11composition_Isotope;
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;
struct __pyx_t_7brainpy_2_c_11composition_Element;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;
struct __pyx_t_7brainpy_2_c_11composition_Composition;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) noexcept nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) noexcept nogil
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) noexcept nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) noexcept nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */
struct __pyx_t_7brainpy_2_c_11composition_Isotope {
  double mass;
  double abundance;
  int neutrons;
  int neutron_shift;
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;
  size_t size;
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */
struct __pyx_t_7brainpy_2_c_11composition_Element {
  char *symbol;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;
  int monoisotopic_isotope_index;
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;
  size_t size;
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */
struct __pyx_t_7brainpy_2_c_11composition_Composition {
  char **elements;
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;
  size_t size;
  size_t used;
  int max_variants;
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {
  int __pyx_n;
  int free_existing;
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {
  double *v;
  size_t used;
  size_t size;
};
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;
  size_t size;
  size_t used;
};
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {
  double mz;
  double intensity;
  int charge;
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;
  size_t used;
  size_t size;
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;
  int order;
  double average_mass;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution) noexcept nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) noexcept nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) noexcept nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {
  int __pyx_n;
  PyObject *npeaks;
  int charge;
  double charge_carrier;
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {
  int __pyx_n;
  int charge;
  double charge_carrier;
};

/* "brainpy/_c/isotopic_distribution.pxd":85
 * 
 * 
 * cdef size_t max_variants_approx(double mass, double lambda_factor=*, size_t maxiter=*, double threshold=*) noexcept nogil             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx {
  int __pyx_n;
  double lambda_factor;
  size_t maxiter;
  double threshold;
};
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp) noexcept
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n) noexcept
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {
  int __pyx_n;
  double tolerance;
  size_t hint_start;
  size_t hint_end;
};
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {
  int __pyx_n;
  int charge;
  double charge_carrier;
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {
  int __pyx_n;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {
  int __pyx_n;
  double truncate_after;
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {
  int __pyx_n;
  PyObject *method;
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {
  int __pyx_n;
  PyObject *method;
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {
  int __pyx_n;
  int charge;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {
  int __pyx_n;
  int charge;
};

/* "ms_deisotope/_c/averagine.pxd":94
 * cpdef double isotopic_shift(int charge=*)
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate {
  int __pyx_n;
  double lambda_factor;
  int charge;
};
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {
  int __pyx_n;
  int step;
  double tolerance;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {
  int __pyx_n;
  double charge_carrier;
  double truncate_after;
  double ignore_below;
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {
  int __pyx_n;
  double error_tolerance;
  PyObject *charge_range;
  int left_search_limit;
  int right_search_limit;
  int recalculate_starting_peak;
  int use_quick_charge;
};
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {
  int __pyx_n;
  PyObject *data;
  int missed_peaks;
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {
  int __pyx_n;
  double mass_error_tolerance;
  double minimum_signal_to_noise;
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {
  int __pyx_n;
  double lower;
  double upper;
};
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t;
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak;
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for;

/* "ms_deisotope/_c/peak_set.pxd":138
 * 
 * 
 * cdef enum CPeaksFlags:             # <<<<<<<<<<<<<<
 *     borrowing = 0
 *     owns_peaks = 1
 */
enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags {
  __pyx_e_12ms_deisotope_2_c_8peak_set_borrowing = 0,
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks = 1,
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_index = 2,
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks_and_index = 3
};

/* "ms_deisotope/_c/peak_set.pxd":87
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {
  int __pyx_n;
  double error_tolerance;
  int use_mz;
};

/* "ms_deisotope/_c/peak_set.pxd":89
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {
  int __pyx_n;
  double error_tolerance;
  int use_mz;
};

/* "ms_deisotope/_c/peak_set.pxd":91
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {
  int __pyx_n;
  double tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":118
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_deisotope/_c/peak_set.pxd":124
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};

/* "ms_deisotope/_c/peak_set.pxd":131
 * 
 * 
 * cdef struct deconvoluted_peak_t:             # <<<<<<<<<<<<<<
 *     double neutral_mass
 *     double intensity
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t {
  double neutral_mass;
  double intensity;
  int charge;
  unsigned int index;
};

/* "ms_deisotope/_c/peak_set.pxd":144
 *     owns_peaks_and_index = 3
 * 
 * cdef struct deconvoluted_peak_set_t:             # <<<<<<<<<<<<<<
 *     deconvoluted_peak_t* peaks
 *     double* mass_index
 */
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t {
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *peaks;
  double *mass_index;
  size_t size;
  enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags flags;
};

/* "ms_deisotope/_c/peak_set.pxd":154
 * cdef int free_deconvoluted_peak_set_t(deconvoluted_peak_set_t* destination) noexcept nogil
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) noexcept nogil
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":155
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) noexcept nogil
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * cdef int deconvoluted_peak_eq(deconvoluted_peak_t* self, deconvoluted_peak_t* other) noexcept nogil
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":167
 *     cdef deconvoluted_peak_t* getitem(self, size_t i) nogil
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) noexcept nogil             # <<<<<<<<<<<<<<
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) noexcept nogil
 * 
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":168
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) noexcept nogil
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":170
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) noexcept nogil
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_deisotope/_c/peak_set.pxd":171
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass;

/* "glypy/composition/ccomposition.pxd":6
 *     cdef tuple _mass_args
 *     cpdef CComposition clone(self)
 *     cpdef double calc_mass(self, int average=?, charge=?, dict mass_data=?) except -1             # <<<<<<<<<<<<<<
 *     cpdef _from_formula(self, str formula, dict mass_data)
 *     cpdef _from_dict(self, comp)
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass {
  int __pyx_n;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};

/* "glypy/composition/ccomposition.pxd":31
 * 
 * 
 * cpdef double calculate_mass(CComposition composition=?, str formula=?, int average=?, charge=?, mass_data=?) except -1             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass {
  int __pyx_n;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *formula;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};
struct __pyx_opt_args_5glypy_2_c_9structure_18glycan_composition_16_CompositionBase__update_from_typed_map;

/* "glypy/_c/structure/glycan_composition.pxd":12
 *     cpdef object _setitem_fast(self, object key, object value)
 * 
 *     cpdef object _update_from_typed_map(self, _CompositionBase template, bint copy_nodes=*)             # <<<<<<<<<<<<<<
 *     cpdef str serialize(self)
 * 
 */
struct __pyx_opt_args_5glypy_2_c_9structure_18glycan_composition_16_CompositionBase__update_from_typed_map {
  int __pyx_n;
  int copy_nodes;
};
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin_cell;
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin;
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table;
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_pop;
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get;

/* "glycopeptidepy/_c/count_table.pxd":3
 * from cpython cimport PyObject
 * 
 * cdef struct count_table_bin_cell:             # <<<<<<<<<<<<<<
 *     long value
 *     PyObject* key
 */
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin_cell {
  long value;
  PyObject *key;
};

/* "glycopeptidepy/_c/count_table.pxd":8
 * 
 * 
 * cdef struct count_table_bin:             # <<<<<<<<<<<<<<
 *     count_table_bin_cell* cells
 *     size_t size
 */
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin {
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin_cell *cells;
  size_t size;
  size_t used;
};

/* "glycopeptidepy/_c/count_table.pxd":14
 * 
 * 
 * cdef struct count_table:             # <<<<<<<<<<<<<<
 *     count_table_bin* bins
 *     size_t size
 */
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table {
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin *bins;
  size_t size;
  size_t count;
};

/* "glycopeptidepy/_c/count_table.pxd":100
 *     cpdef clear(self)
 *     cpdef setdefault(self, key, value)
 *     cpdef long pop(self, object key, object default=?) except *             # <<<<<<<<<<<<<<
 *     cpdef get(self, object key, object default=?)
 * 
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};

/* "glycopeptidepy/_c/count_table.pxd":101
 *     cpdef setdefault(self, key, value)
 *     cpdef long pop(self, object key, object default=?) except *
 *     cpdef get(self, object key, object default=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef void increment(self, object key, long value)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get {
  int __pyx_n;
  PyObject *__pyx_default;
};
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop;

/* "glycopeptidepy/_c/structure/glycan.pxd":48
 *     cpdef values(self)
 *     cpdef items(self)
 *     cpdef pop(self, key, default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create;

/* "glycopeptidepy/_c/structure/fragment.pxd":84
 * 
 *     @staticmethod
 *     cdef StubFragment _create(str name, double mass, IonSeriesBase kind, CComposition composition,             # <<<<<<<<<<<<<<
 *                               ChemicalShiftBase chemical_shift, bint is_glycosylated, object glycosylation,
 *                               bint is_extended, int glycosylation_size=*)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create {
  int __pyx_n;
  int glycosylation_size;
};

/* "glycopeptidepy/_c/structure/fragment.pxd":108
 * 
 *     @staticmethod
 *     cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,             # <<<<<<<<<<<<<<
 *                                  list flanking_amino_acids=*, dict glycosylation=*,
 *                                  ChemicalShiftBase chemical_shift=*, CComposition composition=?, double* delta_mass=?)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create {
  int __pyx_n;
  PyObject *flanking_amino_acids;
  PyObject *glycosylation;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *chemical_shift;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  double *delta_mass;
};
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods__make_terminal_group;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore__init_from_parsed_string;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore__init_from_components;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore_get_sequence;

/* "glycopeptidepy/_c/structure/sequence_methods.pxd":12
 * cdef AminoAcidResidueBase _parse_residue(basestring residue_string)
 * 
 * cdef TerminalGroup _make_terminal_group(basestring base_composition_formula, ModificationBase modification=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods__make_terminal_group {
  int __pyx_n;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *modification;
};

/* "glycopeptidepy/_c/structure/sequence_methods.pxd":17
 * cdef class _PeptideSequenceCore(PeptideSequenceBase):
 * 
 *     cpdef _init_from_parsed_string(self, list seq_list, glycan=*, n_term=*, c_term=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _init_from_string(self, sequence, parser_function)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore__init_from_parsed_string {
  int __pyx_n;
  PyObject *glycan;
  PyObject *n_term;
  PyObject *c_term;
};

/* "glycopeptidepy/_c/structure/sequence_methods.pxd":21
 *     cpdef _init_from_string(self, sequence, parser_function)
 * 
 *     cpdef _init_from_components(self, list seq_list, glycan_composition=*, ModificationBase n_term=*, ModificationBase c_term=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _init_termini(self, TerminalGroup n_term, TerminalGroup c_term)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore__init_from_components {
  int __pyx_n;
  PyObject *glycan_composition;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *n_term;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *c_term;
};

/* "glycopeptidepy/_c/structure/sequence_methods.pxd":33
 *     cdef void set_c_term(self, _value)
 * 
 *     cpdef basestring get_sequence(self, bint include_glycan=*, bint include_termini=*, str implicit_n_term=*, str implicit_c_term=*)             # <<<<<<<<<<<<<<
 *     cpdef clone(self)
 * 
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore_get_sequence {
  int __pyx_n;
  int include_glycan;
  int include_termini;
  PyObject *implicit_n_term;
  PyObject *implicit_c_term;
};
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_n_glycan_composition_fragments;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_o_glycan_composition_fragments;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_gag_linker_composition_fragments;

/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":55
 *     cdef list _combinate_sites_single(self, list per_site_shifts, object glycan)
 * 
 *     cpdef list n_glycan_composition_fragments(self, object glycan, int core_count=*, int iteration_count=*)             # <<<<<<<<<<<<<<
 *     cpdef list o_glycan_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)
 *     cpdef list gag_linker_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_n_glycan_composition_fragments {
  int __pyx_n;
  int core_count;
  int iteration_count;
};

/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":56
 * 
 *     cpdef list n_glycan_composition_fragments(self, object glycan, int core_count=*, int iteration_count=*)
 *     cpdef list o_glycan_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)             # <<<<<<<<<<<<<<
 *     cpdef list gag_linker_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)
 * 
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_o_glycan_composition_fragments {
  int __pyx_n;
  long core_count;
  long iteration_count;
};

/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":57
 *     cpdef list n_glycan_composition_fragments(self, object glycan, int core_count=*, int iteration_count=*)
 *     cpdef list o_glycan_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)
 *     cpdef list gag_linker_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list mixed_stub_fragments(self)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_gag_linker_composition_fragments {
  int __pyx_n;
  long core_count;
  long iteration_count;
};
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments;
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs;
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs;
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs;
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_isclose;
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match;
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match;

/* "glycresoft/_c/tandem/core_search.pyx":305
 * 
 *     @cython.boundscheck(False)
 *     cpdef CoarseGlycanMatch _match_fragments(self, object scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list fragment_matches
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments {
  int __pyx_n;
  double mass_shift_tandem_mass;
};

/* "glycresoft/_c/tandem/core_search.pyx":381
 *         return score
 * 
 *     cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs {
  int __pyx_n;
  double mass_shift_tandem_mass;
};

/* "glycresoft/_c/tandem/core_search.pyx":387
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs {
  int __pyx_n;
  double mass_shift_tandem_mass;
};

/* "glycresoft/_c/tandem/core_search.pyx":393
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                              GlycanCombinationRecordBase glycan_combination,
 *                                              double mass_shift_tandem_mass=0.0):
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs {
  int __pyx_n;
  double mass_shift_tandem_mass;
};

/* "glycresoft/_c/tandem/core_search.pyx":439
 * 
 * 
 * cdef bint isclose(double a, double b, double rtol=1e-05, double atol=1e-08):             # <<<<<<<<<<<<<<
 *     return abs(a - b) <= atol + rtol * abs(b)
 * 
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_isclose {
  int __pyx_n;
  double rtol;
  double atol;
};

/* "glycresoft/_c/tandem/core_search.pyx":530
 * @cython.boundscheck(False)
 * @cython.binding(True)
 * cpdef GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list output, glycan_combination_db
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match {
  int __pyx_n;
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *mass_shift;
  PyObject *query_mass;
};

/* "glycresoft/_c/tandem/core_search.pyx":917
 * 
 *     @cython.cdivision
 *     cpdef list match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double precursor_mass, d
 */
struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match {
  int __pyx_n;
  double error_tolerance;
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *mass_shift;
  PyObject *query_mass;
};

/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;
  double cached_mass;
  int _clean;
};


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;
  double mz;
  double intensity;
  int charge;
};


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;
  double mz;
  double intensity;
  double area;
};


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double signal_to_noise;
  double full_width_at_half_max;
  double left_width;
  double right_width;
  long peak_count;
  long index;
};


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;
  PyObject *peaks;
  int indexed;
};


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  double *mz_index;
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;
};


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;
  PyArrayObject *mz_array;
  PyArrayObject *intensity_array;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;
};


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;
  double base_mass;
  PyObject *base_composition;
};


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;
  PyObject *peaklist;
  double origin;
  double offset;
};


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;
  PyObject *backend;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;
  double cache_truncation;
  int enabled;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;
  int use_subtraction;
  PyObject *scale_method;
  int merge_isobaric_peaks;
  double minimum_intensity;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;
  int verbose;
  PyObject *_slice_cache;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  PyObject *averagines;
};


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;
  int lower;
  int upper;
  int sign;
  int *values;
  size_t size;
  size_t index;
};


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;
  double score;
  int charge;
  PyObject *experimental;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;
  int missed_peaks;
  PyObject *data;
  Py_hash_t _hash;
};


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;
  double minimum_score;
};


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;
};


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double mass_error_tolerance;
};


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double mass_error_tolerance;
  double penalty_factor;
};


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  PyObject *function;
};


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;
};


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;
  double peak_count_scale;
  double domain_scale;
};


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double scale_factor;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;
};


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;
  size_t neutral_mass;
  size_t mz;
};


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;
  double mz;
  double intensity;
};


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;
  PyObject *pairs;
};


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double neutral_mass;
  double signal_to_noise;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;
  double full_width_at_half_max;
  int charge;
  double a_to_a2_ratio;
  double most_abundant_mass;
  double average_mass;
  double score;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;
  int chosen_for_msms;
};


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
  double drift_time;
};


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
  PyObject *solution;
};


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;
  PyObject *peaks;
  PyObject *_mz_ordered;
  int indexed;
};


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;
  double *neutral_mass_array;
  double *mz_array;
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;
  size_t _size;
};


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtab;
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t *ptr;
};


/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition {
  PyDictObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtab;
  PyObject *_mass;
  PyObject *_mass_args;
};


/* "glycresoft/_c/chromatogram_tree/mass_shift.pxd":9
 * 
 * 
 * cdef class MassShiftBase(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase {
  PyObject_HEAD
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *tandem_composition;
  PyObject *name;
  long _hash;
  double mass;
  double tandem_mass;
};


/* "glycresoft/_c/chromatogram_tree/mass_shift.pxd":20
 * 
 * 
 * cdef class MassShift(MassShiftBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift {
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase __pyx_base;
};


/* "glycresoft/_c/chromatogram_tree/mass_shift.pxd":24
 * 
 * 
 * cdef class CompoundMassShift(MassShiftBase):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift {
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase __pyx_base;
  PyObject *counts;
};


/* "glypy/_c/structure/glycan_composition.pxd":3
 * from glypy.composition.ccomposition cimport CComposition
 * 
 * cdef class _CompositionBase(dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object _mass
 */
struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase {
  PyDictObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase *__pyx_vtab;
  PyObject *_mass;
  PyObject *_reducing_end;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_composition_offset;
};


/* "glycopeptidepy/_c/count_table.pxd":52
 * 
 * 
 * cdef class CountTableIterator(object):             # <<<<<<<<<<<<<<
 *     cdef :
 *         count_table* table
 */
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_vtab;
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table *table;
  size_t bin_index;
  size_t cell_index;
  PyObject *next_key;
  long next_value;
};


/* "glycopeptidepy/_c/count_table.pxd":69
 * 
 * 
 * cdef class CountTable(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* table
 */
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_vtab;
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table *table;
};


/* "glycopeptidepy/_c/structure/base.pxd":5
 * 
 * 
 * cdef class PeptideSequenceBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list sequence
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_vtab;
  PyObject *sequence;
  double _mass;
  PyObject *_glycosylation_manager;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *_n_term;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *_c_term;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_total_composition;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_peptide_composition;
};


/* "glycopeptidepy/_c/structure/base.pxd":21
 * 
 * 
 * cdef class TerminalGroup(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition base_composition
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *__pyx_vtab;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *base_composition;
  double mass;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *_modification;
};


/* "glycopeptidepy/_c/structure/base.pxd":35
 * 
 * 
 * cdef class AminoAcidResidueBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *__pyx_vtab;
  PyObject *name;
  PyObject *symbol;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *neutral_loss;
  Py_hash_t _hash;
};


/* "glycopeptidepy/_c/structure/base.pxd":50
 * 
 * 
 * cdef class ModificationBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_vtab;
  PyObject *name;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/base.pxd":69
 * 
 * 
 * cdef class SequencePosition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AminoAcidResidueBase amino_acid
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition *__pyx_vtab;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *amino_acid;
  PyObject *modifications;
};


/* "glycopeptidepy/_c/structure/glycan.pxd":12
 * 
 * 
 * cdef class GlycanCompositionProxy(object):             # <<<<<<<<<<<<<<
 *     '''A mapping-like object that imitates the GlycanComposition interface in
 *     a read-only fashion.
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_vtab;
  struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *obj;
  PyObject *_serialized;
};


/* "glycopeptidepy/_c/structure/glycan.pxd":29
 * 
 * 
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition composition_offset
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy __pyx_base;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition_offset;
};


/* "glycopeptidepy/_c/structure/glycan.pxd":34
 * 
 * 
 * cdef class GlycosylationManager(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict mapping
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_vtab;
  PyObject *mapping;
  PyObject *parent;
  PyObject *_aggregate;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *_proxy;
  PyObject *_type_track;
  int _total_glycosylation_size;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":6
 * 
 * 
 * cdef class ChemicalShiftBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public str name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_vtab;
  PyObject *name;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  double mass;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":15
 * 
 * 
 * cdef class IonSeriesBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public str name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase {
  PyObject_HEAD
  PyObject *name;
  int direction;
  int includes_peptide;
  double mass_shift;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition_shift;
  Py_hash_t _hash;
  int int_code;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":26
 * 
 * 
 * cdef class FragmentBase(object):             # <<<<<<<<<<<<<<
 *     """Base class for all Fragment types. Defines basic
 *     name generation and neutral loss handling functions.
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_vtab;
  PyObject *_name;
  Py_hash_t _hash;
  double mass;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *_chemical_shift;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":63
 * 
 * 
 * cdef class SimpleFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *kind;
  int is_glycosylated;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":72
 * 
 * 
 * cdef class StubFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *kind;
  int is_glycosylated;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *glycosylation;
  int is_extended;
  int _glycosylation_size;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":97
 * 
 * 
 * cdef class PeptideFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *kind;
  int position;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *modification_dict;
  double bare_mass;
  PyObject *flanking_amino_acids;
  PyObject *glycosylation;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":125
 * 
 * @cython.final
 * cdef class _NameTree(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_vtab;
  PyObject *name;
  PyObject *value;
  PyObject *children;
};


/* "glycopeptidepy/_c/structure/sequence_methods.pxd":15
 * 
 * 
 * cdef class _PeptideSequenceCore(PeptideSequenceBase):             # <<<<<<<<<<<<<<
 * 
 *     cpdef _init_from_parsed_string(self, list seq_list, glycan=*, n_term=*, c_term=*)
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase __pyx_base;
};


/* "glycopeptidepy/_c/structure/fragmentation_strategy/base.pxd":28
 * 
 * 
 * cdef class FragmentationStrategyBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public _PeptideSequenceCore peptide
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase *__pyx_vtab;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *peptide;
  int compute_compositions;
};


/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":13
 * 
 * 
 * cdef class GlycanCompositionFragment(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *__pyx_vtab;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *key;
  int is_extended;
  Py_hash_t _hash_key;
  PyObject *cache;
  int _glycosylation_size;
};


/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":31
 * 
 * 
 * cdef class GlycanCompositionFragmentStrategyBase(FragmentationStrategyBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint _use_query
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase __pyx_base;
  int _use_query;
  PyObject *_generator;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *glycosylation_manager;
};


/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":42
 * 
 * 
 * cdef class StubGlycopeptideStrategy(GlycanCompositionFragmentStrategyBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint extended
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase __pyx_base;
  int extended;
  int extended_fucosylation;
};


/* "glycresoft/_c/tandem/core_search.pyx":45
 * 
 * 
 * cdef class GlycanCombinationRecordBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t id
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_vtab;
  size_t id;
  double dehydrated_mass;
  PyObject *composition;
  npy_uint16 size;
  npy_uint16 internal_size_approximation;
  npy_uint16 side_group_count;
  npy_uint16 count;
  PyObject *glycan_types;
  PyObject *_fragment_cache;
  Py_hash_t _hash;
  PyObject *fragment_set_properties;
};


/* "glycresoft/_c/tandem/core_search.pyx":152
 * 
 * 
 * cdef class CoarseStubGlycopeptideFragment(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object key
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment {
  PyObject_HEAD
  PyObject *key;
  double mass;
  int is_core;
};


/* "glycresoft/_c/tandem/core_search.pyx":193
 * @cython.freelist(100000)
 * @cython.final
 * cdef class CoarseStubGlycopeptideMatch(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object key
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_vtab;
  PyObject *key;
  double mass;
  double shift_mass;
  PyObject *peaks_matched;
};


/* "glycresoft/_c/tandem/core_search.pyx":228
 * @cython.freelist(100000)
 * @cython.final
 * cdef class CoarseGlycanMatch(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list fragment_matches
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_vtab;
  PyObject *fragment_matches;
  double n_matched;
  double n_theoretical;
  double core_matched;
  double core_theoretical;
};


/* "glycresoft/_c/tandem/core_search.pyx":291
 * 
 * 
 * cdef class GlycanCoarseScorerBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double product_error_tolerance
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_vtab;
  double product_error_tolerance;
  double fragment_weight;
  double core_weight;
  int recalibrate_peptide_mass;
};


/* "glycresoft/_c/tandem/core_search.pyx":444
 * 
 * @cython.freelist(1000)
 * cdef class GlycanMatchResult(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double peptide_mass
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_vtab;
  double peptide_mass;
  double score;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *match;
  size_t glycan_size;
  PyObject *glycan_types;
  double recalibrated_peptide_mass;
};


/* "glycresoft/_c/tandem/core_search.pyx":493
 * @cython.freelist(1000)
 * @cython.final
 * cdef class ScoreCoarseGlycanMatchPair(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double score
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_vtab;
  double score;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *match;
};


/* "glycresoft/_c/tandem/core_search.pyx":598
 * 
 * 
 * cdef class IndexGlycanCompositionFragment(GlycanCompositionFragment):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int index
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment __pyx_base;
  int index;
  PyObject *name;
};


/* "glycresoft/_c/tandem/core_search.pyx":655
 * 
 * @cython.freelist(1000)
 * cdef class ComplementFragment(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment {
  PyObject_HEAD
  double mass;
  PyObject *keys;
};


/* "glycresoft/_c/tandem/core_search.pyx":669
 * 
 * @cython.freelist(10000)
 * cdef class PartialGlycanSolution(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double peptide_mass
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_vtab;
  double peptide_mass;
  double score;
  PyObject *core_matches;
  PyObject *fragment_matches;
  npy_int32 glycan_index;
};


/* "glycresoft/_c/tandem/core_search.pyx":708
 * 
 * 
 * cdef class GlycanFragmentIndex(object):             # <<<<<<<<<<<<<<
 *     '''A fast and sparse in-memory fragment ion index for quickly matching peaks against multiple
 *     glycan composition complements.
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_vtab;
  PyObject *members;
  PyObject *glycosylation_type;
  PyObject *unique_fragments;
  PyObject *_fragments;
  PyObject *fragment_index;
  long counter;
  double fragment_weight;
  double core_weight;
  double lower_bound;
  double upper_bound;
  double resolution;
};


/* "glycresoft/_c/tandem/core_search.pyx":253
 *         self.core_theoretical = core_theoretical
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield self.fragment_matches
 *         yield self.n_matched
 */
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ {
  PyObject_HEAD
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self;
};



/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*_PyComposition__equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*_PyComposition__equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);
};
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
  size_t (*basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet {
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, size_t);
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak *__pyx_optional_args);
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t (*_all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for *__pyx_optional_args);
  PyObject *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak *__pyx_optional_args);
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet;


/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */

struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*clone)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch);
  double (*calc_mass)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args);
  PyObject *(*_from_formula)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_from_dict)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, int __pyx_skip_dispatch);
  long (*getitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, long);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*add_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*subtract_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*_create)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
};
static struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition;


/* "glypy/_c/structure/glycan_composition.pxd":3
 * from glypy.composition.ccomposition cimport CComposition
 * 
 * cdef class _CompositionBase(dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object _mass
 */

struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase {
  PyObject *(*_getitem_fast)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_setitem_fast)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_update_from_typed_map)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_9structure_18glycan_composition_16_CompositionBase__update_from_typed_map *__pyx_optional_args);
  PyObject *(*serialize)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, int __pyx_skip_dispatch);
  void (*_add_from)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *);
  void (*_subtract_from)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *);
  PyObject *(*set_composition_offset)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*get_composition_offset)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, int __pyx_skip_dispatch);
  PyObject *(*_invalidate)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase *__pyx_vtabptr_5glypy_2_c_9structure_18glycan_composition__CompositionBase;


/* "glycopeptidepy/_c/count_table.pxd":52
 * 
 * 
 * cdef class CountTableIterator(object):             # <<<<<<<<<<<<<<
 *     cdef :
 *         count_table* table
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator {
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*advance)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *);
  int (*has_more)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *);
  int (*_locate_next_value)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *, PyObject **, long *);
  int (*get_next_value)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *, PyObject **, long *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator;


/* "glycopeptidepy/_c/count_table.pxd":69
 * 
 * 
 * cdef class CountTable(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* table
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable {
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *(*_create)(void);
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *(*_create_from)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  int (*_initialize_table)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*_update_from_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  void (*_update_from_count_table)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*_add_from)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*_subtract_from)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  int (*_add_from_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  int (*_subtract_from_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  void (*_scale_by)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, long);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  PyObject *(*_to_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*update)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *(*copy)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*clear)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*setdefault)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  long (*pop)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_pop *__pyx_optional_args);
  PyObject *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get *__pyx_optional_args);
  void (*increment)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, long);
  void (*decrement)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, long);
  long (*getitem)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  int (*setitem)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, long);
  long (*delitem)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  long (*total)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTable;


/* "glycopeptidepy/_c/structure/base.pxd":5
 * 
 * 
 * cdef class PeptideSequenceBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list sequence
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, Py_ssize_t);
  PyObject *(*_invalidate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;


/* "glycopeptidepy/_c/structure/base.pxd":21
 * 
 * 
 * cdef class TerminalGroup(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition base_composition
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *(*get_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*get_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *(*modify)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, int __pyx_skip_dispatch);
  void (*set_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, PyObject *);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;


/* "glycopeptidepy/_c/structure/base.pxd":35
 * 
 * 
 * cdef class AminoAcidResidueBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *(*_create)(PyObject *, PyObject *, double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;


/* "glycopeptidepy/_c/structure/base.pxd":50
 * 
 * 
 * cdef class ModificationBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase {
  int (*is_a)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, int __pyx_skip_dispatch);
  int (*is_tracked_for)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase;


/* "glycopeptidepy/_c/structure/base.pxd":69
 * 
 * 
 * cdef class SequencePosition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AminoAcidResidueBase amino_acid
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *, PyObject *);
  double (*get_mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *);
  int (*has_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  int (*is_modified)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, int __pyx_skip_dispatch);
  PyObject *(*drop_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  size_t (*get_modification_count)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *(*get_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, size_t);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition;


/* "glycopeptidepy/_c/structure/glycan.pxd":12
 * 
 * 
 * cdef class GlycanCompositionProxy(object):             # <<<<<<<<<<<<<<
 *     '''A mapping-like object that imitates the GlycanComposition interface in
 *     a read-only fashion.
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy {
  PyObject *(*_getitem_fast)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*__pyx_fuse_0_create)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*__pyx_fuse_1_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*__pyx_fuse_2_create)(PyObject *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;


/* "glycopeptidepy/_c/structure/glycan.pxd":29
 * 
 * 
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition composition_offset
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy __pyx_base;
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;


/* "glycopeptidepy/_c/structure/glycan.pxd":34
 * 
 * 
 * cdef class GlycosylationManager(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict mapping
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager {
  PyObject *(*update)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*clear)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop *__pyx_optional_args);
  size_t (*get_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*copy)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*invalidate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*get_aggregate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  int (*set_aggregate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *);
  PyObject *(*_patch_aggregate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*get_glycosylation_types)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  PyObject *(*_update_track)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*total_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  double (*mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  int (*is_fully_specified_topologies)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*get_glycan_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*_make_glycan_composition_proxy)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  int (*get_total_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  int (*total_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  int (*count_glycosylation_type)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch);
  void (*_init)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, PyObject *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;


/* "glycopeptidepy/_c/structure/fragment.pxd":6
 * 
 * 
 * cdef class ChemicalShiftBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public str name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;


/* "glycopeptidepy/_c/structure/fragment.pxd":26
 * 
 * 
 * cdef class FragmentBase(object):             # <<<<<<<<<<<<<<
 *     """Base class for all Fragment types. Defines basic
 *     name generation and neutral loss handling functions.
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *(*get_series)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *(*get_chemical_shift)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  PyObject *(*set_chemical_shift)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int __pyx_skip_dispatch);
  PyObject *(*base_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  PyObject *(*get_fragment_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  void (*_update_hash_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *);
  PyObject *(*get_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;


/* "glycopeptidepy/_c/structure/fragment.pxd":63
 * 
 * 
 * cdef class SimpleFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;


/* "glycopeptidepy/_c/structure/fragment.pxd":72
 * 
 * 
 * cdef class StubFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *(*_create)(PyObject *, double, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int, PyObject *, int, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create *__pyx_optional_args);
  int (*get_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;


/* "glycopeptidepy/_c/structure/fragment.pxd":97
 * 
 * 
 * cdef class PeptideFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, int, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, double, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create *__pyx_optional_args);
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *, int __pyx_skip_dispatch);
  void (*_update_mass_with_modifications)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *);
  int (*_is_glycosylated)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *);
  long (*get_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;


/* "glycopeptidepy/_c/structure/fragment.pxd":125
 * 
 * @cython.final
 * cdef class _NameTree(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*_create)(void);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_0traverse)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1traverse)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree;


/* "glycopeptidepy/_c/structure/sequence_methods.pxd":15
 * 
 * 
 * cdef class _PeptideSequenceCore(PeptideSequenceBase):             # <<<<<<<<<<<<<<
 * 
 *     cpdef _init_from_parsed_string(self, list seq_list, glycan=*, n_term=*, c_term=*)
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase __pyx_base;
  PyObject *(*_init_from_parsed_string)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore__init_from_parsed_string *__pyx_optional_args);
  PyObject *(*_init_from_string)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_init_from_components)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore__init_from_components *__pyx_optional_args);
  PyObject *(*_init_termini)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, int __pyx_skip_dispatch);
  PyObject *(*_initialize_fields)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  size_t (*get_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *(*get_n_term)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *);
  void (*set_n_term)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, PyObject *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *(*get_c_term)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *);
  void (*set_c_term)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, PyObject *);
  PyObject *(*get_sequence)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_16sequence_methods_20_PeptideSequenceCore_get_sequence *__pyx_optional_args);
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  int (*base_sequence_equality)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  int (*modified_sequence_equality)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  int (*full_structure_equality)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*peptide_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*total_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *, int __pyx_skip_dispatch);
  double (*get_peptide_backbone_mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *);
  double (*get_total_mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore;


/* "glycopeptidepy/_c/structure/fragmentation_strategy/base.pxd":28
 * 
 * 
 * cdef class FragmentationStrategyBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public _PeptideSequenceCore peptide
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*peptide_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*total_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase;


/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":13
 * 
 * 
 * cdef class GlycanCompositionFragment(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*copy)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*_get_glycan_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, int __pyx_skip_dispatch);
  int (*get_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*_create)(double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment;


/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":31
 * 
 * 
 * cdef class GlycanCompositionFragmentStrategyBase(FragmentationStrategyBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint _use_query
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*glycan_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase *, int __pyx_skip_dispatch);
  int (*_guess_query_mode)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *);
  long (*count_glycosylation_type)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase;


/* "glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pxd":42
 * 
 * 
 * cdef class StubGlycopeptideStrategy(GlycanCompositionFragmentStrategyBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint extended
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*fucosylate_increment)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*xylosylate_increment)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*modified_increment)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*fucosylate_extended)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, long, int __pyx_skip_dispatch);
  int (*_validate_glycan_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_combinate_sites)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_combinate_sites_single)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, PyObject *, PyObject *);
  PyObject *(*n_glycan_composition_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_n_glycan_composition_fragments *__pyx_optional_args);
  PyObject *(*o_glycan_composition_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_o_glycan_composition_fragments *__pyx_optional_args);
  PyObject *(*gag_linker_composition_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_gag_linker_composition_fragments *__pyx_optional_args);
  PyObject *(*mixed_stub_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, int __pyx_skip_dispatch);
  PyObject *(*n_glycan_stub_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, int __pyx_skip_dispatch);
  PyObject *(*o_glycan_stub_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, int __pyx_skip_dispatch);
  PyObject *(*gag_linker_stub_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, int __pyx_skip_dispatch);
  PyObject *(*stub_fragments)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy;


/* "glycresoft/_c/tandem/core_search.pyx":45
 * 
 * 
 * cdef class GlycanCombinationRecordBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t id
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase {
  int (*is_n_glycan)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  int (*is_o_glycan)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  int (*is_gag_linker)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  double (*normalized_size)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  PyObject *(*get_n_glycan_fragments)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  PyObject *(*get_o_glycan_fragments)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  PyObject *(*get_gag_linker_glycan_fragments)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
  PyObject *(*clear)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;


/* "glycresoft/_c/tandem/core_search.pyx":193
 * @cython.freelist(100000)
 * @cython.final
 * cdef class CoarseStubGlycopeptideMatch(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object key
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *(*_create)(PyObject *, double, double, PyObject *);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch__create(PyObject *, double, double, PyObject *);


/* "glycresoft/_c/tandem/core_search.pyx":228
 * @cython.freelist(100000)
 * @cython.final
 * cdef class CoarseGlycanMatch(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list fragment_matches
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_create)(PyObject *, double, double, double, double);
  double (*estimate_peptide_mass)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch__create(PyObject *, double, double, double, double);
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_estimate_peptide_mass(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, int __pyx_skip_dispatch);


/* "glycresoft/_c/tandem/core_search.pyx":291
 * 
 * 
 * cdef class GlycanCoarseScorerBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double product_error_tolerance
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_match_fragments)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments *__pyx_optional_args);
  double (*_calculate_score)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, int __pyx_skip_dispatch);
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_n_glycan_match_stubs)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs *__pyx_optional_args);
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_o_glycan_match_stubs)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs *__pyx_optional_args);
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_gag_match_stubs)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs *__pyx_optional_args);
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_get_n_glycan_coarse_score)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *, double, double *);
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_get_o_glycan_coarse_score)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *, double, double *);
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*_get_gag_coarse_score)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *, double, double *);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;


/* "glycresoft/_c/tandem/core_search.pyx":444
 * 
 * @cython.freelist(1000)
 * cdef class GlycanMatchResult(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double peptide_mass
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *(*_create)(double, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, size_t, PyObject *, double);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;


/* "glycresoft/_c/tandem/core_search.pyx":493
 * @cython.freelist(1000)
 * @cython.final
 * cdef class ScoreCoarseGlycanMatchPair(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double score
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *(*_create)(double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create(double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *);


/* "glycresoft/_c/tandem/core_search.pyx":598
 * 
 * 
 * cdef class IndexGlycanCompositionFragment(GlycanCompositionFragment):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int index
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment __pyx_base;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *(*_create)(double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;


/* "glycresoft/_c/tandem/core_search.pyx":669
 * 
 * @cython.freelist(10000)
 * cdef class PartialGlycanSolution(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double peptide_mass
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *(*_create)(void);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;


/* "glycresoft/_c/tandem/core_search.pyx":708
 * 
 * 
 * cdef class GlycanFragmentIndex(object):             # <<<<<<<<<<<<<<
 *     '''A fast and sparse in-memory fragment ion index for quickly matching peaks against multiple
 *     glycan composition complements.
 */

struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *(*_intern)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  long (*query_width)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, double, double, int __pyx_skip_dispatch);
  long (*bin_index)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, double, int __pyx_skip_dispatch);
  PyObject *(*_get_fragments)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, PyObject *, int __pyx_skip_dispatch);
  int (*_match_fragments)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, double, PyObject *);
  PyObject *(*match)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* py_dict_clear.proto */
#define __Pyx_PyDict_Clear(d) (PyDict_Clear(d), 0)

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyIntFromDouble.proto */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value);
#else
#define __Pyx_PyInt_FromDouble(value) PyLong_FromDouble(value)
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* SoftComplexToDouble.proto */
static double __Pyx_SoftComplexToDouble(__pyx_t_double_complex value, int have_gil);

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pystring_as_double.proto */
static CYTHON_INLINE double __Pyx_PyString_AsDouble(PyObject *obj) {
    #if PY_MAJOR_VERSION >= 3
    (void)__Pyx_PyBytes_AsDouble;
    return __Pyx_PyUnicode_AsDouble(obj);
    #else
    (void)__Pyx_PyUnicode_AsDouble;
    return __Pyx_PyBytes_AsDouble(obj);
    #endif
}

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_11
#define __PYX_HAVE_RT_ImportType_proto_3_0_11
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_11 {
   __Pyx_ImportType_CheckSize_Error_3_0_11 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_11 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_11 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_long__double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_long__double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_long__double(a, b) ((a)-(b))
    #define __Pyx_c_prod_long__double(a, b) ((a)*(b))
    #define __Pyx_c_quot_long__double(a, b) ((a)/(b))
    #define __Pyx_c_neg_long__double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_long__double(z) ((z)==(long double)0)
    #define __Pyx_c_conj_long__double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (::std::abs(z))
        #define __Pyx_c_pow_long__double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_long__double(z) ((z)==0)
    #define __Pyx_c_conj_long__double(z)    (conjl(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (cabsl(z))
        #define __Pyx_c_pow_long__double(a, b)  (cpowl(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex);
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex);
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint16(npy_uint16 value);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint16 __Pyx_PyInt_As_npy_uint16(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_o_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_gag_linker(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_normalized_size(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_n_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_o_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_gag_linker_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_clear(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch__create(PyObject *__pyx_v_key, double __pyx_v_mass, double __pyx_v_shift_mass, PyObject *__pyx_v_peaks_matched); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch__create(PyObject *__pyx_v_fragment_matches, double __pyx_v_n_matched, double __pyx_v_n_theoretical, double __pyx_v_core_matched, double __pyx_v_core_theoretical); /* proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_estimate_peptide_mass(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, PyObject *__pyx_v_shifts, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments *__pyx_optional_args); /* proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__calculate_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs *__pyx_optional_args); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs *__pyx_optional_args); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs *__pyx_optional_args); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_n_glycan_coarse_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, double __pyx_v_peptide_mass, double *__pyx_v_scoreout); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_o_glycan_coarse_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, double __pyx_v_peptide_mass, double *__pyx_v_scoreout); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_gag_coarse_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, double __pyx_v_peptide_mass, double *__pyx_v_scoreout); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_f_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult__create(double __pyx_v_peptide_mass, double __pyx_v_score, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_match, size_t __pyx_v_glycan_size, PyObject *__pyx_v_glycan_types, double __pyx_v_recalibrated_peptide_mass); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create(double __pyx_v_score, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_match); /* proto*/
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create(double __pyx_v_mass, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_key, int __pyx_v_is_extended); /* proto*/
struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create__pyx_wrap_1(double __pyx_v_mass, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_key, int __pyx_v_is_extended); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_copy(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_f_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution__create(void); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__intern(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_fragment, PyObject *__pyx_v_glycosylation_type, int __pyx_skip_dispatch); /* proto*/
static long __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_query_width(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass, double __pyx_v_error_tolerance, int __pyx_skip_dispatch); /* proto*/
static long __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_bin_index(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__get_fragments(CYTHON_UNUSED struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_gcr, PyObject *__pyx_v_glycosylation_type, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__match_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_delta_mass, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, PyObject *__pyx_v_result); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_scan, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match *__pyx_optional_args); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython.version" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "libc" */

/* Module declarations from "libc.math" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from "brainpy._c.composition" */
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from "brainpy._c.double_vector" */

/* Module declarations from "brainpy._c.isotopic_constants" */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from "brainpy._c.isotopic_distribution" */

/* Module declarations from "ms_peak_picker._c.peak_set" */
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from "ms_peak_picker._c.peak_index" */

/* Module declarations from "ms_deisotope._c.averagine" */
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)

/* Module declarations from "ms_deisotope._c.deconvoluter_base" */

/* Module declarations from "ms_deisotope._c.scoring" */
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from "ms_deisotope._c.peak_set" */
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from "glypy.composition.ccomposition" */
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__atom = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__atom (*__pyx_vp_5glypy_11composition_12ccomposition__atom)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__formula = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__formula (*__pyx_vp_5glypy_11composition_12ccomposition__formula)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__isotope_string = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__isotope_string (*__pyx_vp_5glypy_11composition_12ccomposition__isotope_string)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern = 0;
#define __pyx_v_5glypy_11composition_12ccomposition_isotope_pattern (*__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern = 0;
#define __pyx_v_5glypy_11composition_12ccomposition_formula_pattern (*__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern)

/* Module declarations from "glycresoft._c.chromatogram_tree.mass_shift" */
static PyObject **__pyx_vp_10glycresoft_2_c_17chromatogram_tree_10mass_shift_mass_shift_index = 0;
#define __pyx_v_10glycresoft_2_c_17chromatogram_tree_10mass_shift_mass_shift_index (*__pyx_vp_10glycresoft_2_c_17chromatogram_tree_10mass_shift_mass_shift_index)
static struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase **__pyx_vp_10glycresoft_2_c_17chromatogram_tree_10mass_shift_Unmodified = 0;
#define __pyx_v_10glycresoft_2_c_17chromatogram_tree_10mass_shift_Unmodified (*__pyx_vp_10glycresoft_2_c_17chromatogram_tree_10mass_shift_Unmodified)

/* Module declarations from "glypy._c.structure.glycan_composition" */

/* Module declarations from "glycopeptidepy._c.count_table" */

/* Module declarations from "glycopeptidepy._c.structure.base" */

/* Module declarations from "glycopeptidepy._c.structure.glycan" */

/* Module declarations from "glycopeptidepy._c.structure.fragment" */

/* Module declarations from "glycopeptidepy._c.structure.sequence_methods" */

/* Module declarations from "glycopeptidepy._c.structure.fragmentation_strategy.base" */
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase **__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__n_glycosylation = 0;
#define __pyx_v_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__n_glycosylation (*__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__n_glycosylation)
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase **__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__o_glycosylation = 0;
#define __pyx_v_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__o_glycosylation (*__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__o_glycosylation)
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase **__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__gag_linker_glycosylation = 0;
#define __pyx_v_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__gag_linker_glycosylation (*__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__gag_linker_glycosylation)
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase **__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_hexnac = 0;
#define __pyx_v_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_hexnac (*__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_hexnac)
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase **__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_xylose = 0;
#define __pyx_v_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_xylose (*__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_xylose)
static PyObject **__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_glycosylation_type_to_core = 0;
#define __pyx_v_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_glycosylation_type_to_core (*__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_glycosylation_type_to_core)

/* Module declarations from "glycopeptidepy._c.structure.fragmentation_strategy.glycan" */

/* Module declarations from "glycresoft._c.tandem.core_search" */
static PyObject *__pyx_v_10glycresoft_2_c_6tandem_11core_search_HashableGlycanComposition = 0;
static struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified = 0;
static PyObject *__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan = 0;
static PyObject *__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan = 0;
static PyObject *__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker = 0;
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_score_getter(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match *__pyx_optional_args); /*proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_partial_score_getter(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCombinationRecordBase__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_CoarseGlycanMatch__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCoarseScorerBase__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanMatchResult__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ComplementFragment__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_PartialGlycanSolution__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanFragmentIndex__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "glycresoft._c.tandem.core_search"
extern int __pyx_module_is_main_glycresoft___c__tandem__core_search;
int __pyx_module_is_main_glycresoft___c__tandem__core_search = 0;

/* Implementation of "glycresoft._c.tandem.core_search" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_ImportError;
/* #### Code section: string_decls ### */
static const char __pyx_k_d[] = "d";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_x[] = "x";
static const char __pyx_k__7[] = ".";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_Hex[] = "Hex";
static const char __pyx_k__15[] = "*";
static const char __pyx_k__89[] = "?";
static const char __pyx_k_f_r[] = "(%f, %r)";
static const char __pyx_k_gcr[] = "gcr";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_low[] = "low";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_comp[] = "comp";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_frag[] = "frag";
static const char __pyx_k_high[] = "high";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mass[] = "mass";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_scan[] = "scan";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_build[] = "build";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_frags[] = "frags";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_ptemp[] = "ptemp";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_HexNAc[] = "HexNAc";
static const char __pyx_k_comp_i[] = "comp_i";
static const char __pyx_k_comp_n[] = "comp_n";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_frag_i[] = "frag_i";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_intern[] = "_intern";
static const char __pyx_k_member[] = "member";
static const char __pyx_k_n_core[] = "n_core";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_shifts[] = "shifts";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_frags_i[] = "frags_i";
static const char __pyx_k_frags_n[] = "frags_n";
static const char __pyx_k_is_core[] = "is_core";
static const char __pyx_k_members[] = "members";
static const char __pyx_k_n_frags[] = "n_frags";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_s_s_f_r[] = "%s(%s, %f, %r)";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_extended[] = "extended";
static const char __pyx_k_fragment[] = "fragment";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_hash_key[] = "_hash_key";
static const char __pyx_k_n_glycan[] = "n_glycan";
static const char __pyx_k_o_glycan[] = "o_glycan";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_bin_index[] = "bin_index";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_n_matched[] = "n_matched";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_s_s_f_f_r[] = "%s(%s, %f, %f, %r)";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_Unmodified[] = "Unmodified";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_comp_frags[] = "comp_frags";
static const char __pyx_k_gag_linker[] = "gag_linker";
static const char __pyx_k_mass_shift[] = "mass_shift";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_query_mass[] = "query_mass";
static const char __pyx_k_resolution[] = "resolution";
static const char __pyx_k_shift_mass[] = "shift_mass";
static const char __pyx_k_GlycanTypes[] = "GlycanTypes";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_core_weight[] = "core_weight";
static const char __pyx_k_defaultdict[] = "defaultdict";
static const char __pyx_k_glycan_size[] = "glycan_size";
static const char __pyx_k_is_extended[] = "is_extended";
static const char __pyx_k_is_n_glycan[] = "is_n_glycan";
static const char __pyx_k_is_o_glycan[] = "is_o_glycan";
static const char __pyx_k_query_width[] = "query_width";
static const char __pyx_k_Unmodified_2[] = "_Unmodified";
static const char __pyx_k_core_matched[] = "core_matched";
static const char __pyx_k_core_matches[] = "core_matches";
static const char __pyx_k_glycan_index[] = "glycan_index";
static const char __pyx_k_glycan_match[] = "glycan_match";
static const char __pyx_k_glycan_types[] = "glycan_types";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_neutral_mass[] = "neutral_mass";
static const char __pyx_k_peptide_mass[] = "peptide_mass";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_score_getter[] = "score_getter";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_get_fragments[] = "_get_fragments";
static const char __pyx_k_is_gag_linker[] = "is_gag_linker";
static const char __pyx_k_n_theoretical[] = "n_theoretical";
static const char __pyx_k_peaks_matched[] = "peaks_matched";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_calculate_score[] = "_calculate_score";
static const char __pyx_k_dehydrated_mass[] = "dehydrated_mass";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_fragment_weight[] = "fragment_weight";
static const char __pyx_k_gag_match_stubs[] = "_gag_match_stubs";
static const char __pyx_k_match_fragments[] = "_match_fragments";
static const char __pyx_k_normalized_size[] = "normalized_size";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_core_theoretical[] = "core_theoretical";
static const char __pyx_k_fragment_matches[] = "fragment_matches";
static const char __pyx_k_CoarseGlycanMatch[] = "CoarseGlycanMatch";
static const char __pyx_k_GlycanMatchResult[] = "GlycanMatchResult";
static const char __pyx_k_ComplementFragment[] = "ComplementFragment";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_glycan_combination[] = "glycan_combination";
static const char __pyx_k_glycosylation_type[] = "glycosylation_type";
static const char __pyx_k_GlycanFragmentIndex[] = "GlycanFragmentIndex";
static const char __pyx_k_build_locals_lambda[] = "build.<locals>.<lambda>";
static const char __pyx_k_glycresoft_serialize[] = "glycresoft.serialize";
static const char __pyx_k_minimum_peptide_mass[] = "minimum_peptide_mass";
static const char __pyx_k_n_glycan_match_stubs[] = "_n_glycan_match_stubs";
static const char __pyx_k_o_glycan_match_stubs[] = "_o_glycan_match_stubs";
static const char __pyx_k_partial_score_getter[] = "partial_score_getter";
static const char __pyx_k_PartialGlycanSolution[] = "PartialGlycanSolution";
static const char __pyx_k_deconvoluted_peak_set[] = "deconvoluted_peak_set";
static const char __pyx_k_estimate_peptide_mass[] = "estimate_peptide_mass";
static const char __pyx_k_glycan_combination_db[] = "glycan_combination_db";
static const char __pyx_k_precursor_information[] = "precursor_information";
static const char __pyx_k_GlycanCoarseScorerBase[] = "GlycanCoarseScorerBase";
static const char __pyx_k_get_n_glycan_fragments[] = "get_n_glycan_fragments";
static const char __pyx_k_get_o_glycan_fragments[] = "get_o_glycan_fragments";
static const char __pyx_k_mass_shift_tandem_mass[] = "mass_shift_tandem_mass";
static const char __pyx_k_product_error_tolerance[] = "product_error_tolerance";
static const char __pyx_k_CoarseGlycanMatch___iter[] = "CoarseGlycanMatch.__iter__";
static const char __pyx_k_recalibrate_peptide_mass[] = "recalibrate_peptide_mass";
static const char __pyx_k_GlycanFragmentIndex_build[] = "GlycanFragmentIndex.build";
static const char __pyx_k_GlycanFragmentIndex_match[] = "GlycanFragmentIndex.match";
static const char __pyx_k_HashableGlycanComposition[] = "HashableGlycanComposition";
static const char __pyx_k_recalibrated_peptide_mass[] = "recalibrated_peptide_mass";
static const char __pyx_k_ScoreCoarseGlycanMatchPair[] = "ScoreCoarseGlycanMatchPair";
static const char __pyx_k_CoarseStubGlycopeptideMatch[] = "CoarseStubGlycopeptideMatch";
static const char __pyx_k_GlycanCombinationRecordBase[] = "GlycanCombinationRecordBase";
static const char __pyx_k_GlycanFragmentIndex__intern[] = "GlycanFragmentIndex._intern";
static const char __pyx_k_HashableGlycanComposition_2[] = "_HashableGlycanComposition";
static const char __pyx_k_glycresoft_chromatogram_tree[] = "glycresoft.chromatogram_tree";
static const char __pyx_k_GlycanFragmentIndex_bin_index[] = "GlycanFragmentIndex.bin_index";
static const char __pyx_k_CoarseStubGlycopeptideFragment[] = "CoarseStubGlycopeptideFragment";
static const char __pyx_k_IndexGlycanCompositionFragment[] = "IndexGlycanCompositionFragment";
static const char __pyx_k_pyx_unpickle_CoarseGlycanMatch[] = "__pyx_unpickle_CoarseGlycanMatch";
static const char __pyx_k_pyx_unpickle_ComplementFragmen[] = "__pyx_unpickle_ComplementFragment";
static const char __pyx_k_pyx_unpickle_GlycanCoarseScore[] = "__pyx_unpickle_GlycanCoarseScorerBase";
static const char __pyx_k_pyx_unpickle_GlycanCombination[] = "__pyx_unpickle_GlycanCombinationRecordBase";
static const char __pyx_k_pyx_unpickle_GlycanFragmentInd[] = "__pyx_unpickle_GlycanFragmentIndex";
static const char __pyx_k_pyx_unpickle_GlycanMatchResult[] = "__pyx_unpickle_GlycanMatchResult";
static const char __pyx_k_pyx_unpickle_PartialGlycanSolu[] = "__pyx_unpickle_PartialGlycanSolution";
static const char __pyx_k_pyx_unpickle_ScoreCoarseGlycan[] = "__pyx_unpickle_ScoreCoarseGlycanMatchPair";
static const char __pyx_k_self___class_____name___self_n[] = "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, {self.core_matched}, {self.core_theoretical})";
static const char __pyx_k_GlycanCoarseScorerBase___reduce[] = "GlycanCoarseScorerBase.__reduce_cython__";
static const char __pyx_k_GlycanCombinationRecordBase_get[] = "GlycanCombinationRecordBase.get_n_glycan_fragments";
static const char __pyx_k_GlycanFragmentIndex_query_width[] = "GlycanFragmentIndex.query_width";
static const char __pyx_k_Re_estimated_peptide_mass_error[] = "Re-estimated peptide mass error is large: %f vs %f";
static const char __pyx_k_get_gag_linker_glycan_fragments[] = "get_gag_linker_glycan_fragments";
static const char __pyx_k_self___class_____name___self_ma[] = "{self.__class__.__name__}({self.mass})";
static const char __pyx_k_self___class_____name___self_pe[] = "{self.__class__.__name__}({self.peptide_mass}, {self.score}, {self.match}, {self.glycan_size}, {self.glycan_types}, {self.recalibrated_peptide_mass})";
static const char __pyx_k_CoarseGlycanMatch___reduce_cytho[] = "CoarseGlycanMatch.__reduce_cython__";
static const char __pyx_k_CoarseGlycanMatch___setstate_cyt[] = "CoarseGlycanMatch.__setstate_cython__";
static const char __pyx_k_CoarseGlycanMatch_estimate_pepti[] = "CoarseGlycanMatch.estimate_peptide_mass";
static const char __pyx_k_CoarseStubGlycopeptideFragment_2[] = "CoarseStubGlycopeptideFragment.__reduce__";
static const char __pyx_k_CoarseStubGlycopeptideMatch___re[] = "CoarseStubGlycopeptideMatch.__reduce__";
static const char __pyx_k_ComplementFragment___reduce_cyth[] = "ComplementFragment.__reduce_cython__";
static const char __pyx_k_ComplementFragment___setstate_cy[] = "ComplementFragment.__setstate_cython__";
static const char __pyx_k_GlycanCoarseScorerBase___setstat[] = "GlycanCoarseScorerBase.__setstate_cython__";
static const char __pyx_k_GlycanCoarseScorerBase__calculat[] = "GlycanCoarseScorerBase._calculate_score";
static const char __pyx_k_GlycanCoarseScorerBase__gag_matc[] = "GlycanCoarseScorerBase._gag_match_stubs";
static const char __pyx_k_GlycanCoarseScorerBase__match_fr[] = "GlycanCoarseScorerBase._match_fragments";
static const char __pyx_k_GlycanCoarseScorerBase__n_glycan[] = "GlycanCoarseScorerBase._n_glycan_match_stubs";
static const char __pyx_k_GlycanCoarseScorerBase__o_glycan[] = "GlycanCoarseScorerBase._o_glycan_match_stubs";
static const char __pyx_k_GlycanCombinationRecordBase___re[] = "GlycanCombinationRecordBase.__reduce_cython__";
static const char __pyx_k_GlycanCombinationRecordBase___se[] = "GlycanCombinationRecordBase.__setstate_cython__";
static const char __pyx_k_GlycanCombinationRecordBase_clea[] = "GlycanCombinationRecordBase.clear";
static const char __pyx_k_GlycanCombinationRecordBase_is_g[] = "GlycanCombinationRecordBase.is_gag_linker";
static const char __pyx_k_GlycanCombinationRecordBase_is_n[] = "GlycanCombinationRecordBase.is_n_glycan";
static const char __pyx_k_GlycanCombinationRecordBase_is_o[] = "GlycanCombinationRecordBase.is_o_glycan";
static const char __pyx_k_GlycanCombinationRecordBase_norm[] = "GlycanCombinationRecordBase.normalized_size";
static const char __pyx_k_GlycanFilteringPeptideMassEstima[] = "GlycanFilteringPeptideMassEstimator_match";
static const char __pyx_k_GlycanFragmentIndex___reduce_cyt[] = "GlycanFragmentIndex.__reduce_cython__";
static const char __pyx_k_GlycanFragmentIndex___setstate_c[] = "GlycanFragmentIndex.__setstate_cython__";
static const char __pyx_k_GlycanFragmentIndex__get_fragmen[] = "GlycanFragmentIndex._get_fragments";
static const char __pyx_k_GlycanMatchResult___reduce_cytho[] = "GlycanMatchResult.__reduce_cython__";
static const char __pyx_k_GlycanMatchResult___setstate_cyt[] = "GlycanMatchResult.__setstate_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))";
static const char __pyx_k_IndexGlycanCompositionFragment_2[] = "IndexGlycanCompositionFragment.__reduce__";
static const char __pyx_k_IndexGlycanCompositionFragment_3[] = "IndexGlycanCompositionFragment.__getstate__";
static const char __pyx_k_IndexGlycanCompositionFragment_4[] = "IndexGlycanCompositionFragment.__setstate__";
static const char __pyx_k_IndexGlycanCompositionFragment_c[] = "IndexGlycanCompositionFragment.copy";
static const char __pyx_k_PartialGlycanSolution___reduce_c[] = "PartialGlycanSolution.__reduce_cython__";
static const char __pyx_k_PartialGlycanSolution___setstate[] = "PartialGlycanSolution.__setstate_cython__";
static const char __pyx_k_ScoreCoarseGlycanMatchPair___red[] = "ScoreCoarseGlycanMatchPair.__reduce_cython__";
static const char __pyx_k_ScoreCoarseGlycanMatchPair___set[] = "ScoreCoarseGlycanMatchPair.__setstate_cython__";
static const char __pyx_k_glycresoft__c_tandem_core_search[] = "glycresoft._c.tandem.core_search";
static const char __pyx_k_glypy_structure_glycan_compositi[] = "glypy.structure.glycan_composition";
static const char __pyx_k_numpy__core_multiarray_failed_to[] = "numpy._core.multiarray failed to import";
static const char __pyx_k_numpy__core_umath_failed_to_impo[] = "numpy._core.umath failed to import";
static const char __pyx_k_src_glycresoft__c_tandem_core_se[] = "src/glycresoft/_c/tandem/core_search.pyx";
static const char __pyx_k_GlycanCombinationRecordBase_get_2[] = "GlycanCombinationRecordBase.get_o_glycan_fragments";
static const char __pyx_k_GlycanCombinationRecordBase_get_3[] = "GlycanCombinationRecordBase.get_gag_linker_glycan_fragments";
static const char __pyx_k_self___class_____name___self_pe_2[] = "{self.__class__.__name__}({self.peptide_mass}, {self.score}, {self.core_matches}, {self.fragment_matches}, {self.glycan_index})";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_6[] = "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_7[] = "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_8[] = "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2is_o_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4is_gag_linker(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_6normalized_size(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_8get_n_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_10get_o_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12get_gag_linker_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_14clear(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_18__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_mass, PyObject *__pyx_v_is_core); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_2__reduce__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_6__eq__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_8__ne__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_10__hash__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_12__lt__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_14__gt__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_mass, PyObject *__pyx_v_shift_mass, PyObject *__pyx_v_peaks_matched); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_2__reduce__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_fragment_matches, PyObject *__pyx_v_n_matched, PyObject *__pyx_v_n_theoretical, PyObject *__pyx_v_core_matched, PyObject *__pyx_v_core_theoretical); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_2__iter__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_5estimate_peptide_mass(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_7__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_11__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_product_error_tolerance, PyObject *__pyx_v_fragment_weight, PyObject *__pyx_v_core_weight, PyObject *__pyx_v_recalibrate_peptide_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_2_match_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, PyObject *__pyx_v_shifts, double __pyx_v_mass_shift_tandem_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_4_calculate_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_6_n_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, double __pyx_v_mass_shift_tandem_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_8_o_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, double __pyx_v_mass_shift_tandem_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_10_gag_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, double __pyx_v_mass_shift_tandem_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_12__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_14__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_peptide_mass, PyObject *__pyx_v_score, PyObject *__pyx_v_match, PyObject *__pyx_v_glycan_size, PyObject *__pyx_v_glycan_types, PyObject *__pyx_v_recalibrated_peptide_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_20fragment_match_count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_2__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_4__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_6__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_score, PyObject *__pyx_v_match); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_2__getitem__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static Py_ssize_t __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_4__len__(CYTHON_UNUSED struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_6__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_8__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_10__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_score_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_2GlycanFilteringPeptideMassEstimator_match(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, PyObject *__pyx_v_query_mass); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_mass, PyObject *__pyx_v_composition, PyObject *__pyx_v_key, PyObject *__pyx_v_is_extended); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_2copy(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4__reduce__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_6__getstate__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_8__setstate__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v_mass, PyObject *__pyx_v_keys); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_2__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_6__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_peptide_mass, PyObject *__pyx_v_score, PyObject *__pyx_v_core_matches, PyObject *__pyx_v_fragment_matches, npy_int32 __pyx_v_glycan_index); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_2__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_20fragment_match_count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_4__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_6__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_members, PyObject *__pyx_v_fragment_weight, PyObject *__pyx_v_core_weight, PyObject *__pyx_v_resolution); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_2_intern(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_fragment, PyObject *__pyx_v_glycosylation_type); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_4query_width(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_6bin_index(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_8build(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_get_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_gcr, PyObject *__pyx_v_glycosylation_type); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_12match(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_error_tolerance, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, PyObject *__pyx_v_query_mass); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_4partial_score_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_6__pyx_unpickle_GlycanCombinationRecordBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_8__pyx_unpickle_CoarseGlycanMatch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_10__pyx_unpickle_GlycanCoarseScorerBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_12__pyx_unpickle_GlycanMatchResult(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_14__pyx_unpickle_ScoreCoarseGlycanMatchPair(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_16__pyx_unpickle_ComplementFragment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18__pyx_unpickle_PartialGlycanSolution(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_20__pyx_unpickle_GlycanFragmentIndex(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ComplementFragment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase;
  PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak;
  PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet;
  PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
  PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5glypy_11composition_12ccomposition_CComposition;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase;
  PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy;
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
  PyObject *__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__;
  #endif
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__;
  PyObject *__pyx_n_s_AttributeError;
  PyObject *__pyx_n_s_CoarseGlycanMatch;
  PyObject *__pyx_n_s_CoarseGlycanMatch___iter;
  PyObject *__pyx_n_s_CoarseGlycanMatch___reduce_cytho;
  PyObject *__pyx_n_s_CoarseGlycanMatch___setstate_cyt;
  PyObject *__pyx_n_s_CoarseGlycanMatch_estimate_pepti;
  PyObject *__pyx_n_s_CoarseStubGlycopeptideFragment;
  PyObject *__pyx_n_s_CoarseStubGlycopeptideFragment_2;
  PyObject *__pyx_n_s_CoarseStubGlycopeptideMatch;
  PyObject *__pyx_n_s_CoarseStubGlycopeptideMatch___re;
  PyObject *__pyx_n_s_ComplementFragment;
  PyObject *__pyx_n_s_ComplementFragment___reduce_cyth;
  PyObject *__pyx_n_s_ComplementFragment___setstate_cy;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase___reduce;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase___setstat;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase__calculat;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase__gag_matc;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase__match_fr;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase__n_glycan;
  PyObject *__pyx_n_s_GlycanCoarseScorerBase__o_glycan;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase___re;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase___se;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_clea;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_get;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_get_2;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_get_3;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_is_g;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_is_n;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_is_o;
  PyObject *__pyx_n_s_GlycanCombinationRecordBase_norm;
  PyObject *__pyx_n_s_GlycanFilteringPeptideMassEstima;
  PyObject *__pyx_n_s_GlycanFragmentIndex;
  PyObject *__pyx_n_s_GlycanFragmentIndex___reduce_cyt;
  PyObject *__pyx_n_s_GlycanFragmentIndex___setstate_c;
  PyObject *__pyx_n_s_GlycanFragmentIndex__get_fragmen;
  PyObject *__pyx_n_s_GlycanFragmentIndex__intern;
  PyObject *__pyx_n_s_GlycanFragmentIndex_bin_index;
  PyObject *__pyx_n_s_GlycanFragmentIndex_build;
  PyObject *__pyx_n_s_GlycanFragmentIndex_match;
  PyObject *__pyx_n_s_GlycanFragmentIndex_query_width;
  PyObject *__pyx_n_s_GlycanMatchResult;
  PyObject *__pyx_n_s_GlycanMatchResult___reduce_cytho;
  PyObject *__pyx_n_s_GlycanMatchResult___setstate_cyt;
  PyObject *__pyx_n_s_GlycanTypes;
  PyObject *__pyx_n_s_HashableGlycanComposition;
  PyObject *__pyx_n_s_HashableGlycanComposition_2;
  PyObject *__pyx_n_s_Hex;
  PyObject *__pyx_n_s_HexNAc;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_n_s_IndexGlycanCompositionFragment;
  PyObject *__pyx_n_s_IndexGlycanCompositionFragment_2;
  PyObject *__pyx_n_s_IndexGlycanCompositionFragment_3;
  PyObject *__pyx_n_s_IndexGlycanCompositionFragment_4;
  PyObject *__pyx_n_s_IndexGlycanCompositionFragment_c;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_PartialGlycanSolution;
  PyObject *__pyx_n_s_PartialGlycanSolution___reduce_c;
  PyObject *__pyx_n_s_PartialGlycanSolution___setstate;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_kp_s_Re_estimated_peptide_mass_error;
  PyObject *__pyx_n_s_ScoreCoarseGlycanMatchPair;
  PyObject *__pyx_n_s_ScoreCoarseGlycanMatchPair___red;
  PyObject *__pyx_n_s_ScoreCoarseGlycanMatchPair___set;
  PyObject *__pyx_n_s_Unmodified;
  PyObject *__pyx_n_s_Unmodified_2;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s__15;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_n_s__89;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_bin_index;
  PyObject *__pyx_n_s_build;
  PyObject *__pyx_n_s_build_locals_lambda;
  PyObject *__pyx_n_s_calculate_score;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_comp;
  PyObject *__pyx_n_s_comp_frags;
  PyObject *__pyx_n_s_comp_i;
  PyObject *__pyx_n_s_comp_n;
  PyObject *__pyx_n_s_composition;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_core_matched;
  PyObject *__pyx_n_s_core_matches;
  PyObject *__pyx_n_s_core_theoretical;
  PyObject *__pyx_n_s_core_weight;
  PyObject *__pyx_n_s_d;
  PyObject *__pyx_n_s_deconvoluted_peak_set;
  PyObject *__pyx_n_s_defaultdict;
  PyObject *__pyx_n_s_dehydrated_mass;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_error_tolerance;
  PyObject *__pyx_n_s_estimate_peptide_mass;
  PyObject *__pyx_n_s_extended;
  PyObject *__pyx_kp_s_f_r;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_frag;
  PyObject *__pyx_n_s_frag_i;
  PyObject *__pyx_n_s_fragment;
  PyObject *__pyx_n_s_fragment_matches;
  PyObject *__pyx_n_s_fragment_weight;
  PyObject *__pyx_n_s_frags;
  PyObject *__pyx_n_s_frags_i;
  PyObject *__pyx_n_s_frags_n;
  PyObject *__pyx_n_s_gag_linker;
  PyObject *__pyx_n_s_gag_match_stubs;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_gcr;
  PyObject *__pyx_n_s_get_fragments;
  PyObject *__pyx_n_s_get_gag_linker_glycan_fragments;
  PyObject *__pyx_n_s_get_n_glycan_fragments;
  PyObject *__pyx_n_s_get_o_glycan_fragments;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_glycan_combination;
  PyObject *__pyx_n_s_glycan_combination_db;
  PyObject *__pyx_n_s_glycan_index;
  PyObject *__pyx_n_s_glycan_match;
  PyObject *__pyx_n_s_glycan_size;
  PyObject *__pyx_n_s_glycan_types;
  PyObject *__pyx_n_s_glycosylation_type;
  PyObject *__pyx_n_s_glycresoft__c_tandem_core_search;
  PyObject *__pyx_n_s_glycresoft_chromatogram_tree;
  PyObject *__pyx_n_s_glycresoft_serialize;
  PyObject *__pyx_n_s_glypy_structure_glycan_compositi;
  PyObject *__pyx_n_s_hash_key;
  PyObject *__pyx_n_s_high;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_inf;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_intern;
  PyObject *__pyx_n_s_is_core;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_is_extended;
  PyObject *__pyx_n_s_is_gag_linker;
  PyObject *__pyx_n_s_is_n_glycan;
  PyObject *__pyx_n_s_is_o_glycan;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_iter;
  PyObject *__pyx_n_s_j;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_low;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_mass;
  PyObject *__pyx_n_s_mass_shift;
  PyObject *__pyx_n_s_mass_shift_tandem_mass;
  PyObject *__pyx_n_s_match;
  PyObject *__pyx_n_s_match_fragments;
  PyObject *__pyx_n_s_member;
  PyObject *__pyx_n_s_members;
  PyObject *__pyx_n_s_minimum_peptide_mass;
  PyObject *__pyx_n_s_n_core;
  PyObject *__pyx_n_s_n_frags;
  PyObject *__pyx_n_s_n_glycan;
  PyObject *__pyx_n_s_n_glycan_match_stubs;
  PyObject *__pyx_n_s_n_matched;
  PyObject *__pyx_n_s_n_theoretical;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_neutral_mass;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_normalized_size;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_kp_s_numpy__core_multiarray_failed_to;
  PyObject *__pyx_kp_s_numpy__core_umath_failed_to_impo;
  PyObject *__pyx_n_s_o_glycan;
  PyObject *__pyx_n_s_o_glycan_match_stubs;
  PyObject *__pyx_n_s_partial_score_getter;
  PyObject *__pyx_n_s_peaks_matched;
  PyObject *__pyx_n_s_peptide_mass;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_precursor_information;
  PyObject *__pyx_n_s_product_error_tolerance;
  PyObject *__pyx_n_s_ptemp;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_CoarseGlycanMatch;
  PyObject *__pyx_n_s_pyx_unpickle_ComplementFragmen;
  PyObject *__pyx_n_s_pyx_unpickle_GlycanCoarseScore;
  PyObject *__pyx_n_s_pyx_unpickle_GlycanCombination;
  PyObject *__pyx_n_s_pyx_unpickle_GlycanFragmentInd;
  PyObject *__pyx_n_s_pyx_unpickle_GlycanMatchResult;
  PyObject *__pyx_n_s_pyx_unpickle_PartialGlycanSolu;
  PyObject *__pyx_n_s_pyx_unpickle_ScoreCoarseGlycan;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_query_mass;
  PyObject *__pyx_n_s_query_width;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_recalibrate_peptide_mass;
  PyObject *__pyx_n_s_recalibrated_peptide_mass;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_resolution;
  PyObject *__pyx_n_s_reverse;
  PyObject *__pyx_kp_s_s_s_f_f_r;
  PyObject *__pyx_kp_s_s_s_f_r;
  PyObject *__pyx_n_s_scan;
  PyObject *__pyx_n_s_score;
  PyObject *__pyx_n_s_score_getter;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_kp_s_self___class_____name___self_ma;
  PyObject *__pyx_kp_s_self___class_____name___self_n;
  PyObject *__pyx_kp_s_self___class_____name___self_pe;
  PyObject *__pyx_kp_s_self___class_____name___self_pe_2;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shift_mass;
  PyObject *__pyx_n_s_shifts;
  PyObject *__pyx_n_s_sort;
  PyObject *__pyx_n_s_sorted;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_kp_s_src_glycresoft__c_tandem_core_se;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_warn;
  PyObject *__pyx_n_s_warnings;
  PyObject *__pyx_n_s_x;
  PyObject *__pyx_float_0_0;
  PyObject *__pyx_float_0_42;
  PyObject *__pyx_float_0_56;
  PyObject *__pyx_float_1eneg_5;
  PyObject *__pyx_float_100_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_14758241;
  PyObject *__pyx_int_16698535;
  PyObject *__pyx_int_27011869;
  PyObject *__pyx_int_30049220;
  PyObject *__pyx_int_69199841;
  PyObject *__pyx_int_88044796;
  PyObject *__pyx_int_90496148;
  PyObject *__pyx_int_92231173;
  PyObject *__pyx_int_115380717;
  PyObject *__pyx_int_117397179;
  PyObject *__pyx_int_124808126;
  PyObject *__pyx_int_134214914;
  PyObject *__pyx_int_143121537;
  PyObject *__pyx_int_151187253;
  PyObject *__pyx_int_154200376;
  PyObject *__pyx_int_162950457;
  PyObject *__pyx_int_163158466;
  PyObject *__pyx_int_191711013;
  PyObject *__pyx_int_203386150;
  PyObject *__pyx_int_209767785;
  PyObject *__pyx_int_217673008;
  PyObject *__pyx_int_224310728;
  PyObject *__pyx_int_229638544;
  PyObject *__pyx_int_252421743;
  PyObject *__pyx_int_neg_1;
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_k__4;
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_k__5;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__3;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__52;
  PyObject *__pyx_tuple__54;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__67;
  PyObject *__pyx_tuple__69;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__75;
  PyObject *__pyx_tuple__80;
  PyObject *__pyx_codeobj__17;
  PyObject *__pyx_codeobj__18;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__20;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__22;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__24;
  PyObject *__pyx_codeobj__26;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__53;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__60;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__83;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__88;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_7brainpy_2_c_11composition_PyComposition);
  Py_CLEAR(clear_module_state->__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak);
  Py_CLEAR(clear_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak);
  Py_CLEAR(clear_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet);
  Py_CLEAR(clear_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed);
  Py_CLEAR(clear_module_state->__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed);
  Py_CLEAR(clear_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet);
  Py_CLEAR(clear_module_state->__pyx_ptype_5glypy_11composition_12ccomposition_CComposition);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift);
  Py_CLEAR(clear_module_state->__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_n_s_AttributeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseGlycanMatch);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseGlycanMatch___iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseGlycanMatch___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseGlycanMatch___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseGlycanMatch_estimate_pepti);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseStubGlycopeptideFragment);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseStubGlycopeptideFragment_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseStubGlycopeptideMatch);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoarseStubGlycopeptideMatch___re);
  Py_CLEAR(clear_module_state->__pyx_n_s_ComplementFragment);
  Py_CLEAR(clear_module_state->__pyx_n_s_ComplementFragment___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_ComplementFragment___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase___setstat);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase__calculat);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase__gag_matc);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase__match_fr);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase__n_glycan);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCoarseScorerBase__o_glycan);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase___re);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase___se);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_clea);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_get_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_get_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_is_g);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_is_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_is_o);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCombinationRecordBase_norm);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFilteringPeptideMassEstima);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex__get_fragmen);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex__intern);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex_bin_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex_build);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex_match);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanFragmentIndex_query_width);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanMatchResult);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanMatchResult___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanMatchResult___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanTypes);
  Py_CLEAR(clear_module_state->__pyx_n_s_HashableGlycanComposition);
  Py_CLEAR(clear_module_state->__pyx_n_s_HashableGlycanComposition_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Hex);
  Py_CLEAR(clear_module_state->__pyx_n_s_HexNAc);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexGlycanCompositionFragment);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexGlycanCompositionFragment_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexGlycanCompositionFragment_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexGlycanCompositionFragment_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexGlycanCompositionFragment_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PartialGlycanSolution);
  Py_CLEAR(clear_module_state->__pyx_n_s_PartialGlycanSolution___reduce_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_PartialGlycanSolution___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Re_estimated_peptide_mass_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_ScoreCoarseGlycanMatchPair);
  Py_CLEAR(clear_module_state->__pyx_n_s_ScoreCoarseGlycanMatchPair___red);
  Py_CLEAR(clear_module_state->__pyx_n_s_ScoreCoarseGlycanMatchPair___set);
  Py_CLEAR(clear_module_state->__pyx_n_s_Unmodified);
  Py_CLEAR(clear_module_state->__pyx_n_s_Unmodified_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s__15);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_n_s__89);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_bin_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_build);
  Py_CLEAR(clear_module_state->__pyx_n_s_build_locals_lambda);
  Py_CLEAR(clear_module_state->__pyx_n_s_calculate_score);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_comp);
  Py_CLEAR(clear_module_state->__pyx_n_s_comp_frags);
  Py_CLEAR(clear_module_state->__pyx_n_s_comp_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_comp_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_composition);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_core_matched);
  Py_CLEAR(clear_module_state->__pyx_n_s_core_matches);
  Py_CLEAR(clear_module_state->__pyx_n_s_core_theoretical);
  Py_CLEAR(clear_module_state->__pyx_n_s_core_weight);
  Py_CLEAR(clear_module_state->__pyx_n_s_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_deconvoluted_peak_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_defaultdict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dehydrated_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_error_tolerance);
  Py_CLEAR(clear_module_state->__pyx_n_s_estimate_peptide_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_extended);
  Py_CLEAR(clear_module_state->__pyx_kp_s_f_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_frag);
  Py_CLEAR(clear_module_state->__pyx_n_s_frag_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_fragment);
  Py_CLEAR(clear_module_state->__pyx_n_s_fragment_matches);
  Py_CLEAR(clear_module_state->__pyx_n_s_fragment_weight);
  Py_CLEAR(clear_module_state->__pyx_n_s_frags);
  Py_CLEAR(clear_module_state->__pyx_n_s_frags_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_frags_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_gag_linker);
  Py_CLEAR(clear_module_state->__pyx_n_s_gag_match_stubs);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_gcr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_fragments);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_gag_linker_glycan_fragments);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_n_glycan_fragments);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_o_glycan_fragments);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_combination);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_combination_db);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_match);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycosylation_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycresoft__c_tandem_core_search);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycresoft_chromatogram_tree);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycresoft_serialize);
  Py_CLEAR(clear_module_state->__pyx_n_s_glypy_structure_glycan_compositi);
  Py_CLEAR(clear_module_state->__pyx_n_s_hash_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_high);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_inf);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_intern);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_core);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_extended);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_gag_linker);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_n_glycan);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_o_glycan);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_j);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_low);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_mass_shift);
  Py_CLEAR(clear_module_state->__pyx_n_s_mass_shift_tandem_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_match);
  Py_CLEAR(clear_module_state->__pyx_n_s_match_fragments);
  Py_CLEAR(clear_module_state->__pyx_n_s_member);
  Py_CLEAR(clear_module_state->__pyx_n_s_members);
  Py_CLEAR(clear_module_state->__pyx_n_s_minimum_peptide_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_core);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_frags);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_glycan);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_glycan_match_stubs);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_matched);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_theoretical);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_neutral_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_normalized_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_kp_s_numpy__core_multiarray_failed_to);
  Py_CLEAR(clear_module_state->__pyx_kp_s_numpy__core_umath_failed_to_impo);
  Py_CLEAR(clear_module_state->__pyx_n_s_o_glycan);
  Py_CLEAR(clear_module_state->__pyx_n_s_o_glycan_match_stubs);
  Py_CLEAR(clear_module_state->__pyx_n_s_partial_score_getter);
  Py_CLEAR(clear_module_state->__pyx_n_s_peaks_matched);
  Py_CLEAR(clear_module_state->__pyx_n_s_peptide_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_precursor_information);
  Py_CLEAR(clear_module_state->__pyx_n_s_product_error_tolerance);
  Py_CLEAR(clear_module_state->__pyx_n_s_ptemp);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CoarseGlycanMatch);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ComplementFragmen);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_GlycanCoarseScore);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_GlycanCombination);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_GlycanFragmentInd);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_GlycanMatchResult);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PartialGlycanSolu);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ScoreCoarseGlycan);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_query_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_query_width);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_recalibrate_peptide_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_recalibrated_peptide_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_resolution);
  Py_CLEAR(clear_module_state->__pyx_n_s_reverse);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_s_f_f_r);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_s_f_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_scan);
  Py_CLEAR(clear_module_state->__pyx_n_s_score);
  Py_CLEAR(clear_module_state->__pyx_n_s_score_getter);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self___class_____name___self_ma);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self___class_____name___self_n);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self___class_____name___self_pe);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self___class_____name___self_pe_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shift_mass);
  Py_CLEAR(clear_module_state->__pyx_n_s_shifts);
  Py_CLEAR(clear_module_state->__pyx_n_s_sort);
  Py_CLEAR(clear_module_state->__pyx_n_s_sorted);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_glycresoft__c_tandem_core_se);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn);
  Py_CLEAR(clear_module_state->__pyx_n_s_warnings);
  Py_CLEAR(clear_module_state->__pyx_n_s_x);
  Py_CLEAR(clear_module_state->__pyx_float_0_0);
  Py_CLEAR(clear_module_state->__pyx_float_0_42);
  Py_CLEAR(clear_module_state->__pyx_float_0_56);
  Py_CLEAR(clear_module_state->__pyx_float_1eneg_5);
  Py_CLEAR(clear_module_state->__pyx_float_100_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_14758241);
  Py_CLEAR(clear_module_state->__pyx_int_16698535);
  Py_CLEAR(clear_module_state->__pyx_int_27011869);
  Py_CLEAR(clear_module_state->__pyx_int_30049220);
  Py_CLEAR(clear_module_state->__pyx_int_69199841);
  Py_CLEAR(clear_module_state->__pyx_int_88044796);
  Py_CLEAR(clear_module_state->__pyx_int_90496148);
  Py_CLEAR(clear_module_state->__pyx_int_92231173);
  Py_CLEAR(clear_module_state->__pyx_int_115380717);
  Py_CLEAR(clear_module_state->__pyx_int_117397179);
  Py_CLEAR(clear_module_state->__pyx_int_124808126);
  Py_CLEAR(clear_module_state->__pyx_int_134214914);
  Py_CLEAR(clear_module_state->__pyx_int_143121537);
  Py_CLEAR(clear_module_state->__pyx_int_151187253);
  Py_CLEAR(clear_module_state->__pyx_int_154200376);
  Py_CLEAR(clear_module_state->__pyx_int_162950457);
  Py_CLEAR(clear_module_state->__pyx_int_163158466);
  Py_CLEAR(clear_module_state->__pyx_int_191711013);
  Py_CLEAR(clear_module_state->__pyx_int_203386150);
  Py_CLEAR(clear_module_state->__pyx_int_209767785);
  Py_CLEAR(clear_module_state->__pyx_int_217673008);
  Py_CLEAR(clear_module_state->__pyx_int_224310728);
  Py_CLEAR(clear_module_state->__pyx_int_229638544);
  Py_CLEAR(clear_module_state->__pyx_int_252421743);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_k__4);
  Py_CLEAR(clear_module_state->__pyx_k__5);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__3);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__52);
  Py_CLEAR(clear_module_state->__pyx_tuple__54);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__67);
  Py_CLEAR(clear_module_state->__pyx_tuple__69);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__75);
  Py_CLEAR(clear_module_state->__pyx_tuple__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__17);
  Py_CLEAR(clear_module_state->__pyx_codeobj__18);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__20);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__22);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__24);
  Py_CLEAR(clear_module_state->__pyx_codeobj__26);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__53);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__60);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_7brainpy_2_c_11composition_PyComposition);
  Py_VISIT(traverse_module_state->__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak);
  Py_VISIT(traverse_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak);
  Py_VISIT(traverse_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet);
  Py_VISIT(traverse_module_state->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed);
  Py_VISIT(traverse_module_state->__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed);
  Py_VISIT(traverse_module_state->__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet);
  Py_VISIT(traverse_module_state->__pyx_ptype_5glypy_11composition_12ccomposition_CComposition);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift);
  Py_VISIT(traverse_module_state->__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_n_s_AttributeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseGlycanMatch);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseGlycanMatch___iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseGlycanMatch___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseGlycanMatch___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseGlycanMatch_estimate_pepti);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseStubGlycopeptideFragment);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseStubGlycopeptideFragment_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseStubGlycopeptideMatch);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoarseStubGlycopeptideMatch___re);
  Py_VISIT(traverse_module_state->__pyx_n_s_ComplementFragment);
  Py_VISIT(traverse_module_state->__pyx_n_s_ComplementFragment___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_ComplementFragment___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase___setstat);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase__calculat);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase__gag_matc);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase__match_fr);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase__n_glycan);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCoarseScorerBase__o_glycan);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase___re);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase___se);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_clea);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_get_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_get_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_is_g);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_is_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_is_o);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCombinationRecordBase_norm);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFilteringPeptideMassEstima);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex__get_fragmen);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex__intern);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex_bin_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex_build);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex_match);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanFragmentIndex_query_width);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanMatchResult);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanMatchResult___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanMatchResult___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanTypes);
  Py_VISIT(traverse_module_state->__pyx_n_s_HashableGlycanComposition);
  Py_VISIT(traverse_module_state->__pyx_n_s_HashableGlycanComposition_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Hex);
  Py_VISIT(traverse_module_state->__pyx_n_s_HexNAc);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexGlycanCompositionFragment);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexGlycanCompositionFragment_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexGlycanCompositionFragment_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexGlycanCompositionFragment_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexGlycanCompositionFragment_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PartialGlycanSolution);
  Py_VISIT(traverse_module_state->__pyx_n_s_PartialGlycanSolution___reduce_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_PartialGlycanSolution___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Re_estimated_peptide_mass_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_ScoreCoarseGlycanMatchPair);
  Py_VISIT(traverse_module_state->__pyx_n_s_ScoreCoarseGlycanMatchPair___red);
  Py_VISIT(traverse_module_state->__pyx_n_s_ScoreCoarseGlycanMatchPair___set);
  Py_VISIT(traverse_module_state->__pyx_n_s_Unmodified);
  Py_VISIT(traverse_module_state->__pyx_n_s_Unmodified_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s__15);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_n_s__89);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_bin_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_build);
  Py_VISIT(traverse_module_state->__pyx_n_s_build_locals_lambda);
  Py_VISIT(traverse_module_state->__pyx_n_s_calculate_score);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_comp);
  Py_VISIT(traverse_module_state->__pyx_n_s_comp_frags);
  Py_VISIT(traverse_module_state->__pyx_n_s_comp_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_comp_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_composition);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_core_matched);
  Py_VISIT(traverse_module_state->__pyx_n_s_core_matches);
  Py_VISIT(traverse_module_state->__pyx_n_s_core_theoretical);
  Py_VISIT(traverse_module_state->__pyx_n_s_core_weight);
  Py_VISIT(traverse_module_state->__pyx_n_s_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_deconvoluted_peak_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_defaultdict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dehydrated_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_error_tolerance);
  Py_VISIT(traverse_module_state->__pyx_n_s_estimate_peptide_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_extended);
  Py_VISIT(traverse_module_state->__pyx_kp_s_f_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_frag);
  Py_VISIT(traverse_module_state->__pyx_n_s_frag_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_fragment);
  Py_VISIT(traverse_module_state->__pyx_n_s_fragment_matches);
  Py_VISIT(traverse_module_state->__pyx_n_s_fragment_weight);
  Py_VISIT(traverse_module_state->__pyx_n_s_frags);
  Py_VISIT(traverse_module_state->__pyx_n_s_frags_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_frags_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_gag_linker);
  Py_VISIT(traverse_module_state->__pyx_n_s_gag_match_stubs);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_gcr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_fragments);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_gag_linker_glycan_fragments);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_n_glycan_fragments);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_o_glycan_fragments);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_combination);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_combination_db);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_match);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycosylation_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycresoft__c_tandem_core_search);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycresoft_chromatogram_tree);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycresoft_serialize);
  Py_VISIT(traverse_module_state->__pyx_n_s_glypy_structure_glycan_compositi);
  Py_VISIT(traverse_module_state->__pyx_n_s_hash_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_high);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_inf);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_intern);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_core);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_extended);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_gag_linker);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_n_glycan);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_o_glycan);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_j);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_low);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_mass_shift);
  Py_VISIT(traverse_module_state->__pyx_n_s_mass_shift_tandem_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_match);
  Py_VISIT(traverse_module_state->__pyx_n_s_match_fragments);
  Py_VISIT(traverse_module_state->__pyx_n_s_member);
  Py_VISIT(traverse_module_state->__pyx_n_s_members);
  Py_VISIT(traverse_module_state->__pyx_n_s_minimum_peptide_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_core);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_frags);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_glycan);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_glycan_match_stubs);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_matched);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_theoretical);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_neutral_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_normalized_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_kp_s_numpy__core_multiarray_failed_to);
  Py_VISIT(traverse_module_state->__pyx_kp_s_numpy__core_umath_failed_to_impo);
  Py_VISIT(traverse_module_state->__pyx_n_s_o_glycan);
  Py_VISIT(traverse_module_state->__pyx_n_s_o_glycan_match_stubs);
  Py_VISIT(traverse_module_state->__pyx_n_s_partial_score_getter);
  Py_VISIT(traverse_module_state->__pyx_n_s_peaks_matched);
  Py_VISIT(traverse_module_state->__pyx_n_s_peptide_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_precursor_information);
  Py_VISIT(traverse_module_state->__pyx_n_s_product_error_tolerance);
  Py_VISIT(traverse_module_state->__pyx_n_s_ptemp);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CoarseGlycanMatch);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ComplementFragmen);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_GlycanCoarseScore);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_GlycanCombination);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_GlycanFragmentInd);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_GlycanMatchResult);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PartialGlycanSolu);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ScoreCoarseGlycan);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_query_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_query_width);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_recalibrate_peptide_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_recalibrated_peptide_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_resolution);
  Py_VISIT(traverse_module_state->__pyx_n_s_reverse);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_s_f_f_r);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_s_f_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_scan);
  Py_VISIT(traverse_module_state->__pyx_n_s_score);
  Py_VISIT(traverse_module_state->__pyx_n_s_score_getter);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self___class_____name___self_ma);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self___class_____name___self_n);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self___class_____name___self_pe);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self___class_____name___self_pe_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shift_mass);
  Py_VISIT(traverse_module_state->__pyx_n_s_shifts);
  Py_VISIT(traverse_module_state->__pyx_n_s_sort);
  Py_VISIT(traverse_module_state->__pyx_n_s_sorted);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_glycresoft__c_tandem_core_se);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn);
  Py_VISIT(traverse_module_state->__pyx_n_s_warnings);
  Py_VISIT(traverse_module_state->__pyx_n_s_x);
  Py_VISIT(traverse_module_state->__pyx_float_0_0);
  Py_VISIT(traverse_module_state->__pyx_float_0_42);
  Py_VISIT(traverse_module_state->__pyx_float_0_56);
  Py_VISIT(traverse_module_state->__pyx_float_1eneg_5);
  Py_VISIT(traverse_module_state->__pyx_float_100_0);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_14758241);
  Py_VISIT(traverse_module_state->__pyx_int_16698535);
  Py_VISIT(traverse_module_state->__pyx_int_27011869);
  Py_VISIT(traverse_module_state->__pyx_int_30049220);
  Py_VISIT(traverse_module_state->__pyx_int_69199841);
  Py_VISIT(traverse_module_state->__pyx_int_88044796);
  Py_VISIT(traverse_module_state->__pyx_int_90496148);
  Py_VISIT(traverse_module_state->__pyx_int_92231173);
  Py_VISIT(traverse_module_state->__pyx_int_115380717);
  Py_VISIT(traverse_module_state->__pyx_int_117397179);
  Py_VISIT(traverse_module_state->__pyx_int_124808126);
  Py_VISIT(traverse_module_state->__pyx_int_134214914);
  Py_VISIT(traverse_module_state->__pyx_int_143121537);
  Py_VISIT(traverse_module_state->__pyx_int_151187253);
  Py_VISIT(traverse_module_state->__pyx_int_154200376);
  Py_VISIT(traverse_module_state->__pyx_int_162950457);
  Py_VISIT(traverse_module_state->__pyx_int_163158466);
  Py_VISIT(traverse_module_state->__pyx_int_191711013);
  Py_VISIT(traverse_module_state->__pyx_int_203386150);
  Py_VISIT(traverse_module_state->__pyx_int_209767785);
  Py_VISIT(traverse_module_state->__pyx_int_217673008);
  Py_VISIT(traverse_module_state->__pyx_int_224310728);
  Py_VISIT(traverse_module_state->__pyx_int_229638544);
  Py_VISIT(traverse_module_state->__pyx_int_252421743);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_k__4);
  Py_VISIT(traverse_module_state->__pyx_k__5);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__3);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__52);
  Py_VISIT(traverse_module_state->__pyx_tuple__54);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__67);
  Py_VISIT(traverse_module_state->__pyx_tuple__69);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__75);
  Py_VISIT(traverse_module_state->__pyx_tuple__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__17);
  Py_VISIT(traverse_module_state->__pyx_codeobj__18);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__20);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__22);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__24);
  Py_VISIT(traverse_module_state->__pyx_codeobj__26);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__53);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__60);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5numpy_dtype __pyx_mstate_global->__pyx_ptype_5numpy_dtype
#define __pyx_ptype_5numpy_flatiter __pyx_mstate_global->__pyx_ptype_5numpy_flatiter
#define __pyx_ptype_5numpy_broadcast __pyx_mstate_global->__pyx_ptype_5numpy_broadcast
#define __pyx_ptype_5numpy_ndarray __pyx_mstate_global->__pyx_ptype_5numpy_ndarray
#define __pyx_ptype_5numpy_generic __pyx_mstate_global->__pyx_ptype_5numpy_generic
#define __pyx_ptype_5numpy_number __pyx_mstate_global->__pyx_ptype_5numpy_number
#define __pyx_ptype_5numpy_integer __pyx_mstate_global->__pyx_ptype_5numpy_integer
#define __pyx_ptype_5numpy_signedinteger __pyx_mstate_global->__pyx_ptype_5numpy_signedinteger
#define __pyx_ptype_5numpy_unsignedinteger __pyx_mstate_global->__pyx_ptype_5numpy_unsignedinteger
#define __pyx_ptype_5numpy_inexact __pyx_mstate_global->__pyx_ptype_5numpy_inexact
#define __pyx_ptype_5numpy_floating __pyx_mstate_global->__pyx_ptype_5numpy_floating
#define __pyx_ptype_5numpy_complexfloating __pyx_mstate_global->__pyx_ptype_5numpy_complexfloating
#define __pyx_ptype_5numpy_flexible __pyx_mstate_global->__pyx_ptype_5numpy_flexible
#define __pyx_ptype_5numpy_character __pyx_mstate_global->__pyx_ptype_5numpy_character
#define __pyx_ptype_5numpy_ufunc __pyx_mstate_global->__pyx_ptype_5numpy_ufunc
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7brainpy_2_c_11composition_PyComposition __pyx_mstate_global->__pyx_ptype_7brainpy_2_c_11composition_PyComposition
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak __pyx_mstate_global->__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_mstate_global->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase
#define __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak __pyx_mstate_global->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak
#define __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_mstate_global->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet
#define __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed __pyx_mstate_global->__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex __pyx_mstate_global->__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine
#define __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern
#define __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase
#define __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase
#define __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase
#define __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter
#define __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed
#define __pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet __pyx_mstate_global->__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5glypy_11composition_12ccomposition_CComposition __pyx_mstate_global->__pyx_ptype_5glypy_11composition_12ccomposition_CComposition
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase
#define __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift
#define __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase __pyx_mstate_global->__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator
#define __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase
#define __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy __pyx_mstate_global->__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex
#define __pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ __pyx_mstate_global->__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__
#endif
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex
#define __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__
#define __pyx_n_s_AttributeError __pyx_mstate_global->__pyx_n_s_AttributeError
#define __pyx_n_s_CoarseGlycanMatch __pyx_mstate_global->__pyx_n_s_CoarseGlycanMatch
#define __pyx_n_s_CoarseGlycanMatch___iter __pyx_mstate_global->__pyx_n_s_CoarseGlycanMatch___iter
#define __pyx_n_s_CoarseGlycanMatch___reduce_cytho __pyx_mstate_global->__pyx_n_s_CoarseGlycanMatch___reduce_cytho
#define __pyx_n_s_CoarseGlycanMatch___setstate_cyt __pyx_mstate_global->__pyx_n_s_CoarseGlycanMatch___setstate_cyt
#define __pyx_n_s_CoarseGlycanMatch_estimate_pepti __pyx_mstate_global->__pyx_n_s_CoarseGlycanMatch_estimate_pepti
#define __pyx_n_s_CoarseStubGlycopeptideFragment __pyx_mstate_global->__pyx_n_s_CoarseStubGlycopeptideFragment
#define __pyx_n_s_CoarseStubGlycopeptideFragment_2 __pyx_mstate_global->__pyx_n_s_CoarseStubGlycopeptideFragment_2
#define __pyx_n_s_CoarseStubGlycopeptideMatch __pyx_mstate_global->__pyx_n_s_CoarseStubGlycopeptideMatch
#define __pyx_n_s_CoarseStubGlycopeptideMatch___re __pyx_mstate_global->__pyx_n_s_CoarseStubGlycopeptideMatch___re
#define __pyx_n_s_ComplementFragment __pyx_mstate_global->__pyx_n_s_ComplementFragment
#define __pyx_n_s_ComplementFragment___reduce_cyth __pyx_mstate_global->__pyx_n_s_ComplementFragment___reduce_cyth
#define __pyx_n_s_ComplementFragment___setstate_cy __pyx_mstate_global->__pyx_n_s_ComplementFragment___setstate_cy
#define __pyx_n_s_GlycanCoarseScorerBase __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase
#define __pyx_n_s_GlycanCoarseScorerBase___reduce __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase___reduce
#define __pyx_n_s_GlycanCoarseScorerBase___setstat __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase___setstat
#define __pyx_n_s_GlycanCoarseScorerBase__calculat __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase__calculat
#define __pyx_n_s_GlycanCoarseScorerBase__gag_matc __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase__gag_matc
#define __pyx_n_s_GlycanCoarseScorerBase__match_fr __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase__match_fr
#define __pyx_n_s_GlycanCoarseScorerBase__n_glycan __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase__n_glycan
#define __pyx_n_s_GlycanCoarseScorerBase__o_glycan __pyx_mstate_global->__pyx_n_s_GlycanCoarseScorerBase__o_glycan
#define __pyx_n_s_GlycanCombinationRecordBase __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase
#define __pyx_n_s_GlycanCombinationRecordBase___re __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase___re
#define __pyx_n_s_GlycanCombinationRecordBase___se __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase___se
#define __pyx_n_s_GlycanCombinationRecordBase_clea __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_clea
#define __pyx_n_s_GlycanCombinationRecordBase_get __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_get
#define __pyx_n_s_GlycanCombinationRecordBase_get_2 __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_get_2
#define __pyx_n_s_GlycanCombinationRecordBase_get_3 __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_get_3
#define __pyx_n_s_GlycanCombinationRecordBase_is_g __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_is_g
#define __pyx_n_s_GlycanCombinationRecordBase_is_n __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_is_n
#define __pyx_n_s_GlycanCombinationRecordBase_is_o __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_is_o
#define __pyx_n_s_GlycanCombinationRecordBase_norm __pyx_mstate_global->__pyx_n_s_GlycanCombinationRecordBase_norm
#define __pyx_n_s_GlycanFilteringPeptideMassEstima __pyx_mstate_global->__pyx_n_s_GlycanFilteringPeptideMassEstima
#define __pyx_n_s_GlycanFragmentIndex __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex
#define __pyx_n_s_GlycanFragmentIndex___reduce_cyt __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex___reduce_cyt
#define __pyx_n_s_GlycanFragmentIndex___setstate_c __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex___setstate_c
#define __pyx_n_s_GlycanFragmentIndex__get_fragmen __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex__get_fragmen
#define __pyx_n_s_GlycanFragmentIndex__intern __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex__intern
#define __pyx_n_s_GlycanFragmentIndex_bin_index __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex_bin_index
#define __pyx_n_s_GlycanFragmentIndex_build __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex_build
#define __pyx_n_s_GlycanFragmentIndex_match __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex_match
#define __pyx_n_s_GlycanFragmentIndex_query_width __pyx_mstate_global->__pyx_n_s_GlycanFragmentIndex_query_width
#define __pyx_n_s_GlycanMatchResult __pyx_mstate_global->__pyx_n_s_GlycanMatchResult
#define __pyx_n_s_GlycanMatchResult___reduce_cytho __pyx_mstate_global->__pyx_n_s_GlycanMatchResult___reduce_cytho
#define __pyx_n_s_GlycanMatchResult___setstate_cyt __pyx_mstate_global->__pyx_n_s_GlycanMatchResult___setstate_cyt
#define __pyx_n_s_GlycanTypes __pyx_mstate_global->__pyx_n_s_GlycanTypes
#define __pyx_n_s_HashableGlycanComposition __pyx_mstate_global->__pyx_n_s_HashableGlycanComposition
#define __pyx_n_s_HashableGlycanComposition_2 __pyx_mstate_global->__pyx_n_s_HashableGlycanComposition_2
#define __pyx_n_s_Hex __pyx_mstate_global->__pyx_n_s_Hex
#define __pyx_n_s_HexNAc __pyx_mstate_global->__pyx_n_s_HexNAc
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_n_s_IndexGlycanCompositionFragment __pyx_mstate_global->__pyx_n_s_IndexGlycanCompositionFragment
#define __pyx_n_s_IndexGlycanCompositionFragment_2 __pyx_mstate_global->__pyx_n_s_IndexGlycanCompositionFragment_2
#define __pyx_n_s_IndexGlycanCompositionFragment_3 __pyx_mstate_global->__pyx_n_s_IndexGlycanCompositionFragment_3
#define __pyx_n_s_IndexGlycanCompositionFragment_4 __pyx_mstate_global->__pyx_n_s_IndexGlycanCompositionFragment_4
#define __pyx_n_s_IndexGlycanCompositionFragment_c __pyx_mstate_global->__pyx_n_s_IndexGlycanCompositionFragment_c
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_PartialGlycanSolution __pyx_mstate_global->__pyx_n_s_PartialGlycanSolution
#define __pyx_n_s_PartialGlycanSolution___reduce_c __pyx_mstate_global->__pyx_n_s_PartialGlycanSolution___reduce_c
#define __pyx_n_s_PartialGlycanSolution___setstate __pyx_mstate_global->__pyx_n_s_PartialGlycanSolution___setstate
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_kp_s_Re_estimated_peptide_mass_error __pyx_mstate_global->__pyx_kp_s_Re_estimated_peptide_mass_error
#define __pyx_n_s_ScoreCoarseGlycanMatchPair __pyx_mstate_global->__pyx_n_s_ScoreCoarseGlycanMatchPair
#define __pyx_n_s_ScoreCoarseGlycanMatchPair___red __pyx_mstate_global->__pyx_n_s_ScoreCoarseGlycanMatchPair___red
#define __pyx_n_s_ScoreCoarseGlycanMatchPair___set __pyx_mstate_global->__pyx_n_s_ScoreCoarseGlycanMatchPair___set
#define __pyx_n_s_Unmodified __pyx_mstate_global->__pyx_n_s_Unmodified
#define __pyx_n_s_Unmodified_2 __pyx_mstate_global->__pyx_n_s_Unmodified_2
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s__15 __pyx_mstate_global->__pyx_n_s__15
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_n_s__89 __pyx_mstate_global->__pyx_n_s__89
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_bin_index __pyx_mstate_global->__pyx_n_s_bin_index
#define __pyx_n_s_build __pyx_mstate_global->__pyx_n_s_build
#define __pyx_n_s_build_locals_lambda __pyx_mstate_global->__pyx_n_s_build_locals_lambda
#define __pyx_n_s_calculate_score __pyx_mstate_global->__pyx_n_s_calculate_score
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_comp __pyx_mstate_global->__pyx_n_s_comp
#define __pyx_n_s_comp_frags __pyx_mstate_global->__pyx_n_s_comp_frags
#define __pyx_n_s_comp_i __pyx_mstate_global->__pyx_n_s_comp_i
#define __pyx_n_s_comp_n __pyx_mstate_global->__pyx_n_s_comp_n
#define __pyx_n_s_composition __pyx_mstate_global->__pyx_n_s_composition
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_core_matched __pyx_mstate_global->__pyx_n_s_core_matched
#define __pyx_n_s_core_matches __pyx_mstate_global->__pyx_n_s_core_matches
#define __pyx_n_s_core_theoretical __pyx_mstate_global->__pyx_n_s_core_theoretical
#define __pyx_n_s_core_weight __pyx_mstate_global->__pyx_n_s_core_weight
#define __pyx_n_s_d __pyx_mstate_global->__pyx_n_s_d
#define __pyx_n_s_deconvoluted_peak_set __pyx_mstate_global->__pyx_n_s_deconvoluted_peak_set
#define __pyx_n_s_defaultdict __pyx_mstate_global->__pyx_n_s_defaultdict
#define __pyx_n_s_dehydrated_mass __pyx_mstate_global->__pyx_n_s_dehydrated_mass
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_error_tolerance __pyx_mstate_global->__pyx_n_s_error_tolerance
#define __pyx_n_s_estimate_peptide_mass __pyx_mstate_global->__pyx_n_s_estimate_peptide_mass
#define __pyx_n_s_extended __pyx_mstate_global->__pyx_n_s_extended
#define __pyx_kp_s_f_r __pyx_mstate_global->__pyx_kp_s_f_r
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_frag __pyx_mstate_global->__pyx_n_s_frag
#define __pyx_n_s_frag_i __pyx_mstate_global->__pyx_n_s_frag_i
#define __pyx_n_s_fragment __pyx_mstate_global->__pyx_n_s_fragment
#define __pyx_n_s_fragment_matches __pyx_mstate_global->__pyx_n_s_fragment_matches
#define __pyx_n_s_fragment_weight __pyx_mstate_global->__pyx_n_s_fragment_weight
#define __pyx_n_s_frags __pyx_mstate_global->__pyx_n_s_frags
#define __pyx_n_s_frags_i __pyx_mstate_global->__pyx_n_s_frags_i
#define __pyx_n_s_frags_n __pyx_mstate_global->__pyx_n_s_frags_n
#define __pyx_n_s_gag_linker __pyx_mstate_global->__pyx_n_s_gag_linker
#define __pyx_n_s_gag_match_stubs __pyx_mstate_global->__pyx_n_s_gag_match_stubs
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_gcr __pyx_mstate_global->__pyx_n_s_gcr
#define __pyx_n_s_get_fragments __pyx_mstate_global->__pyx_n_s_get_fragments
#define __pyx_n_s_get_gag_linker_glycan_fragments __pyx_mstate_global->__pyx_n_s_get_gag_linker_glycan_fragments
#define __pyx_n_s_get_n_glycan_fragments __pyx_mstate_global->__pyx_n_s_get_n_glycan_fragments
#define __pyx_n_s_get_o_glycan_fragments __pyx_mstate_global->__pyx_n_s_get_o_glycan_fragments
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_glycan_combination __pyx_mstate_global->__pyx_n_s_glycan_combination
#define __pyx_n_s_glycan_combination_db __pyx_mstate_global->__pyx_n_s_glycan_combination_db
#define __pyx_n_s_glycan_index __pyx_mstate_global->__pyx_n_s_glycan_index
#define __pyx_n_s_glycan_match __pyx_mstate_global->__pyx_n_s_glycan_match
#define __pyx_n_s_glycan_size __pyx_mstate_global->__pyx_n_s_glycan_size
#define __pyx_n_s_glycan_types __pyx_mstate_global->__pyx_n_s_glycan_types
#define __pyx_n_s_glycosylation_type __pyx_mstate_global->__pyx_n_s_glycosylation_type
#define __pyx_n_s_glycresoft__c_tandem_core_search __pyx_mstate_global->__pyx_n_s_glycresoft__c_tandem_core_search
#define __pyx_n_s_glycresoft_chromatogram_tree __pyx_mstate_global->__pyx_n_s_glycresoft_chromatogram_tree
#define __pyx_n_s_glycresoft_serialize __pyx_mstate_global->__pyx_n_s_glycresoft_serialize
#define __pyx_n_s_glypy_structure_glycan_compositi __pyx_mstate_global->__pyx_n_s_glypy_structure_glycan_compositi
#define __pyx_n_s_hash_key __pyx_mstate_global->__pyx_n_s_hash_key
#define __pyx_n_s_high __pyx_mstate_global->__pyx_n_s_high
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_inf __pyx_mstate_global->__pyx_n_s_inf
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_intern __pyx_mstate_global->__pyx_n_s_intern
#define __pyx_n_s_is_core __pyx_mstate_global->__pyx_n_s_is_core
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_is_extended __pyx_mstate_global->__pyx_n_s_is_extended
#define __pyx_n_s_is_gag_linker __pyx_mstate_global->__pyx_n_s_is_gag_linker
#define __pyx_n_s_is_n_glycan __pyx_mstate_global->__pyx_n_s_is_n_glycan
#define __pyx_n_s_is_o_glycan __pyx_mstate_global->__pyx_n_s_is_o_glycan
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_iter __pyx_mstate_global->__pyx_n_s_iter
#define __pyx_n_s_j __pyx_mstate_global->__pyx_n_s_j
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_low __pyx_mstate_global->__pyx_n_s_low
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_mass __pyx_mstate_global->__pyx_n_s_mass
#define __pyx_n_s_mass_shift __pyx_mstate_global->__pyx_n_s_mass_shift
#define __pyx_n_s_mass_shift_tandem_mass __pyx_mstate_global->__pyx_n_s_mass_shift_tandem_mass
#define __pyx_n_s_match __pyx_mstate_global->__pyx_n_s_match
#define __pyx_n_s_match_fragments __pyx_mstate_global->__pyx_n_s_match_fragments
#define __pyx_n_s_member __pyx_mstate_global->__pyx_n_s_member
#define __pyx_n_s_members __pyx_mstate_global->__pyx_n_s_members
#define __pyx_n_s_minimum_peptide_mass __pyx_mstate_global->__pyx_n_s_minimum_peptide_mass
#define __pyx_n_s_n_core __pyx_mstate_global->__pyx_n_s_n_core
#define __pyx_n_s_n_frags __pyx_mstate_global->__pyx_n_s_n_frags
#define __pyx_n_s_n_glycan __pyx_mstate_global->__pyx_n_s_n_glycan
#define __pyx_n_s_n_glycan_match_stubs __pyx_mstate_global->__pyx_n_s_n_glycan_match_stubs
#define __pyx_n_s_n_matched __pyx_mstate_global->__pyx_n_s_n_matched
#define __pyx_n_s_n_theoretical __pyx_mstate_global->__pyx_n_s_n_theoretical
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_neutral_mass __pyx_mstate_global->__pyx_n_s_neutral_mass
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_normalized_size __pyx_mstate_global->__pyx_n_s_normalized_size
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_kp_s_numpy__core_multiarray_failed_to __pyx_mstate_global->__pyx_kp_s_numpy__core_multiarray_failed_to
#define __pyx_kp_s_numpy__core_umath_failed_to_impo __pyx_mstate_global->__pyx_kp_s_numpy__core_umath_failed_to_impo
#define __pyx_n_s_o_glycan __pyx_mstate_global->__pyx_n_s_o_glycan
#define __pyx_n_s_o_glycan_match_stubs __pyx_mstate_global->__pyx_n_s_o_glycan_match_stubs
#define __pyx_n_s_partial_score_getter __pyx_mstate_global->__pyx_n_s_partial_score_getter
#define __pyx_n_s_peaks_matched __pyx_mstate_global->__pyx_n_s_peaks_matched
#define __pyx_n_s_peptide_mass __pyx_mstate_global->__pyx_n_s_peptide_mass
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_precursor_information __pyx_mstate_global->__pyx_n_s_precursor_information
#define __pyx_n_s_product_error_tolerance __pyx_mstate_global->__pyx_n_s_product_error_tolerance
#define __pyx_n_s_ptemp __pyx_mstate_global->__pyx_n_s_ptemp
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_CoarseGlycanMatch __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CoarseGlycanMatch
#define __pyx_n_s_pyx_unpickle_ComplementFragmen __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ComplementFragmen
#define __pyx_n_s_pyx_unpickle_GlycanCoarseScore __pyx_mstate_global->__pyx_n_s_pyx_unpickle_GlycanCoarseScore
#define __pyx_n_s_pyx_unpickle_GlycanCombination __pyx_mstate_global->__pyx_n_s_pyx_unpickle_GlycanCombination
#define __pyx_n_s_pyx_unpickle_GlycanFragmentInd __pyx_mstate_global->__pyx_n_s_pyx_unpickle_GlycanFragmentInd
#define __pyx_n_s_pyx_unpickle_GlycanMatchResult __pyx_mstate_global->__pyx_n_s_pyx_unpickle_GlycanMatchResult
#define __pyx_n_s_pyx_unpickle_PartialGlycanSolu __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PartialGlycanSolu
#define __pyx_n_s_pyx_unpickle_ScoreCoarseGlycan __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ScoreCoarseGlycan
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_query_mass __pyx_mstate_global->__pyx_n_s_query_mass
#define __pyx_n_s_query_width __pyx_mstate_global->__pyx_n_s_query_width
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_recalibrate_peptide_mass __pyx_mstate_global->__pyx_n_s_recalibrate_peptide_mass
#define __pyx_n_s_recalibrated_peptide_mass __pyx_mstate_global->__pyx_n_s_recalibrated_peptide_mass
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_resolution __pyx_mstate_global->__pyx_n_s_resolution
#define __pyx_n_s_reverse __pyx_mstate_global->__pyx_n_s_reverse
#define __pyx_kp_s_s_s_f_f_r __pyx_mstate_global->__pyx_kp_s_s_s_f_f_r
#define __pyx_kp_s_s_s_f_r __pyx_mstate_global->__pyx_kp_s_s_s_f_r
#define __pyx_n_s_scan __pyx_mstate_global->__pyx_n_s_scan
#define __pyx_n_s_score __pyx_mstate_global->__pyx_n_s_score
#define __pyx_n_s_score_getter __pyx_mstate_global->__pyx_n_s_score_getter
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_kp_s_self___class_____name___self_ma __pyx_mstate_global->__pyx_kp_s_self___class_____name___self_ma
#define __pyx_kp_s_self___class_____name___self_n __pyx_mstate_global->__pyx_kp_s_self___class_____name___self_n
#define __pyx_kp_s_self___class_____name___self_pe __pyx_mstate_global->__pyx_kp_s_self___class_____name___self_pe
#define __pyx_kp_s_self___class_____name___self_pe_2 __pyx_mstate_global->__pyx_kp_s_self___class_____name___self_pe_2
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shift_mass __pyx_mstate_global->__pyx_n_s_shift_mass
#define __pyx_n_s_shifts __pyx_mstate_global->__pyx_n_s_shifts
#define __pyx_n_s_sort __pyx_mstate_global->__pyx_n_s_sort
#define __pyx_n_s_sorted __pyx_mstate_global->__pyx_n_s_sorted
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_kp_s_src_glycresoft__c_tandem_core_se __pyx_mstate_global->__pyx_kp_s_src_glycresoft__c_tandem_core_se
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_warn __pyx_mstate_global->__pyx_n_s_warn
#define __pyx_n_s_warnings __pyx_mstate_global->__pyx_n_s_warnings
#define __pyx_n_s_x __pyx_mstate_global->__pyx_n_s_x
#define __pyx_float_0_0 __pyx_mstate_global->__pyx_float_0_0
#define __pyx_float_0_42 __pyx_mstate_global->__pyx_float_0_42
#define __pyx_float_0_56 __pyx_mstate_global->__pyx_float_0_56
#define __pyx_float_1eneg_5 __pyx_mstate_global->__pyx_float_1eneg_5
#define __pyx_float_100_0 __pyx_mstate_global->__pyx_float_100_0
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_14758241 __pyx_mstate_global->__pyx_int_14758241
#define __pyx_int_16698535 __pyx_mstate_global->__pyx_int_16698535
#define __pyx_int_27011869 __pyx_mstate_global->__pyx_int_27011869
#define __pyx_int_30049220 __pyx_mstate_global->__pyx_int_30049220
#define __pyx_int_69199841 __pyx_mstate_global->__pyx_int_69199841
#define __pyx_int_88044796 __pyx_mstate_global->__pyx_int_88044796
#define __pyx_int_90496148 __pyx_mstate_global->__pyx_int_90496148
#define __pyx_int_92231173 __pyx_mstate_global->__pyx_int_92231173
#define __pyx_int_115380717 __pyx_mstate_global->__pyx_int_115380717
#define __pyx_int_117397179 __pyx_mstate_global->__pyx_int_117397179
#define __pyx_int_124808126 __pyx_mstate_global->__pyx_int_124808126
#define __pyx_int_134214914 __pyx_mstate_global->__pyx_int_134214914
#define __pyx_int_143121537 __pyx_mstate_global->__pyx_int_143121537
#define __pyx_int_151187253 __pyx_mstate_global->__pyx_int_151187253
#define __pyx_int_154200376 __pyx_mstate_global->__pyx_int_154200376
#define __pyx_int_162950457 __pyx_mstate_global->__pyx_int_162950457
#define __pyx_int_163158466 __pyx_mstate_global->__pyx_int_163158466
#define __pyx_int_191711013 __pyx_mstate_global->__pyx_int_191711013
#define __pyx_int_203386150 __pyx_mstate_global->__pyx_int_203386150
#define __pyx_int_209767785 __pyx_mstate_global->__pyx_int_209767785
#define __pyx_int_217673008 __pyx_mstate_global->__pyx_int_217673008
#define __pyx_int_224310728 __pyx_mstate_global->__pyx_int_224310728
#define __pyx_int_229638544 __pyx_mstate_global->__pyx_int_229638544
#define __pyx_int_252421743 __pyx_mstate_global->__pyx_int_252421743
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_k__4 __pyx_mstate_global->__pyx_k__4
#define __pyx_k__5 __pyx_mstate_global->__pyx_k__5
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__3 __pyx_mstate_global->__pyx_tuple__3
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__52 __pyx_mstate_global->__pyx_tuple__52
#define __pyx_tuple__54 __pyx_mstate_global->__pyx_tuple__54
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__67 __pyx_mstate_global->__pyx_tuple__67
#define __pyx_tuple__69 __pyx_mstate_global->__pyx_tuple__69
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__75 __pyx_mstate_global->__pyx_tuple__75
#define __pyx_tuple__80 __pyx_mstate_global->__pyx_tuple__80
#define __pyx_codeobj__17 __pyx_mstate_global->__pyx_codeobj__17
#define __pyx_codeobj__18 __pyx_mstate_global->__pyx_codeobj__18
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__20 __pyx_mstate_global->__pyx_codeobj__20
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__22 __pyx_mstate_global->__pyx_codeobj__22
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__24 __pyx_mstate_global->__pyx_codeobj__24
#define __pyx_codeobj__26 __pyx_mstate_global->__pyx_codeobj__26
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__53 __pyx_mstate_global->__pyx_codeobj__53
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__60 __pyx_mstate_global->__pyx_codeobj__60
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__83 __pyx_mstate_global->__pyx_codeobj__83
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
/* #### Code section: module_code ### */

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":286
 * 
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ELSIZE(self)
 * 
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":287
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDataType_ELSIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":286
 * 
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ELSIZE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":290
 * 
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ALIGNMENT(self)
 * 
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":291
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use fields/names with care as they may be NULL.  You must check
 */
  __pyx_r = PyDataType_ALIGNMENT(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":290
 * 
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ALIGNMENT(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":296
 *         # for this using PyDataType_HASFIELDS.
 *         @property
 *         cdef inline object fields(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDataType_FIELDS(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("fields", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":297
 *         @property
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_FIELDS(__pyx_v_self);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":296
 *         # for this using PyDataType_HASFIELDS.
 *         @property
 *         cdef inline object fields(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDataType_FIELDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":300
 * 
 *         @property
 *         cdef inline tuple names(self):             # <<<<<<<<<<<<<<
 *             return <tuple>PyDataType_NAMES(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("names", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":301
 *         @property
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use PyDataType_HASSUBARRAY to test whether this field is
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_NAMES(__pyx_v_self);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":300
 * 
 *         @property
 *         cdef inline tuple names(self):             # <<<<<<<<<<<<<<
 *             return <tuple>PyDataType_NAMES(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":307
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_SUBARRAY(self)
 * 
 */

static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self) {
  PyArray_ArrayDescr *__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":308
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDataType_SUBARRAY(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":307
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_SUBARRAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":311
 * 
 *         @property
 *         cdef inline npy_uint64 flags(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)
 */

static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self) {
  npy_uint64 __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":313
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyDataType_FLAGS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":311
 * 
 *         @property
 *         cdef inline npy_uint64 flags(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":323
 * 
 *         @property
 *         cdef inline int numiter(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)
 */

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":325
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_NUMITER(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":323
 * 
 *         @property
 *         cdef inline int numiter(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":328
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":330
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":328
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":333
 * 
 *         @property
 *         cdef inline npy_intp index(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":335
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_INDEX(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":333
 * 
 *         @property
 *         cdef inline npy_intp index(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":338
 * 
 *         @property
 *         cdef inline int nd(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)
 */

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":340
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":338
 * 
 *         @property
 *         cdef inline int nd(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":343
 * 
 *         @property
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":345
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":343
 * 
 *         @property
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":348
 * 
 *         @property
 *         cdef inline void** iters(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 */

static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self) {
  void **__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":351
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 *             return PyArray_MultiIter_ITERS(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_MultiIter_ITERS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":348
 * 
 *         @property
 *         cdef inline void** iters(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":366
 * 
 *         @property
 *         cdef inline PyObject* base(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":369
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":366
 * 
 *         @property
 *         cdef inline PyObject* base(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":372
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":375
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":372
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":378
 * 
 *         @property
 *         cdef inline int ndim(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":381
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":378
 * 
 *         @property
 *         cdef inline int ndim(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":384
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":389
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":384
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":392
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":396
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":392
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":399
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":402
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":399
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":405
 * 
 *         @property
 *         cdef inline char* data(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":411
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":405
 * 
 *         @property
 *         cdef inline char* data(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":807
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":808
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":807
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":810
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":811
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":810
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":813
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":814
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":813
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":816
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":817
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":816
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":819
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":820
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":819
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":822
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":823
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":824
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_5numpy_5dtype_8subarray_subarray(__pyx_v_d)->shape;
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = ((PyObject*)__pyx_t_2);
    goto __pyx_L0;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":823
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":826
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":822
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1010
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1011
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1012
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 1012, __pyx_L1_error)

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1010
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1014
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1015
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1016
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1017
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1016
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1018
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1014
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1022
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1023
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1024
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1024, __pyx_L3_error)

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1023
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1025
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1025, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1026
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1026, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1026, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1023
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1022
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1028
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1030
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1030, __pyx_L3_error)

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1031
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1031, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1032
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1032, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1032, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1028
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1034
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 1);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1035
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1036
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1036, __pyx_L3_error)

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1035
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1037
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1037, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1038
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1035
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1034
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1053
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1068
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1071
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1078
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1071
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1081
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1085
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1081
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1088
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1092
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1088
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":59
 *         public dict fragment_set_properties
 * 
 *     cpdef bint is_n_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_n_glycan", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_n_glycan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 59, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":60
 * 
 *     cpdef bint is_n_glycan(self):
 *         return GlycanTypes_n_glycan in self.glycan_types             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_o_glycan(self):
 */
  __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan, __pyx_v_self->glycan_types, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":59
 *         public dict fragment_set_properties
 * 
 *     cpdef bint is_n_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.is_n_glycan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan, "GlycanCombinationRecordBase.is_n_glycan(self) -> bool");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan = {"is_n_glycan", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_n_glycan (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_n_glycan", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_n_glycan", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_n_glycan", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.is_n_glycan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":62
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 *     cpdef bint is_o_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_o_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_o_glycan", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_o_glycan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":63
 * 
 *     cpdef bint is_o_glycan(self):
 *         return GlycanTypes_o_glycan in self.glycan_types             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_gag_linker(self):
 */
  __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan, __pyx_v_self->glycan_types, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":62
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 *     cpdef bint is_o_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.is_o_glycan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2is_o_glycan, "GlycanCombinationRecordBase.is_o_glycan(self) -> bool");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan = {"is_o_glycan", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2is_o_glycan};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_o_glycan (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_o_glycan", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_o_glycan", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2is_o_glycan(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2is_o_glycan(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_o_glycan", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_o_glycan(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.is_o_glycan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":65
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 *     cpdef bint is_gag_linker(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_gag_linker in self.glycan_types
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_gag_linker(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_gag_linker", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_gag_linker); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 65, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":66
 * 
 *     cpdef bint is_gag_linker(self):
 *         return GlycanTypes_gag_linker in self.glycan_types             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker, __pyx_v_self->glycan_types, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":65
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 *     cpdef bint is_gag_linker(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_gag_linker in self.glycan_types
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.is_gag_linker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4is_gag_linker, "GlycanCombinationRecordBase.is_gag_linker(self) -> bool");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker = {"is_gag_linker", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4is_gag_linker};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_gag_linker (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_gag_linker", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_gag_linker", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4is_gag_linker(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4is_gag_linker(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_gag_linker", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_gag_linker(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.is_gag_linker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":69
 * 
 *     @cython.cdivision(True)
 *     cpdef double normalized_size(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double k
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_normalized_size(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_v_k;
  double __pyx_v_d;
  npy_uint16 __pyx_v_n;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  double __pyx_t_6;
  npy_uint16 __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("normalized_size", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_normalized_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":74
 *             double d
 *             uint16_t n
 *         n = self.internal_size_approximation             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return 0.0
 */
  __pyx_t_7 = __pyx_v_self->internal_size_approximation;
  __pyx_v_n = __pyx_t_7;

  /* "glycresoft/_c/tandem/core_search.pyx":75
 *             uint16_t n
 *         n = self.internal_size_approximation
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         if self.side_group_count > 0:
 */
  __pyx_t_8 = (__pyx_v_n == 0);
  if (__pyx_t_8) {

    /* "glycresoft/_c/tandem/core_search.pyx":76
 *         n = self.internal_size_approximation
 *         if n == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 *         if self.side_group_count > 0:
 *             k = 1.0
 */
    __pyx_r = 0.0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":75
 *             uint16_t n
 *         n = self.internal_size_approximation
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         if self.side_group_count > 0:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":77
 *         if n == 0:
 *             return 0.0
 *         if self.side_group_count > 0:             # <<<<<<<<<<<<<<
 *             k = 1.0
 *         else:
 */
  __pyx_t_8 = (__pyx_v_self->side_group_count > 0);
  if (__pyx_t_8) {

    /* "glycresoft/_c/tandem/core_search.pyx":78
 *             return 0.0
 *         if self.side_group_count > 0:
 *             k = 1.0             # <<<<<<<<<<<<<<
 *         else:
 *             k = 2.0
 */
    __pyx_v_k = 1.0;

    /* "glycresoft/_c/tandem/core_search.pyx":77
 *         if n == 0:
 *             return 0.0
 *         if self.side_group_count > 0:             # <<<<<<<<<<<<<<
 *             k = 1.0
 *         else:
 */
    goto __pyx_L4;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":80
 *             k = 1.0
 *         else:
 *             k = 2.0             # <<<<<<<<<<<<<<
 *         d = max(n * math.log(n) / k, n)
 *         return d
 */
  /*else*/ {
    __pyx_v_k = 2.0;
  }
  __pyx_L4:;

  /* "glycresoft/_c/tandem/core_search.pyx":81
 *         else:
 *             k = 2.0
 *         d = max(n * math.log(n) / k, n)             # <<<<<<<<<<<<<<
 *         return d
 * 
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_6 = ((__pyx_v_n * log(__pyx_v_n)) / __pyx_v_k);
  __pyx_t_8 = (__pyx_t_7 > __pyx_t_6);
  if (__pyx_t_8) {
    __pyx_t_9 = __pyx_t_7;
  } else {
    __pyx_t_9 = __pyx_t_6;
  }
  __pyx_v_d = __pyx_t_9;

  /* "glycresoft/_c/tandem/core_search.pyx":82
 *             k = 2.0
 *         d = max(n * math.log(n) / k, n)
 *         return d             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_n_glycan_fragments(self):
 */
  __pyx_r = __pyx_v_d;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":69
 * 
 *     @cython.cdivision(True)
 *     cpdef double normalized_size(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.normalized_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_6normalized_size, "GlycanCombinationRecordBase.normalized_size(self) -> double");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size = {"normalized_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_6normalized_size};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("normalized_size (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("normalized_size", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "normalized_size", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_6normalized_size(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_6normalized_size(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("normalized_size", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_normalized_size(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.normalized_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":84
 *         return d
 * 
 *     cpdef list get_n_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_n_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_fragment_structs = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *__pyx_v_strategy = NULL;
  PyObject *__pyx_v_shifts = NULL;
  PyObject *__pyx_v_shift = NULL;
  int __pyx_v_is_core;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_n_glycan_composition_fragments __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_n_glycan_fragments", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_n_glycan_fragments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 84, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":88
 *             PyObject* ptemp
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_n_glycan)             # <<<<<<<<<<<<<<
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 */
  __pyx_t_1 = __pyx_v_self->_fragment_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_1, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":89
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_n_glycan)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 */
  __pyx_t_6 = (__pyx_v_ptemp == NULL);
  if (__pyx_t_6) {

    /* "glycresoft/_c/tandem/core_search.pyx":90
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_n_glycan)
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)             # <<<<<<<<<<<<<<
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_extended, Py_True) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy), __pyx_tuple__3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_strategy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":91
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)             # <<<<<<<<<<<<<<
 *             fragment_structs = []
 *             for shift in shifts:
 */
    __pyx_t_1 = __pyx_v_self->composition;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7.__pyx_n = 2;
    __pyx_t_7.core_count = 1;
    __pyx_t_7.iteration_count = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *)__pyx_v_strategy->__pyx_base.__pyx_base.__pyx_vtab)->n_glycan_composition_fragments(__pyx_v_strategy, __pyx_t_1, 0, &__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_shifts = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":92
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []             # <<<<<<<<<<<<<<
 *             for shift in shifts:
 *                 if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_fragment_structs = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":93
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 *             for shift in shifts:             # <<<<<<<<<<<<<<
 *                 if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:
 *                     is_core = True
 */
    if (unlikely(__pyx_v_shifts == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 93, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_shifts; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_8 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 93, __pyx_L1_error)
        #endif
        if (__pyx_t_8 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 93, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_shift, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":94
 *             fragment_structs = []
 *             for shift in shifts:
 *                 if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:             # <<<<<<<<<<<<<<
 *                     is_core = True
 *                 else:
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_s_HexNAc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_int_2, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_9) {
      } else {
        __pyx_t_6 = __pyx_t_9;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_s_Hex); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_int_3, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_6 = __pyx_t_9;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_6) {

        /* "glycresoft/_c/tandem/core_search.pyx":95
 *             for shift in shifts:
 *                 if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:
 *                     is_core = True             # <<<<<<<<<<<<<<
 *                 else:
 *                     is_core = False
 */
        __pyx_v_is_core = 1;

        /* "glycresoft/_c/tandem/core_search.pyx":94
 *             fragment_structs = []
 *             for shift in shifts:
 *                 if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:             # <<<<<<<<<<<<<<
 *                     is_core = True
 *                 else:
 */
        goto __pyx_L6;
      }

      /* "glycresoft/_c/tandem/core_search.pyx":97
 *                     is_core = True
 *                 else:
 *                     is_core = False             # <<<<<<<<<<<<<<
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 */
      /*else*/ {
        __pyx_v_is_core = 0;
      }
      __pyx_L6:;

      /* "glycresoft/_c/tandem/core_search.pyx":100
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], is_core))             # <<<<<<<<<<<<<<
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_n_glycan, result)
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_is_core); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "glycresoft/_c/tandem/core_search.pyx":99
 *                     is_core = False
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(             # <<<<<<<<<<<<<<
 *                         shift['key'], shift['mass'], is_core))
 *             result = sorted(set(fragment_structs))
 */
      __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment), __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":98
 *                 else:
 *                     is_core = False
 *                 fragment_structs.append(             # <<<<<<<<<<<<<<
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], is_core))
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_fragment_structs, __pyx_t_4); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":93
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 *             for shift in shifts:             # <<<<<<<<<<<<<<
 *                 if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:
 *                     is_core = True
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":101
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], is_core))
 *             result = sorted(set(fragment_structs))             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_n_glycan, result)
 *             return result
 */
    __pyx_t_4 = PySet_New(__pyx_v_fragment_structs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_11 = PyList_Sort(__pyx_t_2); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 101, __pyx_L1_error)
    __pyx_v_result = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":102
 *                         shift['key'], shift['mass'], is_core))
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_n_glycan, result)             # <<<<<<<<<<<<<<
 *             return result
 *         else:
 */
    __pyx_t_2 = __pyx_v_self->_fragment_cache;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_10 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan;
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_12 = PyDict_SetItem(__pyx_t_2, __pyx_t_10, __pyx_v_result); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":103
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_n_glycan, result)
 *             return result             # <<<<<<<<<<<<<<
 *         else:
 *             result = <list>ptemp
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":89
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_n_glycan)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":105
 *             return result
 *         else:
 *             result = <list>ptemp             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
  /*else*/ {
    __pyx_t_10 = ((PyObject *)__pyx_v_ptemp);
    __Pyx_INCREF(__pyx_t_10);
    __pyx_v_result = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":106
 *         else:
 *             result = <list>ptemp
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_o_glycan_fragments(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":84
 *         return d
 * 
 *     cpdef list get_n_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.get_n_glycan_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_fragment_structs);
  __Pyx_XDECREF((PyObject *)__pyx_v_strategy);
  __Pyx_XDECREF(__pyx_v_shifts);
  __Pyx_XDECREF(__pyx_v_shift);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_8get_n_glycan_fragments, "GlycanCombinationRecordBase.get_n_glycan_fragments(self) -> list");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments = {"get_n_glycan_fragments", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_8get_n_glycan_fragments};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_n_glycan_fragments (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_n_glycan_fragments", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_n_glycan_fragments", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_8get_n_glycan_fragments(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_8get_n_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_n_glycan_fragments", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_n_glycan_fragments(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.get_n_glycan_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":108
 *             return result
 * 
 *     cpdef list get_o_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_o_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_fragment_structs = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *__pyx_v_strategy = NULL;
  PyObject *__pyx_v_shifts = NULL;
  PyObject *__pyx_v_shift = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_o_glycan_composition_fragments __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_o_glycan_fragments", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_o_glycan_fragments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 108, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":112
 *             PyObject* ptemp
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_o_glycan)             # <<<<<<<<<<<<<<
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 */
  __pyx_t_1 = __pyx_v_self->_fragment_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_1, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":113
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_o_glycan)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
 */
  __pyx_t_6 = (__pyx_v_ptemp == NULL);
  if (__pyx_t_6) {

    /* "glycresoft/_c/tandem/core_search.pyx":114
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_o_glycan)
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)             # <<<<<<<<<<<<<<
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_extended, Py_True) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy), __pyx_tuple__3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_strategy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":115
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)             # <<<<<<<<<<<<<<
 *             fragment_structs = []
 *             for shift in shifts:
 */
    __pyx_t_1 = __pyx_v_self->composition;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7.__pyx_n = 2;
    __pyx_t_7.core_count = 1;
    __pyx_t_7.iteration_count = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *)__pyx_v_strategy->__pyx_base.__pyx_base.__pyx_vtab)->o_glycan_composition_fragments(__pyx_v_strategy, __pyx_t_1, 0, &__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_shifts = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":116
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []             # <<<<<<<<<<<<<<
 *             for shift in shifts:
 *                 fragment_structs.append(
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_fragment_structs = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":117
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 *             for shift in shifts:             # <<<<<<<<<<<<<<
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 */
    if (unlikely(__pyx_v_shifts == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 117, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_shifts; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_8 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 117, __pyx_L1_error)
        #endif
        if (__pyx_t_8 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 117, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_shift, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":120
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], True))             # <<<<<<<<<<<<<<
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_o_glycan, result)
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "glycresoft/_c/tandem/core_search.pyx":119
 *             for shift in shifts:
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(             # <<<<<<<<<<<<<<
 *                         shift['key'], shift['mass'], True))
 *             result = sorted(set(fragment_structs))
 */
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_True)) __PYX_ERR(0, 119, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":118
 *             fragment_structs = []
 *             for shift in shifts:
 *                 fragment_structs.append(             # <<<<<<<<<<<<<<
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], True))
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_fragment_structs, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":117
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 *             for shift in shifts:             # <<<<<<<<<<<<<<
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":121
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], True))
 *             result = sorted(set(fragment_structs))             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_o_glycan, result)
 *             return result
 */
    __pyx_t_3 = PySet_New(__pyx_v_fragment_structs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_9 = PyList_Sort(__pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 121, __pyx_L1_error)
    __pyx_v_result = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":122
 *                         shift['key'], shift['mass'], True))
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_o_glycan, result)             # <<<<<<<<<<<<<<
 *             return result
 *         else:
 */
    __pyx_t_2 = __pyx_v_self->_fragment_cache;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_4 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_10 = PyDict_SetItem(__pyx_t_2, __pyx_t_4, __pyx_v_result); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":123
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_o_glycan, result)
 *             return result             # <<<<<<<<<<<<<<
 *         else:
 *             result = <list>ptemp
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":113
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_o_glycan)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":125
 *             return result
 *         else:
 *             result = <list>ptemp             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
  /*else*/ {
    __pyx_t_4 = ((PyObject *)__pyx_v_ptemp);
    __Pyx_INCREF(__pyx_t_4);
    __pyx_v_result = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":126
 *         else:
 *             result = <list>ptemp
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_gag_linker_glycan_fragments(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":108
 *             return result
 * 
 *     cpdef list get_o_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.get_o_glycan_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_fragment_structs);
  __Pyx_XDECREF((PyObject *)__pyx_v_strategy);
  __Pyx_XDECREF(__pyx_v_shifts);
  __Pyx_XDECREF(__pyx_v_shift);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_10get_o_glycan_fragments, "GlycanCombinationRecordBase.get_o_glycan_fragments(self) -> list");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments = {"get_o_glycan_fragments", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_10get_o_glycan_fragments};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_o_glycan_fragments (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_o_glycan_fragments", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_o_glycan_fragments", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_10get_o_glycan_fragments(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_10get_o_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_o_glycan_fragments", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_o_glycan_fragments(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.get_o_glycan_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":128
 *             return result
 * 
 *     cpdef list get_gag_linker_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_gag_linker_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_fragment_structs = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *__pyx_v_strategy = NULL;
  PyObject *__pyx_v_shifts = NULL;
  PyObject *__pyx_v_shift = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_24StubGlycopeptideStrategy_gag_linker_composition_fragments __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_gag_linker_glycan_fragments", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_gag_linker_glycan_fragments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 128, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":132
 *             PyObject* ptemp
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_gag_linker)             # <<<<<<<<<<<<<<
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 */
  __pyx_t_1 = __pyx_v_self->_fragment_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_1, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":133
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_gag_linker)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
 */
  __pyx_t_6 = (__pyx_v_ptemp == NULL);
  if (__pyx_t_6) {

    /* "glycresoft/_c/tandem/core_search.pyx":134
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_gag_linker)
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)             # <<<<<<<<<<<<<<
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_extended, Py_True) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy), __pyx_tuple__3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_strategy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":135
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)             # <<<<<<<<<<<<<<
 *             fragment_structs = []
 *             for shift in shifts:
 */
    __pyx_t_1 = __pyx_v_self->composition;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7.__pyx_n = 2;
    __pyx_t_7.core_count = 1;
    __pyx_t_7.iteration_count = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy *)__pyx_v_strategy->__pyx_base.__pyx_base.__pyx_vtab)->gag_linker_composition_fragments(__pyx_v_strategy, __pyx_t_1, 0, &__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_shifts = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":136
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []             # <<<<<<<<<<<<<<
 *             for shift in shifts:
 *                 fragment_structs.append(
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_fragment_structs = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":137
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 *             for shift in shifts:             # <<<<<<<<<<<<<<
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 */
    if (unlikely(__pyx_v_shifts == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 137, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_shifts; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_8 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 137, __pyx_L1_error)
        #endif
        if (__pyx_t_8 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 137, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_shift, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":140
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], True))             # <<<<<<<<<<<<<<
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_gag_linker, result)
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_shift, __pyx_n_s_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "glycresoft/_c/tandem/core_search.pyx":139
 *             for shift in shifts:
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(             # <<<<<<<<<<<<<<
 *                         shift['key'], shift['mass'], True))
 *             result = sorted(set(fragment_structs))
 */
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_True)) __PYX_ERR(0, 139, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":138
 *             fragment_structs = []
 *             for shift in shifts:
 *                 fragment_structs.append(             # <<<<<<<<<<<<<<
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], True))
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_fragment_structs, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":137
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 *             for shift in shifts:             # <<<<<<<<<<<<<<
 *                 fragment_structs.append(
 *                     CoarseStubGlycopeptideFragment(
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":141
 *                     CoarseStubGlycopeptideFragment(
 *                         shift['key'], shift['mass'], True))
 *             result = sorted(set(fragment_structs))             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_gag_linker, result)
 *             return result
 */
    __pyx_t_3 = PySet_New(__pyx_v_fragment_structs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_9 = PyList_Sort(__pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 141, __pyx_L1_error)
    __pyx_v_result = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":142
 *                         shift['key'], shift['mass'], True))
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_gag_linker, result)             # <<<<<<<<<<<<<<
 *             return result
 *         else:
 */
    __pyx_t_2 = __pyx_v_self->_fragment_cache;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_4 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_10 = PyDict_SetItem(__pyx_t_2, __pyx_t_4, __pyx_v_result); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":143
 *             result = sorted(set(fragment_structs))
 *             PyDict_SetItem(self._fragment_cache, GlycanTypes_gag_linker, result)
 *             return result             # <<<<<<<<<<<<<<
 *         else:
 *             result = <list>ptemp
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":133
 *             list result, fragment_structs
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_gag_linker)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             strategy = StubGlycopeptideStrategy(None, extended=True)
 *             shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":145
 *             return result
 *         else:
 *             result = <list>ptemp             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
  /*else*/ {
    __pyx_t_4 = ((PyObject *)__pyx_v_ptemp);
    __Pyx_INCREF(__pyx_t_4);
    __pyx_v_result = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":146
 *         else:
 *             result = <list>ptemp
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef clear(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":128
 *             return result
 * 
 *     cpdef list get_gag_linker_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.get_gag_linker_glycan_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_fragment_structs);
  __Pyx_XDECREF((PyObject *)__pyx_v_strategy);
  __Pyx_XDECREF(__pyx_v_shifts);
  __Pyx_XDECREF(__pyx_v_shift);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12get_gag_linker_glycan_fragments, "GlycanCombinationRecordBase.get_gag_linker_glycan_fragments(self) -> list");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments = {"get_gag_linker_glycan_fragments", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12get_gag_linker_glycan_fragments};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_gag_linker_glycan_fragments (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_gag_linker_glycan_fragments", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_gag_linker_glycan_fragments", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12get_gag_linker_glycan_fragments(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12get_gag_linker_glycan_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_gag_linker_glycan_fragments", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_gag_linker_glycan_fragments(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.get_gag_linker_glycan_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":148
 *             return result
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         self._fragment_cache.clear()
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_clear(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":149
 * 
 *     cpdef clear(self):
 *         self._fragment_cache.clear()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_fragment_cache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(0, 149, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyDict_Clear(__pyx_v_self->_fragment_cache); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 149, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":148
 *             return result
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         self._fragment_cache.clear()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_14clear, "GlycanCombinationRecordBase.clear(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear = {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_14clear};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_14clear(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_14clear(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_clear(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":47
 * cdef class GlycanCombinationRecordBase(object):
 *     cdef:
 *         public size_t id             # <<<<<<<<<<<<<<
 *         public double dehydrated_mass
 *         public object composition
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_v_self->id = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":48
 *     cdef:
 *         public size_t id
 *         public double dehydrated_mass             # <<<<<<<<<<<<<<
 *         public object composition
 *         public uint16_t size
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->dehydrated_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.dehydrated_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_v_self->dehydrated_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.dehydrated_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":49
 *         public size_t id
 *         public double dehydrated_mass
 *         public object composition             # <<<<<<<<<<<<<<
 *         public uint16_t size
 *         public uint16_t internal_size_approximation
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->composition);
  __pyx_r = __pyx_v_self->composition;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(__pyx_v_self->composition);
  __pyx_v_self->composition = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(__pyx_v_self->composition);
  __pyx_v_self->composition = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":50
 *         public double dehydrated_mass
 *         public object composition
 *         public uint16_t size             # <<<<<<<<<<<<<<
 *         public uint16_t internal_size_approximation
 *         public uint16_t side_group_count
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  npy_uint16 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint16(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_v_self->size = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":51
 *         public object composition
 *         public uint16_t size
 *         public uint16_t internal_size_approximation             # <<<<<<<<<<<<<<
 *         public uint16_t side_group_count
 *         public uint16_t count
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->internal_size_approximation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.internal_size_approximation.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  npy_uint16 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint16(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_v_self->internal_size_approximation = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.internal_size_approximation.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":52
 *         public uint16_t size
 *         public uint16_t internal_size_approximation
 *         public uint16_t side_group_count             # <<<<<<<<<<<<<<
 *         public uint16_t count
 *         public list glycan_types
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->side_group_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.side_group_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  npy_uint16 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint16(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_v_self->side_group_count = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.side_group_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":53
 *         public uint16_t internal_size_approximation
 *         public uint16_t side_group_count
 *         public uint16_t count             # <<<<<<<<<<<<<<
 *         public list glycan_types
 *         public dict _fragment_cache
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  npy_uint16 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint16(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_v_self->count = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":54
 *         public uint16_t side_group_count
 *         public uint16_t count
 *         public list glycan_types             # <<<<<<<<<<<<<<
 *         public dict _fragment_cache
 *         public Py_hash_t _hash
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->glycan_types);
  __pyx_r = __pyx_v_self->glycan_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(0, 54, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->glycan_types);
  __Pyx_DECREF(__pyx_v_self->glycan_types);
  __pyx_v_self->glycan_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.glycan_types.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->glycan_types);
  __Pyx_DECREF(__pyx_v_self->glycan_types);
  __pyx_v_self->glycan_types = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":55
 *         public uint16_t count
 *         public list glycan_types
 *         public dict _fragment_cache             # <<<<<<<<<<<<<<
 *         public Py_hash_t _hash
 *         public dict fragment_set_properties
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_fragment_cache);
  __pyx_r = __pyx_v_self->_fragment_cache;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fragment_cache);
  __Pyx_DECREF(__pyx_v_self->_fragment_cache);
  __pyx_v_self->_fragment_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase._fragment_cache.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fragment_cache);
  __Pyx_DECREF(__pyx_v_self->_fragment_cache);
  __pyx_v_self->_fragment_cache = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":56
 *         public list glycan_types
 *         public dict _fragment_cache
 *         public Py_hash_t _hash             # <<<<<<<<<<<<<<
 *         public dict fragment_set_properties
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase._hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  Py_hash_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_AsHash_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_v_self->_hash = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase._hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":57
 *         public dict _fragment_cache
 *         public Py_hash_t _hash
 *         public dict fragment_set_properties             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_n_glycan(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fragment_set_properties);
  __pyx_r = __pyx_v_self->fragment_set_properties;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_set_properties);
  __Pyx_DECREF(__pyx_v_self->fragment_set_properties);
  __pyx_v_self->fragment_set_properties = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.fragment_set_properties.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->fragment_set_properties);
  __Pyx_DECREF(__pyx_v_self->fragment_set_properties);
  __pyx_v_self->fragment_set_properties = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16__reduce_cython__, "GlycanCombinationRecordBase.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_17__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._fragment_cache, self._hash, self.composition, self.count, self.dehydrated_mass, self.fragment_set_properties, self.glycan_types, self.id, self.internal_size_approximation, self.side_group_count, self.size)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->count); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->dehydrated_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->id); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->internal_size_approximation); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->side_group_count); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_uint16(__pyx_v_self->size); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(11); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_self->_fragment_cache);
  __Pyx_GIVEREF(__pyx_v_self->_fragment_cache);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_self->_fragment_cache)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->composition);
  __Pyx_GIVEREF(__pyx_v_self->composition);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_self->composition)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fragment_set_properties);
  __Pyx_GIVEREF(__pyx_v_self->fragment_set_properties);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_v_self->fragment_set_properties)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->glycan_types);
  __Pyx_GIVEREF(__pyx_v_self->glycan_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_v_self->glycan_types)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 7, __pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 8, __pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 9, __pyx_t_6)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 10, __pyx_t_7)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._fragment_cache, self._hash, self.composition, self.count, self.dehydrated_mass, self.fragment_set_properties, self.glycan_types, self.id, self.internal_size_approximation, self.side_group_count, self.size)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_8 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v__dict = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "(tree fragment)":7
 *     state = (self._fragment_cache, self._hash, self.composition, self.count, self.dehydrated_mass, self.fragment_set_properties, self.glycan_types, self.id, self.internal_size_approximation, self.side_group_count, self.size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_9 = (__pyx_v__dict != Py_None);
  if (__pyx_t_9) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._fragment_cache is not None or self.composition is not None or self.fragment_set_properties is not None or self.glycan_types is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._fragment_cache, self._hash, self.composition, self.count, self.dehydrated_mass, self.fragment_set_properties, self.glycan_types, self.id, self.internal_size_approximation, self.side_group_count, self.size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._fragment_cache is not None or self.composition is not None or self.fragment_set_properties is not None or self.glycan_types is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, None), state
 */
  /*else*/ {
    __pyx_t_10 = (__pyx_v_self->_fragment_cache != ((PyObject*)Py_None));
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_self->composition != Py_None);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_self->fragment_set_properties != ((PyObject*)Py_None));
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_self->glycan_types != ((PyObject*)Py_None));
    __pyx_t_9 = __pyx_t_10;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_9;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._fragment_cache is not None or self.composition is not None or self.fragment_set_properties is not None or self.glycan_types is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._fragment_cache is not None or self.composition is not None or self.fragment_set_properties is not None or self.glycan_types is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_GlycanCombination); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_90496148);
    __Pyx_GIVEREF(__pyx_int_90496148);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_int_90496148)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_8)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._fragment_cache is not None or self.composition is not None or self.fragment_set_properties is not None or self.glycan_types is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, None), state
 *     else:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCombinationRecordBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_GlycanCombination); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_90496148);
    __Pyx_GIVEREF(__pyx_int_90496148);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_int_90496148)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_8 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCombinationRecordBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_18__setstate_cython__, "GlycanCombinationRecordBase.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_19__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_18__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_18__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_18__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCombinationRecordBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCombinationRecordBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCombinationRecordBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCombinationRecordBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":158
 *         public bint is_core
 * 
 *     def __init__(self, key, mass, is_core):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.mass = mass
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_is_core = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_mass,&__pyx_n_s_is_core,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 158, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 158, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 158, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_is_core)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 158, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 158, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 158, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_key = values[0];
    __pyx_v_mass = values[1];
    __pyx_v_is_core = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 158, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), __pyx_v_key, __pyx_v_mass, __pyx_v_is_core);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_mass, PyObject *__pyx_v_is_core) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":159
 * 
 *     def __init__(self, key, mass, is_core):
 *         self.key = key             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.is_core = is_core
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_key;

  /* "glycresoft/_c/tandem/core_search.pyx":160
 *     def __init__(self, key, mass, is_core):
 *         self.key = key
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.is_core = is_core
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":161
 *         self.key = key
 *         self.mass = mass
 *         self.is_core = is_core             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_is_core); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_v_self->is_core = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":158
 *         public bint is_core
 * 
 *     def __init__(self, key, mass, is_core):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.mass = mass
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":163
 *         self.is_core = is_core
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.is_core)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_2__reduce__, "CoarseStubGlycopeptideFragment.__reduce__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3__reduce__ = {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_2__reduce__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_2__reduce__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_2__reduce__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":164
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.key, self.mass, self.is_core)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->is_core); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->key)) __PYX_ERR(0, 164, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":163
 *         self.is_core = is_core
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.is_core)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":166
 *         return self.__class__, (self.key, self.mass, self.is_core)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%s, %f, %r)" % (
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":167
 * 
 *     def __repr__(self):
 *         return "%s(%s, %f, %r)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.key, self.mass, self.is_core
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycresoft/_c/tandem/core_search.pyx":168
 *     def __repr__(self):
 *         return "%s(%s, %f, %r)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.key, self.mass, self.is_core
 *         )
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":169
 *         return "%s(%s, %f, %r)" % (
 *             self.__class__.__name__,
 *             self.key, self.mass, self.is_core             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->is_core); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "glycresoft/_c/tandem/core_search.pyx":168
 *     def __repr__(self):
 *         return "%s(%s, %f, %r)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.key, self.mass, self.is_core
 *         )
 */
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->key)) __PYX_ERR(0, 168, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":167
 * 
 *     def __repr__(self):
 *         return "%s(%s, %f, %r)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.key, self.mass, self.is_core
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_s_f_r, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":166
 *         return self.__class__, (self.key, self.mass, self.is_core)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%s, %f, %r)" % (
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":172
 *         )
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.key == other.key and self.is_core == other.is_core
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_6__eq__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_6__eq__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":173
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.key == other.key and self.is_core == other.is_core
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glycresoft/_c/tandem/core_search.pyx":174
 *     def __eq__(self, other):
 *         try:
 *             return self.key == other.key and self.is_core == other.is_core             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             return self.key == other
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 174, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyObject_RichCompare(__pyx_v_self->key, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 174, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 174, __pyx_L3_error)
      if (__pyx_t_7) {
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_4 = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_core); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 174, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_is_core); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 174, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 174, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_4 = __pyx_t_8;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "glycresoft/_c/tandem/core_search.pyx":173
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.key == other.key and self.is_core == other.is_core
 *         except AttributeError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":175
 *         try:
 *             return self.key == other.key and self.is_core == other.is_core
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             return self.key == other
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_5) < 0) __PYX_ERR(0, 175, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "glycresoft/_c/tandem/core_search.pyx":176
 *             return self.key == other.key and self.is_core == other.is_core
 *         except AttributeError:
 *             return self.key == other             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = PyObject_RichCompare(__pyx_v_self->key, __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 176, __pyx_L5_except_error)
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "glycresoft/_c/tandem/core_search.pyx":173
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.key == other.key and self.is_core == other.is_core
 *         except AttributeError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":172
 *         )
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.key == other.key and self.is_core == other.is_core
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":178
 *             return self.key == other
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_9__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_9__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_8__ne__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_8__ne__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":179
 * 
 *     def __ne__(self, other):
 *         return not (self == other)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":178
 *             return self.key == other
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":181
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(int(self.mass))
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_11__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_11__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_10__hash__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_10__hash__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":182
 * 
 *     def __hash__(self):
 *         return hash(int(self.mass))             # <<<<<<<<<<<<<<
 * 
 *     def __lt__(self, other):
 */
  __pyx_t_1 = __Pyx_PyInt_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":181
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(int(self.mass))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":184
 *         return hash(int(self.mass))
 * 
 *     def __lt__(self, other):             # <<<<<<<<<<<<<<
 *         return self.mass < other.mass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_13__lt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_13__lt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__lt__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_12__lt__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_12__lt__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__lt__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":185
 * 
 *     def __lt__(self, other):
 *         return self.mass < other.mass             # <<<<<<<<<<<<<<
 * 
 *     def __gt__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":184
 *         return hash(int(self.mass))
 * 
 *     def __lt__(self, other):             # <<<<<<<<<<<<<<
 *         return self.mass < other.mass
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__lt__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":187
 *         return self.mass < other.mass
 * 
 *     def __gt__(self, other):             # <<<<<<<<<<<<<<
 *         return self.mass > other.mass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_15__gt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_15__gt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__gt__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_14__gt__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_14__gt__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__gt__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":188
 * 
 *     def __gt__(self, other):
 *         return self.mass > other.mass             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":187
 *         return self.mass < other.mass
 * 
 *     def __gt__(self, other):             # <<<<<<<<<<<<<<
 *         return self.mass > other.mass
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.__gt__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":154
 * cdef class CoarseStubGlycopeptideFragment(object):
 *     cdef:
 *         public object key             # <<<<<<<<<<<<<<
 *         public double mass
 *         public bint is_core
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->key);
  __pyx_r = __pyx_v_self->key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":155
 *     cdef:
 *         public object key
 *         public double mass             # <<<<<<<<<<<<<<
 *         public bint is_core
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":156
 *         public object key
 *         public double mass
 *         public bint is_core             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, key, mass, is_core):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_core); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.is_core.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __pyx_v_self->is_core = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment.is_core.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":201
 * 
 *     @staticmethod
 *     cdef CoarseStubGlycopeptideMatch _create(object key, double mass, double shift_mass, tuple peaks_matched):             # <<<<<<<<<<<<<<
 *         cdef CoarseStubGlycopeptideMatch instance
 *         instance = CoarseStubGlycopeptideMatch.__new__(CoarseStubGlycopeptideMatch)
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch__create(PyObject *__pyx_v_key, double __pyx_v_mass, double __pyx_v_shift_mass, PyObject *__pyx_v_peaks_matched) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_instance = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":203
 *     cdef CoarseStubGlycopeptideMatch _create(object key, double mass, double shift_mass, tuple peaks_matched):
 *         cdef CoarseStubGlycopeptideMatch instance
 *         instance = CoarseStubGlycopeptideMatch.__new__(CoarseStubGlycopeptideMatch)             # <<<<<<<<<<<<<<
 *         instance.key = key
 *         instance.mass = mass
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_instance = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":204
 *         cdef CoarseStubGlycopeptideMatch instance
 *         instance = CoarseStubGlycopeptideMatch.__new__(CoarseStubGlycopeptideMatch)
 *         instance.key = key             # <<<<<<<<<<<<<<
 *         instance.mass = mass
 *         instance.shift_mass = shift_mass
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_instance->key);
  __Pyx_DECREF(__pyx_v_instance->key);
  __pyx_v_instance->key = __pyx_v_key;

  /* "glycresoft/_c/tandem/core_search.pyx":205
 *         instance = CoarseStubGlycopeptideMatch.__new__(CoarseStubGlycopeptideMatch)
 *         instance.key = key
 *         instance.mass = mass             # <<<<<<<<<<<<<<
 *         instance.shift_mass = shift_mass
 *         instance.peaks_matched = peaks_matched
 */
  __pyx_v_instance->mass = __pyx_v_mass;

  /* "glycresoft/_c/tandem/core_search.pyx":206
 *         instance.key = key
 *         instance.mass = mass
 *         instance.shift_mass = shift_mass             # <<<<<<<<<<<<<<
 *         instance.peaks_matched = peaks_matched
 *         return instance
 */
  __pyx_v_instance->shift_mass = __pyx_v_shift_mass;

  /* "glycresoft/_c/tandem/core_search.pyx":207
 *         instance.mass = mass
 *         instance.shift_mass = shift_mass
 *         instance.peaks_matched = peaks_matched             # <<<<<<<<<<<<<<
 *         return instance
 * 
 */
  __Pyx_INCREF(__pyx_v_peaks_matched);
  __Pyx_GIVEREF(__pyx_v_peaks_matched);
  __Pyx_GOTREF(__pyx_v_instance->peaks_matched);
  __Pyx_DECREF(__pyx_v_instance->peaks_matched);
  __pyx_v_instance->peaks_matched = __pyx_v_peaks_matched;

  /* "glycresoft/_c/tandem/core_search.pyx":208
 *         instance.shift_mass = shift_mass
 *         instance.peaks_matched = peaks_matched
 *         return instance             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, key, mass, shift_mass, peaks_matched):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_instance);
  __pyx_r = __pyx_v_instance;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":201
 * 
 *     @staticmethod
 *     cdef CoarseStubGlycopeptideMatch _create(object key, double mass, double shift_mass, tuple peaks_matched):             # <<<<<<<<<<<<<<
 *         cdef CoarseStubGlycopeptideMatch instance
 *         instance = CoarseStubGlycopeptideMatch.__new__(CoarseStubGlycopeptideMatch)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_instance);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":210
 *         return instance
 * 
 *     def __init__(self, key, mass, shift_mass, peaks_matched):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.mass = mass
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_shift_mass = 0;
  PyObject *__pyx_v_peaks_matched = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_mass,&__pyx_n_s_shift_mass,&__pyx_n_s_peaks_matched,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 210, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 210, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(0, 210, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shift_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 210, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(0, 210, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peaks_matched)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 210, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(0, 210, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 210, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
    }
    __pyx_v_key = values[0];
    __pyx_v_mass = values[1];
    __pyx_v_shift_mass = values[2];
    __pyx_v_peaks_matched = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 210, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self), __pyx_v_key, __pyx_v_mass, __pyx_v_shift_mass, __pyx_v_peaks_matched);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_mass, PyObject *__pyx_v_shift_mass, PyObject *__pyx_v_peaks_matched) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":211
 * 
 *     def __init__(self, key, mass, shift_mass, peaks_matched):
 *         self.key = key             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.shift_mass = shift_mass
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_key;

  /* "glycresoft/_c/tandem/core_search.pyx":212
 *     def __init__(self, key, mass, shift_mass, peaks_matched):
 *         self.key = key
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.shift_mass = shift_mass
 *         self.peaks_matched = peaks_matched
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":213
 *         self.key = key
 *         self.mass = mass
 *         self.shift_mass = shift_mass             # <<<<<<<<<<<<<<
 *         self.peaks_matched = peaks_matched
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_shift_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 213, __pyx_L1_error)
  __pyx_v_self->shift_mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":214
 *         self.mass = mass
 *         self.shift_mass = shift_mass
 *         self.peaks_matched = peaks_matched             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_peaks_matched))||((__pyx_v_peaks_matched) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_peaks_matched))) __PYX_ERR(0, 214, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_peaks_matched;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->peaks_matched);
  __Pyx_DECREF(__pyx_v_self->peaks_matched);
  __pyx_v_self->peaks_matched = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":210
 *         return instance
 * 
 *     def __init__(self, key, mass, shift_mass, peaks_matched):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.mass = mass
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":216
 *         self.peaks_matched = peaks_matched
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_2__reduce__, "CoarseStubGlycopeptideMatch.__reduce__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3__reduce__ = {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_2__reduce__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_2__reduce__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_2__reduce__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":217
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->shift_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->key)) __PYX_ERR(0, 217, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->peaks_matched);
  __Pyx_GIVEREF(__pyx_v_self->peaks_matched);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->peaks_matched)) __PYX_ERR(0, 217, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 217, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":216
 *         self.peaks_matched = peaks_matched
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":219
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%s, %f, %f, %r)" % (
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":220
 * 
 *     def __repr__(self):
 *         return "%s(%s, %f, %f, %r)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.key, self.mass, self.shift_mass, self.peaks_matched
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycresoft/_c/tandem/core_search.pyx":221
 *     def __repr__(self):
 *         return "%s(%s, %f, %f, %r)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.key, self.mass, self.shift_mass, self.peaks_matched
 *         )
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":222
 *         return "%s(%s, %f, %f, %r)" % (
 *             self.__class__.__name__,
 *             self.key, self.mass, self.shift_mass, self.peaks_matched             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->shift_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "glycresoft/_c/tandem/core_search.pyx":221
 *     def __repr__(self):
 *         return "%s(%s, %f, %f, %r)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.key, self.mass, self.shift_mass, self.peaks_matched
 *         )
 */
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->key)) __PYX_ERR(0, 221, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 221, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->peaks_matched);
  __Pyx_GIVEREF(__pyx_v_self->peaks_matched);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_self->peaks_matched)) __PYX_ERR(0, 221, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":220
 * 
 *     def __repr__(self):
 *         return "%s(%s, %f, %f, %r)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.key, self.mass, self.shift_mass, self.peaks_matched
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_s_f_f_r, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":219
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%s, %f, %f, %r)" % (
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":195
 * cdef class CoarseStubGlycopeptideMatch(object):
 *     cdef:
 *         public object key             # <<<<<<<<<<<<<<
 *         public double mass
 *         public double shift_mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->key);
  __pyx_r = __pyx_v_self->key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":196
 *     cdef:
 *         public object key
 *         public double mass             # <<<<<<<<<<<<<<
 *         public double shift_mass
 *         public tuple peaks_matched
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 196, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":197
 *         public object key
 *         public double mass
 *         public double shift_mass             # <<<<<<<<<<<<<<
 *         public tuple peaks_matched
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->shift_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.shift_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L1_error)
  __pyx_v_self->shift_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.shift_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":198
 *         public double mass
 *         public double shift_mass
 *         public tuple peaks_matched             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->peaks_matched);
  __pyx_r = __pyx_v_self->peaks_matched;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_value))) __PYX_ERR(0, 198, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->peaks_matched);
  __Pyx_DECREF(__pyx_v_self->peaks_matched);
  __pyx_v_self->peaks_matched = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch.peaks_matched.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->peaks_matched);
  __Pyx_DECREF(__pyx_v_self->peaks_matched);
  __pyx_v_self->peaks_matched = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":237
 * 
 *     @staticmethod
 *     cdef CoarseGlycanMatch _create(list fragment_matches, double n_matched, double n_theoretical, double core_matched, double core_theoretical):             # <<<<<<<<<<<<<<
 *         cdef CoarseGlycanMatch inst = CoarseGlycanMatch.__new__(CoarseGlycanMatch)
 *         inst.fragment_matches = fragment_matches
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch__create(PyObject *__pyx_v_fragment_matches, double __pyx_v_n_matched, double __pyx_v_n_theoretical, double __pyx_v_core_matched, double __pyx_v_core_theoretical) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_inst = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":238
 *     @staticmethod
 *     cdef CoarseGlycanMatch _create(list fragment_matches, double n_matched, double n_theoretical, double core_matched, double core_theoretical):
 *         cdef CoarseGlycanMatch inst = CoarseGlycanMatch.__new__(CoarseGlycanMatch)             # <<<<<<<<<<<<<<
 *         inst.fragment_matches = fragment_matches
 *         inst.n_matched = n_matched
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_inst = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":239
 *     cdef CoarseGlycanMatch _create(list fragment_matches, double n_matched, double n_theoretical, double core_matched, double core_theoretical):
 *         cdef CoarseGlycanMatch inst = CoarseGlycanMatch.__new__(CoarseGlycanMatch)
 *         inst.fragment_matches = fragment_matches             # <<<<<<<<<<<<<<
 *         inst.n_matched = n_matched
 *         inst.n_theoretical = n_theoretical
 */
  __Pyx_INCREF(__pyx_v_fragment_matches);
  __Pyx_GIVEREF(__pyx_v_fragment_matches);
  __Pyx_GOTREF(__pyx_v_inst->fragment_matches);
  __Pyx_DECREF(__pyx_v_inst->fragment_matches);
  __pyx_v_inst->fragment_matches = __pyx_v_fragment_matches;

  /* "glycresoft/_c/tandem/core_search.pyx":240
 *         cdef CoarseGlycanMatch inst = CoarseGlycanMatch.__new__(CoarseGlycanMatch)
 *         inst.fragment_matches = fragment_matches
 *         inst.n_matched = n_matched             # <<<<<<<<<<<<<<
 *         inst.n_theoretical = n_theoretical
 *         inst.core_matched = core_matched
 */
  __pyx_v_inst->n_matched = __pyx_v_n_matched;

  /* "glycresoft/_c/tandem/core_search.pyx":241
 *         inst.fragment_matches = fragment_matches
 *         inst.n_matched = n_matched
 *         inst.n_theoretical = n_theoretical             # <<<<<<<<<<<<<<
 *         inst.core_matched = core_matched
 *         inst.core_theoretical = core_theoretical
 */
  __pyx_v_inst->n_theoretical = __pyx_v_n_theoretical;

  /* "glycresoft/_c/tandem/core_search.pyx":242
 *         inst.n_matched = n_matched
 *         inst.n_theoretical = n_theoretical
 *         inst.core_matched = core_matched             # <<<<<<<<<<<<<<
 *         inst.core_theoretical = core_theoretical
 *         return inst
 */
  __pyx_v_inst->core_matched = __pyx_v_core_matched;

  /* "glycresoft/_c/tandem/core_search.pyx":243
 *         inst.n_theoretical = n_theoretical
 *         inst.core_matched = core_matched
 *         inst.core_theoretical = core_theoretical             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->core_theoretical = __pyx_v_core_theoretical;

  /* "glycresoft/_c/tandem/core_search.pyx":244
 *         inst.core_matched = core_matched
 *         inst.core_theoretical = core_theoretical
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":237
 * 
 *     @staticmethod
 *     cdef CoarseGlycanMatch _create(list fragment_matches, double n_matched, double n_theoretical, double core_matched, double core_theoretical):             # <<<<<<<<<<<<<<
 *         cdef CoarseGlycanMatch inst = CoarseGlycanMatch.__new__(CoarseGlycanMatch)
 *         inst.fragment_matches = fragment_matches
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":246
 *         return inst
 * 
 *     def __init__(self, fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical):             # <<<<<<<<<<<<<<
 *         self.fragment_matches = list(fragment_matches)
 *         self.n_matched = n_matched
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fragment_matches = 0;
  PyObject *__pyx_v_n_matched = 0;
  PyObject *__pyx_v_n_theoretical = 0;
  PyObject *__pyx_v_core_matched = 0;
  PyObject *__pyx_v_core_theoretical = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fragment_matches,&__pyx_n_s_n_matched,&__pyx_n_s_n_theoretical,&__pyx_n_s_core_matched,&__pyx_n_s_core_theoretical,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fragment_matches)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_n_matched)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 1); __PYX_ERR(0, 246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_n_theoretical)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 2); __PYX_ERR(0, 246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_core_matched)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 3); __PYX_ERR(0, 246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_core_theoretical)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 4); __PYX_ERR(0, 246, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 246, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
    }
    __pyx_v_fragment_matches = values[0];
    __pyx_v_n_matched = values[1];
    __pyx_v_n_theoretical = values[2];
    __pyx_v_core_matched = values[3];
    __pyx_v_core_theoretical = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 246, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), __pyx_v_fragment_matches, __pyx_v_n_matched, __pyx_v_n_theoretical, __pyx_v_core_matched, __pyx_v_core_theoretical);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_fragment_matches, PyObject *__pyx_v_n_matched, PyObject *__pyx_v_n_theoretical, PyObject *__pyx_v_core_matched, PyObject *__pyx_v_core_theoretical) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":247
 * 
 *     def __init__(self, fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical):
 *         self.fragment_matches = list(fragment_matches)             # <<<<<<<<<<<<<<
 *         self.n_matched = n_matched
 *         self.n_theoretical = n_theoretical
 */
  __pyx_t_1 = PySequence_List(__pyx_v_fragment_matches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":248
 *     def __init__(self, fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical):
 *         self.fragment_matches = list(fragment_matches)
 *         self.n_matched = n_matched             # <<<<<<<<<<<<<<
 *         self.n_theoretical = n_theoretical
 *         self.core_matched = core_matched
 */
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_n_matched); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 248, __pyx_L1_error)
  __pyx_v_self->n_matched = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":249
 *         self.fragment_matches = list(fragment_matches)
 *         self.n_matched = n_matched
 *         self.n_theoretical = n_theoretical             # <<<<<<<<<<<<<<
 *         self.core_matched = core_matched
 *         self.core_theoretical = core_theoretical
 */
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_n_theoretical); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_v_self->n_theoretical = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":250
 *         self.n_matched = n_matched
 *         self.n_theoretical = n_theoretical
 *         self.core_matched = core_matched             # <<<<<<<<<<<<<<
 *         self.core_theoretical = core_theoretical
 * 
 */
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_core_matched); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 250, __pyx_L1_error)
  __pyx_v_self->core_matched = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":251
 *         self.n_theoretical = n_theoretical
 *         self.core_matched = core_matched
 *         self.core_theoretical = core_theoretical             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_core_theoretical); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L1_error)
  __pyx_v_self->core_theoretical = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":246
 *         return inst
 * 
 *     def __init__(self, fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical):             # <<<<<<<<<<<<<<
 *         self.fragment_matches = list(fragment_matches)
 *         self.n_matched = n_matched
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "glycresoft/_c/tandem/core_search.pyx":253
 *         self.core_theoretical = core_theoretical
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield self.fragment_matches
 *         yield self.n_matched
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_3__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_2__iter__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_2__iter__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 253, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_4generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_CoarseGlycanMatch___iter, __pyx_n_s_glycresoft__c_tandem_core_search); if (unlikely(!gen)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_yield;
    case 2: goto __pyx_L5_resume_from_yield;
    case 3: goto __pyx_L6_resume_from_yield;
    case 4: goto __pyx_L7_resume_from_yield;
    case 5: goto __pyx_L8_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 253, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":254
 * 
 *     def __iter__(self):
 *         yield self.fragment_matches             # <<<<<<<<<<<<<<
 *         yield self.n_matched
 *         yield self.n_theoretical
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self->fragment_matches);
  __pyx_r = __pyx_cur_scope->__pyx_v_self->fragment_matches;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 1;
  return __pyx_r;
  __pyx_L4_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 254, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":255
 *     def __iter__(self):
 *         yield self.fragment_matches
 *         yield self.n_matched             # <<<<<<<<<<<<<<
 *         yield self.n_theoretical
 *         yield self.core_matched
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->n_matched); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 2;
  return __pyx_r;
  __pyx_L5_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 255, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":256
 *         yield self.fragment_matches
 *         yield self.n_matched
 *         yield self.n_theoretical             # <<<<<<<<<<<<<<
 *         yield self.core_matched
 *         yield self.core_theoretical
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->n_theoretical); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 3;
  return __pyx_r;
  __pyx_L6_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 256, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":257
 *         yield self.n_matched
 *         yield self.n_theoretical
 *         yield self.core_matched             # <<<<<<<<<<<<<<
 *         yield self.core_theoretical
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->core_matched); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 4;
  return __pyx_r;
  __pyx_L7_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 257, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":258
 *         yield self.n_theoretical
 *         yield self.core_matched
 *         yield self.core_theoretical             # <<<<<<<<<<<<<<
 * 
 *     cpdef double estimate_peptide_mass(self):
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->core_theoretical); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 5;
  return __pyx_r;
  __pyx_L8_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 258, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "glycresoft/_c/tandem/core_search.pyx":253
 *         self.core_theoretical = core_theoretical
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield self.fragment_matches
 *         yield self.n_matched
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":260
 *         yield self.core_theoretical
 * 
 *     cpdef double estimate_peptide_mass(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double weighted_mass_acc
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_6estimate_peptide_mass(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_estimate_peptide_mass(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_weighted_mass_acc;
  double __pyx_v_weight_acc;
  double __pyx_v_fmass;
  size_t __pyx_v_i_fmatch;
  size_t __pyx_v_n_fmatch;
  size_t __pyx_v_i_peak;
  size_t __pyx_v_n_peak;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_fmatch = 0;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5;
  double __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("estimate_peptide_mass", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":269
 *             CoarseStubGlycopeptideMatch fmatch
 *             DeconvolutedPeak peak
 *         weighted_mass_acc = 0             # <<<<<<<<<<<<<<
 *         weight_acc = 0
 *         n_fmatch = PyList_GET_SIZE(self.fragment_matches)
 */
  __pyx_v_weighted_mass_acc = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":270
 *             DeconvolutedPeak peak
 *         weighted_mass_acc = 0
 *         weight_acc = 0             # <<<<<<<<<<<<<<
 *         n_fmatch = PyList_GET_SIZE(self.fragment_matches)
 *         for i_fmatch in range(n_fmatch):
 */
  __pyx_v_weight_acc = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":271
 *         weighted_mass_acc = 0
 *         weight_acc = 0
 *         n_fmatch = PyList_GET_SIZE(self.fragment_matches)             # <<<<<<<<<<<<<<
 *         for i_fmatch in range(n_fmatch):
 *             fmatch = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(self.fragment_matches, i_fmatch)
 */
  __pyx_t_1 = __pyx_v_self->fragment_matches;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_n_fmatch = PyList_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":272
 *         weight_acc = 0
 *         n_fmatch = PyList_GET_SIZE(self.fragment_matches)
 *         for i_fmatch in range(n_fmatch):             # <<<<<<<<<<<<<<
 *             fmatch = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(self.fragment_matches, i_fmatch)
 *             fmass = fmatch.shift_mass
 */
  __pyx_t_2 = __pyx_v_n_fmatch;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i_fmatch = __pyx_t_4;

    /* "glycresoft/_c/tandem/core_search.pyx":273
 *         n_fmatch = PyList_GET_SIZE(self.fragment_matches)
 *         for i_fmatch in range(n_fmatch):
 *             fmatch = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(self.fragment_matches, i_fmatch)             # <<<<<<<<<<<<<<
 *             fmass = fmatch.shift_mass
 *             n_peak = PyTuple_GET_SIZE(fmatch.peaks_matched)
 */
    __pyx_t_1 = __pyx_v_self->fragment_matches;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i_fmatch);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)__pyx_t_5);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_fmatch, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":274
 *         for i_fmatch in range(n_fmatch):
 *             fmatch = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(self.fragment_matches, i_fmatch)
 *             fmass = fmatch.shift_mass             # <<<<<<<<<<<<<<
 *             n_peak = PyTuple_GET_SIZE(fmatch.peaks_matched)
 *             for i_peak in range(n_peak):
 */
    __pyx_t_6 = __pyx_v_fmatch->shift_mass;
    __pyx_v_fmass = __pyx_t_6;

    /* "glycresoft/_c/tandem/core_search.pyx":275
 *             fmatch = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(self.fragment_matches, i_fmatch)
 *             fmass = fmatch.shift_mass
 *             n_peak = PyTuple_GET_SIZE(fmatch.peaks_matched)             # <<<<<<<<<<<<<<
 *             for i_peak in range(n_peak):
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(fmatch.peaks_matched, i_peak)
 */
    __pyx_t_1 = __pyx_v_fmatch->peaks_matched;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_n_peak = PyTuple_GET_SIZE(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":276
 *             fmass = fmatch.shift_mass
 *             n_peak = PyTuple_GET_SIZE(fmatch.peaks_matched)
 *             for i_peak in range(n_peak):             # <<<<<<<<<<<<<<
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(fmatch.peaks_matched, i_peak)
 *                 weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
 */
    __pyx_t_7 = __pyx_v_n_peak;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_i_peak = __pyx_t_9;

      /* "glycresoft/_c/tandem/core_search.pyx":277
 *             n_peak = PyTuple_GET_SIZE(fmatch.peaks_matched)
 *             for i_peak in range(n_peak):
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(fmatch.peaks_matched, i_peak)             # <<<<<<<<<<<<<<
 *                 weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
 *                 weight_acc += peak.intensity
 */
      __pyx_t_1 = __pyx_v_fmatch->peaks_matched;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_v_i_peak);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((PyObject *)__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":278
 *             for i_peak in range(n_peak):
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(fmatch.peaks_matched, i_peak)
 *                 weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity             # <<<<<<<<<<<<<<
 *                 weight_acc += peak.intensity
 *         if weight_acc == 0:
 */
      __pyx_v_weighted_mass_acc = (__pyx_v_weighted_mass_acc + ((__pyx_v_peak->neutral_mass - __pyx_v_fmass) * __pyx_v_peak->__pyx_base.intensity));

      /* "glycresoft/_c/tandem/core_search.pyx":279
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(fmatch.peaks_matched, i_peak)
 *                 weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
 *                 weight_acc += peak.intensity             # <<<<<<<<<<<<<<
 *         if weight_acc == 0:
 *             return -1
 */
      __pyx_v_weight_acc = (__pyx_v_weight_acc + __pyx_v_peak->__pyx_base.intensity);
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":280
 *                 weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
 *                 weight_acc += peak.intensity
 *         if weight_acc == 0:             # <<<<<<<<<<<<<<
 *             return -1
 *         return weighted_mass_acc / weight_acc
 */
  __pyx_t_10 = (__pyx_v_weight_acc == 0.0);
  if (__pyx_t_10) {

    /* "glycresoft/_c/tandem/core_search.pyx":281
 *                 weight_acc += peak.intensity
 *         if weight_acc == 0:
 *             return -1             # <<<<<<<<<<<<<<
 *         return weighted_mass_acc / weight_acc
 * 
 */
    __pyx_r = -1.0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":280
 *                 weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
 *                 weight_acc += peak.intensity
 *         if weight_acc == 0:             # <<<<<<<<<<<<<<
 *             return -1
 *         return weighted_mass_acc / weight_acc
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":282
 *         if weight_acc == 0:
 *             return -1
 *         return weighted_mass_acc / weight_acc             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  if (unlikely(__pyx_v_weight_acc == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 282, __pyx_L1_error)
  }
  __pyx_r = (__pyx_v_weighted_mass_acc / __pyx_v_weight_acc);
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":260
 *         yield self.core_theoretical
 * 
 *     cpdef double estimate_peptide_mass(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double weighted_mass_acc
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.estimate_peptide_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fmatch);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_6estimate_peptide_mass(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_5estimate_peptide_mass, "CoarseGlycanMatch.estimate_peptide_mass(self) -> double");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_6estimate_peptide_mass = {"estimate_peptide_mass", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_6estimate_peptide_mass, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_5estimate_peptide_mass};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_6estimate_peptide_mass(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("estimate_peptide_mass (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("estimate_peptide_mass", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "estimate_peptide_mass", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_5estimate_peptide_mass(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_5estimate_peptide_mass(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("estimate_peptide_mass", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_estimate_peptide_mass(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 260, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.estimate_peptide_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":284
 *         return weighted_mass_acc / weight_acc
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = (
 *             "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, "
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_7__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_7__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_v_template = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":286
 *     def __repr__(self):
 *         template = (
 *             "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, "             # <<<<<<<<<<<<<<
 *             "{self.core_matched}, {self.core_theoretical})")
 *         return template.format(self=self)
 */
  __Pyx_INCREF(__pyx_kp_s_self___class_____name___self_n);
  __pyx_v_template = __pyx_kp_s_self___class_____name___self_n;

  /* "glycresoft/_c/tandem/core_search.pyx":288
 *             "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, "
 *             "{self.core_matched}, {self.core_theoretical})")
 *         return template.format(self=self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_template, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 288, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":284
 *         return weighted_mass_acc / weight_acc
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = (
 *             "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, "
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_template);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":230
 * cdef class CoarseGlycanMatch(object):
 *     cdef:
 *         public list fragment_matches             # <<<<<<<<<<<<<<
 *         public double n_matched
 *         public double n_theoretical
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fragment_matches);
  __pyx_r = __pyx_v_self->fragment_matches;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(0, 230, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.fragment_matches.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":231
 *     cdef:
 *         public list fragment_matches
 *         public double n_matched             # <<<<<<<<<<<<<<
 *         public double n_theoretical
 *         public double core_matched
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->n_matched); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.n_matched.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 231, __pyx_L1_error)
  __pyx_v_self->n_matched = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.n_matched.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":232
 *         public list fragment_matches
 *         public double n_matched
 *         public double n_theoretical             # <<<<<<<<<<<<<<
 *         public double core_matched
 *         public double core_theoretical
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->n_theoretical); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.n_theoretical.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 232, __pyx_L1_error)
  __pyx_v_self->n_theoretical = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.n_theoretical.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":233
 *         public double n_matched
 *         public double n_theoretical
 *         public double core_matched             # <<<<<<<<<<<<<<
 *         public double core_theoretical
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_matched); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.core_matched.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_v_self->core_matched = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.core_matched.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":234
 *         public double n_theoretical
 *         public double core_matched
 *         public double core_theoretical             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_theoretical); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.core_theoretical.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_v_self->core_theoretical = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.core_theoretical.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_10__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9__reduce_cython__, "CoarseGlycanMatch.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_10__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_10__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_10__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.core_matched, self.core_theoretical, self.fragment_matches, self.n_matched, self.n_theoretical)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_matched); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->core_theoretical); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->n_matched); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->n_theoretical); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fragment_matches);
  __Pyx_GIVEREF(__pyx_v_self->fragment_matches);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->fragment_matches)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.core_matched, self.core_theoretical, self.fragment_matches, self.n_matched, self.n_theoretical)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.core_matched, self.core_theoretical, self.fragment_matches, self.n_matched, self.n_theoretical)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.fragment_matches is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.core_matched, self.core_theoretical, self.fragment_matches, self.n_matched, self.n_theoretical)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.fragment_matches is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->fragment_matches != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.fragment_matches is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.fragment_matches is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_CoarseGlycanMatch); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_252421743);
    __Pyx_GIVEREF(__pyx_int_252421743);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_252421743)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.fragment_matches is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, None), state
 *     else:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CoarseGlycanMatch__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_CoarseGlycanMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_252421743);
    __Pyx_GIVEREF(__pyx_int_252421743);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_252421743)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoarseGlycanMatch__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_11__setstate_cython__, "CoarseGlycanMatch.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_11__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_11__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_11__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CoarseGlycanMatch__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_CoarseGlycanMatch__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoarseGlycanMatch__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.CoarseGlycanMatch.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":298
 *         public bint recalibrate_peptide_mass
 * 
 *     def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False):             # <<<<<<<<<<<<<<
 *         self.product_error_tolerance = product_error_tolerance
 *         self.fragment_weight = fragment_weight
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_product_error_tolerance = 0;
  PyObject *__pyx_v_fragment_weight = 0;
  PyObject *__pyx_v_core_weight = 0;
  PyObject *__pyx_v_recalibrate_peptide_mass = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_product_error_tolerance,&__pyx_n_s_fragment_weight,&__pyx_n_s_core_weight,&__pyx_n_s_recalibrate_peptide_mass,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_1eneg_5));
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_0_56));
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_0_42));
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_product_error_tolerance);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fragment_weight);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_core_weight);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_recalibrate_peptide_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 298, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_product_error_tolerance = values[0];
    __pyx_v_fragment_weight = values[1];
    __pyx_v_core_weight = values[2];
    __pyx_v_recalibrate_peptide_mass = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 4, __pyx_nargs); __PYX_ERR(0, 298, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v_product_error_tolerance, __pyx_v_fragment_weight, __pyx_v_core_weight, __pyx_v_recalibrate_peptide_mass);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_product_error_tolerance, PyObject *__pyx_v_fragment_weight, PyObject *__pyx_v_core_weight, PyObject *__pyx_v_recalibrate_peptide_mass) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":299
 * 
 *     def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False):
 *         self.product_error_tolerance = product_error_tolerance             # <<<<<<<<<<<<<<
 *         self.fragment_weight = fragment_weight
 *         self.core_weight = core_weight
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_product_error_tolerance); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 299, __pyx_L1_error)
  __pyx_v_self->product_error_tolerance = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":300
 *     def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False):
 *         self.product_error_tolerance = product_error_tolerance
 *         self.fragment_weight = fragment_weight             # <<<<<<<<<<<<<<
 *         self.core_weight = core_weight
 *         self.recalibrate_peptide_mass = recalibrate_peptide_mass
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_fragment_weight); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 300, __pyx_L1_error)
  __pyx_v_self->fragment_weight = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":301
 *         self.product_error_tolerance = product_error_tolerance
 *         self.fragment_weight = fragment_weight
 *         self.core_weight = core_weight             # <<<<<<<<<<<<<<
 *         self.recalibrate_peptide_mass = recalibrate_peptide_mass
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_core_weight); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L1_error)
  __pyx_v_self->core_weight = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":302
 *         self.fragment_weight = fragment_weight
 *         self.core_weight = core_weight
 *         self.recalibrate_peptide_mass = recalibrate_peptide_mass             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_recalibrate_peptide_mass); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L1_error)
  __pyx_v_self->recalibrate_peptide_mass = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":298
 *         public bint recalibrate_peptide_mass
 * 
 *     def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False):             # <<<<<<<<<<<<<<
 *         self.product_error_tolerance = product_error_tolerance
 *         self.fragment_weight = fragment_weight
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":305
 * 
 *     @cython.boundscheck(False)
 *     cpdef CoarseGlycanMatch _match_fragments(self, object scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list fragment_matches
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, PyObject *__pyx_v_shifts, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments *__pyx_optional_args) {
  double __pyx_v_mass_shift_tandem_mass = ((double)0.0);
  PyObject *__pyx_v_fragment_matches = 0;
  double __pyx_v_core_matched;
  double __pyx_v_core_theoretical;
  double __pyx_v_product_error_tolerance;
  double __pyx_v_n_frags_matched;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  int __pyx_v_has_tandem_shift;
  PyObject *__pyx_v_hits = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *__pyx_v_shift = 0;
  int __pyx_v_is_core;
  double __pyx_v_target_mass;
  double __pyx_v_shifted_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  double __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  PyObject *__pyx_t_12;
  struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_match_fragments", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_mass_shift_tandem_mass = __pyx_optional_args->mass_shift_tandem_mass;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_match_fragments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_peptide_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_mass_shift_tandem_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_scan, __pyx_t_3, __pyx_v_shifts, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 4+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 305, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 305, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":316
 *             CoarseStubGlycopeptideFragment shift
 * 
 *         product_error_tolerance = (self.product_error_tolerance)             # <<<<<<<<<<<<<<
 *         fragment_matches = []
 * 
 */
  __pyx_t_8 = __pyx_v_self->product_error_tolerance;
  __pyx_v_product_error_tolerance = __pyx_t_8;

  /* "glycresoft/_c/tandem/core_search.pyx":317
 * 
 *         product_error_tolerance = (self.product_error_tolerance)
 *         fragment_matches = []             # <<<<<<<<<<<<<<
 * 
 *         core_matched = 0.0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":319
 *         fragment_matches = []
 * 
 *         core_matched = 0.0             # <<<<<<<<<<<<<<
 *         core_theoretical = 0.0
 *         n_frags_matched = 0.0
 */
  __pyx_v_core_matched = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":320
 * 
 *         core_matched = 0.0
 *         core_theoretical = 0.0             # <<<<<<<<<<<<<<
 *         n_frags_matched = 0.0
 *         has_tandem_shift = abs(mass_shift_tandem_mass) > 0
 */
  __pyx_v_core_theoretical = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":321
 *         core_matched = 0.0
 *         core_theoretical = 0.0
 *         n_frags_matched = 0.0             # <<<<<<<<<<<<<<
 *         has_tandem_shift = abs(mass_shift_tandem_mass) > 0
 *         peak_set = scan.deconvoluted_peak_set
 */
  __pyx_v_n_frags_matched = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":322
 *         core_theoretical = 0.0
 *         n_frags_matched = 0.0
 *         has_tandem_shift = abs(mass_shift_tandem_mass) > 0             # <<<<<<<<<<<<<<
 *         peak_set = scan.deconvoluted_peak_set
 * 
 */
  __pyx_v_has_tandem_shift = (fabs(__pyx_v_mass_shift_tandem_mass) > 0.0);

  /* "glycresoft/_c/tandem/core_search.pyx":323
 *         n_frags_matched = 0.0
 *         has_tandem_shift = abs(mass_shift_tandem_mass) > 0
 *         peak_set = scan.deconvoluted_peak_set             # <<<<<<<<<<<<<<
 * 
 *         n = PyList_GET_SIZE(shifts)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scan, __pyx_n_s_deconvoluted_peak_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet))))) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_v_peak_set = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":325
 *         peak_set = scan.deconvoluted_peak_set
 * 
 *         n = PyList_GET_SIZE(shifts)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n):
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_shifts);

  /* "glycresoft/_c/tandem/core_search.pyx":327
 *         n = PyList_GET_SIZE(shifts)
 * 
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             shift = <CoarseStubGlycopeptideFragment>PyList_GET_ITEM(shifts, i)
 *             if shift.is_core:
 */
  __pyx_t_9 = __pyx_v_n;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "glycresoft/_c/tandem/core_search.pyx":328
 * 
 *         for i in range(n):
 *             shift = <CoarseStubGlycopeptideFragment>PyList_GET_ITEM(shifts, i)             # <<<<<<<<<<<<<<
 *             if shift.is_core:
 *                 is_core = True
 */
    __pyx_t_12 = PyList_GET_ITEM(__pyx_v_shifts, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_12);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_shift, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":329
 *         for i in range(n):
 *             shift = <CoarseStubGlycopeptideFragment>PyList_GET_ITEM(shifts, i)
 *             if shift.is_core:             # <<<<<<<<<<<<<<
 *                 is_core = True
 *                 core_theoretical += 1
 */
    if (__pyx_v_shift->is_core) {

      /* "glycresoft/_c/tandem/core_search.pyx":330
 *             shift = <CoarseStubGlycopeptideFragment>PyList_GET_ITEM(shifts, i)
 *             if shift.is_core:
 *                 is_core = True             # <<<<<<<<<<<<<<
 *                 core_theoretical += 1
 *             target_mass = shift.mass + peptide_mass
 */
      __pyx_v_is_core = 1;

      /* "glycresoft/_c/tandem/core_search.pyx":331
 *             if shift.is_core:
 *                 is_core = True
 *                 core_theoretical += 1             # <<<<<<<<<<<<<<
 *             target_mass = shift.mass + peptide_mass
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
 */
      __pyx_v_core_theoretical = (__pyx_v_core_theoretical + 1.0);

      /* "glycresoft/_c/tandem/core_search.pyx":329
 *         for i in range(n):
 *             shift = <CoarseStubGlycopeptideFragment>PyList_GET_ITEM(shifts, i)
 *             if shift.is_core:             # <<<<<<<<<<<<<<
 *                 is_core = True
 *                 core_theoretical += 1
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":332
 *                 is_core = True
 *                 core_theoretical += 1
 *             target_mass = shift.mass + peptide_mass             # <<<<<<<<<<<<<<
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
 *             if hits:
 */
    __pyx_v_target_mass = (__pyx_v_shift->mass + __pyx_v_peptide_mass);

    /* "glycresoft/_c/tandem/core_search.pyx":333
 *                 core_theoretical += 1
 *             target_mass = shift.mass + peptide_mass
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)             # <<<<<<<<<<<<<<
 *             if hits:
 *                 if is_core:
 */
    __pyx_t_13.__pyx_n = 1;
    __pyx_t_13.tolerance = __pyx_v_product_error_tolerance;
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set->__pyx_vtab)->all_peaks_for(__pyx_v_peak_set, __pyx_v_target_mass, 0, &__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_hits, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":334
 *             target_mass = shift.mass + peptide_mass
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
 *             if hits:             # <<<<<<<<<<<<<<
 *                 if is_core:
 *                     core_matched += 1
 */
    __pyx_t_14 = (__pyx_v_hits != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_hits) != 0);
    if (__pyx_t_14) {

      /* "glycresoft/_c/tandem/core_search.pyx":335
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
 *             if hits:
 *                 if is_core:             # <<<<<<<<<<<<<<
 *                     core_matched += 1
 *                 fragment_matches.append(
 */
      if (__pyx_v_is_core) {

        /* "glycresoft/_c/tandem/core_search.pyx":336
 *             if hits:
 *                 if is_core:
 *                     core_matched += 1             # <<<<<<<<<<<<<<
 *                 fragment_matches.append(
 *                     CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))
 */
        __pyx_v_core_matched = (__pyx_v_core_matched + 1.0);

        /* "glycresoft/_c/tandem/core_search.pyx":335
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
 *             if hits:
 *                 if is_core:             # <<<<<<<<<<<<<<
 *                     core_matched += 1
 *                 fragment_matches.append(
 */
      }

      /* "glycresoft/_c/tandem/core_search.pyx":338
 *                     core_matched += 1
 *                 fragment_matches.append(
 *                     CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))             # <<<<<<<<<<<<<<
 *                 n_frags_matched += 1
 *             if has_tandem_shift:
 */
      __pyx_t_1 = __pyx_v_shift->key;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_2 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch__create(__pyx_t_1, __pyx_v_target_mass, __pyx_v_shift->mass, __pyx_v_hits)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":337
 *                 if is_core:
 *                     core_matched += 1
 *                 fragment_matches.append(             # <<<<<<<<<<<<<<
 *                     CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))
 *                 n_frags_matched += 1
 */
      __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_fragment_matches, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":339
 *                 fragment_matches.append(
 *                     CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))
 *                 n_frags_matched += 1             # <<<<<<<<<<<<<<
 *             if has_tandem_shift:
 *                 shifted_mass = target_mass + mass_shift_tandem_mass
 */
      __pyx_v_n_frags_matched = (__pyx_v_n_frags_matched + 1.0);

      /* "glycresoft/_c/tandem/core_search.pyx":334
 *             target_mass = shift.mass + peptide_mass
 *             hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
 *             if hits:             # <<<<<<<<<<<<<<
 *                 if is_core:
 *                     core_matched += 1
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":340
 *                     CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))
 *                 n_frags_matched += 1
 *             if has_tandem_shift:             # <<<<<<<<<<<<<<
 *                 shifted_mass = target_mass + mass_shift_tandem_mass
 *                 hits = peak_set.all_peaks_for(
 */
    if (__pyx_v_has_tandem_shift) {

      /* "glycresoft/_c/tandem/core_search.pyx":341
 *                 n_frags_matched += 1
 *             if has_tandem_shift:
 *                 shifted_mass = target_mass + mass_shift_tandem_mass             # <<<<<<<<<<<<<<
 *                 hits = peak_set.all_peaks_for(
 *                     shifted_mass, product_error_tolerance)
 */
      __pyx_v_shifted_mass = (__pyx_v_target_mass + __pyx_v_mass_shift_tandem_mass);

      /* "glycresoft/_c/tandem/core_search.pyx":342
 *             if has_tandem_shift:
 *                 shifted_mass = target_mass + mass_shift_tandem_mass
 *                 hits = peak_set.all_peaks_for(             # <<<<<<<<<<<<<<
 *                     shifted_mass, product_error_tolerance)
 *                 if hits:
 */
      __pyx_t_13.__pyx_n = 1;
      __pyx_t_13.tolerance = __pyx_v_product_error_tolerance;
      __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set->__pyx_vtab)->all_peaks_for(__pyx_v_peak_set, __pyx_v_shifted_mass, 0, &__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_hits, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":344
 *                 hits = peak_set.all_peaks_for(
 *                     shifted_mass, product_error_tolerance)
 *                 if hits:             # <<<<<<<<<<<<<<
 *                     if is_core:
 *                         core_matched += 1
 */
      __pyx_t_14 = (__pyx_v_hits != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_hits) != 0);
      if (__pyx_t_14) {

        /* "glycresoft/_c/tandem/core_search.pyx":345
 *                     shifted_mass, product_error_tolerance)
 *                 if hits:
 *                     if is_core:             # <<<<<<<<<<<<<<
 *                         core_matched += 1
 *                     fragment_matches.append(
 */
        if (__pyx_v_is_core) {

          /* "glycresoft/_c/tandem/core_search.pyx":346
 *                 if hits:
 *                     if is_core:
 *                         core_matched += 1             # <<<<<<<<<<<<<<
 *                     fragment_matches.append(
 *                         CoarseStubGlycopeptideMatch._create(
 */
          __pyx_v_core_matched = (__pyx_v_core_matched + 1.0);

          /* "glycresoft/_c/tandem/core_search.pyx":345
 *                     shifted_mass, product_error_tolerance)
 *                 if hits:
 *                     if is_core:             # <<<<<<<<<<<<<<
 *                         core_matched += 1
 *                     fragment_matches.append(
 */
        }

        /* "glycresoft/_c/tandem/core_search.pyx":349
 *                     fragment_matches.append(
 *                         CoarseStubGlycopeptideMatch._create(
 *                             shift.key, shifted_mass, shift.mass + mass_shift_tandem_mass, hits))             # <<<<<<<<<<<<<<
 *                     n_frags_matched += 1
 * 
 */
        __pyx_t_2 = __pyx_v_shift->key;
        __Pyx_INCREF(__pyx_t_2);

        /* "glycresoft/_c/tandem/core_search.pyx":348
 *                         core_matched += 1
 *                     fragment_matches.append(
 *                         CoarseStubGlycopeptideMatch._create(             # <<<<<<<<<<<<<<
 *                             shift.key, shifted_mass, shift.mass + mass_shift_tandem_mass, hits))
 *                     n_frags_matched += 1
 */
        __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch__create(__pyx_t_2, __pyx_v_shifted_mass, (__pyx_v_shift->mass + __pyx_v_mass_shift_tandem_mass), __pyx_v_hits)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "glycresoft/_c/tandem/core_search.pyx":347
 *                     if is_core:
 *                         core_matched += 1
 *                     fragment_matches.append(             # <<<<<<<<<<<<<<
 *                         CoarseStubGlycopeptideMatch._create(
 *                             shift.key, shifted_mass, shift.mass + mass_shift_tandem_mass, hits))
 */
        __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_fragment_matches, __pyx_t_1); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "glycresoft/_c/tandem/core_search.pyx":350
 *                         CoarseStubGlycopeptideMatch._create(
 *                             shift.key, shifted_mass, shift.mass + mass_shift_tandem_mass, hits))
 *                     n_frags_matched += 1             # <<<<<<<<<<<<<<
 * 
 *         return CoarseGlycanMatch._create(
 */
        __pyx_v_n_frags_matched = (__pyx_v_n_frags_matched + 1.0);

        /* "glycresoft/_c/tandem/core_search.pyx":344
 *                 hits = peak_set.all_peaks_for(
 *                     shifted_mass, product_error_tolerance)
 *                 if hits:             # <<<<<<<<<<<<<<
 *                     if is_core:
 *                         core_matched += 1
 */
      }

      /* "glycresoft/_c/tandem/core_search.pyx":340
 *                     CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))
 *                 n_frags_matched += 1
 *             if has_tandem_shift:             # <<<<<<<<<<<<<<
 *                 shifted_mass = target_mass + mass_shift_tandem_mass
 *                 hits = peak_set.all_peaks_for(
 */
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":352
 *                     n_frags_matched += 1
 * 
 *         return CoarseGlycanMatch._create(             # <<<<<<<<<<<<<<
 *             fragment_matches, n_frags_matched, float(n), core_matched, core_theoretical)
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "glycresoft/_c/tandem/core_search.pyx":353
 * 
 *         return CoarseGlycanMatch._create(
 *             fragment_matches, n_frags_matched, float(n), core_matched, core_theoretical)             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch__create(__pyx_v_fragment_matches, __pyx_v_n_frags_matched, ((double)__pyx_v_n), __pyx_v_core_matched, __pyx_v_core_theoretical)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":305
 * 
 *     @cython.boundscheck(False)
 *     cpdef CoarseGlycanMatch _match_fragments(self, object scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list fragment_matches
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._match_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fragment_matches);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_set);
  __Pyx_XDECREF(__pyx_v_hits);
  __Pyx_XDECREF((PyObject *)__pyx_v_shift);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_2_match_fragments, "GlycanCoarseScorerBase._match_fragments(self, scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0) -> CoarseGlycanMatch");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments = {"_match_fragments", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_2_match_fragments};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_scan = 0;
  double __pyx_v_peptide_mass;
  PyObject *__pyx_v_shifts = 0;
  double __pyx_v_mass_shift_tandem_mass;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_match_fragments (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scan,&__pyx_n_s_peptide_mass,&__pyx_n_s_shifts,&__pyx_n_s_mass_shift_tandem_mass,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scan)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peptide_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_match_fragments", 0, 3, 4, 1); __PYX_ERR(0, 305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shifts)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_match_fragments", 0, 3, 4, 2); __PYX_ERR(0, 305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass_shift_tandem_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_match_fragments") < 0)) __PYX_ERR(0, 305, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_scan = values[0];
    __pyx_v_peptide_mass = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_peptide_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
    __pyx_v_shifts = ((PyObject*)values[2]);
    if (values[3]) {
      __pyx_v_mass_shift_tandem_mass = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_mass_shift_tandem_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
    } else {
      __pyx_v_mass_shift_tandem_mass = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_match_fragments", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 305, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._match_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shifts), (&PyList_Type), 1, "shifts", 1))) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_2_match_fragments(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_shifts, __pyx_v_mass_shift_tandem_mass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_2_match_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, PyObject *__pyx_v_shifts, double __pyx_v_mass_shift_tandem_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_match_fragments", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->_match_fragments(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_shifts, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._match_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":356
 * 
 *     @cython.cdivision(True)
 *     cpdef double _calculate_score(self, CoarseGlycanMatch glycan_match):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j, n, m
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__calculate_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  size_t __pyx_v_n;
  size_t __pyx_v_m;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_v_matched_fragment = 0;
  double __pyx_v_score;
  double __pyx_v_ratio_fragments;
  double __pyx_v_ratio_core;
  double __pyx_v_coverage;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_v_matched_fragments = 0;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  double __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_calculate_score", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_glycan_match)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":365
 *             list matched_fragments
 * 
 *         ratio_fragments = (<double>glycan_match.n_matched) / glycan_match.n_theoretical             # <<<<<<<<<<<<<<
 *         ratio_core = <double>glycan_match.core_matched / glycan_match.core_theoretical
 *         coverage = (ratio_fragments ** (self.fragment_weight)) * (ratio_core ** (self.core_weight))
 */
  __pyx_v_ratio_fragments = (((double)__pyx_v_glycan_match->n_matched) / __pyx_v_glycan_match->n_theoretical);

  /* "glycresoft/_c/tandem/core_search.pyx":366
 * 
 *         ratio_fragments = (<double>glycan_match.n_matched) / glycan_match.n_theoretical
 *         ratio_core = <double>glycan_match.core_matched / glycan_match.core_theoretical             # <<<<<<<<<<<<<<
 *         coverage = (ratio_fragments ** (self.fragment_weight)) * (ratio_core ** (self.core_weight))
 *         score = 0
 */
  __pyx_v_ratio_core = (((double)__pyx_v_glycan_match->core_matched) / __pyx_v_glycan_match->core_theoretical);

  /* "glycresoft/_c/tandem/core_search.pyx":367
 *         ratio_fragments = (<double>glycan_match.n_matched) / glycan_match.n_theoretical
 *         ratio_core = <double>glycan_match.core_matched / glycan_match.core_theoretical
 *         coverage = (ratio_fragments ** (self.fragment_weight)) * (ratio_core ** (self.core_weight))             # <<<<<<<<<<<<<<
 *         score = 0
 *         matched_fragments = glycan_match.fragment_matches
 */
  __pyx_t_6 = __Pyx_SoftComplexToDouble(__Pyx_c_prod_double(__Pyx_c_pow_double(__pyx_t_double_complex_from_parts(__pyx_v_ratio_fragments, 0), __pyx_t_double_complex_from_parts(__pyx_v_self->fragment_weight, 0)), __Pyx_c_pow_double(__pyx_t_double_complex_from_parts(__pyx_v_ratio_core, 0), __pyx_t_double_complex_from_parts(__pyx_v_self->core_weight, 0))), 1); if (unlikely(__pyx_t_6 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L1_error)
  __pyx_v_coverage = __pyx_t_6;

  /* "glycresoft/_c/tandem/core_search.pyx":368
 *         ratio_core = <double>glycan_match.core_matched / glycan_match.core_theoretical
 *         coverage = (ratio_fragments ** (self.fragment_weight)) * (ratio_core ** (self.core_weight))
 *         score = 0             # <<<<<<<<<<<<<<
 *         matched_fragments = glycan_match.fragment_matches
 *         n = PyList_GET_SIZE(matched_fragments)
 */
  __pyx_v_score = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":369
 *         coverage = (ratio_fragments ** (self.fragment_weight)) * (ratio_core ** (self.core_weight))
 *         score = 0
 *         matched_fragments = glycan_match.fragment_matches             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(matched_fragments)
 *         for i in range(n):
 */
  __pyx_t_1 = __pyx_v_glycan_match->fragment_matches;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_matched_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":370
 *         score = 0
 *         matched_fragments = glycan_match.fragment_matches
 *         n = PyList_GET_SIZE(matched_fragments)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             matched_fragment = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(matched_fragments, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_matched_fragments);

  /* "glycresoft/_c/tandem/core_search.pyx":371
 *         matched_fragments = glycan_match.fragment_matches
 *         n = PyList_GET_SIZE(matched_fragments)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             matched_fragment = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(matched_fragments, i)
 *             m = PyTuple_GET_SIZE(matched_fragment.peaks_matched)
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "glycresoft/_c/tandem/core_search.pyx":372
 *         n = PyList_GET_SIZE(matched_fragments)
 *         for i in range(n):
 *             matched_fragment = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(matched_fragments, i)             # <<<<<<<<<<<<<<
 *             m = PyTuple_GET_SIZE(matched_fragment.peaks_matched)
 *             for j in range(m):
 */
    __pyx_t_10 = PyList_GET_ITEM(__pyx_v_matched_fragments, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_10);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_matched_fragment, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":373
 *         for i in range(n):
 *             matched_fragment = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(matched_fragments, i)
 *             m = PyTuple_GET_SIZE(matched_fragment.peaks_matched)             # <<<<<<<<<<<<<<
 *             for j in range(m):
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(matched_fragment.peaks_matched, j)
 */
    __pyx_t_1 = __pyx_v_matched_fragment->peaks_matched;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_m = PyTuple_GET_SIZE(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":374
 *             matched_fragment = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(matched_fragments, i)
 *             m = PyTuple_GET_SIZE(matched_fragment.peaks_matched)
 *             for j in range(m):             # <<<<<<<<<<<<<<
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(matched_fragment.peaks_matched, j)
 *                 score += math.log(peak.intensity) * (1 - (
 */
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_j = __pyx_t_13;

      /* "glycresoft/_c/tandem/core_search.pyx":375
 *             m = PyTuple_GET_SIZE(matched_fragment.peaks_matched)
 *             for j in range(m):
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(matched_fragment.peaks_matched, j)             # <<<<<<<<<<<<<<
 *                 score += math.log(peak.intensity) * (1 - (
 *                     math.fabs(peak.neutral_mass - matched_fragment.mass) / matched_fragment.mass) ** 4)
 */
      __pyx_t_1 = __pyx_v_matched_fragment->peaks_matched;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_v_j);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((PyObject *)__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":376
 *             for j in range(m):
 *                 peak = <DeconvolutedPeak>PyTuple_GET_ITEM(matched_fragment.peaks_matched, j)
 *                 score += math.log(peak.intensity) * (1 - (             # <<<<<<<<<<<<<<
 *                     math.fabs(peak.neutral_mass - matched_fragment.mass) / matched_fragment.mass) ** 4)
 *         score *= coverage
 */
      __pyx_v_score = (__pyx_v_score + (log(__pyx_v_peak->__pyx_base.intensity) * (1.0 - pow((fabs((__pyx_v_peak->neutral_mass - __pyx_v_matched_fragment->mass)) / __pyx_v_matched_fragment->mass), 4.0))));
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":378
 *                 score += math.log(peak.intensity) * (1 - (
 *                     math.fabs(peak.neutral_mass - matched_fragment.mass) / matched_fragment.mass) ** 4)
 *         score *= coverage             # <<<<<<<<<<<<<<
 *         return score
 * 
 */
  __pyx_v_score = (__pyx_v_score * __pyx_v_coverage);

  /* "glycresoft/_c/tandem/core_search.pyx":379
 *                     math.fabs(peak.neutral_mass - matched_fragment.mass) / matched_fragment.mass) ** 4)
 *         score *= coverage
 *         return score             # <<<<<<<<<<<<<<
 * 
 *     cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,
 */
  __pyx_r = __pyx_v_score;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":356
 * 
 *     @cython.cdivision(True)
 *     cpdef double _calculate_score(self, CoarseGlycanMatch glycan_match):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j, n, m
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._calculate_score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_matched_fragment);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XDECREF(__pyx_v_matched_fragments);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_4_calculate_score, "GlycanCoarseScorerBase._calculate_score(self, CoarseGlycanMatch glycan_match) -> double");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score = {"_calculate_score", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_4_calculate_score};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_calculate_score (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_glycan_match,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_match)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_calculate_score") < 0)) __PYX_ERR(0, 356, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_glycan_match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_calculate_score", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 356, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._calculate_score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glycan_match), __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, 1, "glycan_match", 0))) __PYX_ERR(0, 356, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_4_calculate_score(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v_glycan_match);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_4_calculate_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_calculate_score", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__calculate_score(__pyx_v_self, __pyx_v_glycan_match, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._calculate_score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":381
 *         return score
 * 
 *     cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs *__pyx_optional_args) {
  double __pyx_v_mass_shift_tandem_mass = ((double)0.0);
  PyObject *__pyx_v_shifts = NULL;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_n_glycan_match_stubs", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_mass_shift_tandem_mass = __pyx_optional_args->mass_shift_tandem_mass;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_n_glycan_match_stubs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_peptide_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_mass_shift_tandem_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_scan, __pyx_t_3, ((PyObject *)__pyx_v_glycan_combination), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 4+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 381, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":384
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 *         shifts = glycan_combination.get_n_glycan_fragments()             # <<<<<<<<<<<<<<
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_glycan_combination->__pyx_vtab)->get_n_glycan_fragments(__pyx_v_glycan_combination, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shifts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":385
 *                                                   double mass_shift_tandem_mass=0.0):
 *         shifts = glycan_combination.get_n_glycan_fragments()
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)             # <<<<<<<<<<<<<<
 * 
 *     cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_match_fragments(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_shifts, 0, &__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":381
 *         return score
 * 
 *     cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._n_glycan_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shifts);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_6_n_glycan_match_stubs, "GlycanCoarseScorerBase._n_glycan_match_stubs(self, scan, double peptide_mass, GlycanCombinationRecordBase glycan_combination, double mass_shift_tandem_mass=0.0) -> CoarseGlycanMatch");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs = {"_n_glycan_match_stubs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_6_n_glycan_match_stubs};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_scan = 0;
  double __pyx_v_peptide_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination = 0;
  double __pyx_v_mass_shift_tandem_mass;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_n_glycan_match_stubs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scan,&__pyx_n_s_peptide_mass,&__pyx_n_s_glycan_combination,&__pyx_n_s_mass_shift_tandem_mass,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scan)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peptide_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_n_glycan_match_stubs", 0, 3, 4, 1); __PYX_ERR(0, 381, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_combination)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_n_glycan_match_stubs", 0, 3, 4, 2); __PYX_ERR(0, 381, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass_shift_tandem_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_n_glycan_match_stubs") < 0)) __PYX_ERR(0, 381, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_scan = values[0];
    __pyx_v_peptide_mass = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_peptide_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L3_error)
    __pyx_v_glycan_combination = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)values[2]);
    if (values[3]) {
      __pyx_v_mass_shift_tandem_mass = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_mass_shift_tandem_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 383, __pyx_L3_error)
    } else {
      __pyx_v_mass_shift_tandem_mass = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_n_glycan_match_stubs", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 381, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._n_glycan_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glycan_combination), __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, 1, "glycan_combination", 0))) __PYX_ERR(0, 382, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_6_n_glycan_match_stubs(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, __pyx_v_mass_shift_tandem_mass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_6_n_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, double __pyx_v_mass_shift_tandem_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_n_glycan_match_stubs", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->_n_glycan_match_stubs(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._n_glycan_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":387
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs *__pyx_optional_args) {
  double __pyx_v_mass_shift_tandem_mass = ((double)0.0);
  PyObject *__pyx_v_shifts = NULL;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_o_glycan_match_stubs", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_mass_shift_tandem_mass = __pyx_optional_args->mass_shift_tandem_mass;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_o_glycan_match_stubs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_peptide_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_mass_shift_tandem_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_scan, __pyx_t_3, ((PyObject *)__pyx_v_glycan_combination), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 4+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 387, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":390
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 *         shifts = glycan_combination.get_o_glycan_fragments()             # <<<<<<<<<<<<<<
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_glycan_combination->__pyx_vtab)->get_o_glycan_fragments(__pyx_v_glycan_combination, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shifts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":391
 *                                                   double mass_shift_tandem_mass=0.0):
 *         shifts = glycan_combination.get_o_glycan_fragments()
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)             # <<<<<<<<<<<<<<
 * 
 *     cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_match_fragments(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_shifts, 0, &__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":387
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._o_glycan_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shifts);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_8_o_glycan_match_stubs, "GlycanCoarseScorerBase._o_glycan_match_stubs(self, scan, double peptide_mass, GlycanCombinationRecordBase glycan_combination, double mass_shift_tandem_mass=0.0) -> CoarseGlycanMatch");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs = {"_o_glycan_match_stubs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_8_o_glycan_match_stubs};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_scan = 0;
  double __pyx_v_peptide_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination = 0;
  double __pyx_v_mass_shift_tandem_mass;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_o_glycan_match_stubs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scan,&__pyx_n_s_peptide_mass,&__pyx_n_s_glycan_combination,&__pyx_n_s_mass_shift_tandem_mass,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scan)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peptide_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_o_glycan_match_stubs", 0, 3, 4, 1); __PYX_ERR(0, 387, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_combination)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_o_glycan_match_stubs", 0, 3, 4, 2); __PYX_ERR(0, 387, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass_shift_tandem_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_o_glycan_match_stubs") < 0)) __PYX_ERR(0, 387, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_scan = values[0];
    __pyx_v_peptide_mass = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_peptide_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
    __pyx_v_glycan_combination = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)values[2]);
    if (values[3]) {
      __pyx_v_mass_shift_tandem_mass = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_mass_shift_tandem_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    } else {
      __pyx_v_mass_shift_tandem_mass = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_o_glycan_match_stubs", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 387, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._o_glycan_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glycan_combination), __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, 1, "glycan_combination", 0))) __PYX_ERR(0, 388, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_8_o_glycan_match_stubs(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, __pyx_v_mass_shift_tandem_mass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_8_o_glycan_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, double __pyx_v_mass_shift_tandem_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_o_glycan_match_stubs", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->_o_glycan_match_stubs(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._o_glycan_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":393
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                              GlycanCombinationRecordBase glycan_combination,
 *                                              double mass_shift_tandem_mass=0.0):
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs *__pyx_optional_args) {
  double __pyx_v_mass_shift_tandem_mass = ((double)0.0);
  PyObject *__pyx_v_shifts = NULL;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_gag_match_stubs", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_mass_shift_tandem_mass = __pyx_optional_args->mass_shift_tandem_mass;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gag_match_stubs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_peptide_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_mass_shift_tandem_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_scan, __pyx_t_3, ((PyObject *)__pyx_v_glycan_combination), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 4+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 393, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":396
 *                                              GlycanCombinationRecordBase glycan_combination,
 *                                              double mass_shift_tandem_mass=0.0):
 *         shifts = glycan_combination.get_gag_linker_glycan_fragments()             # <<<<<<<<<<<<<<
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_glycan_combination->__pyx_vtab)->get_gag_linker_glycan_fragments(__pyx_v_glycan_combination, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shifts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":397
 *                                              double mass_shift_tandem_mass=0.0):
 *         shifts = glycan_combination.get_gag_linker_glycan_fragments()
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)             # <<<<<<<<<<<<<<
 * 
 *     cdef CoarseGlycanMatch _get_n_glycan_coarse_score(self, object scan,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_match_fragments(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_shifts, 0, &__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":393
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                              GlycanCombinationRecordBase glycan_combination,
 *                                              double mass_shift_tandem_mass=0.0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._gag_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shifts);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_10_gag_match_stubs, "GlycanCoarseScorerBase._gag_match_stubs(self, scan, double peptide_mass, GlycanCombinationRecordBase glycan_combination, double mass_shift_tandem_mass=0.0) -> CoarseGlycanMatch");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs = {"_gag_match_stubs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_10_gag_match_stubs};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_scan = 0;
  double __pyx_v_peptide_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination = 0;
  double __pyx_v_mass_shift_tandem_mass;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_gag_match_stubs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scan,&__pyx_n_s_peptide_mass,&__pyx_n_s_glycan_combination,&__pyx_n_s_mass_shift_tandem_mass,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scan)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peptide_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_gag_match_stubs", 0, 3, 4, 1); __PYX_ERR(0, 393, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_combination)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_gag_match_stubs", 0, 3, 4, 2); __PYX_ERR(0, 393, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass_shift_tandem_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_gag_match_stubs") < 0)) __PYX_ERR(0, 393, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_scan = values[0];
    __pyx_v_peptide_mass = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_peptide_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L3_error)
    __pyx_v_glycan_combination = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)values[2]);
    if (values[3]) {
      __pyx_v_mass_shift_tandem_mass = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_mass_shift_tandem_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L3_error)
    } else {
      __pyx_v_mass_shift_tandem_mass = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_gag_match_stubs", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 393, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._gag_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glycan_combination), __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, 1, "glycan_combination", 0))) __PYX_ERR(0, 394, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_10_gag_match_stubs(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, __pyx_v_mass_shift_tandem_mass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_10_gag_match_stubs(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_peptide_mass, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, double __pyx_v_mass_shift_tandem_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_gag_match_stubs", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.mass_shift_tandem_mass = __pyx_v_mass_shift_tandem_mass;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->_gag_match_stubs(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._gag_match_stubs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":399
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cdef CoarseGlycanMatch _get_n_glycan_coarse_score(self, object scan,             # <<<<<<<<<<<<<<
 *                                                       GlycanCombinationRecordBase glycan_combination,
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_n_glycan_coarse_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, double __pyx_v_peptide_mass, double *__pyx_v_scoreout) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match = NULL;
  double __pyx_v_score;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs __pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_n_glycan_coarse_score", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":403
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 *                                                       double* scoreout):
 *         if peptide_mass < 0:             # <<<<<<<<<<<<<<
 *             scoreout[0] = -1e6
 *             return None
 */
  __pyx_t_1 = (__pyx_v_peptide_mass < 0.0);
  if (__pyx_t_1) {

    /* "glycresoft/_c/tandem/core_search.pyx":404
 *                                                       double* scoreout):
 *         if peptide_mass < 0:
 *             scoreout[0] = -1e6             # <<<<<<<<<<<<<<
 *             return None
 *         glycan_match = self._n_glycan_match_stubs(
 */
    (__pyx_v_scoreout[0]) = -1e6;

    /* "glycresoft/_c/tandem/core_search.pyx":405
 *         if peptide_mass < 0:
 *             scoreout[0] = -1e6
 *             return None             # <<<<<<<<<<<<<<
 *         glycan_match = self._n_glycan_match_stubs(
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":403
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 *                                                       double* scoreout):
 *         if peptide_mass < 0:             # <<<<<<<<<<<<<<
 *             scoreout[0] = -1e6
 *             return None
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":406
 *             scoreout[0] = -1e6
 *             return None
 *         glycan_match = self._n_glycan_match_stubs(             # <<<<<<<<<<<<<<
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.mass_shift_tandem_mass = __pyx_v_mass_shift->tandem_mass;
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_n_glycan_match_stubs(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, 0, &__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_glycan_match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":408
 *         glycan_match = self._n_glycan_match_stubs(
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)             # <<<<<<<<<<<<<<
 *         scoreout[0] = score
 *         return glycan_match
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_calculate_score(__pyx_v_self, __pyx_v_glycan_match, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 408, __pyx_L1_error)
  __pyx_v_score = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":409
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)
 *         scoreout[0] = score             # <<<<<<<<<<<<<<
 *         return glycan_match
 * 
 */
  (__pyx_v_scoreout[0]) = __pyx_v_score;

  /* "glycresoft/_c/tandem/core_search.pyx":410
 *         score = self._calculate_score(glycan_match)
 *         scoreout[0] = score
 *         return glycan_match             # <<<<<<<<<<<<<<
 * 
 *     cdef CoarseGlycanMatch _get_o_glycan_coarse_score(self, object scan,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_glycan_match);
  __pyx_r = __pyx_v_glycan_match;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":399
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cdef CoarseGlycanMatch _get_n_glycan_coarse_score(self, object scan,             # <<<<<<<<<<<<<<
 *                                                       GlycanCombinationRecordBase glycan_combination,
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._get_n_glycan_coarse_score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_glycan_match);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":412
 *         return glycan_match
 * 
 *     cdef CoarseGlycanMatch _get_o_glycan_coarse_score(self, object scan,             # <<<<<<<<<<<<<<
 *                                                       GlycanCombinationRecordBase glycan_combination,
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_o_glycan_coarse_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, double __pyx_v_peptide_mass, double *__pyx_v_scoreout) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match = NULL;
  double __pyx_v_score;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs __pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_o_glycan_coarse_score", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":416
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 *                                                       double* scoreout):
 *         if peptide_mass < 0:             # <<<<<<<<<<<<<<
 *             scoreout[0] = -1e6
 *             return None
 */
  __pyx_t_1 = (__pyx_v_peptide_mass < 0.0);
  if (__pyx_t_1) {

    /* "glycresoft/_c/tandem/core_search.pyx":417
 *                                                       double* scoreout):
 *         if peptide_mass < 0:
 *             scoreout[0] = -1e6             # <<<<<<<<<<<<<<
 *             return None
 *         glycan_match = self._o_glycan_match_stubs(
 */
    (__pyx_v_scoreout[0]) = -1e6;

    /* "glycresoft/_c/tandem/core_search.pyx":418
 *         if peptide_mass < 0:
 *             scoreout[0] = -1e6
 *             return None             # <<<<<<<<<<<<<<
 *         glycan_match = self._o_glycan_match_stubs(
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":416
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 *                                                       double* scoreout):
 *         if peptide_mass < 0:             # <<<<<<<<<<<<<<
 *             scoreout[0] = -1e6
 *             return None
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":419
 *             scoreout[0] = -1e6
 *             return None
 *         glycan_match = self._o_glycan_match_stubs(             # <<<<<<<<<<<<<<
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.mass_shift_tandem_mass = __pyx_v_mass_shift->tandem_mass;
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_o_glycan_match_stubs(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, 0, &__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_glycan_match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":421
 *         glycan_match = self._o_glycan_match_stubs(
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)             # <<<<<<<<<<<<<<
 *         scoreout[0] = score
 *         return glycan_match
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_calculate_score(__pyx_v_self, __pyx_v_glycan_match, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 421, __pyx_L1_error)
  __pyx_v_score = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":422
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)
 *         scoreout[0] = score             # <<<<<<<<<<<<<<
 *         return glycan_match
 * 
 */
  (__pyx_v_scoreout[0]) = __pyx_v_score;

  /* "glycresoft/_c/tandem/core_search.pyx":423
 *         score = self._calculate_score(glycan_match)
 *         scoreout[0] = score
 *         return glycan_match             # <<<<<<<<<<<<<<
 * 
 *     cdef CoarseGlycanMatch _get_gag_coarse_score(self, object scan,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_glycan_match);
  __pyx_r = __pyx_v_glycan_match;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":412
 *         return glycan_match
 * 
 *     cdef CoarseGlycanMatch _get_o_glycan_coarse_score(self, object scan,             # <<<<<<<<<<<<<<
 *                                                       GlycanCombinationRecordBase glycan_combination,
 *                                                       MassShiftBase mass_shift, double peptide_mass,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._get_o_glycan_coarse_score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_glycan_match);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":425
 *         return glycan_match
 * 
 *     cdef CoarseGlycanMatch _get_gag_coarse_score(self, object scan,             # <<<<<<<<<<<<<<
 *                                                  GlycanCombinationRecordBase glycan_combination,
 *                                                  MassShiftBase mass_shift, double peptide_mass,
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_gag_coarse_score(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, double __pyx_v_peptide_mass, double *__pyx_v_scoreout) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_glycan_match = NULL;
  double __pyx_v_score;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs __pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_gag_coarse_score", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":429
 *                                                  MassShiftBase mass_shift, double peptide_mass,
 *                                                  double* scoreout):
 *         if peptide_mass < 0:             # <<<<<<<<<<<<<<
 *             scoreout[0] = -1e6
 *             return None
 */
  __pyx_t_1 = (__pyx_v_peptide_mass < 0.0);
  if (__pyx_t_1) {

    /* "glycresoft/_c/tandem/core_search.pyx":430
 *                                                  double* scoreout):
 *         if peptide_mass < 0:
 *             scoreout[0] = -1e6             # <<<<<<<<<<<<<<
 *             return None
 *         glycan_match = self._gag_match_stubs(
 */
    (__pyx_v_scoreout[0]) = -1e6;

    /* "glycresoft/_c/tandem/core_search.pyx":431
 *         if peptide_mass < 0:
 *             scoreout[0] = -1e6
 *             return None             # <<<<<<<<<<<<<<
 *         glycan_match = self._gag_match_stubs(
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":429
 *                                                  MassShiftBase mass_shift, double peptide_mass,
 *                                                  double* scoreout):
 *         if peptide_mass < 0:             # <<<<<<<<<<<<<<
 *             scoreout[0] = -1e6
 *             return None
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":432
 *             scoreout[0] = -1e6
 *             return None
 *         glycan_match = self._gag_match_stubs(             # <<<<<<<<<<<<<<
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.mass_shift_tandem_mass = __pyx_v_mass_shift->tandem_mass;
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_gag_match_stubs(__pyx_v_self, __pyx_v_scan, __pyx_v_peptide_mass, __pyx_v_glycan_combination, 0, &__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_glycan_match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":434
 *         glycan_match = self._gag_match_stubs(
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)             # <<<<<<<<<<<<<<
 *         scoreout[0] = score
 *         return glycan_match
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_calculate_score(__pyx_v_self, __pyx_v_glycan_match, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 434, __pyx_L1_error)
  __pyx_v_score = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":435
 *             scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
 *         score = self._calculate_score(glycan_match)
 *         scoreout[0] = score             # <<<<<<<<<<<<<<
 *         return glycan_match
 * 
 */
  (__pyx_v_scoreout[0]) = __pyx_v_score;

  /* "glycresoft/_c/tandem/core_search.pyx":436
 *         score = self._calculate_score(glycan_match)
 *         scoreout[0] = score
 *         return glycan_match             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_glycan_match);
  __pyx_r = __pyx_v_glycan_match;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":425
 *         return glycan_match
 * 
 *     cdef CoarseGlycanMatch _get_gag_coarse_score(self, object scan,             # <<<<<<<<<<<<<<
 *                                                  GlycanCombinationRecordBase glycan_combination,
 *                                                  MassShiftBase mass_shift, double peptide_mass,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase._get_gag_coarse_score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_glycan_match);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":293
 * cdef class GlycanCoarseScorerBase(object):
 *     cdef:
 *         public double product_error_tolerance             # <<<<<<<<<<<<<<
 *         public double fragment_weight
 *         public double core_weight
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->product_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.product_error_tolerance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_v_self->product_error_tolerance = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.product_error_tolerance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":294
 *     cdef:
 *         public double product_error_tolerance
 *         public double fragment_weight             # <<<<<<<<<<<<<<
 *         public double core_weight
 *         public bint recalibrate_peptide_mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fragment_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.fragment_weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L1_error)
  __pyx_v_self->fragment_weight = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.fragment_weight.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":295
 *         public double product_error_tolerance
 *         public double fragment_weight
 *         public double core_weight             # <<<<<<<<<<<<<<
 *         public bint recalibrate_peptide_mass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.core_weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_v_self->core_weight = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.core_weight.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":296
 *         public double fragment_weight
 *         public double core_weight
 *         public bint recalibrate_peptide_mass             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->recalibrate_peptide_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.recalibrate_peptide_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_v_self->recalibrate_peptide_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.recalibrate_peptide_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_12__reduce_cython__, "GlycanCoarseScorerBase.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_12__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_12__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_12__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.core_weight, self.fragment_weight, self.product_error_tolerance, self.recalibrate_peptide_mass)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->fragment_weight); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->product_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->recalibrate_peptide_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.core_weight, self.fragment_weight, self.product_error_tolerance, self.recalibrate_peptide_mass)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.core_weight, self.fragment_weight, self.product_error_tolerance, self.recalibrate_peptide_mass)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.core_weight, self.fragment_weight, self.product_error_tolerance, self.recalibrate_peptide_mass)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_GlycanCoarseScore); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_30049220);
    __Pyx_GIVEREF(__pyx_int_30049220);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_30049220)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, None), state
 *     else:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCoarseScorerBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_GlycanCoarseScore); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_30049220);
    __Pyx_GIVEREF(__pyx_int_30049220);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_30049220)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCoarseScorerBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_14__setstate_cython__, "GlycanCoarseScorerBase.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_14__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_14__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_14__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCoarseScorerBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCoarseScorerBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCoarseScorerBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanCoarseScorerBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":439
 * 
 * 
 * cdef bint isclose(double a, double b, double rtol=1e-05, double atol=1e-08):             # <<<<<<<<<<<<<<
 *     return abs(a - b) <= atol + rtol * abs(b)
 * 
 */

static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_isclose(double __pyx_v_a, double __pyx_v_b, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_isclose *__pyx_optional_args) {
  double __pyx_v_rtol = ((double)1e-05);
  double __pyx_v_atol = ((double)1e-08);
  int __pyx_r;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_rtol = __pyx_optional_args->rtol;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_atol = __pyx_optional_args->atol;
      }
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":440
 * 
 * cdef bint isclose(double a, double b, double rtol=1e-05, double atol=1e-08):
 *     return abs(a - b) <= atol + rtol * abs(b)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (fabs((__pyx_v_a - __pyx_v_b)) <= (__pyx_v_atol + (__pyx_v_rtol * fabs(__pyx_v_b))));
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":439
 * 
 * 
 * cdef bint isclose(double a, double b, double rtol=1e-05, double atol=1e-08):             # <<<<<<<<<<<<<<
 *     return abs(a - b) <= atol + rtol * abs(b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":455
 * 
 *     @staticmethod
 *     cdef GlycanMatchResult _create(double peptide_mass, double score, CoarseGlycanMatch match, size_t glycan_size,             # <<<<<<<<<<<<<<
 *                                    dict glycan_types, double recalibrated_peptide_mass):
 *         cdef:
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_f_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult__create(double __pyx_v_peptide_mass, double __pyx_v_score, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_match, size_t __pyx_v_glycan_size, PyObject *__pyx_v_glycan_types, double __pyx_v_recalibrated_peptide_mass) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":460
 *             GlycanMatchResult self
 * 
 *         self = GlycanMatchResult.__new__(GlycanMatchResult)             # <<<<<<<<<<<<<<
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":461
 * 
 *         self = GlycanMatchResult.__new__(GlycanMatchResult)
 *         self.peptide_mass = peptide_mass             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.match = match
 */
  __pyx_v_self->peptide_mass = __pyx_v_peptide_mass;

  /* "glycresoft/_c/tandem/core_search.pyx":462
 *         self = GlycanMatchResult.__new__(GlycanMatchResult)
 *         self.peptide_mass = peptide_mass
 *         self.score = score             # <<<<<<<<<<<<<<
 *         self.match = match
 *         self.glycan_size = glycan_size
 */
  __pyx_v_self->score = __pyx_v_score;

  /* "glycresoft/_c/tandem/core_search.pyx":463
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 *         self.match = match             # <<<<<<<<<<<<<<
 *         self.glycan_size = glycan_size
 *         self.glycan_types = glycan_types
 */
  __Pyx_INCREF((PyObject *)__pyx_v_match);
  __Pyx_GIVEREF((PyObject *)__pyx_v_match);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = __pyx_v_match;

  /* "glycresoft/_c/tandem/core_search.pyx":464
 *         self.score = score
 *         self.match = match
 *         self.glycan_size = glycan_size             # <<<<<<<<<<<<<<
 *         self.glycan_types = glycan_types
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 */
  __pyx_v_self->glycan_size = __pyx_v_glycan_size;

  /* "glycresoft/_c/tandem/core_search.pyx":465
 *         self.match = match
 *         self.glycan_size = glycan_size
 *         self.glycan_types = glycan_types             # <<<<<<<<<<<<<<
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 *         return self
 */
  __Pyx_INCREF(__pyx_v_glycan_types);
  __Pyx_GIVEREF(__pyx_v_glycan_types);
  __Pyx_GOTREF(__pyx_v_self->glycan_types);
  __Pyx_DECREF(__pyx_v_self->glycan_types);
  __pyx_v_self->glycan_types = __pyx_v_glycan_types;

  /* "glycresoft/_c/tandem/core_search.pyx":466
 *         self.glycan_size = glycan_size
 *         self.glycan_types = glycan_types
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_v_self->recalibrated_peptide_mass = __pyx_v_recalibrated_peptide_mass;

  /* "glycresoft/_c/tandem/core_search.pyx":467
 *         self.glycan_types = glycan_types
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":455
 * 
 *     @staticmethod
 *     cdef GlycanMatchResult _create(double peptide_mass, double score, CoarseGlycanMatch match, size_t glycan_size,             # <<<<<<<<<<<<<<
 *                                    dict glycan_types, double recalibrated_peptide_mass):
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":469
 *         return self
 * 
 *     def __init__(self, peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass):             # <<<<<<<<<<<<<<
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peptide_mass = 0;
  PyObject *__pyx_v_score = 0;
  PyObject *__pyx_v_match = 0;
  PyObject *__pyx_v_glycan_size = 0;
  PyObject *__pyx_v_glycan_types = 0;
  PyObject *__pyx_v_recalibrated_peptide_mass = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peptide_mass,&__pyx_n_s_score,&__pyx_n_s_match,&__pyx_n_s_glycan_size,&__pyx_n_s_glycan_types,&__pyx_n_s_recalibrated_peptide_mass,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peptide_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_score)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 1); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_match)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 2); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_size)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 3); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_types)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 4); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_recalibrated_peptide_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[5]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 5); __PYX_ERR(0, 469, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 469, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
      values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
    }
    __pyx_v_peptide_mass = values[0];
    __pyx_v_score = values[1];
    __pyx_v_match = values[2];
    __pyx_v_glycan_size = values[3];
    __pyx_v_glycan_types = values[4];
    __pyx_v_recalibrated_peptide_mass = values[5];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 469, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), __pyx_v_peptide_mass, __pyx_v_score, __pyx_v_match, __pyx_v_glycan_size, __pyx_v_glycan_types, __pyx_v_recalibrated_peptide_mass);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_peptide_mass, PyObject *__pyx_v_score, PyObject *__pyx_v_match, PyObject *__pyx_v_glycan_size, PyObject *__pyx_v_glycan_types, PyObject *__pyx_v_recalibrated_peptide_mass) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":470
 * 
 *     def __init__(self, peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass):
 *         self.peptide_mass = peptide_mass             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.match = match
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_peptide_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_v_self->peptide_mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":471
 *     def __init__(self, peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass):
 *         self.peptide_mass = peptide_mass
 *         self.score = score             # <<<<<<<<<<<<<<
 *         self.match = match
 *         self.glycan_size = glycan_size
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_score); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 471, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":472
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 *         self.match = match             # <<<<<<<<<<<<<<
 *         self.glycan_size = glycan_size
 *         self.glycan_types = glycan_types
 */
  if (!(likely(((__pyx_v_match) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_match, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 472, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_match;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":473
 *         self.score = score
 *         self.match = match
 *         self.glycan_size = glycan_size             # <<<<<<<<<<<<<<
 *         self.glycan_types = glycan_types
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 */
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_v_glycan_size); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)
  __pyx_v_self->glycan_size = __pyx_t_3;

  /* "glycresoft/_c/tandem/core_search.pyx":474
 *         self.match = match
 *         self.glycan_size = glycan_size
 *         self.glycan_types = glycan_types             # <<<<<<<<<<<<<<
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 * 
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_glycan_types))||((__pyx_v_glycan_types) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_glycan_types))) __PYX_ERR(0, 474, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_glycan_types;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->glycan_types);
  __Pyx_DECREF(__pyx_v_self->glycan_types);
  __pyx_v_self->glycan_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":475
 *         self.glycan_size = glycan_size
 *         self.glycan_types = glycan_types
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_recalibrated_peptide_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 475, __pyx_L1_error)
  __pyx_v_self->recalibrated_peptide_mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":469
 *         return self
 * 
 *     def __init__(self, peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass):             # <<<<<<<<<<<<<<
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":477
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fragment_match_count(self):
 *         match = self.match
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_20fragment_match_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_20fragment_match_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_20fragment_match_count___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_20fragment_match_count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_match = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":479
 *     @property
 *     def fragment_match_count(self):
 *         match = self.match             # <<<<<<<<<<<<<<
 *         if match is None:
 *             return 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->match);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":480
 *     def fragment_match_count(self):
 *         match = self.match
 *         if match is None:             # <<<<<<<<<<<<<<
 *             return 0
 *         return match.n_matched
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_match) == Py_None);
  if (__pyx_t_2) {

    /* "glycresoft/_c/tandem/core_search.pyx":481
 *         match = self.match
 *         if match is None:
 *             return 0             # <<<<<<<<<<<<<<
 *         return match.n_matched
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":480
 *     def fragment_match_count(self):
 *         match = self.match
 *         if match is None:             # <<<<<<<<<<<<<<
 *             return 0
 *         return match.n_matched
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":482
 *         if match is None:
 *             return 0
 *         return match.n_matched             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_match->n_matched); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":477
 *         self.recalibrated_peptide_mass = recalibrated_peptide_mass
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fragment_match_count(self):
 *         match = self.match
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.fragment_match_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_match);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":484
 *         return match.n_matched
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = (
 *             "{self.__class__.__name__}({self.peptide_mass}, {self.score},"
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_2__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_2__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_v_template = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":486
 *     def __repr__(self):
 *         template = (
 *             "{self.__class__.__name__}({self.peptide_mass}, {self.score},"             # <<<<<<<<<<<<<<
 *             " {self.match}, {self.glycan_size}, {self.glycan_types}, {self.recalibrated_peptide_mass})")
 *         return template.format(self=self)
 */
  __Pyx_INCREF(__pyx_kp_s_self___class_____name___self_pe);
  __pyx_v_template = __pyx_kp_s_self___class_____name___self_pe;

  /* "glycresoft/_c/tandem/core_search.pyx":488
 *             "{self.__class__.__name__}({self.peptide_mass}, {self.score},"
 *             " {self.match}, {self.glycan_size}, {self.glycan_types}, {self.recalibrated_peptide_mass})")
 *         return template.format(self=self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_template, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 488, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":484
 *         return match.n_matched
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = (
 *             "{self.__class__.__name__}({self.peptide_mass}, {self.score},"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_template);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":446
 * cdef class GlycanMatchResult(object):
 *     cdef:
 *         public double peptide_mass             # <<<<<<<<<<<<<<
 *         public double score
 *         public CoarseGlycanMatch match
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->peptide_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.peptide_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L1_error)
  __pyx_v_self->peptide_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.peptide_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":447
 *     cdef:
 *         public double peptide_mass
 *         public double score             # <<<<<<<<<<<<<<
 *         public CoarseGlycanMatch match
 *         public size_t glycan_size
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":448
 *         public double peptide_mass
 *         public double score
 *         public CoarseGlycanMatch match             # <<<<<<<<<<<<<<
 *         public size_t glycan_size
 *         public dict glycan_types
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->match);
  __pyx_r = ((PyObject *)__pyx_v_self->match);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 448, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.match.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":449
 *         public double score
 *         public CoarseGlycanMatch match
 *         public size_t glycan_size             # <<<<<<<<<<<<<<
 *         public dict glycan_types
 *         public double recalibrated_peptide_mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->glycan_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.glycan_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)
  __pyx_v_self->glycan_size = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.glycan_size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":450
 *         public CoarseGlycanMatch match
 *         public size_t glycan_size
 *         public dict glycan_types             # <<<<<<<<<<<<<<
 *         public double recalibrated_peptide_mass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->glycan_types);
  __pyx_r = __pyx_v_self->glycan_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 450, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->glycan_types);
  __Pyx_DECREF(__pyx_v_self->glycan_types);
  __pyx_v_self->glycan_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.glycan_types.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->glycan_types);
  __Pyx_DECREF(__pyx_v_self->glycan_types);
  __pyx_v_self->glycan_types = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":451
 *         public size_t glycan_size
 *         public dict glycan_types
 *         public double recalibrated_peptide_mass             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->recalibrated_peptide_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.recalibrated_peptide_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_v_self->recalibrated_peptide_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.recalibrated_peptide_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_4__reduce_cython__, "GlycanMatchResult.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_4__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_4__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_4__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.glycan_size, self.glycan_types, self.match, self.peptide_mass, self.recalibrated_peptide_mass, self.score)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->glycan_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->peptide_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->recalibrated_peptide_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->glycan_types);
  __Pyx_GIVEREF(__pyx_v_self->glycan_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->glycan_types)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->match);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->match);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, ((PyObject *)__pyx_v_self->match))) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.glycan_size, self.glycan_types, self.match, self.peptide_mass, self.recalibrated_peptide_mass, self.score)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.glycan_size, self.glycan_types, self.match, self.peptide_mass, self.recalibrated_peptide_mass, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.glycan_types is not None or self.match is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.glycan_size, self.glycan_types, self.match, self.peptide_mass, self.recalibrated_peptide_mass, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.glycan_types is not None or self.match is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, None), state
 */
  /*else*/ {
    __pyx_t_7 = (__pyx_v_self->glycan_types != ((PyObject*)Py_None));
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (((PyObject *)__pyx_v_self->match) != Py_None);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.glycan_types is not None or self.match is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.glycan_types is not None or self.match is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_GlycanMatchResult); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_143121537);
    __Pyx_GIVEREF(__pyx_int_143121537);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_143121537)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.glycan_types is not None or self.match is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, None), state
 *     else:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanMatchResult__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_GlycanMatchResult); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_143121537);
    __Pyx_GIVEREF(__pyx_int_143121537);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_143121537)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanMatchResult__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_6__setstate_cython__, "GlycanMatchResult.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_6__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_6__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_6__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanMatchResult__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanMatchResult__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanMatchResult__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanMatchResult.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":499
 * 
 *     @staticmethod
 *     cdef ScoreCoarseGlycanMatchPair _create(double score, CoarseGlycanMatch match):             # <<<<<<<<<<<<<<
 *         cdef ScoreCoarseGlycanMatchPair self = ScoreCoarseGlycanMatchPair.__new__(ScoreCoarseGlycanMatchPair)
 *         self.score = score
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create(double __pyx_v_score, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_match) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":500
 *     @staticmethod
 *     cdef ScoreCoarseGlycanMatchPair _create(double score, CoarseGlycanMatch match):
 *         cdef ScoreCoarseGlycanMatchPair self = ScoreCoarseGlycanMatchPair.__new__(ScoreCoarseGlycanMatchPair)             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.match = match
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":501
 *     cdef ScoreCoarseGlycanMatchPair _create(double score, CoarseGlycanMatch match):
 *         cdef ScoreCoarseGlycanMatchPair self = ScoreCoarseGlycanMatchPair.__new__(ScoreCoarseGlycanMatchPair)
 *         self.score = score             # <<<<<<<<<<<<<<
 *         self.match = match
 *         return self
 */
  __pyx_v_self->score = __pyx_v_score;

  /* "glycresoft/_c/tandem/core_search.pyx":502
 *         cdef ScoreCoarseGlycanMatchPair self = ScoreCoarseGlycanMatchPair.__new__(ScoreCoarseGlycanMatchPair)
 *         self.score = score
 *         self.match = match             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_match);
  __Pyx_GIVEREF((PyObject *)__pyx_v_match);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = __pyx_v_match;

  /* "glycresoft/_c/tandem/core_search.pyx":503
 *         self.score = score
 *         self.match = match
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, score, match):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":499
 * 
 *     @staticmethod
 *     cdef ScoreCoarseGlycanMatchPair _create(double score, CoarseGlycanMatch match):             # <<<<<<<<<<<<<<
 *         cdef ScoreCoarseGlycanMatchPair self = ScoreCoarseGlycanMatchPair.__new__(ScoreCoarseGlycanMatchPair)
 *         self.score = score
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":505
 *         return self
 * 
 *     def __init__(self, score, match):             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.match = match
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_score = 0;
  PyObject *__pyx_v_match = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_score,&__pyx_n_s_match,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_score)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_match)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 505, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 505, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_score = values[0];
    __pyx_v_match = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 505, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self), __pyx_v_score, __pyx_v_match);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_score, PyObject *__pyx_v_match) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":506
 * 
 *     def __init__(self, score, match):
 *         self.score = score             # <<<<<<<<<<<<<<
 *         self.match = match
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_score); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 506, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":507
 *     def __init__(self, score, match):
 *         self.score = score
 *         self.match = match             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */
  if (!(likely(((__pyx_v_match) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_match, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 507, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_match;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":505
 *         return self
 * 
 *     def __init__(self, score, match):             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.match = match
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":509
 *         self.match = match
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             return self.score
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_2__getitem__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self), ((PyObject *)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_2__getitem__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":510
 * 
 *     def __getitem__(self, i):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             return self.score
 *         elif i == 1:
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_i, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 510, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "glycresoft/_c/tandem/core_search.pyx":511
 *     def __getitem__(self, i):
 *         if i == 0:
 *             return self.score             # <<<<<<<<<<<<<<
 *         elif i == 1:
 *             return self.match
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":510
 * 
 *     def __getitem__(self, i):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             return self.score
 *         elif i == 1:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":512
 *         if i == 0:
 *             return self.score
 *         elif i == 1:             # <<<<<<<<<<<<<<
 *             return self.match
 *         else:
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_i, __pyx_int_1, 1, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 512, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "glycresoft/_c/tandem/core_search.pyx":513
 *             return self.score
 *         elif i == 1:
 *             return self.match             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError(i)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->match);
    __pyx_r = ((PyObject *)__pyx_v_self->match);
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":512
 *         if i == 0:
 *             return self.score
 *         elif i == 1:             # <<<<<<<<<<<<<<
 *             return self.match
 *         else:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":515
 *             return self.match
 *         else:
 *             raise IndexError(i)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 515, __pyx_L1_error)
  }

  /* "glycresoft/_c/tandem/core_search.pyx":509
 *         self.match = match
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             return self.score
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":517
 *             raise IndexError(i)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return 2
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_4__len__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_4__len__(CYTHON_UNUSED struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "glycresoft/_c/tandem/core_search.pyx":518
 * 
 *     def __len__(self):
 *         return 2             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 2;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":517
 *             raise IndexError(i)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return 2
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":520
 *         return 2
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "(%f, %r)" % (self.score, self.match)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_6__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_6__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":521
 * 
 *     def __repr__(self):
 *         return "(%f, %r)" % (self.score, self.match)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->match);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->match);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->match))) __PYX_ERR(0, 521, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_f_r, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":520
 *         return 2
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "(%f, %r)" % (self.score, self.match)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":495
 * cdef class ScoreCoarseGlycanMatchPair(object):
 *     cdef:
 *         public double score             # <<<<<<<<<<<<<<
 *         public CoarseGlycanMatch match
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 495, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":496
 *     cdef:
 *         public double score
 *         public CoarseGlycanMatch match             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->match);
  __pyx_r = ((PyObject *)__pyx_v_self->match);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(0, 496, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.match.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->match);
  __Pyx_DECREF((PyObject *)__pyx_v_self->match);
  __pyx_v_self->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_8__reduce_cython__, "ScoreCoarseGlycanMatchPair.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_8__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_8__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_8__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.match, self.score)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->match);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->match);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->match))) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.match, self.score)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.match, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.match is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.match, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.match is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, None), state
 */
  /*else*/ {
    __pyx_t_3 = (((PyObject *)__pyx_v_self->match) != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.match is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.match is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_ScoreCoarseGlycan); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_88044796);
    __Pyx_GIVEREF(__pyx_int_88044796);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_88044796)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.match is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, None), state
 *     else:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_ScoreCoarseGlycan); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_88044796);
    __Pyx_GIVEREF(__pyx_int_88044796);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_88044796)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_10__setstate_cython__, "ScoreCoarseGlycanMatchPair.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_10__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_10__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_10__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":524
 * 
 * 
 * cpdef double score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<GlycanMatchResult>x).score
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_1score_getter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_score_getter(PyObject *__pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_r;

  /* "glycresoft/_c/tandem/core_search.pyx":525
 * 
 * cpdef double score_getter(object x):
 *     return (<GlycanMatchResult>x).score             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v_x)->score;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":524
 * 
 * 
 * cpdef double score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<GlycanMatchResult>x).score
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_1score_getter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_score_getter, "score_getter(x) -> double");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_1score_getter = {"score_getter", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_1score_getter, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_score_getter};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_1score_getter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("score_getter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "score_getter") < 0)) __PYX_ERR(0, 524, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_x = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("score_getter", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 524, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.score_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_score_getter(__pyx_self, __pyx_v_x);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_score_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("score_getter", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_score_getter(__pyx_v_x, 0); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.score_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":530
 * @cython.boundscheck(False)
 * @cython.binding(True)
 * cpdef GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list output, glycan_combination_db
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_3GlycanFilteringPeptideMassEstimator_match(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match *__pyx_optional_args) {
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift = __pyx_k__4;
  PyObject *__pyx_v_query_mass = ((PyObject *)Py_None);
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_glycan_combination_db = 0;
  double __pyx_v_intact_mass;
  double __pyx_v_threshold_mass;
  CYTHON_UNUSED double __pyx_v_last_peptide_mass;
  double __pyx_v_peptide_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_glycan_combination = 0;
  PyObject *__pyx_v_type_to_score = 0;
  double __pyx_v_best_score;
  double __pyx_v_score;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_best_match = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v_match = 0;
  double __pyx_v_recalibrated_peptide_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GlycanFilteringPeptideMassEstimator_match", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_mass_shift = __pyx_optional_args->mass_shift;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_query_mass = __pyx_optional_args->query_mass;
      }
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":541
 *         CoarseGlycanMatch best_match, match
 * 
 *     output = []             # <<<<<<<<<<<<<<
 *     if query_mass is None:
 *         intact_mass = scan.precursor_information.neutral_mass
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_output = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":542
 * 
 *     output = []
 *     if query_mass is None:             # <<<<<<<<<<<<<<
 *         intact_mass = scan.precursor_information.neutral_mass
 *     else:
 */
  __pyx_t_2 = (__pyx_v_query_mass == Py_None);
  if (__pyx_t_2) {

    /* "glycresoft/_c/tandem/core_search.pyx":543
 *     output = []
 *     if query_mass is None:
 *         intact_mass = scan.precursor_information.neutral_mass             # <<<<<<<<<<<<<<
 *     else:
 *         intact_mass = query_mass
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scan, __pyx_n_s_precursor_information); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_neutral_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_intact_mass = __pyx_t_4;

    /* "glycresoft/_c/tandem/core_search.pyx":542
 * 
 *     output = []
 *     if query_mass is None:             # <<<<<<<<<<<<<<
 *         intact_mass = scan.precursor_information.neutral_mass
 *     else:
 */
    goto __pyx_L3;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":545
 *         intact_mass = scan.precursor_information.neutral_mass
 *     else:
 *         intact_mass = query_mass             # <<<<<<<<<<<<<<
 *     threshold_mass = (intact_mass + 1) - self.minimum_peptide_mass
 *     last_peptide_mass = 0
 */
  /*else*/ {
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_query_mass); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 545, __pyx_L1_error)
    __pyx_v_intact_mass = __pyx_t_4;
  }
  __pyx_L3:;

  /* "glycresoft/_c/tandem/core_search.pyx":546
 *     else:
 *         intact_mass = query_mass
 *     threshold_mass = (intact_mass + 1) - self.minimum_peptide_mass             # <<<<<<<<<<<<<<
 *     last_peptide_mass = 0
 *     glycan_combination_db = self.glycan_combination_db
 */
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_intact_mass + 1.0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_minimum_peptide_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyNumber_Subtract(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_threshold_mass = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":547
 *         intact_mass = query_mass
 *     threshold_mass = (intact_mass + 1) - self.minimum_peptide_mass
 *     last_peptide_mass = 0             # <<<<<<<<<<<<<<
 *     glycan_combination_db = self.glycan_combination_db
 *     n = PyList_Size(glycan_combination_db)
 */
  __pyx_v_last_peptide_mass = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":548
 *     threshold_mass = (intact_mass + 1) - self.minimum_peptide_mass
 *     last_peptide_mass = 0
 *     glycan_combination_db = self.glycan_combination_db             # <<<<<<<<<<<<<<
 *     n = PyList_Size(glycan_combination_db)
 *     for i in range(n):
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_glycan_combination_db); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(PyList_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_5))) __PYX_ERR(0, 548, __pyx_L1_error)
  __pyx_v_glycan_combination_db = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":549
 *     last_peptide_mass = 0
 *     glycan_combination_db = self.glycan_combination_db
 *     n = PyList_Size(glycan_combination_db)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         glycan_combination = <GlycanCombinationRecordBase>PyList_GET_ITEM(glycan_combination_db, i)
 */
  __pyx_t_6 = PyList_Size(__pyx_v_glycan_combination_db); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 549, __pyx_L1_error)
  __pyx_v_n = __pyx_t_6;

  /* "glycresoft/_c/tandem/core_search.pyx":550
 *     glycan_combination_db = self.glycan_combination_db
 *     n = PyList_Size(glycan_combination_db)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         glycan_combination = <GlycanCombinationRecordBase>PyList_GET_ITEM(glycan_combination_db, i)
 *         # Stop searching when the peptide mass would be below the minimum peptide mass
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "glycresoft/_c/tandem/core_search.pyx":551
 *     n = PyList_Size(glycan_combination_db)
 *     for i in range(n):
 *         glycan_combination = <GlycanCombinationRecordBase>PyList_GET_ITEM(glycan_combination_db, i)             # <<<<<<<<<<<<<<
 *         # Stop searching when the peptide mass would be below the minimum peptide mass
 *         if threshold_mass < glycan_combination.dehydrated_mass:
 */
    __pyx_t_10 = PyList_GET_ITEM(__pyx_v_glycan_combination_db, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_10);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_glycan_combination, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":553
 *         glycan_combination = <GlycanCombinationRecordBase>PyList_GET_ITEM(glycan_combination_db, i)
 *         # Stop searching when the peptide mass would be below the minimum peptide mass
 *         if threshold_mass < glycan_combination.dehydrated_mass:             # <<<<<<<<<<<<<<
 *             break
 *         peptide_mass = (
 */
    __pyx_t_2 = (__pyx_v_threshold_mass < __pyx_v_glycan_combination->dehydrated_mass);
    if (__pyx_t_2) {

      /* "glycresoft/_c/tandem/core_search.pyx":554
 *         # Stop searching when the peptide mass would be below the minimum peptide mass
 *         if threshold_mass < glycan_combination.dehydrated_mass:
 *             break             # <<<<<<<<<<<<<<
 *         peptide_mass = (
 *             intact_mass - glycan_combination.dehydrated_mass
 */
      goto __pyx_L5_break;

      /* "glycresoft/_c/tandem/core_search.pyx":553
 *         glycan_combination = <GlycanCombinationRecordBase>PyList_GET_ITEM(glycan_combination_db, i)
 *         # Stop searching when the peptide mass would be below the minimum peptide mass
 *         if threshold_mass < glycan_combination.dehydrated_mass:             # <<<<<<<<<<<<<<
 *             break
 *         peptide_mass = (
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":557
 *         peptide_mass = (
 *             intact_mass - glycan_combination.dehydrated_mass
 *         ) - mass_shift.mass             # <<<<<<<<<<<<<<
 *         best_score = 0
 *         best_match = None
 */
    __pyx_v_peptide_mass = ((__pyx_v_intact_mass - __pyx_v_glycan_combination->dehydrated_mass) - __pyx_v_mass_shift->mass);

    /* "glycresoft/_c/tandem/core_search.pyx":558
 *             intact_mass - glycan_combination.dehydrated_mass
 *         ) - mass_shift.mass
 *         best_score = 0             # <<<<<<<<<<<<<<
 *         best_match = None
 *         type_to_score = {}
 */
    __pyx_v_best_score = 0.0;

    /* "glycresoft/_c/tandem/core_search.pyx":559
 *         ) - mass_shift.mass
 *         best_score = 0
 *         best_match = None             # <<<<<<<<<<<<<<
 *         type_to_score = {}
 *         if glycan_combination.is_n_glycan():
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_best_match, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None));

    /* "glycresoft/_c/tandem/core_search.pyx":560
 *         best_score = 0
 *         best_match = None
 *         type_to_score = {}             # <<<<<<<<<<<<<<
 *         if glycan_combination.is_n_glycan():
 *             score = 0
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_type_to_score, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":561
 *         best_match = None
 *         type_to_score = {}
 *         if glycan_combination.is_n_glycan():             # <<<<<<<<<<<<<<
 *             score = 0
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_glycan_combination->__pyx_vtab)->is_n_glycan(__pyx_v_glycan_combination, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 561, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "glycresoft/_c/tandem/core_search.pyx":562
 *         type_to_score = {}
 *         if glycan_combination.is_n_glycan():
 *             score = 0             # <<<<<<<<<<<<<<
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 */
      __pyx_v_score = 0.0;

      /* "glycresoft/_c/tandem/core_search.pyx":563
 *         if glycan_combination.is_n_glycan():
 *             score = 0
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)             # <<<<<<<<<<<<<<
 *             type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:
 */
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_get_n_glycan_coarse_score(__pyx_v_self, __pyx_v_scan, __pyx_v_glycan_combination, __pyx_v_mass_shift, __pyx_v_peptide_mass, (&__pyx_v_score))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_match, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":564
 *             score = 0
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)             # <<<<<<<<<<<<<<
 *             if score > best_score:
 *                 best_score = score
 */
      __pyx_t_5 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create(__pyx_v_score, __pyx_v_match)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((PyDict_SetItem(__pyx_v_type_to_score, __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan, __pyx_t_5) < 0))) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":565
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = score
 *                 best_match = match
 */
      __pyx_t_2 = (__pyx_v_score > __pyx_v_best_score);
      if (__pyx_t_2) {

        /* "glycresoft/_c/tandem/core_search.pyx":566
 *             type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:
 *                 best_score = score             # <<<<<<<<<<<<<<
 *                 best_match = match
 *         if glycan_combination.is_o_glycan():
 */
        __pyx_v_best_score = __pyx_v_score;

        /* "glycresoft/_c/tandem/core_search.pyx":567
 *             if score > best_score:
 *                 best_score = score
 *                 best_match = match             # <<<<<<<<<<<<<<
 *         if glycan_combination.is_o_glycan():
 *             score = 0
 */
        __Pyx_INCREF((PyObject *)__pyx_v_match);
        __Pyx_DECREF_SET(__pyx_v_best_match, __pyx_v_match);

        /* "glycresoft/_c/tandem/core_search.pyx":565
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = score
 *                 best_match = match
 */
      }

      /* "glycresoft/_c/tandem/core_search.pyx":561
 *         best_match = None
 *         type_to_score = {}
 *         if glycan_combination.is_n_glycan():             # <<<<<<<<<<<<<<
 *             score = 0
 *             match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":568
 *                 best_score = score
 *                 best_match = match
 *         if glycan_combination.is_o_glycan():             # <<<<<<<<<<<<<<
 *             score = 0
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_glycan_combination->__pyx_vtab)->is_o_glycan(__pyx_v_glycan_combination, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 568, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "glycresoft/_c/tandem/core_search.pyx":569
 *                 best_match = match
 *         if glycan_combination.is_o_glycan():
 *             score = 0             # <<<<<<<<<<<<<<
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 */
      __pyx_v_score = 0.0;

      /* "glycresoft/_c/tandem/core_search.pyx":570
 *         if glycan_combination.is_o_glycan():
 *             score = 0
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)             # <<<<<<<<<<<<<<
 *             type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:
 */
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_get_o_glycan_coarse_score(__pyx_v_self, __pyx_v_scan, __pyx_v_glycan_combination, __pyx_v_mass_shift, __pyx_v_peptide_mass, (&__pyx_v_score))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_match, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":571
 *             score = 0
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)             # <<<<<<<<<<<<<<
 *             if score > best_score:
 *                 best_score = score
 */
      __pyx_t_5 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create(__pyx_v_score, __pyx_v_match)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((PyDict_SetItem(__pyx_v_type_to_score, __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan, __pyx_t_5) < 0))) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":572
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = score
 *                 best_match = match
 */
      __pyx_t_2 = (__pyx_v_score > __pyx_v_best_score);
      if (__pyx_t_2) {

        /* "glycresoft/_c/tandem/core_search.pyx":573
 *             type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:
 *                 best_score = score             # <<<<<<<<<<<<<<
 *                 best_match = match
 *         if glycan_combination.is_gag_linker():
 */
        __pyx_v_best_score = __pyx_v_score;

        /* "glycresoft/_c/tandem/core_search.pyx":574
 *             if score > best_score:
 *                 best_score = score
 *                 best_match = match             # <<<<<<<<<<<<<<
 *         if glycan_combination.is_gag_linker():
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 */
        __Pyx_INCREF((PyObject *)__pyx_v_match);
        __Pyx_DECREF_SET(__pyx_v_best_match, __pyx_v_match);

        /* "glycresoft/_c/tandem/core_search.pyx":572
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = score
 *                 best_match = match
 */
      }

      /* "glycresoft/_c/tandem/core_search.pyx":568
 *                 best_score = score
 *                 best_match = match
 *         if glycan_combination.is_o_glycan():             # <<<<<<<<<<<<<<
 *             score = 0
 *             match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":575
 *                 best_score = score
 *                 best_match = match
 *         if glycan_combination.is_gag_linker():             # <<<<<<<<<<<<<<
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_glycan_combination->__pyx_vtab)->is_gag_linker(__pyx_v_glycan_combination, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 575, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "glycresoft/_c/tandem/core_search.pyx":576
 *                 best_match = match
 *         if glycan_combination.is_gag_linker():
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)             # <<<<<<<<<<<<<<
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:
 */
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v_self->__pyx_vtab)->_get_gag_coarse_score(__pyx_v_self, __pyx_v_scan, __pyx_v_glycan_combination, __pyx_v_mass_shift, __pyx_v_peptide_mass, (&__pyx_v_score))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_match, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":577
 *         if glycan_combination.is_gag_linker():
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)             # <<<<<<<<<<<<<<
 *             if score > best_score:
 *                 best_score = score
 */
      __pyx_t_5 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create(__pyx_v_score, __pyx_v_match)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((PyDict_SetItem(__pyx_v_type_to_score, __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker, __pyx_t_5) < 0))) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":578
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = score
 *                 best_match = match
 */
      __pyx_t_2 = (__pyx_v_score > __pyx_v_best_score);
      if (__pyx_t_2) {

        /* "glycresoft/_c/tandem/core_search.pyx":579
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:
 *                 best_score = score             # <<<<<<<<<<<<<<
 *                 best_match = match
 * 
 */
        __pyx_v_best_score = __pyx_v_score;

        /* "glycresoft/_c/tandem/core_search.pyx":580
 *             if score > best_score:
 *                 best_score = score
 *                 best_match = match             # <<<<<<<<<<<<<<
 * 
 *         if best_match is not None and self.recalibrate_peptide_mass:
 */
        __Pyx_INCREF((PyObject *)__pyx_v_match);
        __Pyx_DECREF_SET(__pyx_v_best_match, __pyx_v_match);

        /* "glycresoft/_c/tandem/core_search.pyx":578
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
 *             if score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = score
 *                 best_match = match
 */
      }

      /* "glycresoft/_c/tandem/core_search.pyx":575
 *                 best_score = score
 *                 best_match = match
 *         if glycan_combination.is_gag_linker():             # <<<<<<<<<<<<<<
 *             match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
 *             type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":582
 *                 best_match = match
 * 
 *         if best_match is not None and self.recalibrate_peptide_mass:             # <<<<<<<<<<<<<<
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()
 *             if recalibrated_peptide_mass > 0:
 */
    __pyx_t_11 = (((PyObject *)__pyx_v_best_match) != Py_None);
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = __pyx_v_self->recalibrate_peptide_mass;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_2) {

      /* "glycresoft/_c/tandem/core_search.pyx":583
 * 
 *         if best_match is not None and self.recalibrate_peptide_mass:
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()             # <<<<<<<<<<<<<<
 *             if recalibrated_peptide_mass > 0:
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
 */
      __pyx_t_4 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_estimate_peptide_mass(__pyx_v_best_match, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 583, __pyx_L1_error)
      __pyx_v_recalibrated_peptide_mass = __pyx_t_4;

      /* "glycresoft/_c/tandem/core_search.pyx":584
 *         if best_match is not None and self.recalibrate_peptide_mass:
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()
 *             if recalibrated_peptide_mass > 0:             # <<<<<<<<<<<<<<
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
 */
      __pyx_t_2 = (__pyx_v_recalibrated_peptide_mass > 0.0);
      if (__pyx_t_2) {

        /* "glycresoft/_c/tandem/core_search.pyx":585
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()
 *             if recalibrated_peptide_mass > 0:
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:             # <<<<<<<<<<<<<<
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
 *                         peptide_mass, recalibrated_peptide_mass))
 */
        __pyx_t_2 = (fabs((__pyx_v_recalibrated_peptide_mass - __pyx_v_peptide_mass)) > 0.5);
        if (__pyx_t_2) {

          /* "glycresoft/_c/tandem/core_search.pyx":586
 *             if recalibrated_peptide_mass > 0:
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (             # <<<<<<<<<<<<<<
 *                         peptide_mass, recalibrated_peptide_mass))
 *         else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":587
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
 *                         peptide_mass, recalibrated_peptide_mass))             # <<<<<<<<<<<<<<
 *         else:
 *             recalibrated_peptide_mass = 0
 */
          __pyx_t_1 = PyFloat_FromDouble(__pyx_v_peptide_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = PyFloat_FromDouble(__pyx_v_recalibrated_peptide_mass); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 587, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 587, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_1);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error);
          __Pyx_GIVEREF(__pyx_t_12);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_12)) __PYX_ERR(0, 587, __pyx_L1_error);
          __pyx_t_1 = 0;
          __pyx_t_12 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":586
 *             if recalibrated_peptide_mass > 0:
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (             # <<<<<<<<<<<<<<
 *                         peptide_mass, recalibrated_peptide_mass))
 *         else:
 */
          __pyx_t_12 = __Pyx_PyString_Format(__pyx_kp_s_Re_estimated_peptide_mass_error, __pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 586, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_13 = NULL;
          __pyx_t_14 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_14 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_12};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 586, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":585
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()
 *             if recalibrated_peptide_mass > 0:
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:             # <<<<<<<<<<<<<<
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
 *                         peptide_mass, recalibrated_peptide_mass))
 */
        }

        /* "glycresoft/_c/tandem/core_search.pyx":584
 *         if best_match is not None and self.recalibrate_peptide_mass:
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()
 *             if recalibrated_peptide_mass > 0:             # <<<<<<<<<<<<<<
 *                 if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
 *                     warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
 */
      }

      /* "glycresoft/_c/tandem/core_search.pyx":582
 *                 best_match = match
 * 
 *         if best_match is not None and self.recalibrate_peptide_mass:             # <<<<<<<<<<<<<<
 *             recalibrated_peptide_mass = best_match.estimate_peptide_mass()
 *             if recalibrated_peptide_mass > 0:
 */
      goto __pyx_L13;
    }

    /* "glycresoft/_c/tandem/core_search.pyx":589
 *                         peptide_mass, recalibrated_peptide_mass))
 *         else:
 *             recalibrated_peptide_mass = 0             # <<<<<<<<<<<<<<
 *         result = GlycanMatchResult._create(
 *             peptide_mass,
 */
    /*else*/ {
      __pyx_v_recalibrated_peptide_mass = 0.0;
    }
    __pyx_L13:;

    /* "glycresoft/_c/tandem/core_search.pyx":590
 *         else:
 *             recalibrated_peptide_mass = 0
 *         result = GlycanMatchResult._create(             # <<<<<<<<<<<<<<
 *             peptide_mass,
 *             best_score, best_match, glycan_combination.size, type_to_score, recalibrated_peptide_mass)
 */
    __pyx_t_5 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult__create(__pyx_v_peptide_mass, __pyx_v_best_score, __pyx_v_best_match, __pyx_v_glycan_combination->size, __pyx_v_type_to_score, __pyx_v_recalibrated_peptide_mass)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_result, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":593
 *             peptide_mass,
 *             best_score, best_match, glycan_combination.size, type_to_score, recalibrated_peptide_mass)
 *         output.append(result)             # <<<<<<<<<<<<<<
 *     output = sorted(output, key=score_getter, reverse=1)
 *     return output
 */
    __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_output, ((PyObject *)__pyx_v_result)); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 593, __pyx_L1_error)
  }
  __pyx_L5_break:;

  /* "glycresoft/_c/tandem/core_search.pyx":594
 *             best_score, best_match, glycan_combination.size, type_to_score, recalibrated_peptide_mass)
 *         output.append(result)
 *     output = sorted(output, key=score_getter, reverse=1)             # <<<<<<<<<<<<<<
 *     return output
 * 
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_output);
  __Pyx_GIVEREF(__pyx_v_output);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_output)) __PYX_ERR(0, 594, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_score_getter); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_key, __pyx_t_12) < 0) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_reverse, __pyx_int_1) < 0) __PYX_ERR(0, 594, __pyx_L1_error)
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_12))||((__pyx_t_12) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_12))) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_output, ((PyObject*)__pyx_t_12));
  __pyx_t_12 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":595
 *         output.append(result)
 *     output = sorted(output, key=score_getter, reverse=1)
 *     return output             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_output);
  __pyx_r = __pyx_v_output;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":530
 * @cython.boundscheck(False)
 * @cython.binding(True)
 * cpdef GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list output, glycan_combination_db
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFilteringPeptideMassEstimator_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_output);
  __Pyx_XDECREF(__pyx_v_glycan_combination_db);
  __Pyx_XDECREF((PyObject *)__pyx_v_glycan_combination);
  __Pyx_XDECREF(__pyx_v_type_to_score);
  __Pyx_XDECREF((PyObject *)__pyx_v_best_match);
  __Pyx_XDECREF((PyObject *)__pyx_v_match);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_3GlycanFilteringPeptideMassEstimator_match(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_2GlycanFilteringPeptideMassEstimator_match, "GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_3GlycanFilteringPeptideMassEstimator_match = {"GlycanFilteringPeptideMassEstimator_match", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_3GlycanFilteringPeptideMassEstimator_match, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_2GlycanFilteringPeptideMassEstimator_match};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_3GlycanFilteringPeptideMassEstimator_match(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self = 0;
  PyObject *__pyx_v_scan = 0;
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift = 0;
  PyObject *__pyx_v_query_mass = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GlycanFilteringPeptideMassEstimator_match (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_scan,&__pyx_n_s_mass_shift,&__pyx_n_s_query_mass,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL((PyObject *)__pyx_k__4);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scan)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("GlycanFilteringPeptideMassEstimator_match", 0, 2, 4, 1); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass_shift);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_query_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "GlycanFilteringPeptideMassEstimator_match") < 0)) __PYX_ERR(0, 530, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)values[0]);
    __pyx_v_scan = values[1];
    __pyx_v_mass_shift = ((struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *)values[2]);
    __pyx_v_query_mass = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GlycanFilteringPeptideMassEstimator_match", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 530, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFilteringPeptideMassEstimator_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, 1, "self", 0))) __PYX_ERR(0, 530, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mass_shift), __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase, 1, "mass_shift", 0))) __PYX_ERR(0, 530, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_2GlycanFilteringPeptideMassEstimator_match(__pyx_self, __pyx_v_self, __pyx_v_scan, __pyx_v_mass_shift, __pyx_v_query_mass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_2GlycanFilteringPeptideMassEstimator_match(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v_self, PyObject *__pyx_v_scan, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, PyObject *__pyx_v_query_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GlycanFilteringPeptideMassEstimator_match", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.mass_shift = __pyx_v_mass_shift;
  __pyx_t_2.query_mass = __pyx_v_query_mass;
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_GlycanFilteringPeptideMassEstimator_match(__pyx_v_self, __pyx_v_scan, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFilteringPeptideMassEstimator_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":603
 *         public str name
 * 
 *     def __init__(self, mass, composition, key, is_extended=False):             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.composition = composition
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_composition = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_is_extended = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mass,&__pyx_n_s_composition,&__pyx_n_s_key,&__pyx_n_s_is_extended,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 603, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_composition)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 603, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(0, 603, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 603, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(0, 603, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_is_extended);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 603, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 603, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mass = values[0];
    __pyx_v_composition = values[1];
    __pyx_v_key = values[2];
    __pyx_v_is_extended = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 603, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self), __pyx_v_mass, __pyx_v_composition, __pyx_v_key, __pyx_v_is_extended);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_mass, PyObject *__pyx_v_composition, PyObject *__pyx_v_key, PyObject *__pyx_v_is_extended) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":604
 * 
 *     def __init__(self, mass, composition, key, is_extended=False):
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         self.key = key
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 604, __pyx_L1_error)
  __pyx_v_self->__pyx_base.mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":605
 *     def __init__(self, mass, composition, key, is_extended=False):
 *         self.mass = mass
 *         self.composition = composition             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.is_extended = is_extended
 */
  if (!(likely(((__pyx_v_composition) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_composition, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 605, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_composition;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.composition);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.composition);
  __pyx_v_self->__pyx_base.composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":606
 *         self.mass = mass
 *         self.composition = composition
 *         self.key = key             # <<<<<<<<<<<<<<
 *         self.is_extended = is_extended
 *         self._hash_key = -1
 */
  if (!(likely(((__pyx_v_key) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_key, __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable))))) __PYX_ERR(0, 606, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_key;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.key);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.key);
  __pyx_v_self->__pyx_base.key = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":607
 *         self.composition = composition
 *         self.key = key
 *         self.is_extended = is_extended             # <<<<<<<<<<<<<<
 *         self._hash_key = -1
 *         self.index = -1
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_is_extended); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L1_error)
  __pyx_v_self->__pyx_base.is_extended = __pyx_t_3;

  /* "glycresoft/_c/tandem/core_search.pyx":608
 *         self.key = key
 *         self.is_extended = is_extended
 *         self._hash_key = -1             # <<<<<<<<<<<<<<
 *         self.index = -1
 *         self.name = None
 */
  __pyx_v_self->__pyx_base._hash_key = -1L;

  /* "glycresoft/_c/tandem/core_search.pyx":609
 *         self.is_extended = is_extended
 *         self._hash_key = -1
 *         self.index = -1             # <<<<<<<<<<<<<<
 *         self.name = None
 * 
 */
  __pyx_v_self->index = -1;

  /* "glycresoft/_c/tandem/core_search.pyx":610
 *         self._hash_key = -1
 *         self.index = -1
 *         self.name = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* "glycresoft/_c/tandem/core_search.pyx":603
 *         public str name
 * 
 *     def __init__(self, mass, composition, key, is_extended=False):             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.composition = composition
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":613
 * 
 *     @staticmethod
 *     cdef IndexGlycanCompositionFragment _create(double mass, CComposition composition, CountTable key, bint is_extended):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IndexGlycanCompositionFragment self
 */

struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create(double __pyx_v_mass, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_key, int __pyx_v_is_extended) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":616
 *         cdef:
 *             IndexGlycanCompositionFragment self
 *         self = IndexGlycanCompositionFragment.__new__(IndexGlycanCompositionFragment)             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.composition = composition
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":617
 *             IndexGlycanCompositionFragment self
 *         self = IndexGlycanCompositionFragment.__new__(IndexGlycanCompositionFragment)
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         self.key = key
 */
  __pyx_v_self->__pyx_base.mass = __pyx_v_mass;

  /* "glycresoft/_c/tandem/core_search.pyx":618
 *         self = IndexGlycanCompositionFragment.__new__(IndexGlycanCompositionFragment)
 *         self.mass = mass
 *         self.composition = composition             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.is_extended = is_extended
 */
  __Pyx_INCREF((PyObject *)__pyx_v_composition);
  __Pyx_GIVEREF((PyObject *)__pyx_v_composition);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.composition);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.composition);
  __pyx_v_self->__pyx_base.composition = __pyx_v_composition;

  /* "glycresoft/_c/tandem/core_search.pyx":619
 *         self.mass = mass
 *         self.composition = composition
 *         self.key = key             # <<<<<<<<<<<<<<
 *         self.is_extended = is_extended
 *         self._hash_key = -1
 */
  __Pyx_INCREF((PyObject *)__pyx_v_key);
  __Pyx_GIVEREF((PyObject *)__pyx_v_key);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.key);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.key);
  __pyx_v_self->__pyx_base.key = __pyx_v_key;

  /* "glycresoft/_c/tandem/core_search.pyx":620
 *         self.composition = composition
 *         self.key = key
 *         self.is_extended = is_extended             # <<<<<<<<<<<<<<
 *         self._hash_key = -1
 *         self.index = -1
 */
  __pyx_v_self->__pyx_base.is_extended = __pyx_v_is_extended;

  /* "glycresoft/_c/tandem/core_search.pyx":621
 *         self.key = key
 *         self.is_extended = is_extended
 *         self._hash_key = -1             # <<<<<<<<<<<<<<
 *         self.index = -1
 *         self.name = None
 */
  __pyx_v_self->__pyx_base._hash_key = -1L;

  /* "glycresoft/_c/tandem/core_search.pyx":622
 *         self.is_extended = is_extended
 *         self._hash_key = -1
 *         self.index = -1             # <<<<<<<<<<<<<<
 *         self.name = None
 *         return self
 */
  __pyx_v_self->index = -1;

  /* "glycresoft/_c/tandem/core_search.pyx":623
 *         self._hash_key = -1
 *         self.index = -1
 *         self.name = None             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* "glycresoft/_c/tandem/core_search.pyx":624
 *         self.index = -1
 *         self.name = None
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycanCompositionFragment copy(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":613
 * 
 *     @staticmethod
 *     cdef IndexGlycanCompositionFragment _create(double mass, CComposition composition, CountTable key, bint is_extended):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IndexGlycanCompositionFragment self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create__pyx_wrap_1(double __pyx_v_mass, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_key, int __pyx_v_is_extended) {
  return __pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create(__pyx_v_mass, __pyx_v_composition, __pyx_v_key, __pyx_v_is_extended);
}

/* "glycresoft/_c/tandem/core_search.pyx":626
 *         return self
 * 
 *     cpdef GlycanCompositionFragment copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IndexGlycanCompositionFragment inst
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_copy(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_inst = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_hash_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment))))) __PYX_ERR(0, 626, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":631
 *         inst = IndexGlycanCompositionFragment._create(
 *             self.mass,
 *             self.composition.clone() if self.composition is not None else None, self.key.copy(),             # <<<<<<<<<<<<<<
 *             self.is_extended)
 *         inst._hash_key = self._hash_key
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.composition) != Py_None);
  if (__pyx_t_6) {
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_base.composition->__pyx_vtab)->clone(__pyx_v_self->__pyx_base.composition, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 631, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_base.key->__pyx_vtab)->copy(__pyx_v_self->__pyx_base.key, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "glycresoft/_c/tandem/core_search.pyx":629
 *         cdef:
 *             IndexGlycanCompositionFragment inst
 *         inst = IndexGlycanCompositionFragment._create(             # <<<<<<<<<<<<<<
 *             self.mass,
 *             self.composition.clone() if self.composition is not None else None, self.key.copy(),
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create(__pyx_v_self->__pyx_base.mass, ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1), ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_2), __pyx_v_self->__pyx_base.is_extended)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_inst = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":633
 *             self.composition.clone() if self.composition is not None else None, self.key.copy(),
 *             self.is_extended)
 *         inst._hash_key = self._hash_key             # <<<<<<<<<<<<<<
 *         inst.index = self.index
 *         inst.name = self.name
 */
  __pyx_t_7 = __pyx_v_self->__pyx_base._hash_key;
  __pyx_v_inst->__pyx_base._hash_key = __pyx_t_7;

  /* "glycresoft/_c/tandem/core_search.pyx":634
 *             self.is_extended)
 *         inst._hash_key = self._hash_key
 *         inst.index = self.index             # <<<<<<<<<<<<<<
 *         inst.name = self.name
 *         return inst
 */
  __pyx_t_8 = __pyx_v_self->index;
  __pyx_v_inst->index = __pyx_t_8;

  /* "glycresoft/_c/tandem/core_search.pyx":635
 *         inst._hash_key = self._hash_key
 *         inst.index = self.index
 *         inst.name = self.name             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_3 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_inst->name);
  __Pyx_DECREF(__pyx_v_inst->name);
  __pyx_v_inst->name = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":636
 *         inst.index = self.index
 *         inst.name = self.name
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *)__pyx_v_inst);
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":626
 *         return self
 * 
 *     cpdef GlycanCompositionFragment copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IndexGlycanCompositionFragment inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_2copy, "IndexGlycanCompositionFragment.copy(self) -> GlycanCompositionFragment");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_2copy};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_2copy(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_2copy(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":638
 *         return inst
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4__reduce__, "IndexGlycanCompositionFragment.__reduce__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5__reduce__ = {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4__reduce__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4__reduce__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4__reduce__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":639
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.is_extended); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.composition);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.composition);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self->__pyx_base.composition))) __PYX_ERR(0, 639, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.key);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_self->__pyx_base.key))) __PYX_ERR(0, 639, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":638
 *         return inst
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":641
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {
 *             "index": self.index,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_7__getstate__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_6__getstate__, "IndexGlycanCompositionFragment.__getstate__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_7__getstate__ = {"__getstate__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_7__getstate__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_6__getstate__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_7__getstate__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__getstate__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__getstate__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_6__getstate__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_6__getstate__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":642
 * 
 *     def __getstate__(self):
 *         return {             # <<<<<<<<<<<<<<
 *             "index": self.index,
 *             "name": self.name,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycresoft/_c/tandem/core_search.pyx":643
 *     def __getstate__(self):
 *         return {
 *             "index": self.index,             # <<<<<<<<<<<<<<
 *             "name": self.name,
 *             "_hash_key": self._hash_key
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_index, __pyx_t_2) < 0) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":644
 *         return {
 *             "index": self.index,
 *             "name": self.name,             # <<<<<<<<<<<<<<
 *             "_hash_key": self._hash_key
 *         }
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_name_2, __pyx_v_self->name) < 0) __PYX_ERR(0, 643, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":645
 *             "index": self.index,
 *             "name": self.name,
 *             "_hash_key": self._hash_key             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_FromHash_t(__pyx_v_self->__pyx_base._hash_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_hash_key, __pyx_t_2) < 0) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":641
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {
 *             "index": self.index,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":648
 *         }
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.index = state['index']
 *         self.name = state['name']
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_9__setstate__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_8__setstate__, "IndexGlycanCompositionFragment.__setstate__(self, state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_9__setstate__ = {"__setstate__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_9__setstate__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_8__setstate__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_9__setstate__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate__") < 0)) __PYX_ERR(0, 648, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 648, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_8__setstate__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_8__setstate__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_hash_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":649
 * 
 *     def __setstate__(self, state):
 *         self.index = state['index']             # <<<<<<<<<<<<<<
 *         self.name = state['name']
 *         self._hash_key = state['_hash_key']
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_s_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->index = __pyx_t_2;

  /* "glycresoft/_c/tandem/core_search.pyx":650
 *     def __setstate__(self, state):
 *         self.index = state['index']
 *         self.name = state['name']             # <<<<<<<<<<<<<<
 *         self._hash_key = state['_hash_key']
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":651
 *         self.index = state['index']
 *         self.name = state['name']
 *         self._hash_key = state['_hash_key']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_state, __pyx_n_s_hash_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_3 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->__pyx_base._hash_key = __pyx_t_3;

  /* "glycresoft/_c/tandem/core_search.pyx":648
 *         }
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.index = state['index']
 *         self.name = state['name']
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":600
 * cdef class IndexGlycanCompositionFragment(GlycanCompositionFragment):
 *     cdef:
 *         public int index             # <<<<<<<<<<<<<<
 *         public str name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":601
 *     cdef:
 *         public int index
 *         public str name             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, mass, composition, key, is_extended=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_v_value))) __PYX_ERR(0, 601, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":660
 *         public list keys
 * 
 *     def __init__(self, mass, keys=None):             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.keys = keys or []
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_keys = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mass,&__pyx_n_s_keys,0};
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_keys);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 660, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mass = values[0];
    __pyx_v_keys = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 660, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self), __pyx_v_mass, __pyx_v_keys);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v_mass, PyObject *__pyx_v_keys) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":661
 * 
 *     def __init__(self, mass, keys=None):
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.keys = keys or []
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mass); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 661, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":662
 *     def __init__(self, mass, keys=None):
 *         self.mass = mass
 *         self.keys = keys or []             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_keys); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 662, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    if (!(likely(PyList_CheckExact(__pyx_v_keys))||((__pyx_v_keys) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_keys))) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_keys);
    __pyx_t_2 = __pyx_v_keys;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->keys);
  __Pyx_DECREF(__pyx_v_self->keys);
  __pyx_v_self->keys = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":660
 *         public list keys
 * 
 *     def __init__(self, mass, keys=None):             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.keys = keys or []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":664
 *         self.keys = keys or []
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({self.mass})".format(self=self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_2__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_2__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":665
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({self.mass})".format(self=self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___self_ma, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 665, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":664
 *         self.keys = keys or []
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({self.mass})".format(self=self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":657
 * cdef class ComplementFragment(object):
 *     cdef:
 *         public double mass             # <<<<<<<<<<<<<<
 *         public list keys
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":658
 *     cdef:
 *         public double mass
 *         public list keys             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, mass, keys=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->keys);
  __pyx_r = __pyx_v_self->keys;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(0, 658, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->keys);
  __Pyx_DECREF(__pyx_v_self->keys);
  __pyx_v_self->keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.keys.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->keys);
  __Pyx_DECREF(__pyx_v_self->keys);
  __pyx_v_self->keys = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4__reduce_cython__, "ComplementFragment.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.keys, self.mass)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->keys);
  __Pyx_GIVEREF(__pyx_v_self->keys);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->keys)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.keys, self.mass)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.keys, self.mass)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.keys is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.keys, self.mass)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.keys is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->keys != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.keys is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_ComplementFragmen); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_134214914);
    __Pyx_GIVEREF(__pyx_int_134214914);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_134214914)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, None), state
 *     else:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ComplementFragment__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_ComplementFragmen); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_134214914);
    __Pyx_GIVEREF(__pyx_int_134214914);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_134214914)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ComplementFragment__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_6__setstate_cython__, "ComplementFragment.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_6__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_6__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_6__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ComplementFragment__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ComplementFragment__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ComplementFragment__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.ComplementFragment.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":677
 *         public int32_t glycan_index
 * 
 *     def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):             # <<<<<<<<<<<<<<
 *         if core_matches is None:
 *             core_matches = set()
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peptide_mass = 0;
  PyObject *__pyx_v_score = 0;
  PyObject *__pyx_v_core_matches = 0;
  PyObject *__pyx_v_fragment_matches = 0;
  npy_int32 __pyx_v_glycan_index;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peptide_mass,&__pyx_n_s_score,&__pyx_n_s_core_matches,&__pyx_n_s_fragment_matches,&__pyx_n_s_glycan_index,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_int_neg_1));
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_int_0));
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_peptide_mass);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_score);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_core_matches);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fragment_matches);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_index);
          if (value) { values[4] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 677, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_peptide_mass = values[0];
    __pyx_v_score = values[1];
    __pyx_v_core_matches = values[2];
    __pyx_v_fragment_matches = values[3];
    if (values[4]) {
      __pyx_v_glycan_index = __Pyx_PyInt_As_npy_int32(values[4]); if (unlikely((__pyx_v_glycan_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    } else {
      __pyx_v_glycan_index = ((npy_int32)-1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, __pyx_nargs); __PYX_ERR(0, 677, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), __pyx_v_peptide_mass, __pyx_v_score, __pyx_v_core_matches, __pyx_v_fragment_matches, __pyx_v_glycan_index);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_peptide_mass, PyObject *__pyx_v_score, PyObject *__pyx_v_core_matches, PyObject *__pyx_v_fragment_matches, npy_int32 __pyx_v_glycan_index) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_core_matches);
  __Pyx_INCREF(__pyx_v_fragment_matches);

  /* "glycresoft/_c/tandem/core_search.pyx":678
 * 
 *     def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):
 *         if core_matches is None:             # <<<<<<<<<<<<<<
 *             core_matches = set()
 *         if fragment_matches is None:
 */
  __pyx_t_1 = (__pyx_v_core_matches == Py_None);
  if (__pyx_t_1) {

    /* "glycresoft/_c/tandem/core_search.pyx":679
 *     def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):
 *         if core_matches is None:
 *             core_matches = set()             # <<<<<<<<<<<<<<
 *         if fragment_matches is None:
 *             fragment_matches = set()
 */
    __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_core_matches, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":678
 * 
 *     def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):
 *         if core_matches is None:             # <<<<<<<<<<<<<<
 *             core_matches = set()
 *         if fragment_matches is None:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":680
 *         if core_matches is None:
 *             core_matches = set()
 *         if fragment_matches is None:             # <<<<<<<<<<<<<<
 *             fragment_matches = set()
 *         self.peptide_mass = peptide_mass
 */
  __pyx_t_1 = (__pyx_v_fragment_matches == Py_None);
  if (__pyx_t_1) {

    /* "glycresoft/_c/tandem/core_search.pyx":681
 *             core_matches = set()
 *         if fragment_matches is None:
 *             fragment_matches = set()             # <<<<<<<<<<<<<<
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 */
    __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_fragment_matches, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":680
 *         if core_matches is None:
 *             core_matches = set()
 *         if fragment_matches is None:             # <<<<<<<<<<<<<<
 *             fragment_matches = set()
 *         self.peptide_mass = peptide_mass
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":682
 *         if fragment_matches is None:
 *             fragment_matches = set()
 *         self.peptide_mass = peptide_mass             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.core_matches = core_matches
 */
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_peptide_mass); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L1_error)
  __pyx_v_self->peptide_mass = __pyx_t_3;

  /* "glycresoft/_c/tandem/core_search.pyx":683
 *             fragment_matches = set()
 *         self.peptide_mass = peptide_mass
 *         self.score = score             # <<<<<<<<<<<<<<
 *         self.core_matches = core_matches
 *         self.fragment_matches = fragment_matches
 */
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_score); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_3;

  /* "glycresoft/_c/tandem/core_search.pyx":684
 *         self.peptide_mass = peptide_mass
 *         self.score = score
 *         self.core_matches = core_matches             # <<<<<<<<<<<<<<
 *         self.fragment_matches = fragment_matches
 *         self.glycan_index = glycan_index
 */
  if (!(likely(PySet_CheckExact(__pyx_v_core_matches))||((__pyx_v_core_matches) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_core_matches))) __PYX_ERR(0, 684, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_core_matches;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->core_matches);
  __Pyx_DECREF(__pyx_v_self->core_matches);
  __pyx_v_self->core_matches = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":685
 *         self.score = score
 *         self.core_matches = core_matches
 *         self.fragment_matches = fragment_matches             # <<<<<<<<<<<<<<
 *         self.glycan_index = glycan_index
 * 
 */
  if (!(likely(PySet_CheckExact(__pyx_v_fragment_matches))||((__pyx_v_fragment_matches) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_fragment_matches))) __PYX_ERR(0, 685, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_fragment_matches;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":686
 *         self.core_matches = core_matches
 *         self.fragment_matches = fragment_matches
 *         self.glycan_index = glycan_index             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_v_self->glycan_index = __pyx_v_glycan_index;

  /* "glycresoft/_c/tandem/core_search.pyx":677
 *         public int32_t glycan_index
 * 
 *     def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):             # <<<<<<<<<<<<<<
 *         if core_matches is None:
 *             core_matches = set()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_core_matches);
  __Pyx_XDECREF(__pyx_v_fragment_matches);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":689
 * 
 *     @staticmethod
 *     cdef PartialGlycanSolution _create():             # <<<<<<<<<<<<<<
 *         cdef PartialGlycanSolution self = PartialGlycanSolution.__new__(PartialGlycanSolution)
 *         self.peptide_mass = -1
 */

static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_f_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution__create(void) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":690
 *     @staticmethod
 *     cdef PartialGlycanSolution _create():
 *         cdef PartialGlycanSolution self = PartialGlycanSolution.__new__(PartialGlycanSolution)             # <<<<<<<<<<<<<<
 *         self.peptide_mass = -1
 *         self.score = 0.0
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":691
 *     cdef PartialGlycanSolution _create():
 *         cdef PartialGlycanSolution self = PartialGlycanSolution.__new__(PartialGlycanSolution)
 *         self.peptide_mass = -1             # <<<<<<<<<<<<<<
 *         self.score = 0.0
 *         self.core_matches = set()
 */
  __pyx_v_self->peptide_mass = -1.0;

  /* "glycresoft/_c/tandem/core_search.pyx":692
 *         cdef PartialGlycanSolution self = PartialGlycanSolution.__new__(PartialGlycanSolution)
 *         self.peptide_mass = -1
 *         self.score = 0.0             # <<<<<<<<<<<<<<
 *         self.core_matches = set()
 *         self.fragment_matches = set()
 */
  __pyx_v_self->score = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":693
 *         self.peptide_mass = -1
 *         self.score = 0.0
 *         self.core_matches = set()             # <<<<<<<<<<<<<<
 *         self.fragment_matches = set()
 *         self.glycan_index = -1
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->core_matches);
  __Pyx_DECREF(__pyx_v_self->core_matches);
  __pyx_v_self->core_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":694
 *         self.score = 0.0
 *         self.core_matches = set()
 *         self.fragment_matches = set()             # <<<<<<<<<<<<<<
 *         self.glycan_index = -1
 *         return self
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":695
 *         self.core_matches = set()
 *         self.fragment_matches = set()
 *         self.glycan_index = -1             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_v_self->glycan_index = -1;

  /* "glycresoft/_c/tandem/core_search.pyx":696
 *         self.fragment_matches = set()
 *         self.glycan_index = -1
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":689
 * 
 *     @staticmethod
 *     cdef PartialGlycanSolution _create():             # <<<<<<<<<<<<<<
 *         cdef PartialGlycanSolution self = PartialGlycanSolution.__new__(PartialGlycanSolution)
 *         self.peptide_mass = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":698
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template =("{self.__class__.__name__}({self.peptide_mass}, {self.score}, "
 *                    "{self.core_matches}, {self.fragment_matches}, {self.glycan_index})")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_2__repr__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_2__repr__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_v_template = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":699
 * 
 *     def __repr__(self):
 *         template =("{self.__class__.__name__}({self.peptide_mass}, {self.score}, "             # <<<<<<<<<<<<<<
 *                    "{self.core_matches}, {self.fragment_matches}, {self.glycan_index})")
 *         return template.format(self=self)
 */
  __Pyx_INCREF(__pyx_kp_s_self___class_____name___self_pe_2);
  __pyx_v_template = __pyx_kp_s_self___class_____name___self_pe_2;

  /* "glycresoft/_c/tandem/core_search.pyx":701
 *         template =("{self.__class__.__name__}({self.peptide_mass}, {self.score}, "
 *                    "{self.core_matches}, {self.fragment_matches}, {self.glycan_index})")
 *         return template.format(self=self)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_template, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 701, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":698
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template =("{self.__class__.__name__}({self.peptide_mass}, {self.score}, "
 *                    "{self.core_matches}, {self.fragment_matches}, {self.glycan_index})")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_template);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":703
 *         return template.format(self=self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fragment_match_count(self):
 *         return len(self.fragment_matches)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_20fragment_match_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_20fragment_match_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_20fragment_match_count___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_20fragment_match_count___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":705
 *     @property
 *     def fragment_match_count(self):
 *         return len(self.fragment_matches)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->fragment_matches;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 705, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PySet_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":703
 *         return template.format(self=self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fragment_match_count(self):
 *         return len(self.fragment_matches)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.fragment_match_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":671
 * cdef class PartialGlycanSolution(object):
 *     cdef:
 *         public double peptide_mass             # <<<<<<<<<<<<<<
 *         public double score
 *         public set core_matches
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->peptide_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.peptide_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L1_error)
  __pyx_v_self->peptide_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.peptide_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":672
 *     cdef:
 *         public double peptide_mass
 *         public double score             # <<<<<<<<<<<<<<
 *         public set core_matches
 *         public set fragment_matches
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":673
 *         public double peptide_mass
 *         public double score
 *         public set core_matches             # <<<<<<<<<<<<<<
 *         public set fragment_matches
 *         public int32_t glycan_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->core_matches);
  __pyx_r = __pyx_v_self->core_matches;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_value))) __PYX_ERR(0, 673, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->core_matches);
  __Pyx_DECREF(__pyx_v_self->core_matches);
  __pyx_v_self->core_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.core_matches.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->core_matches);
  __Pyx_DECREF(__pyx_v_self->core_matches);
  __pyx_v_self->core_matches = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":674
 *         public double score
 *         public set core_matches
 *         public set fragment_matches             # <<<<<<<<<<<<<<
 *         public int32_t glycan_index
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fragment_matches);
  __pyx_r = __pyx_v_self->fragment_matches;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_value))) __PYX_ERR(0, 674, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.fragment_matches.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->fragment_matches);
  __Pyx_DECREF(__pyx_v_self->fragment_matches);
  __pyx_v_self->fragment_matches = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":675
 *         public set core_matches
 *         public set fragment_matches
 *         public int32_t glycan_index             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->glycan_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.glycan_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  npy_int32 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 675, __pyx_L1_error)
  __pyx_v_self->glycan_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.glycan_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_4__reduce_cython__, "PartialGlycanSolution.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_4__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_4__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_4__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.core_matches, self.fragment_matches, self.glycan_index, self.peptide_mass, self.score)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->glycan_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->peptide_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->core_matches);
  __Pyx_GIVEREF(__pyx_v_self->core_matches);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->core_matches)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fragment_matches);
  __Pyx_GIVEREF(__pyx_v_self->fragment_matches);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->fragment_matches)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.core_matches, self.fragment_matches, self.glycan_index, self.peptide_mass, self.score)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.core_matches, self.fragment_matches, self.glycan_index, self.peptide_mass, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.core_matches is not None or self.fragment_matches is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.core_matches, self.fragment_matches, self.glycan_index, self.peptide_mass, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.core_matches is not None or self.fragment_matches is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->core_matches != ((PyObject*)Py_None));
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->fragment_matches != ((PyObject*)Py_None));
    __pyx_t_5 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.core_matches is not None or self.fragment_matches is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.core_matches is not None or self.fragment_matches is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PartialGlycanSolu); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_92231173);
    __Pyx_GIVEREF(__pyx_int_92231173);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_92231173)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.core_matches is not None or self.fragment_matches is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, None), state
 *     else:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PartialGlycanSolution__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_PartialGlycanSolu); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_92231173);
    __Pyx_GIVEREF(__pyx_int_92231173);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_92231173)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PartialGlycanSolution__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_6__setstate_cython__, "PartialGlycanSolution.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_6__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_6__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_6__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PartialGlycanSolution__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_PartialGlycanSolution__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PartialGlycanSolution__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.PartialGlycanSolution.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":753
 *         public double resolution
 * 
 *     def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0):             # <<<<<<<<<<<<<<
 *         self.members = members or []
 *         self.unique_fragments = dict()
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_members = 0;
  PyObject *__pyx_v_fragment_weight = 0;
  PyObject *__pyx_v_core_weight = 0;
  PyObject *__pyx_v_resolution = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_members,&__pyx_n_s_fragment_weight,&__pyx_n_s_core_weight,&__pyx_n_s_resolution,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_0_56));
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_0_42));
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_100_0));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_members);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fragment_weight);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_core_weight);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_resolution);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 753, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_members = values[0];
    __pyx_v_fragment_weight = values[1];
    __pyx_v_core_weight = values[2];
    __pyx_v_resolution = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 4, __pyx_nargs); __PYX_ERR(0, 753, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex___init__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v_members, __pyx_v_fragment_weight, __pyx_v_core_weight, __pyx_v_resolution);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex___init__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_members, PyObject *__pyx_v_fragment_weight, PyObject *__pyx_v_core_weight, PyObject *__pyx_v_resolution) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":754
 * 
 *     def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0):
 *         self.members = members or []             # <<<<<<<<<<<<<<
 *         self.unique_fragments = dict()
 *         self._fragments = []
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_members); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 754, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    if (!(likely(PyList_CheckExact(__pyx_v_members))||((__pyx_v_members) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_members))) __PYX_ERR(0, 754, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_members);
    __pyx_t_1 = __pyx_v_members;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->members);
  __Pyx_DECREF(__pyx_v_self->members);
  __pyx_v_self->members = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":755
 *     def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0):
 *         self.members = members or []
 *         self.unique_fragments = dict()             # <<<<<<<<<<<<<<
 *         self._fragments = []
 *         self.fragment_index = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->unique_fragments);
  __Pyx_DECREF(__pyx_v_self->unique_fragments);
  __pyx_v_self->unique_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":756
 *         self.members = members or []
 *         self.unique_fragments = dict()
 *         self._fragments = []             # <<<<<<<<<<<<<<
 *         self.fragment_index = dict()
 *         self.counter = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fragments);
  __Pyx_DECREF(__pyx_v_self->_fragments);
  __pyx_v_self->_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":757
 *         self.unique_fragments = dict()
 *         self._fragments = []
 *         self.fragment_index = dict()             # <<<<<<<<<<<<<<
 *         self.counter = 0
 *         self.fragment_weight = fragment_weight
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_index);
  __Pyx_DECREF(__pyx_v_self->fragment_index);
  __pyx_v_self->fragment_index = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":758
 *         self._fragments = []
 *         self.fragment_index = dict()
 *         self.counter = 0             # <<<<<<<<<<<<<<
 *         self.fragment_weight = fragment_weight
 *         self.core_weight = core_weight
 */
  __pyx_v_self->counter = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":759
 *         self.fragment_index = dict()
 *         self.counter = 0
 *         self.fragment_weight = fragment_weight             # <<<<<<<<<<<<<<
 *         self.core_weight = core_weight
 *         self.lower_bound = float('inf')
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_fragment_weight); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 759, __pyx_L1_error)
  __pyx_v_self->fragment_weight = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":760
 *         self.counter = 0
 *         self.fragment_weight = fragment_weight
 *         self.core_weight = core_weight             # <<<<<<<<<<<<<<
 *         self.lower_bound = float('inf')
 *         self.upper_bound = 0
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_core_weight); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)
  __pyx_v_self->core_weight = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":761
 *         self.fragment_weight = fragment_weight
 *         self.core_weight = core_weight
 *         self.lower_bound = float('inf')             # <<<<<<<<<<<<<<
 *         self.upper_bound = 0
 *         self.resolution = resolution
 */
  __pyx_t_4 = __Pyx_PyString_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 761, __pyx_L1_error)
  __pyx_v_self->lower_bound = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":762
 *         self.core_weight = core_weight
 *         self.lower_bound = float('inf')
 *         self.upper_bound = 0             # <<<<<<<<<<<<<<
 *         self.resolution = resolution
 * 
 */
  __pyx_v_self->upper_bound = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":763
 *         self.lower_bound = float('inf')
 *         self.upper_bound = 0
 *         self.resolution = resolution             # <<<<<<<<<<<<<<
 * 
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_resolution); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 763, __pyx_L1_error)
  __pyx_v_self->resolution = __pyx_t_4;

  /* "glycresoft/_c/tandem/core_search.pyx":753
 *         public double resolution
 * 
 *     def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0):             # <<<<<<<<<<<<<<
 *         self.members = members or []
 *         self.unique_fragments = dict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":765
 *         self.resolution = resolution
 * 
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):             # <<<<<<<<<<<<<<
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__intern(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_fragment, PyObject *__pyx_v_glycosylation_type, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_cache_for_type = NULL;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  double __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_intern", 0);
  __Pyx_INCREF(__pyx_v_fragment);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_intern); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_fragment, __pyx_v_glycosylation_type};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment))))) __PYX_ERR(0, 765, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":766
 * 
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):
 *         key = str(HashableGlycanComposition(fragment.key))             # <<<<<<<<<<<<<<
 *         try:
 *             cache_for_type = self.unique_fragments[glycosylation_type]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fragment, __pyx_n_s_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_HashableGlycanComposition);
  __pyx_t_3 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_HashableGlycanComposition; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_Str(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_key = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":767
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:             # <<<<<<<<<<<<<<
 *             cache_for_type = self.unique_fragments[glycosylation_type]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "glycresoft/_c/tandem/core_search.pyx":768
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:
 *             cache_for_type = self.unique_fragments[glycosylation_type]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()
 */
      if (unlikely(__pyx_v_self->unique_fragments == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 768, __pyx_L3_error)
      }
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_self->unique_fragments, __pyx_v_glycosylation_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 768, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_cache_for_type = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":767
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:             # <<<<<<<<<<<<<<
 *             cache_for_type = self.unique_fragments[glycosylation_type]
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":769
 *         try:
 *             cache_for_type = self.unique_fragments[glycosylation_type]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()
 *         try:
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._intern", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 769, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_2);

      /* "glycresoft/_c/tandem/core_search.pyx":770
 *             cache_for_type = self.unique_fragments[glycosylation_type]
 *         except KeyError:
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()             # <<<<<<<<<<<<<<
 *         try:
 *             return cache_for_type[key]
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 770, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_cache_for_type, __pyx_t_4);
      if (unlikely(__pyx_v_self->unique_fragments == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 770, __pyx_L5_except_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_self->unique_fragments, __pyx_v_glycosylation_type, __pyx_t_4) < 0))) __PYX_ERR(0, 770, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "glycresoft/_c/tandem/core_search.pyx":767
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:             # <<<<<<<<<<<<<<
 *             cache_for_type = self.unique_fragments[glycosylation_type]
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    __pyx_L8_try_end:;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":771
 *         except KeyError:
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()
 *         try:             # <<<<<<<<<<<<<<
 *             return cache_for_type[key]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "glycresoft/_c/tandem/core_search.pyx":772
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()
 *         try:
 *             return cache_for_type[key]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             fragment = IndexGlycanCompositionFragment._create(
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_cache_for_type, __pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 772, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment))))) __PYX_ERR(0, 772, __pyx_L11_error)
      __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_t_2);
      __pyx_t_2 = 0;
      goto __pyx_L15_try_return;

      /* "glycresoft/_c/tandem/core_search.pyx":771
 *         except KeyError:
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()
 *         try:             # <<<<<<<<<<<<<<
 *             return cache_for_type[key]
 *         except KeyError:
 */
    }
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":773
 *         try:
 *             return cache_for_type[key]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             fragment = IndexGlycanCompositionFragment._create(
 *                 fragment.mass, None, fragment.key, not fragment.is_core)
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._intern", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 773, __pyx_L13_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "glycresoft/_c/tandem/core_search.pyx":775
 *         except KeyError:
 *             fragment = IndexGlycanCompositionFragment._create(
 *                 fragment.mass, None, fragment.key, not fragment.is_core)             # <<<<<<<<<<<<<<
 *             cache_for_type[key] = fragment
 *             fragment.name = key
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fragment, __pyx_n_s_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 775, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 775, __pyx_L13_except_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fragment, __pyx_n_s_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 775, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable))))) __PYX_ERR(0, 775, __pyx_L13_except_error)
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_fragment, __pyx_n_s_is_core); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 775, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 775, __pyx_L13_except_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":774
 *             return cache_for_type[key]
 *         except KeyError:
 *             fragment = IndexGlycanCompositionFragment._create(             # <<<<<<<<<<<<<<
 *                 fragment.mass, None, fragment.key, not fragment.is_core)
 *             cache_for_type[key] = fragment
 */
      __pyx_t_11 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create(__pyx_t_10, ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None), ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_4), (!__pyx_t_12))); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 774, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_fragment, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":776
 *             fragment = IndexGlycanCompositionFragment._create(
 *                 fragment.mass, None, fragment.key, not fragment.is_core)
 *             cache_for_type[key] = fragment             # <<<<<<<<<<<<<<
 *             fragment.name = key
 *             fragment.index = self.counter
 */
      if (unlikely((PyObject_SetItem(__pyx_v_cache_for_type, __pyx_v_key, __pyx_v_fragment) < 0))) __PYX_ERR(0, 776, __pyx_L13_except_error)

      /* "glycresoft/_c/tandem/core_search.pyx":777
 *                 fragment.mass, None, fragment.key, not fragment.is_core)
 *             cache_for_type[key] = fragment
 *             fragment.name = key             # <<<<<<<<<<<<<<
 *             fragment.index = self.counter
 *             self._fragments.append(fragment)
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_fragment, __pyx_n_s_name_2, __pyx_v_key) < 0) __PYX_ERR(0, 777, __pyx_L13_except_error)

      /* "glycresoft/_c/tandem/core_search.pyx":778
 *             cache_for_type[key] = fragment
 *             fragment.name = key
 *             fragment.index = self.counter             # <<<<<<<<<<<<<<
 *             self._fragments.append(fragment)
 *             self.counter += 1
 */
      __pyx_t_11 = __Pyx_PyInt_From_long(__pyx_v_self->counter); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 778, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_fragment, __pyx_n_s_index, __pyx_t_11) < 0) __PYX_ERR(0, 778, __pyx_L13_except_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":779
 *             fragment.name = key
 *             fragment.index = self.counter
 *             self._fragments.append(fragment)             # <<<<<<<<<<<<<<
 *             self.counter += 1
 *             return fragment
 */
      if (unlikely(__pyx_v_self->_fragments == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 779, __pyx_L13_except_error)
      }
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_self->_fragments, __pyx_v_fragment); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 779, __pyx_L13_except_error)

      /* "glycresoft/_c/tandem/core_search.pyx":780
 *             fragment.index = self.counter
 *             self._fragments.append(fragment)
 *             self.counter += 1             # <<<<<<<<<<<<<<
 *             return fragment
 * 
 */
      __pyx_v_self->counter = (__pyx_v_self->counter + 1);

      /* "glycresoft/_c/tandem/core_search.pyx":781
 *             self._fragments.append(fragment)
 *             self.counter += 1
 *             return fragment             # <<<<<<<<<<<<<<
 * 
 *     cpdef long query_width(self, double mass, double error_tolerance):
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      if (!(likely(((__pyx_v_fragment) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_fragment, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment))))) __PYX_ERR(0, 781, __pyx_L13_except_error)
      __Pyx_INCREF(__pyx_v_fragment);
      __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_v_fragment);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L14_except_return;
    }
    goto __pyx_L13_except_error;

    /* "glycresoft/_c/tandem/core_search.pyx":771
 *         except KeyError:
 *             cache_for_type = self.unique_fragments[glycosylation_type] = dict()
 *         try:             # <<<<<<<<<<<<<<
 *             return cache_for_type[key]
 *         except KeyError:
 */
    __pyx_L13_except_error:;
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L15_try_return:;
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L14_except_return:;
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
    goto __pyx_L0;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":765
 *         self.resolution = resolution
 * 
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):             # <<<<<<<<<<<<<<
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._intern", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_cache_for_type);
  __Pyx_XDECREF(__pyx_v_fragment);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_2_intern, "GlycanFragmentIndex._intern(self, fragment, glycosylation_type) -> IndexGlycanCompositionFragment");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern = {"_intern", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_2_intern};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fragment = 0;
  PyObject *__pyx_v_glycosylation_type = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_intern (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fragment,&__pyx_n_s_glycosylation_type,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fragment)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 765, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycosylation_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 765, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_intern", 1, 2, 2, 1); __PYX_ERR(0, 765, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_intern") < 0)) __PYX_ERR(0, 765, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_fragment = values[0];
    __pyx_v_glycosylation_type = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_intern", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 765, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._intern", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_2_intern(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v_fragment, __pyx_v_glycosylation_type);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_2_intern(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_fragment, PyObject *__pyx_v_glycosylation_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_intern", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__intern(__pyx_v_self, __pyx_v_fragment, __pyx_v_glycosylation_type, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._intern", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":783
 *             return fragment
 * 
 *     cpdef long query_width(self, double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static long __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_query_width(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass, double __pyx_v_error_tolerance, int __pyx_skip_dispatch) {
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  long __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_width", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query_width); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width)) {
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 783, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 783, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __pyx_t_8 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_8 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_8;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":784
 * 
 *     cpdef long query_width(self, double mass, double error_tolerance):
 *         return <long>(mass * error_tolerance * self.resolution) + 1             # <<<<<<<<<<<<<<
 * 
 *     cpdef long bin_index(self, double mass):
 */
  __pyx_r = (((long)((__pyx_v_mass * __pyx_v_error_tolerance) * __pyx_v_self->resolution)) + 1);
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":783
 *             return fragment
 * 
 *     cpdef long query_width(self, double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.query_width", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_4query_width, "GlycanFragmentIndex.query_width(self, double mass, double error_tolerance) -> long");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width = {"query_width", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_4query_width};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_mass;
  double __pyx_v_error_tolerance;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_width (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mass,&__pyx_n_s_error_tolerance,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_error_tolerance)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("query_width", 1, 2, 2, 1); __PYX_ERR(0, 783, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "query_width") < 0)) __PYX_ERR(0, 783, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_mass = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query_width", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 783, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.query_width", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_4query_width(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v_mass, __pyx_v_error_tolerance);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_4query_width(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass, double __pyx_v_error_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_width", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_query_width(__pyx_v_self, __pyx_v_mass, __pyx_v_error_tolerance, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.query_width", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":786
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 *     cpdef long bin_index(self, double mass):             # <<<<<<<<<<<<<<
 *         return <long>(mass * self.resolution)
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static long __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_bin_index(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass, int __pyx_skip_dispatch) {
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  long __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin_index", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_bin_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index)) {
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 786, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 786, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":787
 * 
 *     cpdef long bin_index(self, double mass):
 *         return <long>(mass * self.resolution)             # <<<<<<<<<<<<<<
 * 
 *     def build(self):
 */
  __pyx_r = ((long)(__pyx_v_mass * __pyx_v_self->resolution));
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":786
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 *     cpdef long bin_index(self, double mass):             # <<<<<<<<<<<<<<
 *         return <long>(mass * self.resolution)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.bin_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_6bin_index, "GlycanFragmentIndex.bin_index(self, double mass) -> long");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index = {"bin_index", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_6bin_index};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_mass;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin_index (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mass,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "bin_index") < 0)) __PYX_ERR(0, 786, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_mass = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mass == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bin_index", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 786, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.bin_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_6bin_index(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v_mass);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_6bin_index(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin_index", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_bin_index(__pyx_v_self, __pyx_v_mass, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.bin_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":789
 *         return <long>(mass * self.resolution)
 * 
 *     def build(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_9build(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_8build, "GlycanFragmentIndex.build(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_9build = {"build", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_9build, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_8build};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_9build(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("build", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "build", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_8build(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":801
 *             GlycanCombinationRecordBase member
 * 
 *         self.members.sort(key=lambda x: (x.dehydrated_mass, x.id))             # <<<<<<<<<<<<<<
 *         self.fragment_index.clear()
 *         j = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5build_lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5build_lambda = {"lambda", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5build_lambda, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5build_lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 801, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "lambda") < 0)) __PYX_ERR(0, 801, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_x = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 801, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.build.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, __pyx_v_x);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_dehydrated_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.build.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":789
 *         return <long>(mass * self.resolution)
 * 
 *     def build(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_8build(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_v_key = 0;
  long __pyx_v_n_frags;
  long __pyx_v_n_core;
  long __pyx_v_j;
  double __pyx_v_mass;
  double __pyx_v_d;
  double __pyx_v_low;
  double __pyx_v_high;
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_comp_frags = 0;
  PyObject *__pyx_v_frags = 0;
  size_t __pyx_v_comp_i;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_frag = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_comp = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_member = 0;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_glycosylation_type = NULL;
  Py_ssize_t __pyx_v_frag_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  double __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12;
  int __pyx_t_13;
  long __pyx_t_14;
  int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  size_t __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":801
 *             GlycanCombinationRecordBase member
 * 
 *         self.members.sort(key=lambda x: (x.dehydrated_mass, x.id))             # <<<<<<<<<<<<<<
 *         self.fragment_index.clear()
 *         j = 0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->members, __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5build_lambda, 0, __pyx_n_s_build_locals_lambda, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":802
 * 
 *         self.members.sort(key=lambda x: (x.dehydrated_mass, x.id))
 *         self.fragment_index.clear()             # <<<<<<<<<<<<<<
 *         j = 0
 *         low = float('inf')
 */
  if (unlikely(__pyx_v_self->fragment_index == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(0, 802, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyDict_Clear(__pyx_v_self->fragment_index); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 802, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":803
 *         self.members.sort(key=lambda x: (x.dehydrated_mass, x.id))
 *         self.fragment_index.clear()
 *         j = 0             # <<<<<<<<<<<<<<
 *         low = float('inf')
 *         high = 0
 */
  __pyx_v_j = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":804
 *         self.fragment_index.clear()
 *         j = 0
 *         low = float('inf')             # <<<<<<<<<<<<<<
 *         high = 0
 *         for i, member in enumerate(self.members):
 */
  __pyx_t_5 = __Pyx_PyString_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_5 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L1_error)
  __pyx_v_low = __pyx_t_5;

  /* "glycresoft/_c/tandem/core_search.pyx":805
 *         j = 0
 *         low = float('inf')
 *         high = 0             # <<<<<<<<<<<<<<
 *         for i, member in enumerate(self.members):
 *             mass = member.dehydrated_mass
 */
  __pyx_v_high = 0.0;

  /* "glycresoft/_c/tandem/core_search.pyx":806
 *         low = float('inf')
 *         high = 0
 *         for i, member in enumerate(self.members):             # <<<<<<<<<<<<<<
 *             mass = member.dehydrated_mass
 *             for glycosylation_type in member.glycan_types:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  __pyx_t_2 = __pyx_v_self->members; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 806, __pyx_L1_error)
      #endif
      if (__pyx_t_6 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 806, __pyx_L1_error)
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase))))) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_member, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_t_1));
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":807
 *         high = 0
 *         for i, member in enumerate(self.members):
 *             mass = member.dehydrated_mass             # <<<<<<<<<<<<<<
 *             for glycosylation_type in member.glycan_types:
 *                 n_core = 0
 */
    __pyx_t_5 = __pyx_v_member->dehydrated_mass;
    __pyx_v_mass = __pyx_t_5;

    /* "glycresoft/_c/tandem/core_search.pyx":808
 *         for i, member in enumerate(self.members):
 *             mass = member.dehydrated_mass
 *             for glycosylation_type in member.glycan_types:             # <<<<<<<<<<<<<<
 *                 n_core = 0
 *                 n_frags = 0
 */
    if (unlikely(__pyx_v_member->glycan_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 808, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_member->glycan_types; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 808, __pyx_L1_error)
        #endif
        if (__pyx_t_7 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 808, __pyx_L1_error)
      #else
      __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 808, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_glycosylation_type, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":809
 *             mass = member.dehydrated_mass
 *             for glycosylation_type in member.glycan_types:
 *                 n_core = 0             # <<<<<<<<<<<<<<
 *                 n_frags = 0
 *                 frags = self._get_fragments(member, glycosylation_type)
 */
      __pyx_v_n_core = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":810
 *             for glycosylation_type in member.glycan_types:
 *                 n_core = 0
 *                 n_frags = 0             # <<<<<<<<<<<<<<
 *                 frags = self._get_fragments(member, glycosylation_type)
 *                 for frag_i in range(PyList_GET_SIZE(frags)):
 */
      __pyx_v_n_frags = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":811
 *                 n_core = 0
 *                 n_frags = 0
 *                 frags = self._get_fragments(member, glycosylation_type)             # <<<<<<<<<<<<<<
 *                 for frag_i in range(PyList_GET_SIZE(frags)):
 *                     frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)
 */
      __pyx_t_8 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self->__pyx_vtab)->_get_fragments(__pyx_v_self, __pyx_v_member, __pyx_v_glycosylation_type, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_frags, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":812
 *                 n_frags = 0
 *                 frags = self._get_fragments(member, glycosylation_type)
 *                 for frag_i in range(PyList_GET_SIZE(frags)):             # <<<<<<<<<<<<<<
 *                     frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)
 *                     if not frag.is_extended:
 */
      __pyx_t_9 = PyList_GET_SIZE(__pyx_v_frags);
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_frag_i = __pyx_t_11;

        /* "glycresoft/_c/tandem/core_search.pyx":813
 *                 frags = self._get_fragments(member, glycosylation_type)
 *                 for frag_i in range(PyList_GET_SIZE(frags)):
 *                     frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)             # <<<<<<<<<<<<<<
 *                     if not frag.is_extended:
 *                         n_core += 1
 */
        __pyx_t_12 = PyList_GET_ITEM(__pyx_v_frags, __pyx_v_frag_i);
        __pyx_t_8 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self->__pyx_vtab)->_intern(__pyx_v_self, ((PyObject *)__pyx_t_12), __pyx_v_glycosylation_type, 0)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 813, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_frag, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_t_8));
        __pyx_t_8 = 0;

        /* "glycresoft/_c/tandem/core_search.pyx":814
 *                 for frag_i in range(PyList_GET_SIZE(frags)):
 *                     frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)
 *                     if not frag.is_extended:             # <<<<<<<<<<<<<<
 *                         n_core += 1
 *                     n_frags += 1
 */
        __pyx_t_13 = (!__pyx_v_frag->__pyx_base.is_extended);
        if (__pyx_t_13) {

          /* "glycresoft/_c/tandem/core_search.pyx":815
 *                     frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)
 *                     if not frag.is_extended:
 *                         n_core += 1             # <<<<<<<<<<<<<<
 *                     n_frags += 1
 *                     d = mass - frag.mass
 */
          __pyx_v_n_core = (__pyx_v_n_core + 1);

          /* "glycresoft/_c/tandem/core_search.pyx":814
 *                 for frag_i in range(PyList_GET_SIZE(frags)):
 *                     frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)
 *                     if not frag.is_extended:             # <<<<<<<<<<<<<<
 *                         n_core += 1
 *                     n_frags += 1
 */
        }

        /* "glycresoft/_c/tandem/core_search.pyx":816
 *                     if not frag.is_extended:
 *                         n_core += 1
 *                     n_frags += 1             # <<<<<<<<<<<<<<
 *                     d = mass - frag.mass
 *                     if d < 0:
 */
        __pyx_v_n_frags = (__pyx_v_n_frags + 1);

        /* "glycresoft/_c/tandem/core_search.pyx":817
 *                         n_core += 1
 *                     n_frags += 1
 *                     d = mass - frag.mass             # <<<<<<<<<<<<<<
 *                     if d < 0:
 *                         d = 0
 */
        __pyx_v_d = (__pyx_v_mass - __pyx_v_frag->__pyx_base.mass);

        /* "glycresoft/_c/tandem/core_search.pyx":818
 *                     n_frags += 1
 *                     d = mass - frag.mass
 *                     if d < 0:             # <<<<<<<<<<<<<<
 *                         d = 0
 *                     if high < d:
 */
        __pyx_t_13 = (__pyx_v_d < 0.0);
        if (__pyx_t_13) {

          /* "glycresoft/_c/tandem/core_search.pyx":819
 *                     d = mass - frag.mass
 *                     if d < 0:
 *                         d = 0             # <<<<<<<<<<<<<<
 *                     if high < d:
 *                         high = d
 */
          __pyx_v_d = 0.0;

          /* "glycresoft/_c/tandem/core_search.pyx":818
 *                     n_frags += 1
 *                     d = mass - frag.mass
 *                     if d < 0:             # <<<<<<<<<<<<<<
 *                         d = 0
 *                     if high < d:
 */
        }

        /* "glycresoft/_c/tandem/core_search.pyx":820
 *                     if d < 0:
 *                         d = 0
 *                     if high < d:             # <<<<<<<<<<<<<<
 *                         high = d
 *                     if low > d:
 */
        __pyx_t_13 = (__pyx_v_high < __pyx_v_d);
        if (__pyx_t_13) {

          /* "glycresoft/_c/tandem/core_search.pyx":821
 *                         d = 0
 *                     if high < d:
 *                         high = d             # <<<<<<<<<<<<<<
 *                     if low > d:
 *                         low = d
 */
          __pyx_v_high = __pyx_v_d;

          /* "glycresoft/_c/tandem/core_search.pyx":820
 *                     if d < 0:
 *                         d = 0
 *                     if high < d:             # <<<<<<<<<<<<<<
 *                         high = d
 *                     if low > d:
 */
        }

        /* "glycresoft/_c/tandem/core_search.pyx":822
 *                     if high < d:
 *                         high = d
 *                     if low > d:             # <<<<<<<<<<<<<<
 *                         low = d
 *                     key = PyInt_FromLong(self.bin_index(d))
 */
        __pyx_t_13 = (__pyx_v_low > __pyx_v_d);
        if (__pyx_t_13) {

          /* "glycresoft/_c/tandem/core_search.pyx":823
 *                         high = d
 *                     if low > d:
 *                         low = d             # <<<<<<<<<<<<<<
 *                     key = PyInt_FromLong(self.bin_index(d))
 *                     ptemp = PyDict_GetItem(self.fragment_index, key)
 */
          __pyx_v_low = __pyx_v_d;

          /* "glycresoft/_c/tandem/core_search.pyx":822
 *                     if high < d:
 *                         high = d
 *                     if low > d:             # <<<<<<<<<<<<<<
 *                         low = d
 *                     key = PyInt_FromLong(self.bin_index(d))
 */
        }

        /* "glycresoft/_c/tandem/core_search.pyx":824
 *                     if low > d:
 *                         low = d
 *                     key = PyInt_FromLong(self.bin_index(d))             # <<<<<<<<<<<<<<
 *                     ptemp = PyDict_GetItem(self.fragment_index, key)
 *                     if ptemp == NULL:
 */
        __pyx_t_14 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self->__pyx_vtab)->bin_index(__pyx_v_self, __pyx_v_d, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L1_error)
        __pyx_t_8 = PyInt_FromLong(__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 824, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "glycresoft/_c/tandem/core_search.pyx":825
 *                         low = d
 *                     key = PyInt_FromLong(self.bin_index(d))
 *                     ptemp = PyDict_GetItem(self.fragment_index, key)             # <<<<<<<<<<<<<<
 *                     if ptemp == NULL:
 *                         comp_frags = []
 */
        __pyx_t_8 = __pyx_v_self->fragment_index;
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_ptemp = PyDict_GetItem(__pyx_t_8, __pyx_v_key);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "glycresoft/_c/tandem/core_search.pyx":826
 *                     key = PyInt_FromLong(self.bin_index(d))
 *                     ptemp = PyDict_GetItem(self.fragment_index, key)
 *                     if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                         comp_frags = []
 *                         PyDict_SetItem(self.fragment_index, key, comp_frags)
 */
        __pyx_t_13 = (__pyx_v_ptemp == NULL);
        if (__pyx_t_13) {

          /* "glycresoft/_c/tandem/core_search.pyx":827
 *                     ptemp = PyDict_GetItem(self.fragment_index, key)
 *                     if ptemp == NULL:
 *                         comp_frags = []             # <<<<<<<<<<<<<<
 *                         PyDict_SetItem(self.fragment_index, key, comp_frags)
 *                     else:
 */
          __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 827, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_XDECREF_SET(__pyx_v_comp_frags, ((PyObject*)__pyx_t_8));
          __pyx_t_8 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":828
 *                     if ptemp == NULL:
 *                         comp_frags = []
 *                         PyDict_SetItem(self.fragment_index, key, comp_frags)             # <<<<<<<<<<<<<<
 *                     else:
 *                         comp_frags = <list>ptemp
 */
          __pyx_t_8 = __pyx_v_self->fragment_index;
          __Pyx_INCREF(__pyx_t_8);
          __pyx_t_15 = PyDict_SetItem(__pyx_t_8, __pyx_v_key, __pyx_v_comp_frags); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 828, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":826
 *                     key = PyInt_FromLong(self.bin_index(d))
 *                     ptemp = PyDict_GetItem(self.fragment_index, key)
 *                     if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                         comp_frags = []
 *                         PyDict_SetItem(self.fragment_index, key, comp_frags)
 */
          goto __pyx_L13;
        }

        /* "glycresoft/_c/tandem/core_search.pyx":830
 *                         PyDict_SetItem(self.fragment_index, key, comp_frags)
 *                     else:
 *                         comp_frags = <list>ptemp             # <<<<<<<<<<<<<<
 *                     for comp_i in range(PyList_GET_SIZE(comp_frags)):
 *                         comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 */
        /*else*/ {
          __pyx_t_8 = ((PyObject *)__pyx_v_ptemp);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_XDECREF_SET(__pyx_v_comp_frags, ((PyObject*)__pyx_t_8));
          __pyx_t_8 = 0;
        }
        __pyx_L13:;

        /* "glycresoft/_c/tandem/core_search.pyx":831
 *                     else:
 *                         comp_frags = <list>ptemp
 *                     for comp_i in range(PyList_GET_SIZE(comp_frags)):             # <<<<<<<<<<<<<<
 *                         comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                         if abs(comp.mass - d) <= 1e-3:
 */
        __pyx_t_16 = PyList_GET_SIZE(__pyx_v_comp_frags);
        __pyx_t_17 = __pyx_t_16;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
          __pyx_v_comp_i = __pyx_t_18;

          /* "glycresoft/_c/tandem/core_search.pyx":832
 *                         comp_frags = <list>ptemp
 *                     for comp_i in range(PyList_GET_SIZE(comp_frags)):
 *                         comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)             # <<<<<<<<<<<<<<
 *                         if abs(comp.mass - d) <= 1e-3:
 *                             PyList_Append(comp.keys, (frag, i, glycosylation_type))
 */
          __pyx_t_12 = PyList_GET_ITEM(__pyx_v_comp_frags, __pyx_v_comp_i);
          __pyx_t_8 = ((PyObject *)__pyx_t_12);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_XDECREF_SET(__pyx_v_comp, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_t_8));
          __pyx_t_8 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":833
 *                     for comp_i in range(PyList_GET_SIZE(comp_frags)):
 *                         comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                         if abs(comp.mass - d) <= 1e-3:             # <<<<<<<<<<<<<<
 *                             PyList_Append(comp.keys, (frag, i, glycosylation_type))
 *                             break
 */
          __pyx_t_13 = (fabs((__pyx_v_comp->mass - __pyx_v_d)) <= 1e-3);
          if (__pyx_t_13) {

            /* "glycresoft/_c/tandem/core_search.pyx":834
 *                         comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                         if abs(comp.mass - d) <= 1e-3:
 *                             PyList_Append(comp.keys, (frag, i, glycosylation_type))             # <<<<<<<<<<<<<<
 *                             break
 *                     else:
 */
            __pyx_t_8 = __pyx_v_comp->keys;
            __Pyx_INCREF(__pyx_t_8);
            __pyx_t_19 = PyTuple_New(3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 834, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_INCREF((PyObject *)__pyx_v_frag);
            __Pyx_GIVEREF((PyObject *)__pyx_v_frag);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 0, ((PyObject *)__pyx_v_frag))) __PYX_ERR(0, 834, __pyx_L1_error);
            __Pyx_INCREF(__pyx_v_i);
            __Pyx_GIVEREF(__pyx_v_i);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_v_i)) __PYX_ERR(0, 834, __pyx_L1_error);
            __Pyx_INCREF(__pyx_v_glycosylation_type);
            __Pyx_GIVEREF(__pyx_v_glycosylation_type);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_v_glycosylation_type)) __PYX_ERR(0, 834, __pyx_L1_error);
            __pyx_t_15 = PyList_Append(__pyx_t_8, __pyx_t_19); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 834, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

            /* "glycresoft/_c/tandem/core_search.pyx":835
 *                         if abs(comp.mass - d) <= 1e-3:
 *                             PyList_Append(comp.keys, (frag, i, glycosylation_type))
 *                             break             # <<<<<<<<<<<<<<
 *                     else:
 *                         comp_frags.append(ComplementFragment(d, [(frag, i, glycosylation_type)]))
 */
            goto __pyx_L15_break;

            /* "glycresoft/_c/tandem/core_search.pyx":833
 *                     for comp_i in range(PyList_GET_SIZE(comp_frags)):
 *                         comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                         if abs(comp.mass - d) <= 1e-3:             # <<<<<<<<<<<<<<
 *                             PyList_Append(comp.keys, (frag, i, glycosylation_type))
 *                             break
 */
          }
        }
        /*else*/ {

          /* "glycresoft/_c/tandem/core_search.pyx":837
 *                             break
 *                     else:
 *                         comp_frags.append(ComplementFragment(d, [(frag, i, glycosylation_type)]))             # <<<<<<<<<<<<<<
 *                         j += 1
 *                 member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)
 */
          if (unlikely(__pyx_v_comp_frags == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
            __PYX_ERR(0, 837, __pyx_L1_error)
          }
          __pyx_t_19 = PyFloat_FromDouble(__pyx_v_d); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_INCREF((PyObject *)__pyx_v_frag);
          __Pyx_GIVEREF((PyObject *)__pyx_v_frag);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_v_frag))) __PYX_ERR(0, 837, __pyx_L1_error);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_i)) __PYX_ERR(0, 837, __pyx_L1_error);
          __Pyx_INCREF(__pyx_v_glycosylation_type);
          __Pyx_GIVEREF(__pyx_v_glycosylation_type);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_glycosylation_type)) __PYX_ERR(0, 837, __pyx_L1_error);
          __pyx_t_20 = PyList_New(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_GIVEREF(__pyx_t_8);
          if (__Pyx_PyList_SET_ITEM(__pyx_t_20, 0, __pyx_t_8)) __PYX_ERR(0, 837, __pyx_L1_error);
          __pyx_t_8 = 0;
          __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_19);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_19)) __PYX_ERR(0, 837, __pyx_L1_error);
          __Pyx_GIVEREF(__pyx_t_20);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_20)) __PYX_ERR(0, 837, __pyx_L1_error);
          __pyx_t_19 = 0;
          __pyx_t_20 = 0;
          __pyx_t_20 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment), __pyx_t_8, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_comp_frags, __pyx_t_20); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 837, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":838
 *                     else:
 *                         comp_frags.append(ComplementFragment(d, [(frag, i, glycosylation_type)]))
 *                         j += 1             # <<<<<<<<<<<<<<
 *                 member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)
 *             member.clear()
 */
          __pyx_v_j = (__pyx_v_j + 1);
        }
        __pyx_L15_break:;
      }

      /* "glycresoft/_c/tandem/core_search.pyx":839
 *                         comp_frags.append(ComplementFragment(d, [(frag, i, glycosylation_type)]))
 *                         j += 1
 *                 member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)             # <<<<<<<<<<<<<<
 *             member.clear()
 *         self.lower_bound = low
 */
      __pyx_t_20 = __Pyx_PyInt_From_long(__pyx_v_n_core); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_8 = __Pyx_PyInt_From_long(__pyx_v_n_frags); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_GIVEREF(__pyx_t_20);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_20)) __PYX_ERR(0, 839, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_8)) __PYX_ERR(0, 839, __pyx_L1_error);
      __pyx_t_20 = 0;
      __pyx_t_8 = 0;
      if (unlikely(__pyx_v_member->fragment_set_properties == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 839, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_member->fragment_set_properties, __pyx_v_glycosylation_type, __pyx_t_19) < 0))) __PYX_ERR(0, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":808
 *         for i, member in enumerate(self.members):
 *             mass = member.dehydrated_mass
 *             for glycosylation_type in member.glycan_types:             # <<<<<<<<<<<<<<
 *                 n_core = 0
 *                 n_frags = 0
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":840
 *                         j += 1
 *                 member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)
 *             member.clear()             # <<<<<<<<<<<<<<
 *         self.lower_bound = low
 *         self.upper_bound = high
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_member->__pyx_vtab)->clear(__pyx_v_member, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":806
 *         low = float('inf')
 *         high = 0
 *         for i, member in enumerate(self.members):             # <<<<<<<<<<<<<<
 *             mass = member.dehydrated_mass
 *             for glycosylation_type in member.glycan_types:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":841
 *                 member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)
 *             member.clear()
 *         self.lower_bound = low             # <<<<<<<<<<<<<<
 *         self.upper_bound = high
 *         return j
 */
  __pyx_v_self->lower_bound = __pyx_v_low;

  /* "glycresoft/_c/tandem/core_search.pyx":842
 *             member.clear()
 *         self.lower_bound = low
 *         self.upper_bound = high             # <<<<<<<<<<<<<<
 *         return j
 * 
 */
  __pyx_v_self->upper_bound = __pyx_v_high;

  /* "glycresoft/_c/tandem/core_search.pyx":843
 *         self.lower_bound = low
 *         self.upper_bound = high
 *         return j             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":789
 *         return <long>(mass * self.resolution)
 * 
 *     def build(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_comp_frags);
  __Pyx_XDECREF(__pyx_v_frags);
  __Pyx_XDECREF((PyObject *)__pyx_v_frag);
  __Pyx_XDECREF((PyObject *)__pyx_v_comp);
  __Pyx_XDECREF((PyObject *)__pyx_v_member);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_glycosylation_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":845
 *         return j
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):             # <<<<<<<<<<<<<<
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__get_fragments(CYTHON_UNUSED struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_gcr, PyObject *__pyx_v_glycosylation_type, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_fragments", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_fragments); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_v_gcr), __pyx_v_glycosylation_type};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 845, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":846
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):
 *         if glycosylation_type == GlycanTypes_n_glycan:             # <<<<<<<<<<<<<<
 *             return gcr.get_n_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_o_glycan:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_glycosylation_type, __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 846, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "glycresoft/_c/tandem/core_search.pyx":847
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()             # <<<<<<<<<<<<<<
 *         elif glycosylation_type == GlycanTypes_o_glycan:
 *             return gcr.get_o_glycan_fragments()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_gcr->__pyx_vtab)->get_n_glycan_fragments(__pyx_v_gcr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":846
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):
 *         if glycosylation_type == GlycanTypes_n_glycan:             # <<<<<<<<<<<<<<
 *             return gcr.get_n_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_o_glycan:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":848
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_o_glycan:             # <<<<<<<<<<<<<<
 *             return gcr.get_o_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_gag_linker:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_glycosylation_type, __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 848, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "glycresoft/_c/tandem/core_search.pyx":849
 *             return gcr.get_n_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_o_glycan:
 *             return gcr.get_o_glycan_fragments()             # <<<<<<<<<<<<<<
 *         elif glycosylation_type == GlycanTypes_gag_linker:
 *             return gcr.get_gag_linker_glycan_fragments()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_gcr->__pyx_vtab)->get_o_glycan_fragments(__pyx_v_gcr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":848
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_o_glycan:             # <<<<<<<<<<<<<<
 *             return gcr.get_o_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_gag_linker:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":850
 *         elif glycosylation_type == GlycanTypes_o_glycan:
 *             return gcr.get_o_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_gag_linker:             # <<<<<<<<<<<<<<
 *             return gcr.get_gag_linker_glycan_fragments()
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_glycosylation_type, __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_6)) {

    /* "glycresoft/_c/tandem/core_search.pyx":851
 *             return gcr.get_o_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_gag_linker:
 *             return gcr.get_gag_linker_glycan_fragments()             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError(glycosylation_type)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v_gcr->__pyx_vtab)->get_gag_linker_glycan_fragments(__pyx_v_gcr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "glycresoft/_c/tandem/core_search.pyx":850
 *         elif glycosylation_type == GlycanTypes_o_glycan:
 *             return gcr.get_o_glycan_fragments()
 *         elif glycosylation_type == GlycanTypes_gag_linker:             # <<<<<<<<<<<<<<
 *             return gcr.get_gag_linker_glycan_fragments()
 *         else:
 */
  }

  /* "glycresoft/_c/tandem/core_search.pyx":853
 *             return gcr.get_gag_linker_glycan_fragments()
 *         else:
 *             raise ValueError(glycosylation_type)             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_glycosylation_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 853, __pyx_L1_error)
  }

  /* "glycresoft/_c/tandem/core_search.pyx":845
 *         return j
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):             # <<<<<<<<<<<<<<
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._get_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_get_fragments, "GlycanFragmentIndex._get_fragments(self, GlycanCombinationRecordBase gcr, glycosylation_type) -> list");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments = {"_get_fragments", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_get_fragments};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_gcr = 0;
  PyObject *__pyx_v_glycosylation_type = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_fragments (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gcr,&__pyx_n_s_glycosylation_type,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gcr)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 845, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycosylation_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 845, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_get_fragments", 1, 2, 2, 1); __PYX_ERR(0, 845, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_fragments") < 0)) __PYX_ERR(0, 845, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_gcr = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)values[0]);
    __pyx_v_glycosylation_type = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_fragments", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 845, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._get_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gcr), __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, 1, "gcr", 0))) __PYX_ERR(0, 845, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_get_fragments(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v_gcr, __pyx_v_glycosylation_type);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_get_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_gcr, PyObject *__pyx_v_glycosylation_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_fragments", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__get_fragments(__pyx_v_self, __pyx_v_gcr, __pyx_v_glycosylation_type, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._get_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":856
 * 
 *     @cython.cdivision
 *     cdef int _match_fragments(self, double delta_mass, DeconvolutedPeak peak, double error_tolerance, dict result):             # <<<<<<<<<<<<<<
 *             cdef:
 *                 object idx, glycosylation_type
 */

static int __pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__match_fragments(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, double __pyx_v_delta_mass, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak, double __pyx_v_error_tolerance, PyObject *__pyx_v_result) {
  PyObject *__pyx_v_idx = 0;
  PyObject *__pyx_v_glycosylation_type = 0;
  PyObject *__pyx_v_frag_id = 0;
  PyObject *__pyx_v_comp_frags = 0;
  PyObject *__pyx_v_keys = 0;
  PyObject *__pyx_v_sols_for_glycan = 0;
  long __pyx_v_key;
  long __pyx_v_width;
  long __pyx_v_off;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v_comp_frag = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *__pyx_v_frag = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_sol = 0;
  size_t __pyx_v_comp_i;
  size_t __pyx_v_comp_n;
  size_t __pyx_v_key_i;
  size_t __pyx_v_key_n;
  PyObject *__pyx_v_ptemp;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  long __pyx_t_2;
  long __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  int __pyx_t_14;
  npy_int32 __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_match_fragments", 1);

  /* "glycresoft/_c/tandem/core_search.pyx":870
 *                 PyObject* ptemp
 * 
 *             key = self.bin_index(delta_mass)             # <<<<<<<<<<<<<<
 *             width = <long>(peak.neutral_mass * error_tolerance * self.resolution) + 1
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self->__pyx_vtab)->bin_index(__pyx_v_self, __pyx_v_delta_mass, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L1_error)
  __pyx_v_key = __pyx_t_1;

  /* "glycresoft/_c/tandem/core_search.pyx":871
 * 
 *             key = self.bin_index(delta_mass)
 *             width = <long>(peak.neutral_mass * error_tolerance * self.resolution) + 1             # <<<<<<<<<<<<<<
 * 
 *             for off in range(-width, width + 1):
 */
  __pyx_v_width = (((long)((__pyx_v_peak->neutral_mass * __pyx_v_error_tolerance) * __pyx_v_self->resolution)) + 1);

  /* "glycresoft/_c/tandem/core_search.pyx":873
 *             width = <long>(peak.neutral_mass * error_tolerance * self.resolution) + 1
 * 
 *             for off in range(-width, width + 1):             # <<<<<<<<<<<<<<
 *                 ptemp = PyDict_GetItem(self.fragment_index, key + off)
 *                 if ptemp == NULL:
 */
  __pyx_t_1 = (__pyx_v_width + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = (-__pyx_v_width); __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_off = __pyx_t_3;

    /* "glycresoft/_c/tandem/core_search.pyx":874
 * 
 *             for off in range(-width, width + 1):
 *                 ptemp = PyDict_GetItem(self.fragment_index, key + off)             # <<<<<<<<<<<<<<
 *                 if ptemp == NULL:
 *                     continue
 */
    __pyx_t_4 = __pyx_v_self->fragment_index;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_long((__pyx_v_key + __pyx_v_off)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_ptemp = PyDict_GetItem(__pyx_t_4, __pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":875
 *             for off in range(-width, width + 1):
 *                 ptemp = PyDict_GetItem(self.fragment_index, key + off)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
    __pyx_t_6 = (__pyx_v_ptemp == NULL);
    if (__pyx_t_6) {

      /* "glycresoft/_c/tandem/core_search.pyx":876
 *                 ptemp = PyDict_GetItem(self.fragment_index, key + off)
 *                 if ptemp == NULL:
 *                     continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     comp_frags = <list>ptemp
 */
      goto __pyx_L3_continue;

      /* "glycresoft/_c/tandem/core_search.pyx":875
 *             for off in range(-width, width + 1):
 *                 ptemp = PyDict_GetItem(self.fragment_index, key + off)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
    }

    /* "glycresoft/_c/tandem/core_search.pyx":878
 *                     continue
 *                 else:
 *                     comp_frags = <list>ptemp             # <<<<<<<<<<<<<<
 *                 comp_n = PyList_GET_SIZE(comp_frags)
 *                 for comp_i in range(comp_n):
 */
    /*else*/ {
      __pyx_t_5 = ((PyObject *)__pyx_v_ptemp);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_comp_frags, ((PyObject*)__pyx_t_5));
      __pyx_t_5 = 0;
    }

    /* "glycresoft/_c/tandem/core_search.pyx":879
 *                 else:
 *                     comp_frags = <list>ptemp
 *                 comp_n = PyList_GET_SIZE(comp_frags)             # <<<<<<<<<<<<<<
 *                 for comp_i in range(comp_n):
 *                     comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 */
    __pyx_v_comp_n = PyList_GET_SIZE(__pyx_v_comp_frags);

    /* "glycresoft/_c/tandem/core_search.pyx":880
 *                     comp_frags = <list>ptemp
 *                 comp_n = PyList_GET_SIZE(comp_frags)
 *                 for comp_i in range(comp_n):             # <<<<<<<<<<<<<<
 *                     comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                     if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:
 */
    __pyx_t_7 = __pyx_v_comp_n;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_comp_i = __pyx_t_9;

      /* "glycresoft/_c/tandem/core_search.pyx":881
 *                 comp_n = PyList_GET_SIZE(comp_frags)
 *                 for comp_i in range(comp_n):
 *                     comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)             # <<<<<<<<<<<<<<
 *                     if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:
 *                         keys = comp_frag.keys
 */
      __pyx_t_10 = PyList_GET_ITEM(__pyx_v_comp_frags, __pyx_v_comp_i);
      __pyx_t_5 = ((PyObject *)__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_comp_frag, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":882
 *                 for comp_i in range(comp_n):
 *                     comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                     if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:             # <<<<<<<<<<<<<<
 *                         keys = comp_frag.keys
 *                         key_n = PyList_GET_SIZE(keys)
 */
      __pyx_t_6 = ((fabs((__pyx_v_comp_frag->mass - __pyx_v_delta_mass)) / __pyx_v_peak->neutral_mass) <= __pyx_v_error_tolerance);
      if (__pyx_t_6) {

        /* "glycresoft/_c/tandem/core_search.pyx":883
 *                     comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                     if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:
 *                         keys = comp_frag.keys             # <<<<<<<<<<<<<<
 *                         key_n = PyList_GET_SIZE(keys)
 *                         for key_i in range(key_n):
 */
        __pyx_t_5 = __pyx_v_comp_frag->keys;
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_keys, ((PyObject*)__pyx_t_5));
        __pyx_t_5 = 0;

        /* "glycresoft/_c/tandem/core_search.pyx":884
 *                     if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:
 *                         keys = comp_frag.keys
 *                         key_n = PyList_GET_SIZE(keys)             # <<<<<<<<<<<<<<
 *                         for key_i in range(key_n):
 *                             frag_id = <tuple>PyList_GET_ITEM(keys, key_i)
 */
        __pyx_v_key_n = PyList_GET_SIZE(__pyx_v_keys);

        /* "glycresoft/_c/tandem/core_search.pyx":885
 *                         keys = comp_frag.keys
 *                         key_n = PyList_GET_SIZE(keys)
 *                         for key_i in range(key_n):             # <<<<<<<<<<<<<<
 *                             frag_id = <tuple>PyList_GET_ITEM(keys, key_i)
 *                             frag = <IndexGlycanCompositionFragment>PyTuple_GET_ITEM(frag_id, 0)
 */
        __pyx_t_11 = __pyx_v_key_n;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_key_i = __pyx_t_13;

          /* "glycresoft/_c/tandem/core_search.pyx":886
 *                         key_n = PyList_GET_SIZE(keys)
 *                         for key_i in range(key_n):
 *                             frag_id = <tuple>PyList_GET_ITEM(keys, key_i)             # <<<<<<<<<<<<<<
 *                             frag = <IndexGlycanCompositionFragment>PyTuple_GET_ITEM(frag_id, 0)
 *                             idx = <object>PyTuple_GET_ITEM(frag_id, 1)
 */
          __pyx_t_10 = PyList_GET_ITEM(__pyx_v_keys, __pyx_v_key_i);
          __pyx_t_5 = ((PyObject *)__pyx_t_10);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_v_frag_id, ((PyObject*)__pyx_t_5));
          __pyx_t_5 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":887
 *                         for key_i in range(key_n):
 *                             frag_id = <tuple>PyList_GET_ITEM(keys, key_i)
 *                             frag = <IndexGlycanCompositionFragment>PyTuple_GET_ITEM(frag_id, 0)             # <<<<<<<<<<<<<<
 *                             idx = <object>PyTuple_GET_ITEM(frag_id, 1)
 *                             glycosylation_type = <object>PyTuple_GET_ITEM(frag_id, 2)
 */
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_v_frag_id, 0);
          __pyx_t_5 = ((PyObject *)__pyx_t_10);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_v_frag, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)__pyx_t_5));
          __pyx_t_5 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":888
 *                             frag_id = <tuple>PyList_GET_ITEM(keys, key_i)
 *                             frag = <IndexGlycanCompositionFragment>PyTuple_GET_ITEM(frag_id, 0)
 *                             idx = <object>PyTuple_GET_ITEM(frag_id, 1)             # <<<<<<<<<<<<<<
 *                             glycosylation_type = <object>PyTuple_GET_ITEM(frag_id, 2)
 * 
 */
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_v_frag_id, 1);
          __pyx_t_5 = ((PyObject *)__pyx_t_10);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":889
 *                             frag = <IndexGlycanCompositionFragment>PyTuple_GET_ITEM(frag_id, 0)
 *                             idx = <object>PyTuple_GET_ITEM(frag_id, 1)
 *                             glycosylation_type = <object>PyTuple_GET_ITEM(frag_id, 2)             # <<<<<<<<<<<<<<
 * 
 *                             # Fetch the dict for the glycan's solutions across glycosylation types.
 */
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_v_frag_id, 2);
          __pyx_t_5 = ((PyObject *)__pyx_t_10);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_v_glycosylation_type, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "glycresoft/_c/tandem/core_search.pyx":893
 *                             # Fetch the dict for the glycan's solutions across glycosylation types.
 *                             # If it doesn't exist, populate it now.
 *                             ptemp = PyDict_GetItem(result, idx)             # <<<<<<<<<<<<<<
 *                             if ptemp == NULL:
 *                                 sols_for_glycan = dict()
 */
          __pyx_v_ptemp = PyDict_GetItem(__pyx_v_result, __pyx_v_idx);

          /* "glycresoft/_c/tandem/core_search.pyx":894
 *                             # If it doesn't exist, populate it now.
 *                             ptemp = PyDict_GetItem(result, idx)
 *                             if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                                 sols_for_glycan = dict()
 *                                 PyDict_SetItem(result, idx, sols_for_glycan)
 */
          __pyx_t_6 = (__pyx_v_ptemp == NULL);
          if (__pyx_t_6) {

            /* "glycresoft/_c/tandem/core_search.pyx":895
 *                             ptemp = PyDict_GetItem(result, idx)
 *                             if ptemp == NULL:
 *                                 sols_for_glycan = dict()             # <<<<<<<<<<<<<<
 *                                 PyDict_SetItem(result, idx, sols_for_glycan)
 *                             else:
 */
            __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 895, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_XDECREF_SET(__pyx_v_sols_for_glycan, ((PyObject*)__pyx_t_5));
            __pyx_t_5 = 0;

            /* "glycresoft/_c/tandem/core_search.pyx":896
 *                             if ptemp == NULL:
 *                                 sols_for_glycan = dict()
 *                                 PyDict_SetItem(result, idx, sols_for_glycan)             # <<<<<<<<<<<<<<
 *                             else:
 *                                 sols_for_glycan = <dict>ptemp
 */
            __pyx_t_14 = PyDict_SetItem(__pyx_v_result, __pyx_v_idx, __pyx_v_sols_for_glycan); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 896, __pyx_L1_error)

            /* "glycresoft/_c/tandem/core_search.pyx":894
 *                             # If it doesn't exist, populate it now.
 *                             ptemp = PyDict_GetItem(result, idx)
 *                             if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                                 sols_for_glycan = dict()
 *                                 PyDict_SetItem(result, idx, sols_for_glycan)
 */
            goto __pyx_L11;
          }

          /* "glycresoft/_c/tandem/core_search.pyx":898
 *                                 PyDict_SetItem(result, idx, sols_for_glycan)
 *                             else:
 *                                 sols_for_glycan = <dict>ptemp             # <<<<<<<<<<<<<<
 *                             # From solutions for the current glycan across glycosylation types,
 *                             # fetch the solution for the matched glycosylation type. If it does not
 */
          /*else*/ {
            __pyx_t_5 = ((PyObject *)__pyx_v_ptemp);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_XDECREF_SET(__pyx_v_sols_for_glycan, ((PyObject*)__pyx_t_5));
            __pyx_t_5 = 0;
          }
          __pyx_L11:;

          /* "glycresoft/_c/tandem/core_search.pyx":902
 *                             # fetch the solution for the matched glycosylation type. If it does not
 *                             # exist, populate it now.
 *                             ptemp = PyDict_GetItem(sols_for_glycan, glycosylation_type)             # <<<<<<<<<<<<<<
 *                             if ptemp == NULL:
 *                                 sol = PartialGlycanSolution._create()
 */
          __pyx_v_ptemp = PyDict_GetItem(__pyx_v_sols_for_glycan, __pyx_v_glycosylation_type);

          /* "glycresoft/_c/tandem/core_search.pyx":903
 *                             # exist, populate it now.
 *                             ptemp = PyDict_GetItem(sols_for_glycan, glycosylation_type)
 *                             if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                                 sol = PartialGlycanSolution._create()
 *                                 sol.glycan_index = idx
 */
          __pyx_t_6 = (__pyx_v_ptemp == NULL);
          if (__pyx_t_6) {

            /* "glycresoft/_c/tandem/core_search.pyx":904
 *                             ptemp = PyDict_GetItem(sols_for_glycan, glycosylation_type)
 *                             if ptemp == NULL:
 *                                 sol = PartialGlycanSolution._create()             # <<<<<<<<<<<<<<
 *                                 sol.glycan_index = idx
 *                                 PyDict_SetItem(sols_for_glycan, glycosylation_type, sol)
 */
            __pyx_t_5 = ((PyObject *)__pyx_f_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution__create()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_XDECREF_SET(__pyx_v_sol, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_t_5));
            __pyx_t_5 = 0;

            /* "glycresoft/_c/tandem/core_search.pyx":905
 *                             if ptemp == NULL:
 *                                 sol = PartialGlycanSolution._create()
 *                                 sol.glycan_index = idx             # <<<<<<<<<<<<<<
 *                                 PyDict_SetItem(sols_for_glycan, glycosylation_type, sol)
 *                             else:
 */
            __pyx_t_15 = __Pyx_PyInt_As_npy_int32(__pyx_v_idx); if (unlikely((__pyx_t_15 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L1_error)
            __pyx_v_sol->glycan_index = __pyx_t_15;

            /* "glycresoft/_c/tandem/core_search.pyx":906
 *                                 sol = PartialGlycanSolution._create()
 *                                 sol.glycan_index = idx
 *                                 PyDict_SetItem(sols_for_glycan, glycosylation_type, sol)             # <<<<<<<<<<<<<<
 *                             else:
 *                                 sol = <PartialGlycanSolution>ptemp
 */
            __pyx_t_14 = PyDict_SetItem(__pyx_v_sols_for_glycan, __pyx_v_glycosylation_type, ((PyObject *)__pyx_v_sol)); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 906, __pyx_L1_error)

            /* "glycresoft/_c/tandem/core_search.pyx":903
 *                             # exist, populate it now.
 *                             ptemp = PyDict_GetItem(sols_for_glycan, glycosylation_type)
 *                             if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                                 sol = PartialGlycanSolution._create()
 *                                 sol.glycan_index = idx
 */
            goto __pyx_L12;
          }

          /* "glycresoft/_c/tandem/core_search.pyx":908
 *                                 PyDict_SetItem(sols_for_glycan, glycosylation_type, sol)
 *                             else:
 *                                 sol = <PartialGlycanSolution>ptemp             # <<<<<<<<<<<<<<
 *                             sol.score += math.log10(peak.intensity) * (
 *                                 1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
 */
          /*else*/ {
            __pyx_t_5 = ((PyObject *)__pyx_v_ptemp);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_XDECREF_SET(__pyx_v_sol, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_t_5));
            __pyx_t_5 = 0;
          }
          __pyx_L12:;

          /* "glycresoft/_c/tandem/core_search.pyx":909
 *                             else:
 *                                 sol = <PartialGlycanSolution>ptemp
 *                             sol.score += math.log10(peak.intensity) * (             # <<<<<<<<<<<<<<
 *                                 1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
 *                             if not frag.is_extended:
 */
          __pyx_v_sol->score = (__pyx_v_sol->score + (log10(__pyx_v_peak->__pyx_base.intensity) * (1.0 - pow(((fabs((__pyx_v_delta_mass - __pyx_v_comp_frag->mass)) / __pyx_v_peak->neutral_mass) / __pyx_v_error_tolerance), 4.0))));

          /* "glycresoft/_c/tandem/core_search.pyx":911
 *                             sol.score += math.log10(peak.intensity) * (
 *                                 1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
 *                             if not frag.is_extended:             # <<<<<<<<<<<<<<
 *                                 PySet_Add(sol.core_matches, frag.index)
 *                             PySet_Add(sol.fragment_matches, frag.index)
 */
          __pyx_t_6 = (!__pyx_v_frag->__pyx_base.is_extended);
          if (__pyx_t_6) {

            /* "glycresoft/_c/tandem/core_search.pyx":912
 *                                 1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
 *                             if not frag.is_extended:
 *                                 PySet_Add(sol.core_matches, frag.index)             # <<<<<<<<<<<<<<
 *                             PySet_Add(sol.fragment_matches, frag.index)
 *             return 0
 */
            __pyx_t_5 = __pyx_v_sol->core_matches;
            __Pyx_INCREF(__pyx_t_5);
            __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_frag->index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 912, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_14 = PySet_Add(__pyx_t_5, __pyx_t_4); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 912, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "glycresoft/_c/tandem/core_search.pyx":911
 *                             sol.score += math.log10(peak.intensity) * (
 *                                 1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
 *                             if not frag.is_extended:             # <<<<<<<<<<<<<<
 *                                 PySet_Add(sol.core_matches, frag.index)
 *                             PySet_Add(sol.fragment_matches, frag.index)
 */
          }

          /* "glycresoft/_c/tandem/core_search.pyx":913
 *                             if not frag.is_extended:
 *                                 PySet_Add(sol.core_matches, frag.index)
 *                             PySet_Add(sol.fragment_matches, frag.index)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
          __pyx_t_4 = __pyx_v_sol->fragment_matches;
          __Pyx_INCREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_frag->index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 913, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_14 = PySet_Add(__pyx_t_4, __pyx_t_5); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 913, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }

        /* "glycresoft/_c/tandem/core_search.pyx":882
 *                 for comp_i in range(comp_n):
 *                     comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
 *                     if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:             # <<<<<<<<<<<<<<
 *                         keys = comp_frag.keys
 *                         key_n = PyList_GET_SIZE(keys)
 */
      }
    }
    __pyx_L3_continue:;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":914
 *                                 PySet_Add(sol.core_matches, frag.index)
 *                             PySet_Add(sol.fragment_matches, frag.index)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":856
 * 
 *     @cython.cdivision
 *     cdef int _match_fragments(self, double delta_mass, DeconvolutedPeak peak, double error_tolerance, dict result):             # <<<<<<<<<<<<<<
 *             cdef:
 *                 object idx, glycosylation_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._match_fragments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_glycosylation_type);
  __Pyx_XDECREF(__pyx_v_frag_id);
  __Pyx_XDECREF(__pyx_v_comp_frags);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_sols_for_glycan);
  __Pyx_XDECREF((PyObject *)__pyx_v_comp_frag);
  __Pyx_XDECREF((PyObject *)__pyx_v_frag);
  __Pyx_XDECREF((PyObject *)__pyx_v_sol);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":917
 * 
 *     @cython.cdivision
 *     cpdef list match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double precursor_mass, d
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_scan, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)1e-5);
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift = __pyx_k__5;
  PyObject *__pyx_v_query_mass = ((PyObject *)Py_None);
  double __pyx_v_precursor_mass;
  double __pyx_v_d;
  double __pyx_v_best_score;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_sol = 0;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v_best_solution = 0;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_sols_for_glycan = 0;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set = 0;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_v_glycosylation_type = 0;
  PyObject *__pyx_v_id_sol = 0;
  PyObject *__pyx_v_glytype_sol = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_solutions = 0;
  double __pyx_v_coverage;
  double __pyx_v_mass_shift_tandem_mass;
  double __pyx_v_mass_shift_mass;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v_rec = 0;
  size_t __pyx_v_peak_i;
  size_t __pyx_v_n_peaks;
  size_t __pyx_v_n_core;
  size_t __pyx_v_n_frag;
  size_t __pyx_v_sols_i;
  size_t __pyx_v_sols_n;
  PyObject *__pyx_v_keys = NULL;
  Py_ssize_t __pyx_v_key_n;
  Py_ssize_t __pyx_v_key_i;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15;
  Py_ssize_t __pyx_t_16;
  long __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("match", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_mass_shift = __pyx_optional_args->mass_shift;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_query_mass = __pyx_optional_args->query_mass;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_match); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_5, __pyx_v_scan, __pyx_t_3, ((PyObject *)__pyx_v_mass_shift), __pyx_v_query_mass};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 4+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 917, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/tandem/core_search.pyx":933
 *             PyObject* ptemp
 * 
 *         if query_mass is None:             # <<<<<<<<<<<<<<
 *             precursor_mass = scan.precursor_information.neutral_mass
 *         else:
 */
  __pyx_t_7 = (__pyx_v_query_mass == Py_None);
  if (__pyx_t_7) {

    /* "glycresoft/_c/tandem/core_search.pyx":934
 * 
 *         if query_mass is None:
 *             precursor_mass = scan.precursor_information.neutral_mass             # <<<<<<<<<<<<<<
 *         else:
 *             precursor_mass = query_mass
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_scan, __pyx_n_s_precursor_information); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_neutral_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_precursor_mass = __pyx_t_8;

    /* "glycresoft/_c/tandem/core_search.pyx":933
 *             PyObject* ptemp
 * 
 *         if query_mass is None:             # <<<<<<<<<<<<<<
 *             precursor_mass = scan.precursor_information.neutral_mass
 *         else:
 */
    goto __pyx_L3;
  }

  /* "glycresoft/_c/tandem/core_search.pyx":936
 *             precursor_mass = scan.precursor_information.neutral_mass
 *         else:
 *             precursor_mass = query_mass             # <<<<<<<<<<<<<<
 * 
 *         mass_shift_mass = mass_shift.mass
 */
  /*else*/ {
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_v_query_mass); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L1_error)
    __pyx_v_precursor_mass = __pyx_t_8;
  }
  __pyx_L3:;

  /* "glycresoft/_c/tandem/core_search.pyx":938
 *             precursor_mass = query_mass
 * 
 *         mass_shift_mass = mass_shift.mass             # <<<<<<<<<<<<<<
 *         mass_shift_tandem_mass = mass_shift.tandem_mass
 * 
 */
  __pyx_t_8 = __pyx_v_mass_shift->mass;
  __pyx_v_mass_shift_mass = __pyx_t_8;

  /* "glycresoft/_c/tandem/core_search.pyx":939
 * 
 *         mass_shift_mass = mass_shift.mass
 *         mass_shift_tandem_mass = mass_shift.tandem_mass             # <<<<<<<<<<<<<<
 * 
 *         peak_set = scan.deconvoluted_peak_set
 */
  __pyx_t_8 = __pyx_v_mass_shift->tandem_mass;
  __pyx_v_mass_shift_tandem_mass = __pyx_t_8;

  /* "glycresoft/_c/tandem/core_search.pyx":941
 *         mass_shift_tandem_mass = mass_shift.tandem_mass
 * 
 *         peak_set = scan.deconvoluted_peak_set             # <<<<<<<<<<<<<<
 *         n_peaks = peak_set.get_size()
 *         result = dict()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_scan, __pyx_n_s_deconvoluted_peak_set); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet))))) __PYX_ERR(0, 941, __pyx_L1_error)
  __pyx_v_peak_set = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":942
 * 
 *         peak_set = scan.deconvoluted_peak_set
 *         n_peaks = peak_set.get_size()             # <<<<<<<<<<<<<<
 *         result = dict()
 *         for peak_i in range(n_peaks):
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set->__pyx_vtab)->get_size(__pyx_v_peak_set); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_v_n_peaks = __pyx_t_9;

  /* "glycresoft/_c/tandem/core_search.pyx":943
 *         peak_set = scan.deconvoluted_peak_set
 *         n_peaks = peak_set.get_size()
 *         result = dict()             # <<<<<<<<<<<<<<
 *         for peak_i in range(n_peaks):
 *             peak = peak_set.getitem(peak_i)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_result = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":944
 *         n_peaks = peak_set.get_size()
 *         result = dict()
 *         for peak_i in range(n_peaks):             # <<<<<<<<<<<<<<
 *             peak = peak_set.getitem(peak_i)
 *             # Subtract the precursor mass shift mass to match unmodified peaks
 */
  __pyx_t_9 = __pyx_v_n_peaks;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_peak_i = __pyx_t_11;

    /* "glycresoft/_c/tandem/core_search.pyx":945
 *         result = dict()
 *         for peak_i in range(n_peaks):
 *             peak = peak_set.getitem(peak_i)             # <<<<<<<<<<<<<<
 *             # Subtract the precursor mass shift mass to match unmodified peaks
 *             d = (precursor_mass - peak.neutral_mass - mass_shift_mass)
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set->__pyx_vtab)->getitem(__pyx_v_peak_set, __pyx_v_peak_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":947
 *             peak = peak_set.getitem(peak_i)
 *             # Subtract the precursor mass shift mass to match unmodified peaks
 *             d = (precursor_mass - peak.neutral_mass - mass_shift_mass)             # <<<<<<<<<<<<<<
 *             self._match_fragments(d, peak, error_tolerance, result)
 *             if mass_shift_tandem_mass != 0:
 */
    __pyx_v_d = ((__pyx_v_precursor_mass - __pyx_v_peak->neutral_mass) - __pyx_v_mass_shift_mass);

    /* "glycresoft/_c/tandem/core_search.pyx":948
 *             # Subtract the precursor mass shift mass to match unmodified peaks
 *             d = (precursor_mass - peak.neutral_mass - mass_shift_mass)
 *             self._match_fragments(d, peak, error_tolerance, result)             # <<<<<<<<<<<<<<
 *             if mass_shift_tandem_mass != 0:
 *                 # Add back the tandem mass shift to match modified peaks
 */
    ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self->__pyx_vtab)->_match_fragments(__pyx_v_self, __pyx_v_d, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_result); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L1_error)

    /* "glycresoft/_c/tandem/core_search.pyx":949
 *             d = (precursor_mass - peak.neutral_mass - mass_shift_mass)
 *             self._match_fragments(d, peak, error_tolerance, result)
 *             if mass_shift_tandem_mass != 0:             # <<<<<<<<<<<<<<
 *                 # Add back the tandem mass shift to match modified peaks
 *                 d = (precursor_mass - peak.neutral_mass - mass_shift_mass + mass_shift_tandem_mass)
 */
    __pyx_t_7 = (__pyx_v_mass_shift_tandem_mass != 0.0);
    if (__pyx_t_7) {

      /* "glycresoft/_c/tandem/core_search.pyx":951
 *             if mass_shift_tandem_mass != 0:
 *                 # Add back the tandem mass shift to match modified peaks
 *                 d = (precursor_mass - peak.neutral_mass - mass_shift_mass + mass_shift_tandem_mass)             # <<<<<<<<<<<<<<
 *                 self._match_fragments(d, peak, error_tolerance, result)
 * 
 */
      __pyx_v_d = (((__pyx_v_precursor_mass - __pyx_v_peak->neutral_mass) - __pyx_v_mass_shift_mass) + __pyx_v_mass_shift_tandem_mass);

      /* "glycresoft/_c/tandem/core_search.pyx":952
 *                 # Add back the tandem mass shift to match modified peaks
 *                 d = (precursor_mass - peak.neutral_mass - mass_shift_mass + mass_shift_tandem_mass)
 *                 self._match_fragments(d, peak, error_tolerance, result)             # <<<<<<<<<<<<<<
 * 
 *         out = []
 */
      ((struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self->__pyx_vtab)->_match_fragments(__pyx_v_self, __pyx_v_d, __pyx_v_peak, __pyx_v_error_tolerance, __pyx_v_result); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L1_error)

      /* "glycresoft/_c/tandem/core_search.pyx":949
 *             d = (precursor_mass - peak.neutral_mass - mass_shift_mass)
 *             self._match_fragments(d, peak, error_tolerance, result)
 *             if mass_shift_tandem_mass != 0:             # <<<<<<<<<<<<<<
 *                 # Add back the tandem mass shift to match modified peaks
 *                 d = (precursor_mass - peak.neutral_mass - mass_shift_mass + mass_shift_tandem_mass)
 */
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":954
 *                 self._match_fragments(d, peak, error_tolerance, result)
 * 
 *         out = []             # <<<<<<<<<<<<<<
 *         keys = PyDict_Items(result)
 *         key_n = PyList_GET_SIZE(keys)
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_out = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":955
 * 
 *         out = []
 *         keys = PyDict_Items(result)             # <<<<<<<<<<<<<<
 *         key_n = PyList_GET_SIZE(keys)
 *         for key_i in range(key_n):
 */
  __pyx_t_2 = PyDict_Items(__pyx_v_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_keys = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":956
 *         out = []
 *         keys = PyDict_Items(result)
 *         key_n = PyList_GET_SIZE(keys)             # <<<<<<<<<<<<<<
 *         for key_i in range(key_n):
 *             id_sol = <tuple>PyList_GET_ITEM(keys, key_i)
 */
  __pyx_v_key_n = PyList_GET_SIZE(__pyx_v_keys);

  /* "glycresoft/_c/tandem/core_search.pyx":957
 *         keys = PyDict_Items(result)
 *         key_n = PyList_GET_SIZE(keys)
 *         for key_i in range(key_n):             # <<<<<<<<<<<<<<
 *             id_sol = <tuple>PyList_GET_ITEM(keys, key_i)
 *             idx = <object>PyTuple_GET_ITEM(id_sol, 0)
 */
  __pyx_t_12 = __pyx_v_key_n;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_key_i = __pyx_t_14;

    /* "glycresoft/_c/tandem/core_search.pyx":958
 *         key_n = PyList_GET_SIZE(keys)
 *         for key_i in range(key_n):
 *             id_sol = <tuple>PyList_GET_ITEM(keys, key_i)             # <<<<<<<<<<<<<<
 *             idx = <object>PyTuple_GET_ITEM(id_sol, 0)
 *             sols_for_glycan = <dict>PyTuple_GET_ITEM(id_sol, 1)
 */
    __pyx_t_15 = PyList_GET_ITEM(__pyx_v_keys, __pyx_v_key_i);
    __pyx_t_2 = ((PyObject *)__pyx_t_15);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_id_sol, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":959
 *         for key_i in range(key_n):
 *             id_sol = <tuple>PyList_GET_ITEM(keys, key_i)
 *             idx = <object>PyTuple_GET_ITEM(id_sol, 0)             # <<<<<<<<<<<<<<
 *             sols_for_glycan = <dict>PyTuple_GET_ITEM(id_sol, 1)
 *             rec = <GlycanCombinationRecordBase>PyList_GET_ITEM(self.members, idx)
 */
    __pyx_t_15 = PyTuple_GET_ITEM(__pyx_v_id_sol, 0);
    __pyx_t_2 = ((PyObject *)__pyx_t_15);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":960
 *             id_sol = <tuple>PyList_GET_ITEM(keys, key_i)
 *             idx = <object>PyTuple_GET_ITEM(id_sol, 0)
 *             sols_for_glycan = <dict>PyTuple_GET_ITEM(id_sol, 1)             # <<<<<<<<<<<<<<
 *             rec = <GlycanCombinationRecordBase>PyList_GET_ITEM(self.members, idx)
 * 
 */
    __pyx_t_15 = PyTuple_GET_ITEM(__pyx_v_id_sol, 1);
    __pyx_t_2 = ((PyObject *)__pyx_t_15);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_sols_for_glycan, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":961
 *             idx = <object>PyTuple_GET_ITEM(id_sol, 0)
 *             sols_for_glycan = <dict>PyTuple_GET_ITEM(id_sol, 1)
 *             rec = <GlycanCombinationRecordBase>PyList_GET_ITEM(self.members, idx)             # <<<<<<<<<<<<<<
 * 
 *             best_score = -1e6
 */
    __pyx_t_2 = __pyx_v_self->members;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_16 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_16 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 961, __pyx_L1_error)
    __pyx_t_15 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_16);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = ((PyObject *)__pyx_t_15);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_rec, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":963
 *             rec = <GlycanCombinationRecordBase>PyList_GET_ITEM(self.members, idx)
 * 
 *             best_score = -1e6             # <<<<<<<<<<<<<<
 *             best_solution = None
 *             solutions = PyDict_Items(sols_for_glycan)
 */
    __pyx_v_best_score = -1e6;

    /* "glycresoft/_c/tandem/core_search.pyx":964
 * 
 *             best_score = -1e6
 *             best_solution = None             # <<<<<<<<<<<<<<
 *             solutions = PyDict_Items(sols_for_glycan)
 *             sols_n = PyList_GET_SIZE(solutions)
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_best_solution, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)Py_None));

    /* "glycresoft/_c/tandem/core_search.pyx":965
 *             best_score = -1e6
 *             best_solution = None
 *             solutions = PyDict_Items(sols_for_glycan)             # <<<<<<<<<<<<<<
 *             sols_n = PyList_GET_SIZE(solutions)
 *             for sols_i in range(sols_n):
 */
    __pyx_t_2 = PyDict_Items(__pyx_v_sols_for_glycan); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_solutions, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glycresoft/_c/tandem/core_search.pyx":966
 *             best_solution = None
 *             solutions = PyDict_Items(sols_for_glycan)
 *             sols_n = PyList_GET_SIZE(solutions)             # <<<<<<<<<<<<<<
 *             for sols_i in range(sols_n):
 *                 glytype_sol = <tuple>PyList_GET_ITEM(solutions, sols_i)
 */
    __pyx_v_sols_n = PyList_GET_SIZE(__pyx_v_solutions);

    /* "glycresoft/_c/tandem/core_search.pyx":967
 *             solutions = PyDict_Items(sols_for_glycan)
 *             sols_n = PyList_GET_SIZE(solutions)
 *             for sols_i in range(sols_n):             # <<<<<<<<<<<<<<
 *                 glytype_sol = <tuple>PyList_GET_ITEM(solutions, sols_i)
 * 
 */
    __pyx_t_9 = __pyx_v_sols_n;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_sols_i = __pyx_t_11;

      /* "glycresoft/_c/tandem/core_search.pyx":968
 *             sols_n = PyList_GET_SIZE(solutions)
 *             for sols_i in range(sols_n):
 *                 glytype_sol = <tuple>PyList_GET_ITEM(solutions, sols_i)             # <<<<<<<<<<<<<<
 * 
 *                 glycosylation_type = <object>PyTuple_GET_ITEM(glytype_sol, 0)
 */
      __pyx_t_15 = PyList_GET_ITEM(__pyx_v_solutions, __pyx_v_sols_i);
      __pyx_t_2 = ((PyObject *)__pyx_t_15);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_glytype_sol, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":970
 *                 glytype_sol = <tuple>PyList_GET_ITEM(solutions, sols_i)
 * 
 *                 glycosylation_type = <object>PyTuple_GET_ITEM(glytype_sol, 0)             # <<<<<<<<<<<<<<
 *                 sol = <PartialGlycanSolution>PyTuple_GET_ITEM(glytype_sol, 1)
 *                 # Reusing container variable lazily
 */
      __pyx_t_15 = PyTuple_GET_ITEM(__pyx_v_glytype_sol, 0);
      __pyx_t_2 = ((PyObject *)__pyx_t_15);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_glycosylation_type, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":971
 * 
 *                 glycosylation_type = <object>PyTuple_GET_ITEM(glytype_sol, 0)
 *                 sol = <PartialGlycanSolution>PyTuple_GET_ITEM(glytype_sol, 1)             # <<<<<<<<<<<<<<
 *                 # Reusing container variable lazily
 *                 id_sol = <tuple>PyDict_GetItem(rec.fragment_set_properties, glycosylation_type)
 */
      __pyx_t_15 = PyTuple_GET_ITEM(__pyx_v_glytype_sol, 1);
      __pyx_t_2 = ((PyObject *)__pyx_t_15);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_sol, ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":973
 *                 sol = <PartialGlycanSolution>PyTuple_GET_ITEM(glytype_sol, 1)
 *                 # Reusing container variable lazily
 *                 id_sol = <tuple>PyDict_GetItem(rec.fragment_set_properties, glycosylation_type)             # <<<<<<<<<<<<<<
 *                 n_core = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 0))
 *                 n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))
 */
      __pyx_t_2 = __pyx_v_rec->fragment_set_properties;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_15 = PyDict_GetItem(__pyx_t_2, __pyx_v_glycosylation_type);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = ((PyObject *)__pyx_t_15);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_id_sol, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycresoft/_c/tandem/core_search.pyx":974
 *                 # Reusing container variable lazily
 *                 id_sol = <tuple>PyDict_GetItem(rec.fragment_set_properties, glycosylation_type)
 *                 n_core = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 0))             # <<<<<<<<<<<<<<
 *                 n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))
 *                 coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (
 */
      __pyx_t_15 = PyTuple_GET_ITEM(__pyx_v_id_sol, 0);
      __pyx_t_17 = PyInt_AsLong(((PyObject *)__pyx_t_15)); if (unlikely(__pyx_t_17 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 974, __pyx_L1_error)
      __pyx_v_n_core = __pyx_t_17;

      /* "glycresoft/_c/tandem/core_search.pyx":975
 *                 id_sol = <tuple>PyDict_GetItem(rec.fragment_set_properties, glycosylation_type)
 *                 n_core = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 0))
 *                 n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))             # <<<<<<<<<<<<<<
 *                 coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (
 *                     PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
 */
      __pyx_t_15 = PyTuple_GET_ITEM(__pyx_v_id_sol, 1);
      __pyx_t_17 = PyInt_AsLong(((PyObject *)__pyx_t_15)); if (unlikely(__pyx_t_17 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 975, __pyx_L1_error)
      __pyx_v_n_frag = __pyx_t_17;

      /* "glycresoft/_c/tandem/core_search.pyx":976
 *                 n_core = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 0))
 *                 n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))
 *                 coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (             # <<<<<<<<<<<<<<
 *                     PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
 *                 sol.score *= coverage
 */
      __pyx_t_2 = __pyx_v_sol->core_matches;
      __Pyx_INCREF(__pyx_t_2);

      /* "glycresoft/_c/tandem/core_search.pyx":977
 *                 n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))
 *                 coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (
 *                     PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight             # <<<<<<<<<<<<<<
 *                 sol.score *= coverage
 *                 sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
 */
      __pyx_t_1 = __pyx_v_sol->fragment_matches;
      __Pyx_INCREF(__pyx_t_1);

      /* "glycresoft/_c/tandem/core_search.pyx":976
 *                 n_core = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 0))
 *                 n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))
 *                 coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (             # <<<<<<<<<<<<<<
 *                     PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
 *                 sol.score *= coverage
 */
      __pyx_t_8 = __Pyx_SoftComplexToDouble(__Pyx_c_prod_double(__Pyx_c_pow_double(__pyx_t_double_complex_from_parts(((PySet_GET_SIZE(__pyx_t_2) * 1.0) / ((double)__pyx_v_n_core)), 0), __pyx_t_double_complex_from_parts(__pyx_v_self->core_weight, 0)), __Pyx_c_pow_double(__pyx_t_double_complex_from_parts(((PySet_GET_SIZE(__pyx_t_1) * 1.0) / ((double)__pyx_v_n_frag)), 0), __pyx_t_double_complex_from_parts(__pyx_v_self->fragment_weight, 0))), 1); if (unlikely(__pyx_t_8 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 976, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_coverage = __pyx_t_8;

      /* "glycresoft/_c/tandem/core_search.pyx":978
 *                 coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (
 *                     PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
 *                 sol.score *= coverage             # <<<<<<<<<<<<<<
 *                 sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
 *                 if best_score < sol.score and sol.peptide_mass > 0:
 */
      __pyx_v_sol->score = (__pyx_v_sol->score * __pyx_v_coverage);

      /* "glycresoft/_c/tandem/core_search.pyx":979
 *                     PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
 *                 sol.score *= coverage
 *                 sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass             # <<<<<<<<<<<<<<
 *                 if best_score < sol.score and sol.peptide_mass > 0:
 *                     best_score = sol.score
 */
      __pyx_v_sol->peptide_mass = ((__pyx_v_precursor_mass - __pyx_v_rec->dehydrated_mass) - __pyx_v_mass_shift_mass);

      /* "glycresoft/_c/tandem/core_search.pyx":980
 *                 sol.score *= coverage
 *                 sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
 *                 if best_score < sol.score and sol.peptide_mass > 0:             # <<<<<<<<<<<<<<
 *                     best_score = sol.score
 *                     best_solution = sol
 */
      __pyx_t_18 = (__pyx_v_best_score < __pyx_v_sol->score);
      if (__pyx_t_18) {
      } else {
        __pyx_t_7 = __pyx_t_18;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_18 = (__pyx_v_sol->peptide_mass > 0.0);
      __pyx_t_7 = __pyx_t_18;
      __pyx_L12_bool_binop_done:;
      if (__pyx_t_7) {

        /* "glycresoft/_c/tandem/core_search.pyx":981
 *                 sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
 *                 if best_score < sol.score and sol.peptide_mass > 0:
 *                     best_score = sol.score             # <<<<<<<<<<<<<<
 *                     best_solution = sol
 *             if best_solution is not None:
 */
        __pyx_t_8 = __pyx_v_sol->score;
        __pyx_v_best_score = __pyx_t_8;

        /* "glycresoft/_c/tandem/core_search.pyx":982
 *                 if best_score < sol.score and sol.peptide_mass > 0:
 *                     best_score = sol.score
 *                     best_solution = sol             # <<<<<<<<<<<<<<
 *             if best_solution is not None:
 *                 out.append(best_solution)
 */
        __Pyx_INCREF((PyObject *)__pyx_v_sol);
        __Pyx_DECREF_SET(__pyx_v_best_solution, __pyx_v_sol);

        /* "glycresoft/_c/tandem/core_search.pyx":980
 *                 sol.score *= coverage
 *                 sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
 *                 if best_score < sol.score and sol.peptide_mass > 0:             # <<<<<<<<<<<<<<
 *                     best_score = sol.score
 *                     best_solution = sol
 */
      }
    }

    /* "glycresoft/_c/tandem/core_search.pyx":983
 *                     best_score = sol.score
 *                     best_solution = sol
 *             if best_solution is not None:             # <<<<<<<<<<<<<<
 *                 out.append(best_solution)
 * 
 */
    __pyx_t_7 = (((PyObject *)__pyx_v_best_solution) != Py_None);
    if (__pyx_t_7) {

      /* "glycresoft/_c/tandem/core_search.pyx":984
 *                     best_solution = sol
 *             if best_solution is not None:
 *                 out.append(best_solution)             # <<<<<<<<<<<<<<
 * 
 *         out.sort(key=partial_score_getter, reverse=True)
 */
      __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_out, ((PyObject *)__pyx_v_best_solution)); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 984, __pyx_L1_error)

      /* "glycresoft/_c/tandem/core_search.pyx":983
 *                     best_score = sol.score
 *                     best_solution = sol
 *             if best_solution is not None:             # <<<<<<<<<<<<<<
 *                 out.append(best_solution)
 * 
 */
    }
  }

  /* "glycresoft/_c/tandem/core_search.pyx":986
 *                 out.append(best_solution)
 * 
 *         out.sort(key=partial_score_getter, reverse=True)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_partial_score_getter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_4) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":987
 * 
 *         out.sort(key=partial_score_getter, reverse=True)
 *         return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":917
 * 
 *     @cython.cdivision
 *     cpdef list match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double precursor_mass, d
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sol);
  __Pyx_XDECREF((PyObject *)__pyx_v_best_solution);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_sols_for_glycan);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_set);
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XDECREF(__pyx_v_glycosylation_type);
  __Pyx_XDECREF(__pyx_v_id_sol);
  __Pyx_XDECREF(__pyx_v_glytype_sol);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_solutions);
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_12match, "GlycanFragmentIndex.match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None) -> list");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match = {"match", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_12match};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_scan = 0;
  double __pyx_v_error_tolerance;
  struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift = 0;
  PyObject *__pyx_v_query_mass = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("match (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scan,&__pyx_n_s_error_tolerance,&__pyx_n_s_mass_shift,&__pyx_n_s_query_mass,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL((PyObject *)__pyx_k__5);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scan)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_error_tolerance);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mass_shift);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_query_mass);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "match") < 0)) __PYX_ERR(0, 917, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_scan = values[0];
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)1e-5);
    }
    __pyx_v_mass_shift = ((struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *)values[2]);
    __pyx_v_query_mass = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("match", 0, 1, 4, __pyx_nargs); __PYX_ERR(0, 917, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mass_shift), __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase, 1, "mass_shift", 0))) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_12match(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v_scan, __pyx_v_error_tolerance, __pyx_v_mass_shift, __pyx_v_query_mass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_12match(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_scan, double __pyx_v_error_tolerance, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *__pyx_v_mass_shift, PyObject *__pyx_v_query_mass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("match", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_2.mass_shift = __pyx_v_mass_shift;
  __pyx_t_2.query_mass = __pyx_v_query_mass;
  __pyx_t_1 = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex->match(__pyx_v_self, __pyx_v_scan, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":741
 * 
 *     cdef:
 *         public list members             # <<<<<<<<<<<<<<
 *         public object glycosylation_type
 *         public dict unique_fragments
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->members);
  __pyx_r = __pyx_v_self->members;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->members);
  __Pyx_DECREF(__pyx_v_self->members);
  __pyx_v_self->members = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.members.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->members);
  __Pyx_DECREF(__pyx_v_self->members);
  __pyx_v_self->members = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":742
 *     cdef:
 *         public list members
 *         public object glycosylation_type             # <<<<<<<<<<<<<<
 *         public dict unique_fragments
 *         public list _fragments
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->glycosylation_type);
  __pyx_r = __pyx_v_self->glycosylation_type;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->glycosylation_type);
  __Pyx_DECREF(__pyx_v_self->glycosylation_type);
  __pyx_v_self->glycosylation_type = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->glycosylation_type);
  __Pyx_DECREF(__pyx_v_self->glycosylation_type);
  __pyx_v_self->glycosylation_type = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":743
 *         public list members
 *         public object glycosylation_type
 *         public dict unique_fragments             # <<<<<<<<<<<<<<
 *         public list _fragments
 *         public dict fragment_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->unique_fragments);
  __pyx_r = __pyx_v_self->unique_fragments;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 743, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->unique_fragments);
  __Pyx_DECREF(__pyx_v_self->unique_fragments);
  __pyx_v_self->unique_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.unique_fragments.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->unique_fragments);
  __Pyx_DECREF(__pyx_v_self->unique_fragments);
  __pyx_v_self->unique_fragments = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":744
 *         public object glycosylation_type
 *         public dict unique_fragments
 *         public list _fragments             # <<<<<<<<<<<<<<
 *         public dict fragment_index
 *         public long counter
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_fragments);
  __pyx_r = __pyx_v_self->_fragments;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(0, 744, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fragments);
  __Pyx_DECREF(__pyx_v_self->_fragments);
  __pyx_v_self->_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex._fragments.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fragments);
  __Pyx_DECREF(__pyx_v_self->_fragments);
  __pyx_v_self->_fragments = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":745
 *         public dict unique_fragments
 *         public list _fragments
 *         public dict fragment_index             # <<<<<<<<<<<<<<
 *         public long counter
 *         public double fragment_weight
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fragment_index);
  __pyx_r = __pyx_v_self->fragment_index;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 745, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fragment_index);
  __Pyx_DECREF(__pyx_v_self->fragment_index);
  __pyx_v_self->fragment_index = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.fragment_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_4__del__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_4__del__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->fragment_index);
  __Pyx_DECREF(__pyx_v_self->fragment_index);
  __pyx_v_self->fragment_index = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":746
 *         public list _fragments
 *         public dict fragment_index
 *         public long counter             # <<<<<<<<<<<<<<
 *         public double fragment_weight
 *         public double core_weight
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->counter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.counter.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L1_error)
  __pyx_v_self->counter = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.counter.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":747
 *         public dict fragment_index
 *         public long counter
 *         public double fragment_weight             # <<<<<<<<<<<<<<
 *         public double core_weight
 *         public double lower_bound
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fragment_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.fragment_weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L1_error)
  __pyx_v_self->fragment_weight = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.fragment_weight.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":748
 *         public long counter
 *         public double fragment_weight
 *         public double core_weight             # <<<<<<<<<<<<<<
 *         public double lower_bound
 *         public double upper_bound
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.core_weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 748, __pyx_L1_error)
  __pyx_v_self->core_weight = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.core_weight.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":749
 *         public double fragment_weight
 *         public double core_weight
 *         public double lower_bound             # <<<<<<<<<<<<<<
 *         public double upper_bound
 *         public double resolution
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->lower_bound); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.lower_bound.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 749, __pyx_L1_error)
  __pyx_v_self->lower_bound = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.lower_bound.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":750
 *         public double core_weight
 *         public double lower_bound
 *         public double upper_bound             # <<<<<<<<<<<<<<
 *         public double resolution
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->upper_bound); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.upper_bound.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L1_error)
  __pyx_v_self->upper_bound = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.upper_bound.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":751
 *         public double lower_bound
 *         public double upper_bound
 *         public double resolution             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution___get__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution___get__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->resolution); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.resolution.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_2__set__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_2__set__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 751, __pyx_L1_error)
  __pyx_v_self->resolution = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.resolution.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14__reduce_cython__, "GlycanFragmentIndex.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._fragments, self.core_weight, self.counter, self.fragment_index, self.fragment_weight, self.glycosylation_type, self.lower_bound, self.members, self.resolution, self.unique_fragments, self.upper_bound)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->core_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v_self->counter); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->fragment_weight); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->lower_bound); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->resolution); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->upper_bound); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(11); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_self->_fragments);
  __Pyx_GIVEREF(__pyx_v_self->_fragments);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_self->_fragments)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_2)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fragment_index);
  __Pyx_GIVEREF(__pyx_v_self->fragment_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_v_self->fragment_index)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->glycosylation_type);
  __Pyx_GIVEREF(__pyx_v_self->glycosylation_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_v_self->glycosylation_type)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->members);
  __Pyx_GIVEREF(__pyx_v_self->members);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_v_self->members)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 8, __pyx_t_5)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->unique_fragments);
  __Pyx_GIVEREF(__pyx_v_self->unique_fragments);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 9, __pyx_v_self->unique_fragments)) __PYX_ERR(3, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 10, __pyx_t_6)) __PYX_ERR(3, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._fragments, self.core_weight, self.counter, self.fragment_index, self.fragment_weight, self.glycosylation_type, self.lower_bound, self.members, self.resolution, self.unique_fragments, self.upper_bound)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_7 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v__dict = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "(tree fragment)":7
 *     state = (self._fragments, self.core_weight, self.counter, self.fragment_index, self.fragment_weight, self.glycosylation_type, self.lower_bound, self.members, self.resolution, self.unique_fragments, self.upper_bound)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_8 = (__pyx_v__dict != Py_None);
  if (__pyx_t_8) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v__dict)) __PYX_ERR(3, 8, __pyx_L1_error);
    __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._fragments is not None or self.fragment_index is not None or self.glycosylation_type is not None or self.members is not None or self.unique_fragments is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._fragments, self.core_weight, self.counter, self.fragment_index, self.fragment_weight, self.glycosylation_type, self.lower_bound, self.members, self.resolution, self.unique_fragments, self.upper_bound)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._fragments is not None or self.fragment_index is not None or self.glycosylation_type is not None or self.members is not None or self.unique_fragments is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, None), state
 */
  /*else*/ {
    __pyx_t_9 = (__pyx_v_self->_fragments != ((PyObject*)Py_None));
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->fragment_index != ((PyObject*)Py_None));
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->glycosylation_type != Py_None);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->members != ((PyObject*)Py_None));
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->unique_fragments != ((PyObject*)Py_None));
    __pyx_t_8 = __pyx_t_9;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_8;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._fragments is not None or self.fragment_index is not None or self.glycosylation_type is not None or self.members is not None or self.unique_fragments is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._fragments is not None or self.fragment_index is not None or self.glycosylation_type is not None or self.members is not None or self.unique_fragments is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_GlycanFragmentInd); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_16698535);
    __Pyx_GIVEREF(__pyx_int_16698535);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_16698535)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, Py_None)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7)) __PYX_ERR(3, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(3, 13, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._fragments is not None or self.fragment_index is not None or self.glycosylation_type is not None or self.members is not None or self.unique_fragments is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, None), state
 *     else:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanFragmentIndex__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_GlycanFragmentInd); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_16698535);
    __Pyx_GIVEREF(__pyx_int_16698535);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_16698535)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7)) __PYX_ERR(3, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanFragmentIndex__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16__setstate_cython__, "GlycanFragmentIndex.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(3, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanFragmentIndex__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanFragmentIndex__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanFragmentIndex__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.GlycanFragmentIndex.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/tandem/core_search.pyx":990
 * 
 * 
 * cpdef double partial_score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<PartialGlycanSolution>x).score
 */

static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_5partial_score_getter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_10glycresoft_2_c_6tandem_11core_search_partial_score_getter(PyObject *__pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_r;

  /* "glycresoft/_c/tandem/core_search.pyx":991
 * 
 * cpdef double partial_score_getter(object x):
 *     return (<PartialGlycanSolution>x).score             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v_x)->score;
  goto __pyx_L0;

  /* "glycresoft/_c/tandem/core_search.pyx":990
 * 
 * 
 * cpdef double partial_score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<PartialGlycanSolution>x).score
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_5partial_score_getter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_4partial_score_getter, "partial_score_getter(x) -> double");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_5partial_score_getter = {"partial_score_getter", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_5partial_score_getter, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_4partial_score_getter};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_5partial_score_getter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("partial_score_getter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "partial_score_getter") < 0)) __PYX_ERR(0, 990, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_x = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("partial_score_getter", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 990, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.partial_score_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_4partial_score_getter(__pyx_self, __pyx_v_x);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_4partial_score_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("partial_score_getter", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search_partial_score_getter(__pyx_v_x, 0); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.partial_score_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCombinationRecordBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_7__pyx_unpickle_GlycanCombinationRecordBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_6__pyx_unpickle_GlycanCombinationRecordBase, "__pyx_unpickle_GlycanCombinationRecordBase(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_7__pyx_unpickle_GlycanCombinationRecordBase = {"__pyx_unpickle_GlycanCombinationRecordBase", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_7__pyx_unpickle_GlycanCombinationRecordBase, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_6__pyx_unpickle_GlycanCombinationRecordBase};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_7__pyx_unpickle_GlycanCombinationRecordBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCombinationRecordBase (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCombinationRecordBase", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCombinationRecordBase", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_GlycanCombinationRecordBase") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCombinationRecordBase", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanCombinationRecordBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_6__pyx_unpickle_GlycanCombinationRecordBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_6__pyx_unpickle_GlycanCombinationRecordBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCombinationRecordBase", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x564dc94, 0x19c2b1d, 0x930e938):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__6, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x564dc94, 0x19c2b1d, 0x930e938):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 *     __pyx_result = GlycanCombinationRecordBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x564dc94, 0x19c2b1d, 0x930e938):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycanCombinationRecordBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x564dc94, 0x19c2b1d, 0x930e938):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 *     __pyx_result = GlycanCombinationRecordBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 *     __pyx_result = GlycanCombinationRecordBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycanCombinationRecordBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCombinationRecordBase__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 *     __pyx_result = GlycanCombinationRecordBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCombinationRecordBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanCombinationRecordBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCombinationRecordBase__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  npy_uint16 __pyx_t_3;
  double __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  unsigned int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCombinationRecordBase__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_fragment_cache);
  __Pyx_DECREF(__pyx_v___pyx_result->_fragment_cache);
  __pyx_v___pyx_result->_fragment_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_hash = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->composition);
  __Pyx_DECREF(__pyx_v___pyx_result->composition);
  __pyx_v___pyx_result->composition = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint16(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->count = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->dehydrated_mass = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fragment_set_properties);
  __Pyx_DECREF(__pyx_v___pyx_result->fragment_set_properties);
  __pyx_v___pyx_result->fragment_set_properties = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->glycan_types);
  __Pyx_DECREF(__pyx_v___pyx_result->glycan_types);
  __pyx_v___pyx_result->glycan_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint16(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->internal_size_approximation = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint16(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->side_group_count = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint16(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_uint16)-1)) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->size = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 > 11);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_6 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "(tree fragment)":14
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = NULL;
    __pyx_t_12 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_12 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanCombinationRecordBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CoarseGlycanMatch(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_9__pyx_unpickle_CoarseGlycanMatch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_8__pyx_unpickle_CoarseGlycanMatch, "__pyx_unpickle_CoarseGlycanMatch(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_9__pyx_unpickle_CoarseGlycanMatch = {"__pyx_unpickle_CoarseGlycanMatch", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_9__pyx_unpickle_CoarseGlycanMatch, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_8__pyx_unpickle_CoarseGlycanMatch};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_9__pyx_unpickle_CoarseGlycanMatch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoarseGlycanMatch (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoarseGlycanMatch", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoarseGlycanMatch", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CoarseGlycanMatch") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoarseGlycanMatch", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_CoarseGlycanMatch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_8__pyx_unpickle_CoarseGlycanMatch(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_8__pyx_unpickle_CoarseGlycanMatch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoarseGlycanMatch", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf0ba66f, 0x6ff56bb, 0xc80cd69):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf0ba66f, 0x6ff56bb, 0xc80cd69):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum
 *     __pyx_result = CoarseGlycanMatch.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xf0ba66f, 0x6ff56bb, 0xc80cd69):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CoarseGlycanMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf0ba66f, 0x6ff56bb, 0xc80cd69):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum
 *     __pyx_result = CoarseGlycanMatch.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum
 *     __pyx_result = CoarseGlycanMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CoarseGlycanMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_CoarseGlycanMatch__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf0ba66f, 0x6ff56bb, 0xc80cd69) = (core_matched, core_theoretical, fragment_matches, n_matched, n_theoretical))" % __pyx_checksum
 *     __pyx_result = CoarseGlycanMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CoarseGlycanMatch(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_CoarseGlycanMatch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_CoarseGlycanMatch__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoarseGlycanMatch__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->core_matched = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->core_theoretical = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fragment_matches);
  __Pyx_DECREF(__pyx_v___pyx_result->fragment_matches);
  __pyx_v___pyx_result->fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_matched = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_theoretical = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 5);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CoarseGlycanMatch__set_state(<CoarseGlycanMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CoarseGlycanMatch__set_state(CoarseGlycanMatch __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_matched = __pyx_state[0]; __pyx_result.core_theoretical = __pyx_state[1]; __pyx_result.fragment_matches = __pyx_state[2]; __pyx_result.n_matched = __pyx_state[3]; __pyx_result.n_theoretical = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_CoarseGlycanMatch__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCoarseScorerBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_11__pyx_unpickle_GlycanCoarseScorerBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_10__pyx_unpickle_GlycanCoarseScorerBase, "__pyx_unpickle_GlycanCoarseScorerBase(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_11__pyx_unpickle_GlycanCoarseScorerBase = {"__pyx_unpickle_GlycanCoarseScorerBase", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_11__pyx_unpickle_GlycanCoarseScorerBase, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_10__pyx_unpickle_GlycanCoarseScorerBase};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_11__pyx_unpickle_GlycanCoarseScorerBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCoarseScorerBase (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCoarseScorerBase", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCoarseScorerBase", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_GlycanCoarseScorerBase") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCoarseScorerBase", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanCoarseScorerBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_10__pyx_unpickle_GlycanCoarseScorerBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_10__pyx_unpickle_GlycanCoarseScorerBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCoarseScorerBase", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1ca83c4, 0x41fe7e1, 0x9b999c2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__9, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1ca83c4, 0x41fe7e1, 0x9b999c2):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum
 *     __pyx_result = GlycanCoarseScorerBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1ca83c4, 0x41fe7e1, 0x9b999c2):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycanCoarseScorerBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1ca83c4, 0x41fe7e1, 0x9b999c2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum
 *     __pyx_result = GlycanCoarseScorerBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum
 *     __pyx_result = GlycanCoarseScorerBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycanCoarseScorerBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCoarseScorerBase__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1ca83c4, 0x41fe7e1, 0x9b999c2) = (core_weight, fragment_weight, product_error_tolerance, recalibrate_peptide_mass))" % __pyx_checksum
 *     __pyx_result = GlycanCoarseScorerBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCoarseScorerBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanCoarseScorerBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanCoarseScorerBase__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCoarseScorerBase__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->core_weight = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->fragment_weight = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->product_error_tolerance = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->recalibrate_peptide_mass = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 4);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanCoarseScorerBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycanMatchResult(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_13__pyx_unpickle_GlycanMatchResult(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_12__pyx_unpickle_GlycanMatchResult, "__pyx_unpickle_GlycanMatchResult(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_13__pyx_unpickle_GlycanMatchResult = {"__pyx_unpickle_GlycanMatchResult", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_13__pyx_unpickle_GlycanMatchResult, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_12__pyx_unpickle_GlycanMatchResult};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_13__pyx_unpickle_GlycanMatchResult(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanMatchResult (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanMatchResult", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanMatchResult", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_GlycanMatchResult") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanMatchResult", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanMatchResult", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_12__pyx_unpickle_GlycanMatchResult(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_12__pyx_unpickle_GlycanMatchResult(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanMatchResult", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x887dc81, 0xd5eb5c8, 0x902ef35):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__10, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x887dc81, 0xd5eb5c8, 0x902ef35):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = GlycanMatchResult.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x887dc81, 0xd5eb5c8, 0x902ef35):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycanMatchResult.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x887dc81, 0xd5eb5c8, 0x902ef35):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = GlycanMatchResult.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = GlycanMatchResult.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycanMatchResult.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanMatchResult__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x887dc81, 0xd5eb5c8, 0x902ef35) = (glycan_size, glycan_types, match, peptide_mass, recalibrated_peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = GlycanMatchResult.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanMatchResult(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanMatchResult", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanMatchResult__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanMatchResult__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->glycan_size = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->glycan_types);
  __Pyx_DECREF(__pyx_v___pyx_result->glycan_types);
  __pyx_v___pyx_result->glycan_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->match);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->match);
  __pyx_v___pyx_result->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->peptide_mass = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->recalibrated_peptide_mass = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->score = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 6);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanMatchResult__set_state(<GlycanMatchResult> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanMatchResult__set_state(GlycanMatchResult __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.glycan_size = __pyx_state[0]; __pyx_result.glycan_types = __pyx_state[1]; __pyx_result.match = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.recalibrated_peptide_mass = __pyx_state[4]; __pyx_result.score = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanMatchResult__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ScoreCoarseGlycanMatchPair(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_15__pyx_unpickle_ScoreCoarseGlycanMatchPair(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_14__pyx_unpickle_ScoreCoarseGlycanMatchPair, "__pyx_unpickle_ScoreCoarseGlycanMatchPair(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_15__pyx_unpickle_ScoreCoarseGlycanMatchPair = {"__pyx_unpickle_ScoreCoarseGlycanMatchPair", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_15__pyx_unpickle_ScoreCoarseGlycanMatchPair, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_14__pyx_unpickle_ScoreCoarseGlycanMatchPair};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_15__pyx_unpickle_ScoreCoarseGlycanMatchPair(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ScoreCoarseGlycanMatchPair (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ScoreCoarseGlycanMatchPair", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ScoreCoarseGlycanMatchPair", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ScoreCoarseGlycanMatchPair") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ScoreCoarseGlycanMatchPair", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_ScoreCoarseGlycanMatchPair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_14__pyx_unpickle_ScoreCoarseGlycanMatchPair(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_14__pyx_unpickle_ScoreCoarseGlycanMatchPair(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ScoreCoarseGlycanMatchPair", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x53f74fc, 0x9b66d39, 0x7706bbe):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__11, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x53f74fc, 0x9b66d39, 0x7706bbe):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum
 *     __pyx_result = ScoreCoarseGlycanMatchPair.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x53f74fc, 0x9b66d39, 0x7706bbe):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ScoreCoarseGlycanMatchPair.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x53f74fc, 0x9b66d39, 0x7706bbe):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum
 *     __pyx_result = ScoreCoarseGlycanMatchPair.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum
 *     __pyx_result = ScoreCoarseGlycanMatchPair.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ScoreCoarseGlycanMatchPair.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x53f74fc, 0x9b66d39, 0x7706bbe) = (match, score))" % __pyx_checksum
 *     __pyx_result = ScoreCoarseGlycanMatchPair.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ScoreCoarseGlycanMatchPair(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_ScoreCoarseGlycanMatchPair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch))))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->match);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->match);
  __pyx_v___pyx_result->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->score = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ComplementFragment(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17__pyx_unpickle_ComplementFragment(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_16__pyx_unpickle_ComplementFragment, "__pyx_unpickle_ComplementFragment(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17__pyx_unpickle_ComplementFragment = {"__pyx_unpickle_ComplementFragment", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17__pyx_unpickle_ComplementFragment, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_16__pyx_unpickle_ComplementFragment};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17__pyx_unpickle_ComplementFragment(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ComplementFragment (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ComplementFragment", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ComplementFragment", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ComplementFragment") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ComplementFragment", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_ComplementFragment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_16__pyx_unpickle_ComplementFragment(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_16__pyx_unpickle_ComplementFragment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ComplementFragment", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x7fff502, 0x0e13161, 0x6e091ed):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__12, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x7fff502, 0x0e13161, 0x6e091ed):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum
 *     __pyx_result = ComplementFragment.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x7fff502, 0x0e13161, 0x6e091ed):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ComplementFragment.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x7fff502, 0x0e13161, 0x6e091ed):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum
 *     __pyx_result = ComplementFragment.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum
 *     __pyx_result = ComplementFragment.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ComplementFragment.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ComplementFragment__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x7fff502, 0x0e13161, 0x6e091ed) = (keys, mass))" % __pyx_checksum
 *     __pyx_result = ComplementFragment.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ComplementFragment(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_ComplementFragment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_ComplementFragment__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ComplementFragment__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->keys);
  __Pyx_DECREF(__pyx_v___pyx_result->keys);
  __pyx_v___pyx_result->keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->mass = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ComplementFragment__set_state(<ComplementFragment> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ComplementFragment__set_state(ComplementFragment __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.keys = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_ComplementFragment__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PartialGlycanSolution(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19__pyx_unpickle_PartialGlycanSolution(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_18__pyx_unpickle_PartialGlycanSolution, "__pyx_unpickle_PartialGlycanSolution(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19__pyx_unpickle_PartialGlycanSolution = {"__pyx_unpickle_PartialGlycanSolution", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19__pyx_unpickle_PartialGlycanSolution, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_18__pyx_unpickle_PartialGlycanSolution};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19__pyx_unpickle_PartialGlycanSolution(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PartialGlycanSolution (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PartialGlycanSolution", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PartialGlycanSolution", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PartialGlycanSolution") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PartialGlycanSolution", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_PartialGlycanSolution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_18__pyx_unpickle_PartialGlycanSolution(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_18__pyx_unpickle_PartialGlycanSolution(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PartialGlycanSolution", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x57f5605, 0xcf96d30, 0xc1f6d26):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x57f5605, 0xcf96d30, 0xc1f6d26):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = PartialGlycanSolution.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x57f5605, 0xcf96d30, 0xc1f6d26):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PartialGlycanSolution.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x57f5605, 0xcf96d30, 0xc1f6d26):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = PartialGlycanSolution.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = PartialGlycanSolution.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PartialGlycanSolution.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_PartialGlycanSolution__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x57f5605, 0xcf96d30, 0xc1f6d26) = (core_matches, fragment_matches, glycan_index, peptide_mass, score))" % __pyx_checksum
 *     __pyx_result = PartialGlycanSolution.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PartialGlycanSolution(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_PartialGlycanSolution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_PartialGlycanSolution__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  npy_int32 __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PartialGlycanSolution__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->core_matches);
  __Pyx_DECREF(__pyx_v___pyx_result->core_matches);
  __pyx_v___pyx_result->core_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fragment_matches);
  __Pyx_DECREF(__pyx_v___pyx_result->fragment_matches);
  __pyx_v___pyx_result->fragment_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->glycan_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->peptide_mass = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->score = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 5);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_PartialGlycanSolution__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycanFragmentIndex(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21__pyx_unpickle_GlycanFragmentIndex(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_6tandem_11core_search_20__pyx_unpickle_GlycanFragmentIndex, "__pyx_unpickle_GlycanFragmentIndex(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_6tandem_11core_search_21__pyx_unpickle_GlycanFragmentIndex = {"__pyx_unpickle_GlycanFragmentIndex", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21__pyx_unpickle_GlycanFragmentIndex, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_20__pyx_unpickle_GlycanFragmentIndex};
static PyObject *__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21__pyx_unpickle_GlycanFragmentIndex(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanFragmentIndex (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanFragmentIndex", 1, 3, 3, 1); __PYX_ERR(3, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanFragmentIndex", 1, 3, 3, 2); __PYX_ERR(3, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_GlycanFragmentIndex") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanFragmentIndex", 1, 3, 3, __pyx_nargs); __PYX_ERR(3, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanFragmentIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_6tandem_11core_search_20__pyx_unpickle_GlycanFragmentIndex(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_6tandem_11core_search_20__pyx_unpickle_GlycanFragmentIndex(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanFragmentIndex", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0fecca7, 0xb6d4725, 0xdb00190):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__14, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0fecca7, 0xb6d4725, 0xdb00190):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum
 *     __pyx_result = GlycanFragmentIndex.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(3, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x0fecca7, 0xb6d4725, 0xdb00190):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycanFragmentIndex.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0fecca7, 0xb6d4725, 0xdb00190):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum
 *     __pyx_result = GlycanFragmentIndex.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum
 *     __pyx_result = GlycanFragmentIndex.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycanFragmentIndex.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(3, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanFragmentIndex__set_state(((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0fecca7, 0xb6d4725, 0xdb00190) = (_fragments, core_weight, counter, fragment_index, fragment_weight, glycosylation_type, lower_bound, members, resolution, unique_fragments, upper_bound))" % __pyx_checksum
 *     __pyx_result = GlycanFragmentIndex.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanFragmentIndex(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanFragmentIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_6tandem_11core_search___pyx_unpickle_GlycanFragmentIndex__set_state(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  long __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanFragmentIndex__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_fragments);
  __Pyx_DECREF(__pyx_v___pyx_result->_fragments);
  __pyx_v___pyx_result->_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->core_weight = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_3 == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->counter = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fragment_index);
  __Pyx_DECREF(__pyx_v___pyx_result->fragment_index);
  __pyx_v___pyx_result->fragment_index = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->fragment_weight = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->glycosylation_type);
  __Pyx_DECREF(__pyx_v___pyx_result->glycosylation_type);
  __pyx_v___pyx_result->glycosylation_type = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->lower_bound = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->members);
  __Pyx_DECREF(__pyx_v___pyx_result->members);
  __pyx_v___pyx_result->members = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->resolution = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->unique_fragments);
  __Pyx_DECREF(__pyx_v___pyx_result->unique_fragments);
  __pyx_v___pyx_result->unique_fragments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(3, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->upper_bound = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 11);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanFragmentIndex__set_state(<GlycanFragmentIndex> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanFragmentIndex__set_state(GlycanFragmentIndex __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fragments = __pyx_state[0]; __pyx_result.core_weight = __pyx_state[1]; __pyx_result.counter = __pyx_state[2]; __pyx_result.fragment_index = __pyx_state[3]; __pyx_result.fragment_weight = __pyx_state[4]; __pyx_result.glycosylation_type = __pyx_state[5]; __pyx_result.lower_bound = __pyx_state[6]; __pyx_result.members = __pyx_state[7]; __pyx_result.resolution = __pyx_state[8]; __pyx_result.unique_fragments = __pyx_state[9]; __pyx_result.upper_bound = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glycresoft._c.tandem.core_search.__pyx_unpickle_GlycanFragmentIndex__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;
  p->composition = Py_None; Py_INCREF(Py_None);
  p->glycan_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_fragment_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fragment_set_properties = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->composition);
  Py_CLEAR(p->glycan_types);
  Py_CLEAR(p->_fragment_cache);
  Py_CLEAR(p->fragment_set_properties);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)o;
  if (p->composition) {
    e = (*v)(p->composition, a); if (e) return e;
  }
  if (p->glycan_types) {
    e = (*v)(p->glycan_types, a); if (e) return e;
  }
  if (p->_fragment_cache) {
    e = (*v)(p->_fragment_cache, a); if (e) return e;
  }
  if (p->fragment_set_properties) {
    e = (*v)(p->fragment_set_properties, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *)o;
  tmp = ((PyObject*)p->composition);
  p->composition = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->glycan_types);
  p->glycan_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_fragment_cache);
  p->_fragment_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fragment_set_properties);
  p->fragment_set_properties = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_2id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_dehydrated_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_dehydrated_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15dehydrated_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_composition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11composition_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_4size_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_internal_size_approximation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_internal_size_approximation(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_27internal_size_approximation_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_side_group_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_side_group_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16side_group_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_glycan_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_glycan_types(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_12glycan_types_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__fragment_cache(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__fragment_cache(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15_fragment_cache_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5_hash_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_fragment_set_properties(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_fragment_set_properties(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_23fragment_set_properties_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_18__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase[] = {
  {(char *)"id", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_id, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_id, (char *)PyDoc_STR("id: 'size_t'"), 0},
  {(char *)"dehydrated_mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_dehydrated_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_dehydrated_mass, (char *)PyDoc_STR("dehydrated_mass: 'double'"), 0},
  {(char *)"composition", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_composition, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_composition, (char *)PyDoc_STR("composition: object"), 0},
  {(char *)"size", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_size, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_size, (char *)PyDoc_STR("size: 'npy_uint16'"), 0},
  {(char *)"internal_size_approximation", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_internal_size_approximation, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_internal_size_approximation, (char *)PyDoc_STR("internal_size_approximation: 'npy_uint16'"), 0},
  {(char *)"side_group_count", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_side_group_count, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_side_group_count, (char *)PyDoc_STR("side_group_count: 'npy_uint16'"), 0},
  {(char *)"count", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_count, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_count, (char *)PyDoc_STR("count: 'npy_uint16'"), 0},
  {(char *)"glycan_types", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_glycan_types, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_glycan_types, (char *)PyDoc_STR("glycan_types: list"), 0},
  {(char *)"_fragment_cache", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__fragment_cache, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__fragment_cache, (char *)PyDoc_STR("_fragment_cache: dict"), 0},
  {(char *)"_hash", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__hash, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase__hash, (char *)PyDoc_STR("_hash: 'Py_hash_t'"), 0},
  {(char *)"fragment_set_properties", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_fragment_set_properties, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_fragment_set_properties, (char *)PyDoc_STR("fragment_set_properties: dict"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase_spec = {
  "glycresoft._c.tandem.core_search.GlycanCombinationRecordBase",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""GlycanCombinationRecordBase", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)o);
  p->key = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)o;
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment *)o;
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_9__ne__(o1, o2);
    }
    case Py_LT: {
      return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_13__lt__(o1, o2);
    }
    case Py_GT: {
      return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_15__gt__(o1, o2);
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_key(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3key_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_4mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_is_core(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_is_core(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_7is_core_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment[] = {
  {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_2__reduce__},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__, METH_NOARGS|METH_COEXIST, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment[] = {
  {(char *)"key", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_key, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_key, (char *)PyDoc_STR("key: object"), 0},
  {(char *)"mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_mass, (char *)PyDoc_STR("mass: 'double'"), 0},
  {(char *)"is_core", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_is_core, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_is_core, (char *)PyDoc_STR("is_core: 'bool'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__},
  {Py_tp_hash, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_11__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("CoarseStubGlycopeptideFragment(key, mass, is_core)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment_spec = {
  "glycresoft._c.tandem.core_search.CoarseStubGlycopeptideFragment",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""CoarseStubGlycopeptideFragment", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_11__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("CoarseStubGlycopeptideFragment(key, mass, is_core)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_clear*/
  __pyx_tp_richcompare_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch[100000];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch)))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
  p->key = Py_None; Py_INCREF(Py_None);
  p->peaks_matched = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key);
  Py_CLEAR(p->peaks_matched);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch < 100000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch)))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)o;
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  if (p->peaks_matched) {
    e = (*v)(p->peaks_matched, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *)o;
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->peaks_matched);
  p->peaks_matched = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_key(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3key_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_4mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_shift_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_shift_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_10shift_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_peaks_matched(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_peaks_matched(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_13peaks_matched_5__del__(o);
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch[] = {
  {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_2__reduce__},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__, METH_NOARGS|METH_COEXIST, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch[] = {
  {(char *)"key", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_key, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_key, (char *)PyDoc_STR("key: object"), 0},
  {(char *)"mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_mass, (char *)PyDoc_STR("mass: 'double'"), 0},
  {(char *)"shift_mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_shift_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_shift_mass, (char *)PyDoc_STR("shift_mass: 'double'"), 0},
  {(char *)"peaks_matched", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_peaks_matched, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_peaks_matched, (char *)PyDoc_STR("peaks_matched: tuple"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("CoarseStubGlycopeptideMatch(key, mass, shift_mass, peaks_matched)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch_spec = {
  "glycresoft._c.tandem.core_search.CoarseStubGlycopeptideMatch",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""CoarseStubGlycopeptideMatch", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("CoarseStubGlycopeptideMatch(key, mass, shift_mass, peaks_matched)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch[100000];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch)))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
  p->fragment_matches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->fragment_matches);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch < 100000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch)))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)o;
  if (p->fragment_matches) {
    e = (*v)(p->fragment_matches, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)o;
  tmp = ((PyObject*)p->fragment_matches);
  p->fragment_matches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_fragment_matches(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_fragment_matches(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16fragment_matches_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_matched(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_matched(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9n_matched_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_theoretical(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_theoretical(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_13n_theoretical_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_matched(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_matched(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12core_matched_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_theoretical(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_theoretical(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_16core_theoretical_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_10__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_9__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_11__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch[] = {
  {(char *)"fragment_matches", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_fragment_matches, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_fragment_matches, (char *)PyDoc_STR("fragment_matches: list"), 0},
  {(char *)"n_matched", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_matched, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_matched, (char *)PyDoc_STR("n_matched: 'double'"), 0},
  {(char *)"n_theoretical", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_theoretical, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_n_theoretical, (char *)PyDoc_STR("n_theoretical: 'double'"), 0},
  {(char *)"core_matched", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_matched, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_matched, (char *)PyDoc_STR("core_matched: 'double'"), 0},
  {(char *)"core_theoretical", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_theoretical, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_core_theoretical, (char *)PyDoc_STR("core_theoretical: 'double'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("CoarseGlycanMatch(fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch},
  {Py_tp_iter, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_3__iter__},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch_spec = {
  "glycresoft._c.tandem.core_search.CoarseGlycanMatch",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""CoarseGlycanMatch", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_8__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("CoarseGlycanMatch(fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_product_error_tolerance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_product_error_tolerance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_23product_error_tolerance_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_fragment_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_fragment_weight(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15fragment_weight_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_core_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_core_weight(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11core_weight_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_recalibrate_peptide_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_recalibrate_peptide_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_24recalibrate_peptide_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_12__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_14__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase[] = {
  {(char *)"product_error_tolerance", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_product_error_tolerance, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_product_error_tolerance, (char *)PyDoc_STR("product_error_tolerance: 'double'"), 0},
  {(char *)"fragment_weight", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_fragment_weight, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_fragment_weight, (char *)PyDoc_STR("fragment_weight: 'double'"), 0},
  {(char *)"core_weight", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_core_weight, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_core_weight, (char *)PyDoc_STR("core_weight: 'double'"), 0},
  {(char *)"recalibrate_peptide_mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_recalibrate_peptide_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_recalibrate_peptide_mass, (char *)PyDoc_STR("recalibrate_peptide_mass: 'bool'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase},
  {Py_tp_doc, (void *)PyDoc_STR("GlycanCoarseScorerBase(product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False)")},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase_spec = {
  "glycresoft._c.tandem.core_search.GlycanCoarseScorerBase",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""GlycanCoarseScorerBase", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("GlycanCoarseScorerBase(product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False)"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult[1000];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;
  p->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); Py_INCREF(Py_None);
  p->glycan_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->match);
  Py_CLEAR(p->glycan_types);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult < 1000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)o;
  if (p->match) {
    e = (*v)(((PyObject *)p->match), a); if (e) return e;
  }
  if (p->glycan_types) {
    e = (*v)(p->glycan_types, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *)o;
  tmp = ((PyObject*)p->match);
  p->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->glycan_types);
  p->glycan_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_fragment_match_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_20fragment_match_count_1__get__(o);
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_peptide_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_peptide_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12peptide_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_match(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_match(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5match_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_11glycan_size_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_types(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_12glycan_types_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_recalibrated_peptide_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_recalibrated_peptide_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_25recalibrated_peptide_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult[] = {
  {(char *)"fragment_match_count", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_fragment_match_count, 0, (char *)0, 0},
  {(char *)"peptide_mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_peptide_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_peptide_mass, (char *)PyDoc_STR("peptide_mass: 'double'"), 0},
  {(char *)"score", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_score, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_score, (char *)PyDoc_STR("score: 'double'"), 0},
  {(char *)"match", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_match, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_match, (char *)PyDoc_STR("match: glycresoft._c.tandem.core_search.CoarseGlycanMatch"), 0},
  {(char *)"glycan_size", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_size, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_size, (char *)PyDoc_STR("glycan_size: 'size_t'"), 0},
  {(char *)"glycan_types", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_types, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_glycan_types, (char *)PyDoc_STR("glycan_types: dict"), 0},
  {(char *)"recalibrated_peptide_mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_recalibrated_peptide_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_recalibrated_peptide_mass, (char *)PyDoc_STR("recalibrated_peptide_mass: 'double'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("GlycanMatchResult(peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult_spec = {
  "glycresoft._c.tandem.core_search.GlycanMatchResult",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""GlycanMatchResult", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("GlycanMatchResult(peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair[1000];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair)))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
  p->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->match);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair < 1000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair)))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)o;
  if (p->match) {
    e = (*v)(((PyObject *)p->match), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *)o;
  tmp = ((PyObject*)p->match);
  p->match = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_match(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_match(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5match_5__del__(o);
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_10__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair[] = {
  {(char *)"score", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_score, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_score, (char *)PyDoc_STR("score: 'double'"), 0},
  {(char *)"match", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_match, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_match, (char *)PyDoc_STR("match: glycresoft._c.tandem.core_search.CoarseGlycanMatch"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__},
  {Py_sq_length, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5__len__},
  {Py_sq_item, (void *)__pyx_sq_item_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {Py_mp_length, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5__len__},
  {Py_mp_subscript, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_3__getitem__},
  {Py_tp_doc, (void *)PyDoc_STR("ScoreCoarseGlycanMatchPair(score, match)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair_spec = {
  "glycresoft._c.tandem.core_search.ScoreCoarseGlycanMatchPair",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_ScoreCoarseGlycanMatchPair = {
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_ScoreCoarseGlycanMatchPair = {
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_5__len__, /*mp_length*/
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_3__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""ScoreCoarseGlycanMatchPair", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_7__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_ScoreCoarseGlycanMatchPair, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_ScoreCoarseGlycanMatchPair, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ScoreCoarseGlycanMatchPair(score, match)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment*)__pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  #if PY_MAJOR_VERSION < 3
  if (!(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment) || PyType_IS_GC(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment)) PyObject_GC_Track(o);
  #else
  if (PyType_IS_GC(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment)) PyObject_GC_Track(o);
  #endif
  if (likely(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment)) __Pyx_PyType_GetSlot(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment, tp_dealloc, destructor)(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment(PyObject *o, visitproc v, void *a) {
  int e;
  e = ((likely(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment)) ? ((__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment->tp_traverse) ? __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment)); if (e) return e;
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment(PyObject *o) {
  if (likely(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment)) { if (__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment->tp_clear) __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4name_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment[] = {
  {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_4__reduce__},
  {"__getstate__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_7__getstate__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_6__getstate__},
  {"__setstate__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_9__setstate__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_8__setstate__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment[] = {
  {(char *)"index", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_index, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_index, (char *)PyDoc_STR("index: 'int'"), 0},
  {(char *)"name", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_name, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_name, (char *)PyDoc_STR("name: str"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment},
  {Py_tp_doc, (void *)PyDoc_STR("IndexGlycanCompositionFragment(mass, composition, key, is_extended=False)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment_spec = {
  "glycresoft._c.tandem.core_search.IndexGlycanCompositionFragment",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""IndexGlycanCompositionFragment", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  PyDoc_STR("IndexGlycanCompositionFragment(mass, composition, key, is_extended=False)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_ComplementFragment[1000];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ComplementFragment = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ComplementFragment(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ComplementFragment > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_ComplementFragment[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ComplementFragment];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)o);
  p->keys = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ComplementFragment(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ComplementFragment) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->keys);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ComplementFragment < 1000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search_ComplementFragment[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_ComplementFragment++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_ComplementFragment(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)o;
  if (p->keys) {
    e = (*v)(p->keys, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_ComplementFragment(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment *)o;
  tmp = ((PyObject*)p->keys);
  p->keys = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_keys(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_keys(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4keys_5__del__(o);
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_ComplementFragment[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_ComplementFragment[] = {
  {(char *)"mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_mass, (char *)PyDoc_STR("mass: 'double'"), 0},
  {(char *)"keys", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_keys, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_keys, (char *)PyDoc_STR("keys: list"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ComplementFragment},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("ComplementFragment(mass, keys=None)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_ComplementFragment},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_ComplementFragment},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_ComplementFragment},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_ComplementFragment},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ComplementFragment},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment_spec = {
  "glycresoft._c.tandem.core_search.ComplementFragment",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""ComplementFragment", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ComplementFragment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("ComplementFragment(mass, keys=None)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution[10000];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;
  p->core_matches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fragment_matches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->core_matches);
  Py_CLEAR(p->fragment_matches);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution < 10000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)o;
  if (p->core_matches) {
    e = (*v)(p->core_matches, a); if (e) return e;
  }
  if (p->fragment_matches) {
    e = (*v)(p->fragment_matches, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *)o;
  tmp = ((PyObject*)p->core_matches);
  p->core_matches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fragment_matches);
  p->fragment_matches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_fragment_match_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_20fragment_match_count_1__get__(o);
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_peptide_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_peptide_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12peptide_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_core_matches(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_core_matches(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12core_matches_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_fragment_matches(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_fragment_matches(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_16fragment_matches_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_glycan_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_glycan_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_12glycan_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution[] = {
  {(char *)"fragment_match_count", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_fragment_match_count, 0, (char *)0, 0},
  {(char *)"peptide_mass", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_peptide_mass, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_peptide_mass, (char *)PyDoc_STR("peptide_mass: 'double'"), 0},
  {(char *)"score", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_score, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_score, (char *)PyDoc_STR("score: 'double'"), 0},
  {(char *)"core_matches", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_core_matches, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_core_matches, (char *)PyDoc_STR("core_matches: set"), 0},
  {(char *)"fragment_matches", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_fragment_matches, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_fragment_matches, (char *)PyDoc_STR("fragment_matches: set"), 0},
  {(char *)"glycan_index", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_glycan_index, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_glycan_index, (char *)PyDoc_STR("glycan_index: 'npy_int32'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PartialGlycanSolution(peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, npy_int32 glycan_index=-1)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution_spec = {
  "glycresoft._c.tandem.core_search.PartialGlycanSolution",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""PartialGlycanSolution", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PartialGlycanSolution(peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, npy_int32 glycan_index=-1)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
  p->members = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->glycosylation_type = Py_None; Py_INCREF(Py_None);
  p->unique_fragments = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_fragments = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fragment_index = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->members);
  Py_CLEAR(p->glycosylation_type);
  Py_CLEAR(p->unique_fragments);
  Py_CLEAR(p->_fragments);
  Py_CLEAR(p->fragment_index);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)o;
  if (p->members) {
    e = (*v)(p->members, a); if (e) return e;
  }
  if (p->glycosylation_type) {
    e = (*v)(p->glycosylation_type, a); if (e) return e;
  }
  if (p->unique_fragments) {
    e = (*v)(p->unique_fragments, a); if (e) return e;
  }
  if (p->_fragments) {
    e = (*v)(p->_fragments, a); if (e) return e;
  }
  if (p->fragment_index) {
    e = (*v)(p->fragment_index, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *)o;
  tmp = ((PyObject*)p->members);
  p->members = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->glycosylation_type);
  p->glycosylation_type = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->unique_fragments);
  p->unique_fragments = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_fragments);
  p->_fragments = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fragment_index);
  p->fragment_index = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_members(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_members(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7members_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_glycosylation_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_glycosylation_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_18glycosylation_type_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_unique_fragments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_unique_fragments(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16unique_fragments_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__fragments(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__fragments(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10_fragments_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14fragment_index_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_counter(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_counter(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7counter_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_weight(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15fragment_weight_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_core_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_core_weight(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11core_weight_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_lower_bound(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_lower_bound(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11lower_bound_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_upper_bound(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_upper_bound(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11upper_bound_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_resolution(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_resolution(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_10resolution_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex[] = {
  {"build", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_9build, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_8build},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_16__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex[] = {
  {(char *)"members", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_members, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_members, (char *)PyDoc_STR("members: list"), 0},
  {(char *)"glycosylation_type", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_glycosylation_type, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_glycosylation_type, (char *)PyDoc_STR("glycosylation_type: object"), 0},
  {(char *)"unique_fragments", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_unique_fragments, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_unique_fragments, (char *)PyDoc_STR("unique_fragments: dict"), 0},
  {(char *)"_fragments", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__fragments, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__fragments, (char *)PyDoc_STR("_fragments: list"), 0},
  {(char *)"fragment_index", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_index, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_index, (char *)PyDoc_STR("fragment_index: dict"), 0},
  {(char *)"counter", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_counter, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_counter, (char *)PyDoc_STR("counter: 'long'"), 0},
  {(char *)"fragment_weight", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_weight, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_fragment_weight, (char *)PyDoc_STR("fragment_weight: 'double'"), 0},
  {(char *)"core_weight", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_core_weight, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_core_weight, (char *)PyDoc_STR("core_weight: 'double'"), 0},
  {(char *)"lower_bound", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_lower_bound, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_lower_bound, (char *)PyDoc_STR("lower_bound: 'double'"), 0},
  {(char *)"upper_bound", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_upper_bound, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_upper_bound, (char *)PyDoc_STR("upper_bound: 'double'"), 0},
  {(char *)"resolution", __pyx_getprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_resolution, __pyx_setprop_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_resolution, (char *)PyDoc_STR("resolution: 'double'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex},
  {Py_tp_doc, (void *)PyDoc_STR("GlycanFragmentIndex(members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0)\nA fast and sparse in-memory fragment ion index for quickly matching peaks against multiple\n    glycan composition complements.\n\n    Based upon the complement ion indexing strategy described in [1]_.\n\n    Attributes\n    ----------\n    members : :class:`list` of :class:`GlycanCombinationRecord`\n        The glycan composition combinations in this index.\n    unique_fragments : :class:`dict` of :class:`str` -> :class:`dict` of :class:`str` -> :class:`IndexGlycanCompositionFragment`\n        An internment table for each glycosylation type mapping fragment name to glycan composition fragments\n    fragment_index : :class:`dict` of :class:`int` -> :class:`list` of :class:`ComplementFragment`\n        A sparse index mapping :attr:`resolution` scaled masses to bins. The mass values and binned fragments are complements of\n        true fragments.\n    counter : int\n        A counter to assign each unique fragment a unique integer index.\n    fragment_weight : float\n        A scoring parameter to weight overall coverage with.\n    core_weight : float\n        A scoring parameter to weight core motif coverage with.\n    resolution : float\n        A scaling factor to convert real masses into truncated bin indices.\n\n\n    References\n    ----------\n    ..[1] Zeng, W., Cao, W., Liu, M., He, S., & Yang, P. (2021). Precise, Fast and\n      Comprehensive Analysis of Intact Glycopeptides and Monosaccharide-Modifications with pGlyco3.\n      Bioarxiv. https://doi.org/https://doi.org/10.1101/2021.02.06.430063\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex_spec = {
  "glycresoft._c.tandem.core_search.GlycanFragmentIndex",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""GlycanFragmentIndex", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("GlycanFragmentIndex(members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0)\nA fast and sparse in-memory fragment ion index for quickly matching peaks against multiple\n    glycan composition complements.\n\n    Based upon the complement ion indexing strategy described in [1]_.\n\n    Attributes\n    ----------\n    members : :class:`list` of :class:`GlycanCombinationRecord`\n        The glycan composition combinations in this index.\n    unique_fragments : :class:`dict` of :class:`str` -> :class:`dict` of :class:`str` -> :class:`IndexGlycanCompositionFragment`\n        An internment table for each glycosylation type mapping fragment name to glycan composition fragments\n    fragment_index : :class:`dict` of :class:`int` -> :class:`list` of :class:`ComplementFragment`\n        A sparse index mapping :attr:`resolution` scaled masses to bins. The mass values and binned fragments are complements of\n        true fragments.\n    counter : int\n        A counter to assign each unique fragment a unique integer index.\n    fragment_weight : float\n        A scoring parameter to weight overall coverage with.\n    core_weight : float\n        A scoring parameter to weight core motif coverage with.\n    resolution : float\n        A scaling factor to convert real masses into truncated bin indices.\n\n\n    References\n    ----------\n    ..[1] Zeng, W., Cao, W., Liu, M., He, S., & Yang, P. (2021). Precise, Fast and\n      Comprehensive Analysis of Intact Glycopeptides and Monosaccharide-Modifications with pGlyco3.\n      Bioarxiv. https://doi.org/https://doi.org/10.1101/2021.02.06.430063\n    "), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *__pyx_freelist_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__[8];
static int __pyx_freecount_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__[--__pyx_freecount_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__)))) {
    __pyx_freelist_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__[__pyx_freecount_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__++] = ((struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *p = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter___spec = {
  "glycresoft._c.tandem.core_search.__pyx_scope_struct____iter__",
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter___slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.tandem.core_search.""__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseGlycanMatch, __pyx_k_CoarseGlycanMatch, sizeof(__pyx_k_CoarseGlycanMatch), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseGlycanMatch___iter, __pyx_k_CoarseGlycanMatch___iter, sizeof(__pyx_k_CoarseGlycanMatch___iter), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseGlycanMatch___reduce_cytho, __pyx_k_CoarseGlycanMatch___reduce_cytho, sizeof(__pyx_k_CoarseGlycanMatch___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseGlycanMatch___setstate_cyt, __pyx_k_CoarseGlycanMatch___setstate_cyt, sizeof(__pyx_k_CoarseGlycanMatch___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseGlycanMatch_estimate_pepti, __pyx_k_CoarseGlycanMatch_estimate_pepti, sizeof(__pyx_k_CoarseGlycanMatch_estimate_pepti), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseStubGlycopeptideFragment, __pyx_k_CoarseStubGlycopeptideFragment, sizeof(__pyx_k_CoarseStubGlycopeptideFragment), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseStubGlycopeptideFragment_2, __pyx_k_CoarseStubGlycopeptideFragment_2, sizeof(__pyx_k_CoarseStubGlycopeptideFragment_2), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseStubGlycopeptideMatch, __pyx_k_CoarseStubGlycopeptideMatch, sizeof(__pyx_k_CoarseStubGlycopeptideMatch), 0, 0, 1, 1},
    {&__pyx_n_s_CoarseStubGlycopeptideMatch___re, __pyx_k_CoarseStubGlycopeptideMatch___re, sizeof(__pyx_k_CoarseStubGlycopeptideMatch___re), 0, 0, 1, 1},
    {&__pyx_n_s_ComplementFragment, __pyx_k_ComplementFragment, sizeof(__pyx_k_ComplementFragment), 0, 0, 1, 1},
    {&__pyx_n_s_ComplementFragment___reduce_cyth, __pyx_k_ComplementFragment___reduce_cyth, sizeof(__pyx_k_ComplementFragment___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_ComplementFragment___setstate_cy, __pyx_k_ComplementFragment___setstate_cy, sizeof(__pyx_k_ComplementFragment___setstate_cy), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase, __pyx_k_GlycanCoarseScorerBase, sizeof(__pyx_k_GlycanCoarseScorerBase), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase___reduce, __pyx_k_GlycanCoarseScorerBase___reduce, sizeof(__pyx_k_GlycanCoarseScorerBase___reduce), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase___setstat, __pyx_k_GlycanCoarseScorerBase___setstat, sizeof(__pyx_k_GlycanCoarseScorerBase___setstat), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase__calculat, __pyx_k_GlycanCoarseScorerBase__calculat, sizeof(__pyx_k_GlycanCoarseScorerBase__calculat), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase__gag_matc, __pyx_k_GlycanCoarseScorerBase__gag_matc, sizeof(__pyx_k_GlycanCoarseScorerBase__gag_matc), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase__match_fr, __pyx_k_GlycanCoarseScorerBase__match_fr, sizeof(__pyx_k_GlycanCoarseScorerBase__match_fr), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase__n_glycan, __pyx_k_GlycanCoarseScorerBase__n_glycan, sizeof(__pyx_k_GlycanCoarseScorerBase__n_glycan), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCoarseScorerBase__o_glycan, __pyx_k_GlycanCoarseScorerBase__o_glycan, sizeof(__pyx_k_GlycanCoarseScorerBase__o_glycan), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase, __pyx_k_GlycanCombinationRecordBase, sizeof(__pyx_k_GlycanCombinationRecordBase), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase___re, __pyx_k_GlycanCombinationRecordBase___re, sizeof(__pyx_k_GlycanCombinationRecordBase___re), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase___se, __pyx_k_GlycanCombinationRecordBase___se, sizeof(__pyx_k_GlycanCombinationRecordBase___se), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_clea, __pyx_k_GlycanCombinationRecordBase_clea, sizeof(__pyx_k_GlycanCombinationRecordBase_clea), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_get, __pyx_k_GlycanCombinationRecordBase_get, sizeof(__pyx_k_GlycanCombinationRecordBase_get), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_get_2, __pyx_k_GlycanCombinationRecordBase_get_2, sizeof(__pyx_k_GlycanCombinationRecordBase_get_2), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_get_3, __pyx_k_GlycanCombinationRecordBase_get_3, sizeof(__pyx_k_GlycanCombinationRecordBase_get_3), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_is_g, __pyx_k_GlycanCombinationRecordBase_is_g, sizeof(__pyx_k_GlycanCombinationRecordBase_is_g), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_is_n, __pyx_k_GlycanCombinationRecordBase_is_n, sizeof(__pyx_k_GlycanCombinationRecordBase_is_n), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_is_o, __pyx_k_GlycanCombinationRecordBase_is_o, sizeof(__pyx_k_GlycanCombinationRecordBase_is_o), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCombinationRecordBase_norm, __pyx_k_GlycanCombinationRecordBase_norm, sizeof(__pyx_k_GlycanCombinationRecordBase_norm), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFilteringPeptideMassEstima, __pyx_k_GlycanFilteringPeptideMassEstima, sizeof(__pyx_k_GlycanFilteringPeptideMassEstima), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex, __pyx_k_GlycanFragmentIndex, sizeof(__pyx_k_GlycanFragmentIndex), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex___reduce_cyt, __pyx_k_GlycanFragmentIndex___reduce_cyt, sizeof(__pyx_k_GlycanFragmentIndex___reduce_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex___setstate_c, __pyx_k_GlycanFragmentIndex___setstate_c, sizeof(__pyx_k_GlycanFragmentIndex___setstate_c), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex__get_fragmen, __pyx_k_GlycanFragmentIndex__get_fragmen, sizeof(__pyx_k_GlycanFragmentIndex__get_fragmen), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex__intern, __pyx_k_GlycanFragmentIndex__intern, sizeof(__pyx_k_GlycanFragmentIndex__intern), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex_bin_index, __pyx_k_GlycanFragmentIndex_bin_index, sizeof(__pyx_k_GlycanFragmentIndex_bin_index), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex_build, __pyx_k_GlycanFragmentIndex_build, sizeof(__pyx_k_GlycanFragmentIndex_build), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex_match, __pyx_k_GlycanFragmentIndex_match, sizeof(__pyx_k_GlycanFragmentIndex_match), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanFragmentIndex_query_width, __pyx_k_GlycanFragmentIndex_query_width, sizeof(__pyx_k_GlycanFragmentIndex_query_width), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanMatchResult, __pyx_k_GlycanMatchResult, sizeof(__pyx_k_GlycanMatchResult), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanMatchResult___reduce_cytho, __pyx_k_GlycanMatchResult___reduce_cytho, sizeof(__pyx_k_GlycanMatchResult___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanMatchResult___setstate_cyt, __pyx_k_GlycanMatchResult___setstate_cyt, sizeof(__pyx_k_GlycanMatchResult___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanTypes, __pyx_k_GlycanTypes, sizeof(__pyx_k_GlycanTypes), 0, 0, 1, 1},
    {&__pyx_n_s_HashableGlycanComposition, __pyx_k_HashableGlycanComposition, sizeof(__pyx_k_HashableGlycanComposition), 0, 0, 1, 1},
    {&__pyx_n_s_HashableGlycanComposition_2, __pyx_k_HashableGlycanComposition_2, sizeof(__pyx_k_HashableGlycanComposition_2), 0, 0, 1, 1},
    {&__pyx_n_s_Hex, __pyx_k_Hex, sizeof(__pyx_k_Hex), 0, 0, 1, 1},
    {&__pyx_n_s_HexNAc, __pyx_k_HexNAc, sizeof(__pyx_k_HexNAc), 0, 0, 1, 1},
    {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_k_Incompatible_checksums_0x_x_vs_0_6, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_6), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_k_Incompatible_checksums_0x_x_vs_0_7, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_7), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_k_Incompatible_checksums_0x_x_vs_0_8, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_8), 0, 0, 1, 0},
    {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
    {&__pyx_n_s_IndexGlycanCompositionFragment, __pyx_k_IndexGlycanCompositionFragment, sizeof(__pyx_k_IndexGlycanCompositionFragment), 0, 0, 1, 1},
    {&__pyx_n_s_IndexGlycanCompositionFragment_2, __pyx_k_IndexGlycanCompositionFragment_2, sizeof(__pyx_k_IndexGlycanCompositionFragment_2), 0, 0, 1, 1},
    {&__pyx_n_s_IndexGlycanCompositionFragment_3, __pyx_k_IndexGlycanCompositionFragment_3, sizeof(__pyx_k_IndexGlycanCompositionFragment_3), 0, 0, 1, 1},
    {&__pyx_n_s_IndexGlycanCompositionFragment_4, __pyx_k_IndexGlycanCompositionFragment_4, sizeof(__pyx_k_IndexGlycanCompositionFragment_4), 0, 0, 1, 1},
    {&__pyx_n_s_IndexGlycanCompositionFragment_c, __pyx_k_IndexGlycanCompositionFragment_c, sizeof(__pyx_k_IndexGlycanCompositionFragment_c), 0, 0, 1, 1},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_PartialGlycanSolution, __pyx_k_PartialGlycanSolution, sizeof(__pyx_k_PartialGlycanSolution), 0, 0, 1, 1},
    {&__pyx_n_s_PartialGlycanSolution___reduce_c, __pyx_k_PartialGlycanSolution___reduce_c, sizeof(__pyx_k_PartialGlycanSolution___reduce_c), 0, 0, 1, 1},
    {&__pyx_n_s_PartialGlycanSolution___setstate, __pyx_k_PartialGlycanSolution___setstate, sizeof(__pyx_k_PartialGlycanSolution___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_kp_s_Re_estimated_peptide_mass_error, __pyx_k_Re_estimated_peptide_mass_error, sizeof(__pyx_k_Re_estimated_peptide_mass_error), 0, 0, 1, 0},
    {&__pyx_n_s_ScoreCoarseGlycanMatchPair, __pyx_k_ScoreCoarseGlycanMatchPair, sizeof(__pyx_k_ScoreCoarseGlycanMatchPair), 0, 0, 1, 1},
    {&__pyx_n_s_ScoreCoarseGlycanMatchPair___red, __pyx_k_ScoreCoarseGlycanMatchPair___red, sizeof(__pyx_k_ScoreCoarseGlycanMatchPair___red), 0, 0, 1, 1},
    {&__pyx_n_s_ScoreCoarseGlycanMatchPair___set, __pyx_k_ScoreCoarseGlycanMatchPair___set, sizeof(__pyx_k_ScoreCoarseGlycanMatchPair___set), 0, 0, 1, 1},
    {&__pyx_n_s_Unmodified, __pyx_k_Unmodified, sizeof(__pyx_k_Unmodified), 0, 0, 1, 1},
    {&__pyx_n_s_Unmodified_2, __pyx_k_Unmodified_2, sizeof(__pyx_k_Unmodified_2), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 0, 1, 1},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_n_s__89, __pyx_k__89, sizeof(__pyx_k__89), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_bin_index, __pyx_k_bin_index, sizeof(__pyx_k_bin_index), 0, 0, 1, 1},
    {&__pyx_n_s_build, __pyx_k_build, sizeof(__pyx_k_build), 0, 0, 1, 1},
    {&__pyx_n_s_build_locals_lambda, __pyx_k_build_locals_lambda, sizeof(__pyx_k_build_locals_lambda), 0, 0, 1, 1},
    {&__pyx_n_s_calculate_score, __pyx_k_calculate_score, sizeof(__pyx_k_calculate_score), 0, 0, 1, 1},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_n_s_comp, __pyx_k_comp, sizeof(__pyx_k_comp), 0, 0, 1, 1},
    {&__pyx_n_s_comp_frags, __pyx_k_comp_frags, sizeof(__pyx_k_comp_frags), 0, 0, 1, 1},
    {&__pyx_n_s_comp_i, __pyx_k_comp_i, sizeof(__pyx_k_comp_i), 0, 0, 1, 1},
    {&__pyx_n_s_comp_n, __pyx_k_comp_n, sizeof(__pyx_k_comp_n), 0, 0, 1, 1},
    {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_core_matched, __pyx_k_core_matched, sizeof(__pyx_k_core_matched), 0, 0, 1, 1},
    {&__pyx_n_s_core_matches, __pyx_k_core_matches, sizeof(__pyx_k_core_matches), 0, 0, 1, 1},
    {&__pyx_n_s_core_theoretical, __pyx_k_core_theoretical, sizeof(__pyx_k_core_theoretical), 0, 0, 1, 1},
    {&__pyx_n_s_core_weight, __pyx_k_core_weight, sizeof(__pyx_k_core_weight), 0, 0, 1, 1},
    {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
    {&__pyx_n_s_deconvoluted_peak_set, __pyx_k_deconvoluted_peak_set, sizeof(__pyx_k_deconvoluted_peak_set), 0, 0, 1, 1},
    {&__pyx_n_s_defaultdict, __pyx_k_defaultdict, sizeof(__pyx_k_defaultdict), 0, 0, 1, 1},
    {&__pyx_n_s_dehydrated_mass, __pyx_k_dehydrated_mass, sizeof(__pyx_k_dehydrated_mass), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
    {&__pyx_n_s_estimate_peptide_mass, __pyx_k_estimate_peptide_mass, sizeof(__pyx_k_estimate_peptide_mass), 0, 0, 1, 1},
    {&__pyx_n_s_extended, __pyx_k_extended, sizeof(__pyx_k_extended), 0, 0, 1, 1},
    {&__pyx_kp_s_f_r, __pyx_k_f_r, sizeof(__pyx_k_f_r), 0, 0, 1, 0},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_frag, __pyx_k_frag, sizeof(__pyx_k_frag), 0, 0, 1, 1},
    {&__pyx_n_s_frag_i, __pyx_k_frag_i, sizeof(__pyx_k_frag_i), 0, 0, 1, 1},
    {&__pyx_n_s_fragment, __pyx_k_fragment, sizeof(__pyx_k_fragment), 0, 0, 1, 1},
    {&__pyx_n_s_fragment_matches, __pyx_k_fragment_matches, sizeof(__pyx_k_fragment_matches), 0, 0, 1, 1},
    {&__pyx_n_s_fragment_weight, __pyx_k_fragment_weight, sizeof(__pyx_k_fragment_weight), 0, 0, 1, 1},
    {&__pyx_n_s_frags, __pyx_k_frags, sizeof(__pyx_k_frags), 0, 0, 1, 1},
    {&__pyx_n_s_frags_i, __pyx_k_frags_i, sizeof(__pyx_k_frags_i), 0, 0, 1, 1},
    {&__pyx_n_s_frags_n, __pyx_k_frags_n, sizeof(__pyx_k_frags_n), 0, 0, 1, 1},
    {&__pyx_n_s_gag_linker, __pyx_k_gag_linker, sizeof(__pyx_k_gag_linker), 0, 0, 1, 1},
    {&__pyx_n_s_gag_match_stubs, __pyx_k_gag_match_stubs, sizeof(__pyx_k_gag_match_stubs), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_gcr, __pyx_k_gcr, sizeof(__pyx_k_gcr), 0, 0, 1, 1},
    {&__pyx_n_s_get_fragments, __pyx_k_get_fragments, sizeof(__pyx_k_get_fragments), 0, 0, 1, 1},
    {&__pyx_n_s_get_gag_linker_glycan_fragments, __pyx_k_get_gag_linker_glycan_fragments, sizeof(__pyx_k_get_gag_linker_glycan_fragments), 0, 0, 1, 1},
    {&__pyx_n_s_get_n_glycan_fragments, __pyx_k_get_n_glycan_fragments, sizeof(__pyx_k_get_n_glycan_fragments), 0, 0, 1, 1},
    {&__pyx_n_s_get_o_glycan_fragments, __pyx_k_get_o_glycan_fragments, sizeof(__pyx_k_get_o_glycan_fragments), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_combination, __pyx_k_glycan_combination, sizeof(__pyx_k_glycan_combination), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_combination_db, __pyx_k_glycan_combination_db, sizeof(__pyx_k_glycan_combination_db), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_index, __pyx_k_glycan_index, sizeof(__pyx_k_glycan_index), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_match, __pyx_k_glycan_match, sizeof(__pyx_k_glycan_match), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_size, __pyx_k_glycan_size, sizeof(__pyx_k_glycan_size), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_types, __pyx_k_glycan_types, sizeof(__pyx_k_glycan_types), 0, 0, 1, 1},
    {&__pyx_n_s_glycosylation_type, __pyx_k_glycosylation_type, sizeof(__pyx_k_glycosylation_type), 0, 0, 1, 1},
    {&__pyx_n_s_glycresoft__c_tandem_core_search, __pyx_k_glycresoft__c_tandem_core_search, sizeof(__pyx_k_glycresoft__c_tandem_core_search), 0, 0, 1, 1},
    {&__pyx_n_s_glycresoft_chromatogram_tree, __pyx_k_glycresoft_chromatogram_tree, sizeof(__pyx_k_glycresoft_chromatogram_tree), 0, 0, 1, 1},
    {&__pyx_n_s_glycresoft_serialize, __pyx_k_glycresoft_serialize, sizeof(__pyx_k_glycresoft_serialize), 0, 0, 1, 1},
    {&__pyx_n_s_glypy_structure_glycan_compositi, __pyx_k_glypy_structure_glycan_compositi, sizeof(__pyx_k_glypy_structure_glycan_compositi), 0, 0, 1, 1},
    {&__pyx_n_s_hash_key, __pyx_k_hash_key, sizeof(__pyx_k_hash_key), 0, 0, 1, 1},
    {&__pyx_n_s_high, __pyx_k_high, sizeof(__pyx_k_high), 0, 0, 1, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_intern, __pyx_k_intern, sizeof(__pyx_k_intern), 0, 0, 1, 1},
    {&__pyx_n_s_is_core, __pyx_k_is_core, sizeof(__pyx_k_is_core), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_is_extended, __pyx_k_is_extended, sizeof(__pyx_k_is_extended), 0, 0, 1, 1},
    {&__pyx_n_s_is_gag_linker, __pyx_k_is_gag_linker, sizeof(__pyx_k_is_gag_linker), 0, 0, 1, 1},
    {&__pyx_n_s_is_n_glycan, __pyx_k_is_n_glycan, sizeof(__pyx_k_is_n_glycan), 0, 0, 1, 1},
    {&__pyx_n_s_is_o_glycan, __pyx_k_is_o_glycan, sizeof(__pyx_k_is_o_glycan), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
    {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_low, __pyx_k_low, sizeof(__pyx_k_low), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_mass, __pyx_k_mass, sizeof(__pyx_k_mass), 0, 0, 1, 1},
    {&__pyx_n_s_mass_shift, __pyx_k_mass_shift, sizeof(__pyx_k_mass_shift), 0, 0, 1, 1},
    {&__pyx_n_s_mass_shift_tandem_mass, __pyx_k_mass_shift_tandem_mass, sizeof(__pyx_k_mass_shift_tandem_mass), 0, 0, 1, 1},
    {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
    {&__pyx_n_s_match_fragments, __pyx_k_match_fragments, sizeof(__pyx_k_match_fragments), 0, 0, 1, 1},
    {&__pyx_n_s_member, __pyx_k_member, sizeof(__pyx_k_member), 0, 0, 1, 1},
    {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
    {&__pyx_n_s_minimum_peptide_mass, __pyx_k_minimum_peptide_mass, sizeof(__pyx_k_minimum_peptide_mass), 0, 0, 1, 1},
    {&__pyx_n_s_n_core, __pyx_k_n_core, sizeof(__pyx_k_n_core), 0, 0, 1, 1},
    {&__pyx_n_s_n_frags, __pyx_k_n_frags, sizeof(__pyx_k_n_frags), 0, 0, 1, 1},
    {&__pyx_n_s_n_glycan, __pyx_k_n_glycan, sizeof(__pyx_k_n_glycan), 0, 0, 1, 1},
    {&__pyx_n_s_n_glycan_match_stubs, __pyx_k_n_glycan_match_stubs, sizeof(__pyx_k_n_glycan_match_stubs), 0, 0, 1, 1},
    {&__pyx_n_s_n_matched, __pyx_k_n_matched, sizeof(__pyx_k_n_matched), 0, 0, 1, 1},
    {&__pyx_n_s_n_theoretical, __pyx_k_n_theoretical, sizeof(__pyx_k_n_theoretical), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_neutral_mass, __pyx_k_neutral_mass, sizeof(__pyx_k_neutral_mass), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_n_s_normalized_size, __pyx_k_normalized_size, sizeof(__pyx_k_normalized_size), 0, 0, 1, 1},
    {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
    {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
    {&__pyx_kp_s_numpy__core_multiarray_failed_to, __pyx_k_numpy__core_multiarray_failed_to, sizeof(__pyx_k_numpy__core_multiarray_failed_to), 0, 0, 1, 0},
    {&__pyx_kp_s_numpy__core_umath_failed_to_impo, __pyx_k_numpy__core_umath_failed_to_impo, sizeof(__pyx_k_numpy__core_umath_failed_to_impo), 0, 0, 1, 0},
    {&__pyx_n_s_o_glycan, __pyx_k_o_glycan, sizeof(__pyx_k_o_glycan), 0, 0, 1, 1},
    {&__pyx_n_s_o_glycan_match_stubs, __pyx_k_o_glycan_match_stubs, sizeof(__pyx_k_o_glycan_match_stubs), 0, 0, 1, 1},
    {&__pyx_n_s_partial_score_getter, __pyx_k_partial_score_getter, sizeof(__pyx_k_partial_score_getter), 0, 0, 1, 1},
    {&__pyx_n_s_peaks_matched, __pyx_k_peaks_matched, sizeof(__pyx_k_peaks_matched), 0, 0, 1, 1},
    {&__pyx_n_s_peptide_mass, __pyx_k_peptide_mass, sizeof(__pyx_k_peptide_mass), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_precursor_information, __pyx_k_precursor_information, sizeof(__pyx_k_precursor_information), 0, 0, 1, 1},
    {&__pyx_n_s_product_error_tolerance, __pyx_k_product_error_tolerance, sizeof(__pyx_k_product_error_tolerance), 0, 0, 1, 1},
    {&__pyx_n_s_ptemp, __pyx_k_ptemp, sizeof(__pyx_k_ptemp), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CoarseGlycanMatch, __pyx_k_pyx_unpickle_CoarseGlycanMatch, sizeof(__pyx_k_pyx_unpickle_CoarseGlycanMatch), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ComplementFragmen, __pyx_k_pyx_unpickle_ComplementFragmen, sizeof(__pyx_k_pyx_unpickle_ComplementFragmen), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_GlycanCoarseScore, __pyx_k_pyx_unpickle_GlycanCoarseScore, sizeof(__pyx_k_pyx_unpickle_GlycanCoarseScore), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_GlycanCombination, __pyx_k_pyx_unpickle_GlycanCombination, sizeof(__pyx_k_pyx_unpickle_GlycanCombination), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_GlycanFragmentInd, __pyx_k_pyx_unpickle_GlycanFragmentInd, sizeof(__pyx_k_pyx_unpickle_GlycanFragmentInd), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_GlycanMatchResult, __pyx_k_pyx_unpickle_GlycanMatchResult, sizeof(__pyx_k_pyx_unpickle_GlycanMatchResult), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PartialGlycanSolu, __pyx_k_pyx_unpickle_PartialGlycanSolu, sizeof(__pyx_k_pyx_unpickle_PartialGlycanSolu), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ScoreCoarseGlycan, __pyx_k_pyx_unpickle_ScoreCoarseGlycan, sizeof(__pyx_k_pyx_unpickle_ScoreCoarseGlycan), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_query_mass, __pyx_k_query_mass, sizeof(__pyx_k_query_mass), 0, 0, 1, 1},
    {&__pyx_n_s_query_width, __pyx_k_query_width, sizeof(__pyx_k_query_width), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_recalibrate_peptide_mass, __pyx_k_recalibrate_peptide_mass, sizeof(__pyx_k_recalibrate_peptide_mass), 0, 0, 1, 1},
    {&__pyx_n_s_recalibrated_peptide_mass, __pyx_k_recalibrated_peptide_mass, sizeof(__pyx_k_recalibrated_peptide_mass), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_resolution, __pyx_k_resolution, sizeof(__pyx_k_resolution), 0, 0, 1, 1},
    {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
    {&__pyx_kp_s_s_s_f_f_r, __pyx_k_s_s_f_f_r, sizeof(__pyx_k_s_s_f_f_r), 0, 0, 1, 0},
    {&__pyx_kp_s_s_s_f_r, __pyx_k_s_s_f_r, sizeof(__pyx_k_s_s_f_r), 0, 0, 1, 0},
    {&__pyx_n_s_scan, __pyx_k_scan, sizeof(__pyx_k_scan), 0, 0, 1, 1},
    {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
    {&__pyx_n_s_score_getter, __pyx_k_score_getter, sizeof(__pyx_k_score_getter), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_kp_s_self___class_____name___self_ma, __pyx_k_self___class_____name___self_ma, sizeof(__pyx_k_self___class_____name___self_ma), 0, 0, 1, 0},
    {&__pyx_kp_s_self___class_____name___self_n, __pyx_k_self___class_____name___self_n, sizeof(__pyx_k_self___class_____name___self_n), 0, 0, 1, 0},
    {&__pyx_kp_s_self___class_____name___self_pe, __pyx_k_self___class_____name___self_pe, sizeof(__pyx_k_self___class_____name___self_pe), 0, 0, 1, 0},
    {&__pyx_kp_s_self___class_____name___self_pe_2, __pyx_k_self___class_____name___self_pe_2, sizeof(__pyx_k_self___class_____name___self_pe_2), 0, 0, 1, 0},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_shift_mass, __pyx_k_shift_mass, sizeof(__pyx_k_shift_mass), 0, 0, 1, 1},
    {&__pyx_n_s_shifts, __pyx_k_shifts, sizeof(__pyx_k_shifts), 0, 0, 1, 1},
    {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
    {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_k_src_glycresoft__c_tandem_core_se, sizeof(__pyx_k_src_glycresoft__c_tandem_core_se), 0, 0, 1, 0},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
    {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
    {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 272, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 515, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 594, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 769, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 806, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 853, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 1026, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1026
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_numpy__core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../../../../private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1032
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_numpy__core_umath_failed_to_impo); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "glycresoft/_c/tandem/core_search.pyx":90
 *         ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_n_glycan)
 *         if ptemp == NULL:
 *             strategy = StubGlycopeptideStrategy(None, extended=True)             # <<<<<<<<<<<<<<
 *             shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
 *             fragment_structs = []
 */
  __pyx_tuple__3 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x564dc94, 0x19c2b1d, 0x930e938):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x564dc94, 0x19c2b1d, 0x930e938) = (_fragment_cache, _hash, composition, count, dehydrated_mass, fragment_set_properties, glycan_types, id, internal_size_approximation, side_group_count, size))" % __pyx_checksum
 */
  __pyx_tuple__6 = PyTuple_Pack(3, __pyx_int_90496148, __pyx_int_27011869, __pyx_int_154200376); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_252421743, __pyx_int_117397179, __pyx_int_209767785); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_30049220, __pyx_int_69199841, __pyx_int_163158466); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(3, __pyx_int_143121537, __pyx_int_224310728, __pyx_int_151187253); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_int_88044796, __pyx_int_162950457, __pyx_int_124808126); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_tuple__12 = PyTuple_Pack(3, __pyx_int_134214914, __pyx_int_14758241, __pyx_int_115380717); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_int_92231173, __pyx_int_217673008, __pyx_int_203386150); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_int_16698535, __pyx_int_191711013, __pyx_int_229638544); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "glycresoft/_c/tandem/core_search.pyx":59
 *         public dict fragment_set_properties
 * 
 *     cpdef bint is_n_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_is_n_glycan, 59, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 59, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":62
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 *     cpdef bint is_o_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 */
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_is_o_glycan, 62, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 62, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":65
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 *     cpdef bint is_gag_linker(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_gag_linker in self.glycan_types
 * 
 */
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_is_gag_linker, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 65, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":69
 * 
 *     @cython.cdivision(True)
 *     cpdef double normalized_size(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double k
 */
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_normalized_size, 69, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 69, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":84
 *         return d
 * 
 *     cpdef list get_n_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_get_n_glycan_fragments, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":108
 *             return result
 * 
 *     cpdef list get_o_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_get_o_glycan_fragments, 108, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 108, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":128
 *             return result
 * 
 *     cpdef list get_gag_linker_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_get_gag_linker_glycan_fragments, 128, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 128, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":148
 *             return result
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         self._fragment_cache.clear()
 * 
 */
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_clear, 148, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 148, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__25 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCombinationRecordBase__set_state(self, __pyx_state)
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":163
 *         self.is_core = is_core
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.is_core)
 * 
 */
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_reduce, 163, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 163, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":216
 *         self.peaks_matched = peaks_matched
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)
 * 
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_reduce, 216, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 216, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":260
 *         yield self.core_theoretical
 * 
 *     cpdef double estimate_peptide_mass(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double weighted_mass_acc
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_estimate_peptide_mass, 260, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 260, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoarseGlycanMatch__set_state(self, __pyx_state)
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":305
 * 
 *     @cython.boundscheck(False)
 *     cpdef CoarseGlycanMatch _match_fragments(self, object scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list fragment_matches
 */
  __pyx_tuple__34 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_scan, __pyx_n_s_peptide_mass, __pyx_n_s_shifts, __pyx_n_s_mass_shift_tandem_mass); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_match_fragments, 305, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_float_0_0); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "glycresoft/_c/tandem/core_search.pyx":356
 * 
 *     @cython.cdivision(True)
 *     cpdef double _calculate_score(self, CoarseGlycanMatch glycan_match):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j, n, m
 */
  __pyx_tuple__37 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_glycan_match); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_calculate_score, 356, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 356, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":381
 *         return score
 * 
 *     cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */
  __pyx_tuple__39 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_scan, __pyx_n_s_peptide_mass, __pyx_n_s_glycan_combination, __pyx_n_s_mass_shift_tandem_mass); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_n_glycan_match_stubs, 381, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 381, __pyx_L1_error)
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_float_0_0); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "glycresoft/_c/tandem/core_search.pyx":387
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_o_glycan_match_stubs, 387, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 387, __pyx_L1_error)
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_float_0_0); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "glycresoft/_c/tandem/core_search.pyx":393
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                              GlycanCombinationRecordBase glycan_combination,
 *                                              double mass_shift_tandem_mass=0.0):
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_gag_match_stubs, 393, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_float_0_0); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCoarseScorerBase__set_state(self, __pyx_state)
 */
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanMatchResult__set_state(self, __pyx_state)
 */
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(self, __pyx_state)
 */
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":524
 * 
 * 
 * cpdef double score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<GlycanMatchResult>x).score
 * 
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_n_s_x); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_score_getter, 524, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 524, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":530
 * @cython.boundscheck(False)
 * @cython.binding(True)
 * cpdef GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list output, glycan_combination_db
 */
  __pyx_tuple__54 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_scan, __pyx_n_s_mass_shift, __pyx_n_s_query_mass); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_GlycanFilteringPeptideMassEstima, 530, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 530, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":626
 *         return self
 * 
 *     cpdef GlycanCompositionFragment copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IndexGlycanCompositionFragment inst
 */
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_copy, 626, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 626, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":638
 *         return inst
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_reduce, 638, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 638, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":641
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {
 *             "index": self.index,
 */
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_getstate, 641, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 641, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":648
 *         }
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.index = state['index']
 *         self.name = state['name']
 */
  __pyx_tuple__59 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_state); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_setstate, 648, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 648, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ComplementFragment__set_state(self, __pyx_state)
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PartialGlycanSolution__set_state(self, __pyx_state)
 */
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":765
 *         self.resolution = resolution
 * 
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):             # <<<<<<<<<<<<<<
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:
 */
  __pyx_tuple__65 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_fragment, __pyx_n_s_glycosylation_type); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_intern, 765, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 765, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":783
 *             return fragment
 * 
 *     cpdef long query_width(self, double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_mass, __pyx_n_s_error_tolerance); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_query_width, 783, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 783, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":786
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 *     cpdef long bin_index(self, double mass):             # <<<<<<<<<<<<<<
 *         return <long>(mass * self.resolution)
 * 
 */
  __pyx_tuple__69 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_mass); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_bin_index, 786, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 786, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":789
 *         return <long>(mass * self.resolution)
 * 
 *     def build(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */
  __pyx_tuple__71 = PyTuple_Pack(22, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_n_frags, __pyx_n_s_n_core, __pyx_n_s_j, __pyx_n_s_mass, __pyx_n_s_d, __pyx_n_s_low, __pyx_n_s_high, __pyx_n_s_ptemp, __pyx_n_s_comp_frags, __pyx_n_s_frags, __pyx_n_s_frags_i, __pyx_n_s_frags_n, __pyx_n_s_comp_i, __pyx_n_s_comp_n, __pyx_n_s_frag, __pyx_n_s_comp, __pyx_n_s_member, __pyx_n_s_i, __pyx_n_s_glycosylation_type, __pyx_n_s_frag_i); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 22, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_build, 789, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":845
 *         return j
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):             # <<<<<<<<<<<<<<
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()
 */
  __pyx_tuple__73 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_gcr, __pyx_n_s_glycosylation_type); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_get_fragments, 845, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 845, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":917
 * 
 *     @cython.cdivision
 *     cpdef list match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double precursor_mass, d
 */
  __pyx_tuple__75 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_scan, __pyx_n_s_error_tolerance, __pyx_n_s_mass_shift, __pyx_n_s_query_mass); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_match, 917, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 917, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(3, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanFragmentIndex__set_state(self, __pyx_state)
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":990
 * 
 * 
 * cpdef double partial_score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<PartialGlycanSolution>x).score
 */
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_tandem_core_se, __pyx_n_s_partial_score_getter, 990, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 990, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCombinationRecordBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__80 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycanCombination, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CoarseGlycanMatch, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycanCoarseScore, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycanMatchResult, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ScoreCoarseGlycan, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ComplementFragmen, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PartialGlycanSolu, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(3, 1, __pyx_L1_error)
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycanFragmentInd, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_42 = PyFloat_FromDouble(0.42); if (unlikely(!__pyx_float_0_42)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_56 = PyFloat_FromDouble(0.56); if (unlikely(!__pyx_float_0_56)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_5 = PyFloat_FromDouble(1e-5); if (unlikely(!__pyx_float_1eneg_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_100_0 = PyFloat_FromDouble(100.0); if (unlikely(!__pyx_float_100_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_14758241 = PyInt_FromLong(14758241L); if (unlikely(!__pyx_int_14758241)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16698535 = PyInt_FromLong(16698535L); if (unlikely(!__pyx_int_16698535)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_27011869 = PyInt_FromLong(27011869L); if (unlikely(!__pyx_int_27011869)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_30049220 = PyInt_FromLong(30049220L); if (unlikely(!__pyx_int_30049220)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_69199841 = PyInt_FromLong(69199841L); if (unlikely(!__pyx_int_69199841)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_88044796 = PyInt_FromLong(88044796L); if (unlikely(!__pyx_int_88044796)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_90496148 = PyInt_FromLong(90496148L); if (unlikely(!__pyx_int_90496148)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_92231173 = PyInt_FromLong(92231173L); if (unlikely(!__pyx_int_92231173)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_115380717 = PyInt_FromLong(115380717L); if (unlikely(!__pyx_int_115380717)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_117397179 = PyInt_FromLong(117397179L); if (unlikely(!__pyx_int_117397179)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_124808126 = PyInt_FromLong(124808126L); if (unlikely(!__pyx_int_124808126)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_134214914 = PyInt_FromLong(134214914L); if (unlikely(!__pyx_int_134214914)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_143121537 = PyInt_FromLong(143121537L); if (unlikely(!__pyx_int_143121537)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_151187253 = PyInt_FromLong(151187253L); if (unlikely(!__pyx_int_151187253)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_154200376 = PyInt_FromLong(154200376L); if (unlikely(!__pyx_int_154200376)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_162950457 = PyInt_FromLong(162950457L); if (unlikely(!__pyx_int_162950457)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_163158466 = PyInt_FromLong(163158466L); if (unlikely(!__pyx_int_163158466)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_191711013 = PyInt_FromLong(191711013L); if (unlikely(!__pyx_int_191711013)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_203386150 = PyInt_FromLong(203386150L); if (unlikely(!__pyx_int_203386150)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_209767785 = PyInt_FromLong(209767785L); if (unlikely(!__pyx_int_209767785)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_217673008 = PyInt_FromLong(217673008L); if (unlikely(!__pyx_int_217673008)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_224310728 = PyInt_FromLong(224310728L); if (unlikely(!__pyx_int_224310728)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_229638544 = PyInt_FromLong(229638544L); if (unlikely(!__pyx_int_229638544)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_252421743 = PyInt_FromLong(252421743L); if (unlikely(!__pyx_int_252421743)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_10glycresoft_2_c_6tandem_11core_search_HashableGlycanComposition = Py_None; Py_INCREF(Py_None);
  __pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified = ((struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *)Py_None); Py_INCREF(Py_None);
  __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan = Py_None; Py_INCREF(Py_None);
  __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan = Py_None; Py_INCREF(Py_None);
  __pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.is_n_glycan = (int (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_n_glycan;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.is_o_glycan = (int (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_o_glycan;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.is_gag_linker = (int (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_is_gag_linker;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.normalized_size = (double (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_normalized_size;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.get_n_glycan_fragments = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_n_glycan_fragments;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.get_o_glycan_fragments = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_o_glycan_fragments;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.get_gag_linker_glycan_fragments = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_get_gag_linker_glycan_fragments;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase.clear = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_clear;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase)) __PYX_ERR(0, 45, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanCombinationRecordBase, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment)) __PYX_ERR(0, 152, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CoarseStubGlycopeptideFragment, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch._create = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch *(*)(PyObject *, double, double, PyObject *))__pyx_f_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch__create;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch)) __PYX_ERR(0, 193, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CoarseStubGlycopeptideMatch, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch._create = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(PyObject *, double, double, double, double))__pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch__create;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch.estimate_peptide_mass = (double (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_estimate_peptide_mass;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch)) __PYX_ERR(0, 228, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CoarseGlycanMatch, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._match_fragments = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments *__pyx_optional_args))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__match_fragments;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._calculate_score = (double (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__calculate_score;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._n_glycan_match_stubs = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs *__pyx_optional_args))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__n_glycan_match_stubs;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._o_glycan_match_stubs = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs *__pyx_optional_args))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__o_glycan_match_stubs;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._gag_match_stubs = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs *__pyx_optional_args))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__gag_match_stubs;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._get_n_glycan_coarse_score = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *, double, double *))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_n_glycan_coarse_score;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._get_o_glycan_coarse_score = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *, double, double *))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_o_glycan_coarse_score;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase._get_gag_coarse_score = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase *, PyObject *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *, double, double *))__pyx_f_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase__get_gag_coarse_score;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase)) __PYX_ERR(0, 291, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanCoarseScorerBase, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult._create = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult *(*)(double, double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *, size_t, PyObject *, double))__pyx_f_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult__create;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult)) __PYX_ERR(0, 444, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanMatchResult, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair._create = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair *(*)(double, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch *))__pyx_f_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair__create;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair)) __PYX_ERR(0, 493, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ScoreCoarseGlycanMatchPair, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  #endif
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.fragmentation_strategy.glycan"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragmentation_strategy.glycan", "GlycanCompositionFragment", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment.__pyx_base = *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment.__pyx_base.copy = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_copy;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment.__pyx_base._create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment *(*)(double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int))__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create__pyx_wrap_1;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment._create = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *(*)(double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int))__pyx_f_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment__create;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment)) __PYX_ERR(0, 598, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment->tp_base = __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragment;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IndexGlycanCompositionFragment, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment)) __PYX_ERR(0, 655, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_ComplementFragment;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ComplementFragment, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution._create = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution *(*)(void))__pyx_f_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution__create;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution)) __PYX_ERR(0, 669, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PartialGlycanSolution, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex = &__pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex._intern = (struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__intern;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex.query_width = (long (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, double, double, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_query_width;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex.bin_index = (long (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, double, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_bin_index;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex._get_fragments = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__get_fragments;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex._match_fragments = (int (*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, double, PyObject *))__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex__match_fragments;
  __pyx_vtable_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex.match = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match *__pyx_optional_args))__pyx_f_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_match;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex)) __PYX_ERR(0, 708, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex_spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex = &__pyx_type_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_vtabptr_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanFragmentIndex, (PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter___spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__)) __PYX_ERR(0, 253, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter___spec, __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 253, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__ = &__pyx_type_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 253, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_6tandem_11core_search___pyx_scope_struct____iter__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_3_0_11); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 271, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_11); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 316, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_11); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 320, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_3_0_11); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 359, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 848, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 850, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 852, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 854, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 856, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 858, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 860, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 862, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 864, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 866, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_0_11(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_3_0_11); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType_3_0_11(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(7, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(7, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType_3_0_11(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(8, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(8, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(9, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(9, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(9, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(9, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(9, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(9, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(10, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(10, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(11, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(11, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(12, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(12, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(12, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(12, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(12, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(12, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(13, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(13, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(13, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(13, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(13, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(13, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(13, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(13, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(13, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(13, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(13, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(13, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(13, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(13, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(13, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(13, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(13, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(13, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(13, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(13, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(13, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(13, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(13, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(13, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(14, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(14, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(14, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(14, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(14, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(14, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(14, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(14, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "IonMobilityDeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak) __PYX_ERR(14, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak)) __PYX_ERR(14, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(14, 71, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(14, 71, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(14, 75, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(14, 75, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(14, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(14, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = __Pyx_ImportType_3_0_11(__pyx_t_1, "ms_deisotope._c.peak_set", "_CPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet) __PYX_ERR(14, 161, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet)) __PYX_ERR(14, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glypy.composition.ccomposition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5glypy_11composition_12ccomposition_CComposition = __Pyx_ImportType_3_0_11(__pyx_t_1, "glypy.composition.ccomposition", "CComposition", sizeof(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5glypy_11composition_12ccomposition_CComposition) __PYX_ERR(15, 2, __pyx_L1_error)
  __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition = (struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition*)__Pyx_GetVtable(__pyx_ptype_5glypy_11composition_12ccomposition_CComposition); if (unlikely(!__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition)) __PYX_ERR(15, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycresoft._c.chromatogram_tree.mass_shift"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycresoft._c.chromatogram_tree.mass_shift", "MassShiftBase", sizeof(struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase) __PYX_ERR(16, 9, __pyx_L1_error)
  __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycresoft._c.chromatogram_tree.mass_shift", "MassShift", sizeof(struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShift) __PYX_ERR(16, 20, __pyx_L1_error)
  __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycresoft._c.chromatogram_tree.mass_shift", "CompoundMassShift", sizeof(struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_CompoundMassShift) __PYX_ERR(16, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glypy._c.structure.glycan_composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glypy._c.structure.glycan_composition", "_CompositionBase", sizeof(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase) __PYX_ERR(17, 3, __pyx_L1_error)
  __pyx_vtabptr_5glypy_2_c_9structure_18glycan_composition__CompositionBase = (struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase*)__Pyx_GetVtable(__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase); if (unlikely(!__pyx_vtabptr_5glypy_2_c_9structure_18glycan_composition__CompositionBase)) __PYX_ERR(17, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.count_table"); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.count_table", "CountTableIterator", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator) __PYX_ERR(18, 52, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator)) __PYX_ERR(18, 52, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.count_table", "CountTable", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable) __PYX_ERR(18, 69, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTable = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTable)) __PYX_ERR(18, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.base", "PeptideSequenceBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase) __PYX_ERR(19, 5, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase)) __PYX_ERR(19, 5, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.base", "TerminalGroup", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup) __PYX_ERR(19, 21, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup)) __PYX_ERR(19, 21, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.base", "AminoAcidResidueBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase) __PYX_ERR(19, 35, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase)) __PYX_ERR(19, 35, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.base", "ModificationBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase) __PYX_ERR(19, 50, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase)) __PYX_ERR(19, 50, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.base", "SequencePosition", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition) __PYX_ERR(19, 69, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition)) __PYX_ERR(19, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.glycan"); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.glycan", "GlycanCompositionProxy", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy) __PYX_ERR(20, 12, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy)) __PYX_ERR(20, 12, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.glycan", "GlycanCompositionWithOffsetProxy", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy) __PYX_ERR(20, 29, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy)) __PYX_ERR(20, 29, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.glycan", "GlycosylationManager", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager) __PYX_ERR(20, 34, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager)) __PYX_ERR(20, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.fragment"); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "ChemicalShiftBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase) __PYX_ERR(21, 6, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase)) __PYX_ERR(21, 6, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "IonSeriesBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase) __PYX_ERR(21, 15, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "FragmentBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase) __PYX_ERR(21, 26, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase)) __PYX_ERR(21, 26, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "SimpleFragment", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment) __PYX_ERR(21, 63, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment)) __PYX_ERR(21, 63, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "StubFragment", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment) __PYX_ERR(21, 72, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment)) __PYX_ERR(21, 72, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "PeptideFragment", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment) __PYX_ERR(21, 97, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment)) __PYX_ERR(21, 97, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragment", "_NameTree", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree) __PYX_ERR(21, 125, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree)) __PYX_ERR(21, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.sequence_methods"); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.sequence_methods", "_PeptideSequenceCore", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore) __PYX_ERR(22, 15, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_16sequence_methods__PeptideSequenceCore)) __PYX_ERR(22, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.fragmentation_strategy.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragmentation_strategy.base", "FragmentationStrategyBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase) __PYX_ERR(23, 28, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_FragmentationStrategyBase)) __PYX_ERR(23, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.fragmentation_strategy.glycan"); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragmentation_strategy.glycan", "GlycanCompositionFragmentStrategyBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase) __PYX_ERR(24, 31, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_GlycanCompositionFragmentStrategyBase)) __PYX_ERR(24, 31, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy = __Pyx_ImportType_3_0_11(__pyx_t_1, "glycopeptidepy._c.structure.fragmentation_strategy.glycan", "StubGlycopeptideStrategy", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy) __PYX_ERR(24, 42, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_6glycan_StubGlycopeptideStrategy)) __PYX_ERR(24, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glypy.composition.ccomposition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_atom", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__atom, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_formula", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__formula, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_isotope_string", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__isotope_string, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "isotope_pattern", (void **)&__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "formula_pattern", (void **)&__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycresoft._c.chromatogram_tree.mass_shift"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "mass_shift_index", (void **)&__pyx_vp_10glycresoft_2_c_17chromatogram_tree_10mass_shift_mass_shift_index, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "Unmodified", (void **)&__pyx_vp_10glycresoft_2_c_17chromatogram_tree_10mass_shift_Unmodified, "struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.fragmentation_strategy.base"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_n_glycosylation", (void **)&__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__n_glycosylation, "struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_o_glycosylation", (void **)&__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__o_glycosylation, "struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_gag_linker_glycosylation", (void **)&__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__gag_linker_glycosylation, "struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_modification_hexnac", (void **)&__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_hexnac, "struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "_modification_xylose", (void **)&__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base__modification_xylose, "struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "glycosylation_type_to_core", (void **)&__pyx_vp_14glycopeptidepy_2_c_9structure_22fragmentation_strategy_4base_glycosylation_type_to_core, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_core_search(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_core_search},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "core_search",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcore_search(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcore_search(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_core_search(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_core_search(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_core_search(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'core_search' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("core_search", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "core_search" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_core_search(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_glycresoft___c__tandem__core_search) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "glycresoft._c.tandem.core_search")) {
      if (unlikely((PyDict_SetItemString(modules, "glycresoft._c.tandem.core_search", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "glycresoft/_c/tandem/core_search.pyx":2
 * # cython: embedsignature=True
 * from collections import defaultdict             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_defaultdict);
  __Pyx_GIVEREF(__pyx_n_s_defaultdict);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_defaultdict)) __PYX_ERR(0, 2, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultdict, __pyx_t_2) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":16
 * from libc cimport math
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * from numpy cimport npy_uint16 as uint16_t, npy_int32 as int32_t
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":23
 * from glycresoft._c.chromatogram_tree.mass_shift cimport MassShiftBase
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * import warnings
 */
  __pyx_t_4 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 23, __pyx_L1_error)

  /* "glycresoft/_c/tandem/core_search.pyx":25
 * np.import_array()
 * 
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * from glypy.composition.ccomposition cimport CComposition
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_warnings, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_3) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":30
 * from glycopeptidepy._c.structure.fragmentation_strategy.glycan cimport StubGlycopeptideStrategy, GlycanCompositionFragment
 * from glycopeptidepy._c.count_table cimport CountTable
 * from glycresoft.chromatogram_tree import Unmodified as _Unmodified             # <<<<<<<<<<<<<<
 * from glycresoft.serialize import GlycanTypes
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Unmodified);
  __Pyx_GIVEREF(__pyx_n_s_Unmodified);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Unmodified)) __PYX_ERR(0, 30, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_glycresoft_chromatogram_tree, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Unmodified); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Unmodified_2, __pyx_t_3) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":31
 * from glycopeptidepy._c.count_table cimport CountTable
 * from glycresoft.chromatogram_tree import Unmodified as _Unmodified
 * from glycresoft.serialize import GlycanTypes             # <<<<<<<<<<<<<<
 * 
 * from glypy.structure.glycan_composition import HashableGlycanComposition as _HashableGlycanComposition
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_GlycanTypes);
  __Pyx_GIVEREF(__pyx_n_s_GlycanTypes);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_GlycanTypes)) __PYX_ERR(0, 31, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_glycresoft_serialize, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_GlycanTypes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GlycanTypes, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":33
 * from glycresoft.serialize import GlycanTypes
 * 
 * from glypy.structure.glycan_composition import HashableGlycanComposition as _HashableGlycanComposition             # <<<<<<<<<<<<<<
 * 
 * cdef object HashableGlycanComposition = _HashableGlycanComposition
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_HashableGlycanComposition);
  __Pyx_GIVEREF(__pyx_n_s_HashableGlycanComposition);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_HashableGlycanComposition)) __PYX_ERR(0, 33, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_glypy_structure_glycan_compositi, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_HashableGlycanComposition); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HashableGlycanComposition_2, __pyx_t_3) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":35
 * from glypy.structure.glycan_composition import HashableGlycanComposition as _HashableGlycanComposition
 * 
 * cdef object HashableGlycanComposition = _HashableGlycanComposition             # <<<<<<<<<<<<<<
 * cdef MassShiftBase Unmodified = _Unmodified
 * cdef object GlycanTypes_n_glycan, GlycanTypes_o_glycan, GlycanTypes_gag_linker
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HashableGlycanComposition_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_HashableGlycanComposition);
  __Pyx_DECREF_SET(__pyx_v_10glycresoft_2_c_6tandem_11core_search_HashableGlycanComposition, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":36
 * 
 * cdef object HashableGlycanComposition = _HashableGlycanComposition
 * cdef MassShiftBase Unmodified = _Unmodified             # <<<<<<<<<<<<<<
 * cdef object GlycanTypes_n_glycan, GlycanTypes_o_glycan, GlycanTypes_gag_linker
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Unmodified_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase))))) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_XGOTREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __Pyx_DECREF_SET(__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified, ((struct __pyx_obj_10glycresoft_2_c_17chromatogram_tree_10mass_shift_MassShiftBase *)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":40
 * 
 * 
 * GlycanTypes_n_glycan = GlycanTypes.n_glycan             # <<<<<<<<<<<<<<
 * GlycanTypes_o_glycan = GlycanTypes.o_glycan
 * GlycanTypes_gag_linker = GlycanTypes.gag_linker
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_GlycanTypes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_n_glycan); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan);
  __Pyx_DECREF_SET(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_n_glycan, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":41
 * 
 * GlycanTypes_n_glycan = GlycanTypes.n_glycan
 * GlycanTypes_o_glycan = GlycanTypes.o_glycan             # <<<<<<<<<<<<<<
 * GlycanTypes_gag_linker = GlycanTypes.gag_linker
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_GlycanTypes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_o_glycan); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan);
  __Pyx_DECREF_SET(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_o_glycan, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":42
 * GlycanTypes_n_glycan = GlycanTypes.n_glycan
 * GlycanTypes_o_glycan = GlycanTypes.o_glycan
 * GlycanTypes_gag_linker = GlycanTypes.gag_linker             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_GlycanTypes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_gag_linker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker);
  __Pyx_DECREF_SET(__pyx_v_10glycresoft_2_c_6tandem_11core_search_GlycanTypes_gag_linker, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":59
 *         public dict fragment_set_properties
 * 
 *     cpdef bint is_n_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_1is_n_glycan, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_is_n, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_is_n_glycan, __pyx_t_3) < 0) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":62
 *         return GlycanTypes_n_glycan in self.glycan_types
 * 
 *     cpdef bint is_o_glycan(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_3is_o_glycan, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_is_o, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_is_o_glycan, __pyx_t_3) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":65
 *         return GlycanTypes_o_glycan in self.glycan_types
 * 
 *     cpdef bint is_gag_linker(self):             # <<<<<<<<<<<<<<
 *         return GlycanTypes_gag_linker in self.glycan_types
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_5is_gag_linker, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_is_g, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_is_gag_linker, __pyx_t_3) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":69
 * 
 *     @cython.cdivision(True)
 *     cpdef double normalized_size(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double k
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_7normalized_size, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_norm, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_normalized_size, __pyx_t_3) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":84
 *         return d
 * 
 *     cpdef list get_n_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_9get_n_glycan_fragments, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_get, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_get_n_glycan_fragments, __pyx_t_3) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":108
 *             return result
 * 
 *     cpdef list get_o_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_11get_o_glycan_fragments, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_get_2, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_get_o_glycan_fragments, __pyx_t_3) < 0) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":128
 *             return result
 * 
 *     cpdef list get_gag_linker_glycan_fragments(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_13get_gag_linker_glycan_fragments, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_get_3, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_get_gag_linker_glycan_fragments, __pyx_t_3) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":148
 *             return result
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         self._fragment_cache.clear()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_15clear, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase_clea, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_clear, __pyx_t_3) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_17__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase___re, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCombinationRecordBase, (type(self), 0x564dc94, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCombinationRecordBase__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27GlycanCombinationRecordBase_19__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCombinationRecordBase___se, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCombinationRecordBase);

  /* "glycresoft/_c/tandem/core_search.pyx":163
 *         self.is_core = is_core
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.is_core)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30CoarseStubGlycopeptideFragment_3__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoarseStubGlycopeptideFragment_2, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment, __pyx_n_s_reduce, __pyx_t_3) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideFragment);

  /* "glycresoft/_c/tandem/core_search.pyx":216
 *         self.peaks_matched = peaks_matched
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_27CoarseStubGlycopeptideMatch_3__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoarseStubGlycopeptideMatch___re, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch, __pyx_n_s_reduce, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseStubGlycopeptideMatch);

  /* "glycresoft/_c/tandem/core_search.pyx":260
 *         yield self.core_theoretical
 * 
 *     cpdef double estimate_peptide_mass(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double weighted_mass_acc
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_6estimate_peptide_mass, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoarseGlycanMatch_estimate_pepti, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, __pyx_n_s_estimate_peptide_mass, __pyx_t_3) < 0) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_10__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoarseGlycanMatch___reduce_cytho, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoarseGlycanMatch, (type(self), 0xf0ba66f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoarseGlycanMatch__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17CoarseGlycanMatch_12__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoarseGlycanMatch___setstate_cyt, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_CoarseGlycanMatch);

  /* "glycresoft/_c/tandem/core_search.pyx":305
 * 
 *     @cython.boundscheck(False)
 *     cpdef CoarseGlycanMatch _match_fragments(self, object scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list fragment_matches
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_3_match_fragments, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase__match_fr, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__36);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_match_fragments, __pyx_t_3) < 0) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "glycresoft/_c/tandem/core_search.pyx":356
 * 
 *     @cython.cdivision(True)
 *     cpdef double _calculate_score(self, CoarseGlycanMatch glycan_match):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j, n, m
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_5_calculate_score, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase__calculat, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_calculate_score, __pyx_t_3) < 0) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "glycresoft/_c/tandem/core_search.pyx":381
 *         return score
 * 
 *     cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_7_n_glycan_match_stubs, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase__n_glycan, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__41);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_n_glycan_match_stubs, __pyx_t_3) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "glycresoft/_c/tandem/core_search.pyx":387
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                                   GlycanCombinationRecordBase glycan_combination,
 *                                                   double mass_shift_tandem_mass=0.0):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_9_o_glycan_match_stubs, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase__o_glycan, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__43);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_o_glycan_match_stubs, __pyx_t_3) < 0) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "glycresoft/_c/tandem/core_search.pyx":393
 *         return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)
 * 
 *     cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,             # <<<<<<<<<<<<<<
 *                                              GlycanCombinationRecordBase glycan_combination,
 *                                              double mass_shift_tandem_mass=0.0):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_11_gag_match_stubs, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase__gag_matc, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__45);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_gag_match_stubs, __pyx_t_3) < 0) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase___reduce, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCoarseScorerBase, (type(self), 0x1ca83c4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCoarseScorerBase__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_22GlycanCoarseScorerBase_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCoarseScorerBase___setstat, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanCoarseScorerBase);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanMatchResult___reduce_cytho, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanMatchResult, (type(self), 0x887dc81, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanMatchResult__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17GlycanMatchResult_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanMatchResult___setstate_cyt, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanMatchResult);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ScoreCoarseGlycanMatchPair___red, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ScoreCoarseGlycanMatchPair, (type(self), 0x53f74fc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_26ScoreCoarseGlycanMatchPair_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ScoreCoarseGlycanMatchPair___set, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ScoreCoarseGlycanMatchPair);

  /* "glycresoft/_c/tandem/core_search.pyx":524
 * 
 * 
 * cpdef double score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<GlycanMatchResult>x).score
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_1score_getter, 0, __pyx_n_s_score_getter, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_score_getter, __pyx_t_3) < 0) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":530
 * @cython.boundscheck(False)
 * @cython.binding(True)
 * cpdef GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list output, glycan_combination_db
 */
  __Pyx_INCREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __pyx_k__4 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified;
  __Pyx_GIVEREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __Pyx_GIVEREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified))) __PYX_ERR(0, 530, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, Py_None)) __PYX_ERR(0, 530, __pyx_L1_error);
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_3GlycanFilteringPeptideMassEstimator_match, 0, __pyx_n_s_GlycanFilteringPeptideMassEstima, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GlycanFilteringPeptideMassEstima, __pyx_t_2) < 0) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":626
 *         return self
 * 
 *     cpdef GlycanCompositionFragment copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IndexGlycanCompositionFragment inst
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IndexGlycanCompositionFragment_c, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, __pyx_n_s_copy, __pyx_t_2) < 0) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);

  /* "glycresoft/_c/tandem/core_search.pyx":638
 *         return inst
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_5__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IndexGlycanCompositionFragment_2, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, __pyx_n_s_reduce, __pyx_t_2) < 0) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);

  /* "glycresoft/_c/tandem/core_search.pyx":641
 *         return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {
 *             "index": self.index,
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_7__getstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IndexGlycanCompositionFragment_3, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, __pyx_n_s_getstate, __pyx_t_2) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);

  /* "glycresoft/_c/tandem/core_search.pyx":648
 *         }
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.index = state['index']
 *         self.name = state['name']
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_30IndexGlycanCompositionFragment_9__setstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IndexGlycanCompositionFragment_4, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment, __pyx_n_s_setstate, __pyx_t_2) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_IndexGlycanCompositionFragment);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ComplementFragment___reduce_cyth, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ComplementFragment, (type(self), 0x7fff502, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ComplementFragment__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_18ComplementFragment_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ComplementFragment___setstate_cy, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_ComplementFragment);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PartialGlycanSolution___reduce_c, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PartialGlycanSolution, (type(self), 0x57f5605, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PartialGlycanSolution__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_21PartialGlycanSolution_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PartialGlycanSolution___setstate, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_PartialGlycanSolution);

  /* "glycresoft/_c/tandem/core_search.pyx":765
 *         self.resolution = resolution
 * 
 *     cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):             # <<<<<<<<<<<<<<
 *         key = str(HashableGlycanComposition(fragment.key))
 *         try:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_3_intern, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex__intern, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_intern, __pyx_t_2) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "glycresoft/_c/tandem/core_search.pyx":783
 *             return fragment
 * 
 *     cpdef long query_width(self, double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_5query_width, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex_query_width, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_query_width, __pyx_t_2) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "glycresoft/_c/tandem/core_search.pyx":786
 *         return <long>(mass * error_tolerance * self.resolution) + 1
 * 
 *     cpdef long bin_index(self, double mass):             # <<<<<<<<<<<<<<
 *         return <long>(mass * self.resolution)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_7bin_index, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex_bin_index, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_bin_index, __pyx_t_2) < 0) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "glycresoft/_c/tandem/core_search.pyx":789
 *         return <long>(mass * self.resolution)
 * 
 *     def build(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_9build, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex_build, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_build, __pyx_t_2) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "glycresoft/_c/tandem/core_search.pyx":845
 *         return j
 * 
 *     cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):             # <<<<<<<<<<<<<<
 *         if glycosylation_type == GlycanTypes_n_glycan:
 *             return gcr.get_n_glycan_fragments()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_11_get_fragments, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex__get_fragmen, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_get_fragments, __pyx_t_2) < 0) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "glycresoft/_c/tandem/core_search.pyx":917
 * 
 *     @cython.cdivision
 *     cpdef list match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double precursor_mass, d
 */
  __Pyx_INCREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __pyx_k__5 = __pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified;
  __Pyx_GIVEREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_float_1eneg_5);
  __Pyx_GIVEREF(__pyx_float_1eneg_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_float_1eneg_5)) __PYX_ERR(0, 917, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  __Pyx_GIVEREF((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_10glycresoft_2_c_6tandem_11core_search_Unmodified))) __PYX_ERR(0, 917, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(0, 917, __pyx_L1_error);
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_13match, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex_match, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_match, __pyx_t_3) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex___reduce_cyt, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanFragmentIndex, (type(self), 0x0fecca7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanFragmentIndex__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19GlycanFragmentIndex_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanFragmentIndex___setstate_c, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_6tandem_11core_search_GlycanFragmentIndex);

  /* "glycresoft/_c/tandem/core_search.pyx":990
 * 
 * 
 * cpdef double partial_score_getter(object x):             # <<<<<<<<<<<<<<
 *     return (<PartialGlycanSolution>x).score
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_5partial_score_getter, 0, __pyx_n_s_partial_score_getter, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_partial_score_getter, __pyx_t_3) < 0) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCombinationRecordBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_7__pyx_unpickle_GlycanCombinationRecordBase, 0, __pyx_n_s_pyx_unpickle_GlycanCombination, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycanCombination, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCombinationRecordBase__set_state(<GlycanCombinationRecordBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCombinationRecordBase__set_state(GlycanCombinationRecordBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fragment_cache = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result.composition = __pyx_state[2]; __pyx_result.count = __pyx_state[3]; __pyx_result.dehydrated_mass = __pyx_state[4]; __pyx_result.fragment_set_properties = __pyx_state[5]; __pyx_result.glycan_types = __pyx_state[6]; __pyx_result.id = __pyx_state[7]; __pyx_result.internal_size_approximation = __pyx_state[8]; __pyx_result.side_group_count = __pyx_state[9]; __pyx_result.size = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_9__pyx_unpickle_CoarseGlycanMatch, 0, __pyx_n_s_pyx_unpickle_CoarseGlycanMatch, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CoarseGlycanMatch, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCoarseScorerBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_11__pyx_unpickle_GlycanCoarseScorerBase, 0, __pyx_n_s_pyx_unpickle_GlycanCoarseScore, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycanCoarseScore, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCoarseScorerBase__set_state(<GlycanCoarseScorerBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCoarseScorerBase__set_state(GlycanCoarseScorerBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_weight = __pyx_state[0]; __pyx_result.fragment_weight = __pyx_state[1]; __pyx_result.product_error_tolerance = __pyx_state[2]; __pyx_result.recalibrate_peptide_mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_13__pyx_unpickle_GlycanMatchResult, 0, __pyx_n_s_pyx_unpickle_GlycanMatchResult, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycanMatchResult, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ScoreCoarseGlycanMatchPair(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_15__pyx_unpickle_ScoreCoarseGlycanMatchPair, 0, __pyx_n_s_pyx_unpickle_ScoreCoarseGlycan, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ScoreCoarseGlycan, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(<ScoreCoarseGlycanMatchPair> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ScoreCoarseGlycanMatchPair__set_state(ScoreCoarseGlycanMatchPair __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.match = __pyx_state[0]; __pyx_result.score = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_17__pyx_unpickle_ComplementFragment, 0, __pyx_n_s_pyx_unpickle_ComplementFragmen, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ComplementFragmen, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PartialGlycanSolution(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_19__pyx_unpickle_PartialGlycanSolution, 0, __pyx_n_s_pyx_unpickle_PartialGlycanSolu, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PartialGlycanSolu, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PartialGlycanSolution__set_state(<PartialGlycanSolution> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PartialGlycanSolution__set_state(PartialGlycanSolution __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.core_matches = __pyx_state[0]; __pyx_result.fragment_matches = __pyx_state[1]; __pyx_result.glycan_index = __pyx_state[2]; __pyx_result.peptide_mass = __pyx_state[3]; __pyx_result.score = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_6tandem_11core_search_21__pyx_unpickle_GlycanFragmentIndex, 0, __pyx_n_s_pyx_unpickle_GlycanFragmentInd, NULL, __pyx_n_s_glycresoft__c_tandem_core_search, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycanFragmentInd, __pyx_t_3) < 0) __PYX_ERR(3, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/tandem/core_search.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * from collections import defaultdict
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init glycresoft._c.tandem.core_search", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init glycresoft._c.tandem.core_search");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* PyIntFromDouble */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value) {
    if (value >= (double)LONG_MIN && value <= (double)LONG_MAX) {
        return PyInt_FromLong((long)value);
    }
    return PyLong_FromDouble(value);
}
#endif

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* SoftComplexToDouble */
static double __Pyx_SoftComplexToDouble(__pyx_t_double_complex value, int have_gil) {
    if (unlikely(__Pyx_CIMAG(value))) {
        PyGILState_STATE gilstate;
        if (!have_gil)
            gilstate = PyGILState_Ensure();
        PyErr_SetString(PyExc_TypeError,
            "Cannot convert 'complex' with non-zero imaginary component to 'double' "
            "(this most likely comes from the '**' operator; "
            "use 'cython.cpow(True)' to return 'nan' instead of a "
            "complex number).");
        if (!have_gil)
            PyGILState_Release(gilstate);
        return -1.;
    }
    return __Pyx_CREAL(value);
}

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a == b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* pybytes_as_double */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
#if CYTHON_ASSUME_SAFE_MACROS
        double value = PyFloat_AS_DOUBLE(float_value);
#else
        double value = PyFloat_AsDouble(float_value);
#endif
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__7);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_11
#define __PYX_HAVE_RT_ImportType_3_0_11
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_11 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_11 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__15;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return ::std::complex< long double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return x + y*(__pyx_t_long_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      __pyx_t_long_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsl(b.real) >= fabsl(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                long double r = b.imag / b.real;
                long double s = (long double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_long_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            long double r = b.real / b.imag;
            long double s = (long double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_long_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            long double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_long_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtl(z.real*z.real + z.imag*z.imag);
          #else
            return hypotl(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
            __pyx_t_long_double_complex z;
            long double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    long double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_long__double(a, a);
                    case 3:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, a);
                    case 4:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powl(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2l(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_long__double(a);
                theta = atan2l(a.imag, a.real);
            }
            lnr = logl(r);
            z_r = expl(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosl(z_theta);
            z.imag = z_r * sinl(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint16(npy_uint16 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint16 neg_one = (npy_uint16) -1, const_zero = (npy_uint16) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint16) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint16) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint16) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint16) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint16) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint16),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(npy_uint16));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE npy_uint16 __Pyx_PyInt_As_npy_uint16(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint16 neg_one = (npy_uint16) -1, const_zero = (npy_uint16) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(npy_uint16) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(npy_uint16, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint16) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        npy_uint16 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint16) -1;
        val = __Pyx_PyInt_As_npy_uint16(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(npy_uint16, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(npy_uint16) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) >= 2 * PyLong_SHIFT)) {
                            return (npy_uint16) (((((npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(npy_uint16) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) >= 3 * PyLong_SHIFT)) {
                            return (npy_uint16) (((((((npy_uint16)digits[2]) << PyLong_SHIFT) | (npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(npy_uint16) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) >= 4 * PyLong_SHIFT)) {
                            return (npy_uint16) (((((((((npy_uint16)digits[3]) << PyLong_SHIFT) | (npy_uint16)digits[2]) << PyLong_SHIFT) | (npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (npy_uint16) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(npy_uint16) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_uint16, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(npy_uint16) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_uint16, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(npy_uint16, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(npy_uint16) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) - 1 > 2 * PyLong_SHIFT)) {
                            return (npy_uint16) (((npy_uint16)-1)*(((((npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(npy_uint16) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) - 1 > 2 * PyLong_SHIFT)) {
                            return (npy_uint16) ((((((npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(npy_uint16) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) - 1 > 3 * PyLong_SHIFT)) {
                            return (npy_uint16) (((npy_uint16)-1)*(((((((npy_uint16)digits[2]) << PyLong_SHIFT) | (npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(npy_uint16) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) - 1 > 3 * PyLong_SHIFT)) {
                            return (npy_uint16) ((((((((npy_uint16)digits[2]) << PyLong_SHIFT) | (npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(npy_uint16) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) - 1 > 4 * PyLong_SHIFT)) {
                            return (npy_uint16) (((npy_uint16)-1)*(((((((((npy_uint16)digits[3]) << PyLong_SHIFT) | (npy_uint16)digits[2]) << PyLong_SHIFT) | (npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(npy_uint16) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_uint16, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_uint16) - 1 > 4 * PyLong_SHIFT)) {
                            return (npy_uint16) ((((((((((npy_uint16)digits[3]) << PyLong_SHIFT) | (npy_uint16)digits[2]) << PyLong_SHIFT) | (npy_uint16)digits[1]) << PyLong_SHIFT) | (npy_uint16)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(npy_uint16) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_uint16, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(npy_uint16) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_uint16, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        npy_uint16 val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (npy_uint16) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (npy_uint16) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (npy_uint16) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (npy_uint16) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(npy_uint16) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((npy_uint16) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(npy_uint16) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((npy_uint16) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((npy_uint16) 1) << (sizeof(npy_uint16) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (npy_uint16) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint16");
    return (npy_uint16) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint16");
    return (npy_uint16) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int32 neg_one = (npy_int32) -1, const_zero = (npy_int32) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int32) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int32) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int32) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int32),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(npy_int32));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int32 neg_one = (npy_int32) -1, const_zero = (npy_int32) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(npy_int32) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(npy_int32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int32) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        npy_int32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int32) -1;
        val = __Pyx_PyInt_As_npy_int32(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(npy_int32, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(npy_int32) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) >= 2 * PyLong_SHIFT)) {
                            return (npy_int32) (((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(npy_int32) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) >= 3 * PyLong_SHIFT)) {
                            return (npy_int32) (((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(npy_int32) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) >= 4 * PyLong_SHIFT)) {
                            return (npy_int32) (((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (npy_int32) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(npy_int32) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(npy_int32, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(npy_int32) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT)) {
                            return (npy_int32) (((npy_int32)-1)*(((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(npy_int32) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT)) {
                            return (npy_int32) ((((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT)) {
                            return (npy_int32) (((npy_int32)-1)*(((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(npy_int32) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT)) {
                            return (npy_int32) ((((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT)) {
                            return (npy_int32) (((npy_int32)-1)*(((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(npy_int32) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT)) {
                            return (npy_int32) ((((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(npy_int32) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_int32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(npy_int32) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(npy_int32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        npy_int32 val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (npy_int32) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (npy_int32) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (npy_int32) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (npy_int32) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(npy_int32) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((npy_int32) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(npy_int32) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((npy_int32) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((npy_int32) 1) << (sizeof(npy_int32) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (npy_int32) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int32");
    return (npy_int32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int32");
    return (npy_int32) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__89);
    }
    return name;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_3_0_11
#define __PYX_HAVE_RT_ImportVoidPtr_3_0_11
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
