/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "/private/var/folders/m_/cksx93ys47x4621g0zbw_m4m0000gn/T/pip-build-env-_ainf6v_/overlay/lib/python3.9/site-packages/numpy/_core/include"
        ],
        "name": "glycresoft._c.composition_network.graph",
        "sources": [
            "src/glycresoft/_c/composition_network/graph.pyx"
        ]
    },
    "module_name": "glycresoft._c.composition_network.graph"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__glycresoft___c__composition_network__graph
#define __PYX_HAVE_API__glycresoft___c__composition_network__graph
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/glycresoft/_c/composition_network/graph.pyx",
  "src/glycresoft/_c/composition_network/graph.pxd",
  "<stringsource>",
  "type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector;

/* "glycresoft/_c/composition_network/graph.pxd":53
 * 
 * 
 * cdef struct glycan_composition_vector:             # <<<<<<<<<<<<<<
 *     int* counts
 *     size_t size
 */
struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector {
  int *counts;
  size_t size;
};

/* "glycresoft/_c/composition_network/graph.pxd":1
 * cdef class CompositionGraphNode(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object composition
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode {
  PyObject_HEAD
  PyObject *composition;
  Py_ssize_t index;
  PyObject *_str;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *edges;
  Py_hash_t _hash;
  double _score;
  double internal_score;
  int marked;
};


/* "glycresoft/_c/composition_network/graph.pxd":12
 *         public bint marked
 * 
 * cdef class EdgeSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict store
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_vtab;
  PyObject *store;
};


/* "glycresoft/_c/composition_network/graph.pxd":20
 * 
 * 
 * cdef class CompositionGraphEdge(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CompositionGraphNode node1
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_vtab;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *node1;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *node2;
  long order;
  double weight;
  Py_hash_t _hash;
  PyObject *_str;
};


/* "glycresoft/_c/composition_network/graph.pxd":38
 * 
 * 
 * cdef class CompositionGraphBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list nodes
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_vtab;
  PyObject *nodes;
  PyObject *node_map;
  PyObject *_composition_normalizer;
  PyObject *distance_fn;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *edges;
  PyObject *neighborhoods;
};


/* "glycresoft/_c/composition_network/graph.pxd":68
 * 
 * 
 * cdef class GlycanCompositionVectorContext:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list components
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_vtab;
  PyObject *components;
  size_t component_count;
};


/* "glycresoft/_c/composition_network/graph.pxd":78
 * 
 * 
 * cdef class GlycanCompositionVector:             # <<<<<<<<<<<<<<
 *     cdef:
 *         glycan_composition_vector* ptr
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_vtab;
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *ptr;
};


/* "glycresoft/_c/composition_network/graph.pxd":93
 * 
 * 
 * cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict encode_cache
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext __pyx_base;
  PyObject *encode_cache;
  PyObject *decode_cache;
};


/* "glycresoft/_c/composition_network/graph.pyx":263
 * 
 * 
 * cdef class DijkstraPathFinder(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder {
  PyObject_HEAD
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_vtab;
  PyObject *graph;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *start;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *end;
  PyObject *distance;
  PyObject *unvisited_finite_distance;
  double limit;
};



/* "glycresoft/_c/composition_network/graph.pyx":88
 * 
 * 
 * cdef class EdgeSet(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, store=None):
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet {
  int (*add_if_shorter)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *, int __pyx_skip_dispatch);
  PyObject *(*get_edges)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_EdgeSet;


/* "glycresoft/_c/composition_network/graph.pyx":143
 * 
 * @cython.freelist(15000)
 * cdef class CompositionGraphEdge(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, node1, node2, order, weight=1.0):
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *(*_create)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, long, double);
  void (*_init)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *);
  PyObject *(*copy_for)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;


/* "glycresoft/_c/composition_network/graph.pyx":232
 * 
 * 
 * cdef class CompositionGraphBase(object):             # <<<<<<<<<<<<<<
 *     cdef list get_edges(self):
 *         return self.edges.get_edges()
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase {
  PyObject *(*get_edges)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *);
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *(*copy)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;


/* "glycresoft/_c/composition_network/graph.pyx":483
 * 
 * 
 * cdef class GlycanCompositionVectorContext:             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, components):
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *(*encode_raw)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, PyObject *);
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*encode)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*decode)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;


/* "glycresoft/_c/composition_network/graph.pyx":535
 * @cython.final
 * @cython.freelist(2000)
 * cdef class GlycanCompositionVector:             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*_create)(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *);
  double (*distance)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*difference)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*addition)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*clone)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *);
static double __pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_distance(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_difference(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_addition(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch);


/* "glycresoft/_c/composition_network/graph.pyx":670
 * 
 * 
 * cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, components):
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext {
  struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext __pyx_base;
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;


/* "glycresoft/_c/composition_network/graph.pyx":263
 * 
 * 
 * cdef class DijkstraPathFinder(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder {
  PyObject *(*_build_initial_key_set)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, int __pyx_skip_dispatch);
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *(*find_smallest_unvisited)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*find_path)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyStr_Str.proto */
static CYTHON_INLINE PyObject* __Pyx_PyStr_Str(PyObject *obj);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* py_dict_values.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_11
#define __PYX_HAVE_RT_ImportType_proto_3_0_11
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_11 {
   __Pyx_ImportType_CheckSize_Error_3_0_11 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_11 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_11 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pystring_as_double.proto */
static CYTHON_INLINE double __Pyx_PyString_AsDouble(PyObject *obj) {
    #if PY_MAJOR_VERSION >= 3
    (void)__Pyx_PyBytes_AsDouble;
    return __Pyx_PyUnicode_AsDouble(obj);
    #else
    (void)__Pyx_PyUnicode_AsDouble;
    return __Pyx_PyBytes_AsDouble(obj);
    #endif
}

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_add_if_shorter(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_get_edges(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__create(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2, long __pyx_v_order, double __pyx_v_weight); /* proto*/
static void __pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__init(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_copy_for(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_get_edges(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder__build_initial_key_set(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_smallest_unvisited(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_unvisited, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_path(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode_raw(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_glycan_composition); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_glycan_composition, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_ptr); /* proto*/
static double __pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_distance(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_difference(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_addition(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_gc, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdlib" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.float" */

/* Module declarations from "cpython.int" */

/* Module declarations from "glycresoft._c.composition_network.graph" */
static double __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_reindex_graph(PyObject *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(size_t, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *); /*proto*/
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_destroy_glycan_composition_vector(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *); /*proto*/
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_copy_glycan_composition_vector(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *); /*proto*/
static double __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_distance(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *); /*proto*/
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_difference(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *); /*proto*/
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_addition(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph__reconstruct_GlycanCompositionVector(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphNode__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_EdgeSet__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphBase__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_DijkstraPathFinder__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "glycresoft._c.composition_network.graph"
extern int __pyx_module_is_main_glycresoft___c__composition_network__graph;
int __pyx_module_is_main_glycresoft___c__composition_network__graph = 0;

/* Implementation of "glycresoft._c.composition_network.graph" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
static const char __pyx_k__3[] = ":\n";
static const char __pyx_k__4[] = "->";
static const char __pyx_k__7[] = "(";
static const char __pyx_k__8[] = ")";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k__12[] = ".";
static const char __pyx_k__73[] = "?";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_dup[] = "dup";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_str[] = "_str";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_edge[] = "edge";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_gcvec[] = "gcvec";
static const char __pyx_k_graph[] = "graph";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_limit[] = "limit";
static const char __pyx_k_node1[] = "node1";
static const char __pyx_k_node2[] = "node2";
static const char __pyx_k_nodes[] = "nodes";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_ptemp[] = "ptemp";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_s_s_s[] = "(%s, %s, %s)";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_store[] = "store";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_marked[] = "marked";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_search[] = "search";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_weight[] = "weight";
static const char __pyx_k_EdgeSet[] = "EdgeSet";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_edge_to[] = "edge_to";
static const char __pyx_k_popleft[] = "popleft";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_add_node[] = "add_node";
static const char __pyx_k_addition[] = "addition";
static const char __pyx_k_copy_for[] = "copy_for";
static const char __pyx_k_distance[] = "distance";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_traverse[] = "_traverse";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_find_path[] = "find_path";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_neighbors[] = "neighbors";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_unvisited[] = "unvisited";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_components[] = "components";
static const char __pyx_k_difference[] = "difference";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_EdgeSet_add[] = "EdgeSet.add";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_looking_for[] = "\nlooking for ";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_reindex_graph[] = "reindex_graph";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_EdgeSet_remove[] = "EdgeSet.remove";
static const char __pyx_k_add_if_shorter[] = "add_if_shorter";
static const char __pyx_k_EdgeSet_edge_to[] = "EdgeSet.edge_to";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_DijkstraPathFinder[] = "DijkstraPathFinder";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_glycan_composition[] = "glycan_composition";
static const char __pyx_k_CompositionGraphBase[] = "CompositionGraphBase";
static const char __pyx_k_CompositionGraphEdge[] = "CompositionGraphEdge";
static const char __pyx_k_CompositionGraphNode[] = "CompositionGraphNode";
static const char __pyx_k_pyx_unpickle_EdgeSet[] = "__pyx_unpickle_EdgeSet";
static const char __pyx_k_build_initial_key_set[] = "_build_initial_key_set";
static const char __pyx_k_EdgeSet_add_if_shorter[] = "EdgeSet.add_if_shorter";
static const char __pyx_k_with_smallest_distance[] = " with smallest distance ";
static const char __pyx_k_EdgeSet___reduce_cython[] = "EdgeSet.__reduce_cython__";
static const char __pyx_k_GlycanCompositionVector[] = "GlycanCompositionVector";
static const char __pyx_k_Object_in_invalid_state[] = "Object in invalid state";
static const char __pyx_k_find_smallest_unvisited[] = "find_smallest_unvisited";
static const char __pyx_k_CompositionGraphBase_copy[] = "CompositionGraphBase.copy";
static const char __pyx_k_CompositionGraphNode_copy[] = "CompositionGraphNode.copy";
static const char __pyx_k_DijkstraPathFinder_search[] = "DijkstraPathFinder.search";
static const char __pyx_k_EdgeSet___setstate_cython[] = "EdgeSet.__setstate_cython__";
static const char __pyx_k_HashableGlycanComposition[] = "HashableGlycanComposition";
static const char __pyx_k_CompositionGraphNode_clone[] = "CompositionGraphNode.clone";
static const char __pyx_k_CompositionGraphEdge_remove[] = "CompositionGraphEdge.remove";
static const char __pyx_k_FrozenMonosaccharideResidue[] = "FrozenMonosaccharideResidue";
static const char __pyx_k_CompositionGraphNode_edge_to[] = "CompositionGraphNode.edge_to";
static const char __pyx_k_DijkstraPathFinder_find_path[] = "DijkstraPathFinder.find_path";
static const char __pyx_k_CompositionGraphEdge___reduce[] = "CompositionGraphEdge.__reduce__";
static const char __pyx_k_CompositionGraphEdge_copy_for[] = "CompositionGraphEdge.copy_for";
static const char __pyx_k_CompositionGraphNode_s_d_0_2f[] = "CompositionGraphNode(%s, %d, %0.2f)";
static const char __pyx_k_Dimensionality_does_not_match[] = "Dimensionality does not match";
static const char __pyx_k_GlycanCompositionVector_clone[] = "GlycanCompositionVector.clone";
static const char __pyx_k_CompositionGraphEdge__traverse[] = "CompositionGraphEdge._traverse";
static const char __pyx_k_CompositionGraphNode_neighbors[] = "CompositionGraphNode.neighbors";
static const char __pyx_k_GlycanCompositionVectorContext[] = "GlycanCompositionVectorContext";
static const char __pyx_k_pyx_unpickle_CompositionGraphB[] = "__pyx_unpickle_CompositionGraphBase";
static const char __pyx_k_pyx_unpickle_CompositionGraphN[] = "__pyx_unpickle_CompositionGraphNode";
static const char __pyx_k_pyx_unpickle_DijkstraPathFinde[] = "__pyx_unpickle_DijkstraPathFinder";
static const char __pyx_k_CompositionGraphBase___setstate[] = "CompositionGraphBase.__setstate_cython__";
static const char __pyx_k_CompositionGraphNode___setstate[] = "CompositionGraphNode.__setstate_cython__";
static const char __pyx_k_reconstruct_GlycanCompositionVe[] = "_reconstruct_GlycanCompositionVector";
static const char __pyx_k_CachingGlycanCompositionVectorCo[] = "CachingGlycanCompositionVectorContext";
static const char __pyx_k_Cannot_compute_addition_from_Non[] = "Cannot compute addition from `None`";
static const char __pyx_k_Cannot_compute_difference_from_N[] = "Cannot compute difference from `None`";
static const char __pyx_k_CompositionGraphBase___reduce_cy[] = "CompositionGraphBase.__reduce_cython__";
static const char __pyx_k_CompositionGraphNode___reduce_cy[] = "CompositionGraphNode.__reduce_cython__";
static const char __pyx_k_DijkstraPathFinder___reduce_cyth[] = "DijkstraPathFinder.__reduce_cython__";
static const char __pyx_k_DijkstraPathFinder___setstate_cy[] = "DijkstraPathFinder.__setstate_cython__";
static const char __pyx_k_DijkstraPathFinder__build_initia[] = "DijkstraPathFinder._build_initial_key_set";
static const char __pyx_k_DijkstraPathFinder_find_smallest[] = "DijkstraPathFinder.find_smallest_unvisited";
static const char __pyx_k_Failed_to_select_a_destination_f[] = "Failed to select a destination from unvisited set of size ";
static const char __pyx_k_GlycanCompositionVectorContext_2[] = "GlycanCompositionVectorContext.__reduce__";
static const char __pyx_k_GlycanCompositionVectorContext_d[] = "GlycanCompositionVectorContext.decode";
static const char __pyx_k_GlycanCompositionVectorContext_e[] = "GlycanCompositionVectorContext.encode";
static const char __pyx_k_GlycanCompositionVector___reduce[] = "GlycanCompositionVector.__reduce__";
static const char __pyx_k_GlycanCompositionVector_addition[] = "GlycanCompositionVector.addition";
static const char __pyx_k_GlycanCompositionVector_differen[] = "GlycanCompositionVector.difference";
static const char __pyx_k_GlycanCompositionVector_distance[] = "GlycanCompositionVector.distance";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))";
static const char __pyx_k_The_unvisited_set_must_not_be_em[] = "The unvisited set must not be empty";
static const char __pyx_k_glycresoft__c_composition_networ[] = "glycresoft._c.composition_network.graph";
static const char __pyx_k_glypy_structure_glycan_compositi[] = "glypy.structure.glycan_composition";
static const char __pyx_k_src_glycresoft__c_composition_ne[] = "src/glycresoft/_c/composition_network/graph.pyx";
static const char __pyx_k_CachingGlycanCompositionVectorCo_2[] = "CachingGlycanCompositionVectorContext.encode";
static const char __pyx_k_CachingGlycanCompositionVectorCo_3[] = "CachingGlycanCompositionVectorContext.decode";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))";
/* #### Code section: decls ### */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_composition, PyObject *__pyx_v_index, PyObject *__pyx_v_score, PyObject *__pyx_v_marked, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18glycan_composition___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5order___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_2edge_to(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4neighbors(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_8__str__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_10__hash__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_12__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14copy(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_16clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_20__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v_store); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_2edge_to(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_4add(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_6add_if_shorter(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_8remove(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_10__iter__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_12__len__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_14__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_16__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_18__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_20__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_node1, PyObject *__pyx_v_node2, PyObject *__pyx_v_order, PyObject *__pyx_v_weight); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_2__getitem__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_traverse(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_8__str__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_10__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_12__hash__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_14__reduce__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_16copy_for(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_18remove(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_reindex_graph(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_2__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_4__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_graph, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_limit); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_2_build_initial_key_set(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_4find_smallest_unvisited(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_unvisited); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_6find_path(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8search(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_10__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_12__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_components); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_2__reduce__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_4encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_glycan_composition); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_6decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_2_reconstruct_GlycanCompositionVector(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector___reduce__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static void __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_2__dealloc__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_4__getitem__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, int __pyx_v_i); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_6__setitem__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, int __pyx_v_i, int __pyx_v_value); /* proto */
static Py_ssize_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_8__len__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_10__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_12distance(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_14__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_16__ne__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_18difference(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_20addition(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_22__sub__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_24__add__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_26__neg__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_28__hash__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_30clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_components); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_2encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_gc); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_4decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_4__pyx_unpickle_CompositionGraphNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_6__pyx_unpickle_EdgeSet(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_8__pyx_unpickle_CompositionGraphBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_10__pyx_unpickle_DijkstraPathFinder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_EdgeSet(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;
  PyObject *__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
  #endif
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;
  PyTypeObject *__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_CachingGlycanCompositionVectorCo;
  PyObject *__pyx_n_s_CachingGlycanCompositionVectorCo_2;
  PyObject *__pyx_n_s_CachingGlycanCompositionVectorCo_3;
  PyObject *__pyx_kp_s_Cannot_compute_addition_from_Non;
  PyObject *__pyx_kp_s_Cannot_compute_difference_from_N;
  PyObject *__pyx_n_s_CompositionGraphBase;
  PyObject *__pyx_n_s_CompositionGraphBase___reduce_cy;
  PyObject *__pyx_n_s_CompositionGraphBase___setstate;
  PyObject *__pyx_n_s_CompositionGraphBase_copy;
  PyObject *__pyx_n_s_CompositionGraphEdge;
  PyObject *__pyx_n_s_CompositionGraphEdge___reduce;
  PyObject *__pyx_n_s_CompositionGraphEdge__traverse;
  PyObject *__pyx_n_s_CompositionGraphEdge_copy_for;
  PyObject *__pyx_n_s_CompositionGraphEdge_remove;
  PyObject *__pyx_n_s_CompositionGraphNode;
  PyObject *__pyx_n_s_CompositionGraphNode___reduce_cy;
  PyObject *__pyx_n_s_CompositionGraphNode___setstate;
  PyObject *__pyx_n_s_CompositionGraphNode_clone;
  PyObject *__pyx_n_s_CompositionGraphNode_copy;
  PyObject *__pyx_n_s_CompositionGraphNode_edge_to;
  PyObject *__pyx_n_s_CompositionGraphNode_neighbors;
  PyObject *__pyx_kp_s_CompositionGraphNode_s_d_0_2f;
  PyObject *__pyx_n_s_DijkstraPathFinder;
  PyObject *__pyx_n_s_DijkstraPathFinder___reduce_cyth;
  PyObject *__pyx_n_s_DijkstraPathFinder___setstate_cy;
  PyObject *__pyx_n_s_DijkstraPathFinder__build_initia;
  PyObject *__pyx_n_s_DijkstraPathFinder_find_path;
  PyObject *__pyx_n_s_DijkstraPathFinder_find_smallest;
  PyObject *__pyx_n_s_DijkstraPathFinder_search;
  PyObject *__pyx_kp_s_Dimensionality_does_not_match;
  PyObject *__pyx_n_s_EdgeSet;
  PyObject *__pyx_n_s_EdgeSet___reduce_cython;
  PyObject *__pyx_n_s_EdgeSet___setstate_cython;
  PyObject *__pyx_n_s_EdgeSet_add;
  PyObject *__pyx_n_s_EdgeSet_add_if_shorter;
  PyObject *__pyx_n_s_EdgeSet_edge_to;
  PyObject *__pyx_n_s_EdgeSet_remove;
  PyObject *__pyx_kp_u_Failed_to_select_a_destination_f;
  PyObject *__pyx_n_s_FrozenMonosaccharideResidue;
  PyObject *__pyx_n_s_GlycanCompositionVector;
  PyObject *__pyx_n_s_GlycanCompositionVectorContext;
  PyObject *__pyx_n_s_GlycanCompositionVectorContext_2;
  PyObject *__pyx_n_s_GlycanCompositionVectorContext_d;
  PyObject *__pyx_n_s_GlycanCompositionVectorContext_e;
  PyObject *__pyx_n_s_GlycanCompositionVector___reduce;
  PyObject *__pyx_n_s_GlycanCompositionVector_addition;
  PyObject *__pyx_n_s_GlycanCompositionVector_clone;
  PyObject *__pyx_n_s_GlycanCompositionVector_differen;
  PyObject *__pyx_n_s_GlycanCompositionVector_distance;
  PyObject *__pyx_n_s_HashableGlycanComposition;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_n_s_None;
  PyObject *__pyx_kp_s_Object_in_invalid_state;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_kp_s_The_unvisited_set_must_not_be_em;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_kp_u__12;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_kp_u__4;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_n_s__73;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_n_s_add;
  PyObject *__pyx_n_s_add_if_shorter;
  PyObject *__pyx_n_s_add_node;
  PyObject *__pyx_n_s_addition;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_build_initial_key_set;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_clone;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_components;
  PyObject *__pyx_n_s_composition;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_copy_for;
  PyObject *__pyx_n_s_decode;
  PyObject *__pyx_n_s_deque;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_n_s_difference;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_distance;
  PyObject *__pyx_n_s_dup;
  PyObject *__pyx_n_s_edge;
  PyObject *__pyx_n_s_edge_to;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_end;
  PyObject *__pyx_n_s_find_path;
  PyObject *__pyx_n_s_find_smallest_unvisited;
  PyObject *__pyx_n_s_gc;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_gcvec;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_glycan_composition;
  PyObject *__pyx_n_s_glycresoft__c_composition_networ;
  PyObject *__pyx_n_s_glypy_structure_glycan_compositi;
  PyObject *__pyx_n_s_graph;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_inf;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_limit;
  PyObject *__pyx_kp_u_looking_for;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_marked;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_neighbors;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_node;
  PyObject *__pyx_n_s_node1;
  PyObject *__pyx_n_s_node2;
  PyObject *__pyx_n_s_nodes;
  PyObject *__pyx_n_s_order;
  PyObject *__pyx_n_s_other;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_popleft;
  PyObject *__pyx_n_s_ptemp;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_CompositionGraphB;
  PyObject *__pyx_n_s_pyx_unpickle_CompositionGraphN;
  PyObject *__pyx_n_s_pyx_unpickle_DijkstraPathFinde;
  PyObject *__pyx_n_s_pyx_unpickle_EdgeSet;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_reconstruct_GlycanCompositionVe;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_reindex_graph;
  PyObject *__pyx_n_s_remove;
  PyObject *__pyx_n_s_result;
  PyObject *__pyx_kp_s_s_s_s;
  PyObject *__pyx_n_s_score;
  PyObject *__pyx_n_s_search;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_kp_s_src_glycresoft__c_composition_ne;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_store;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_traverse;
  PyObject *__pyx_n_s_unvisited;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_n_s_weight;
  PyObject *__pyx_kp_u_with_smallest_distance;
  PyObject *__pyx_float_0_;
  PyObject *__pyx_float_1_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_19304315;
  PyObject *__pyx_int_25300956;
  PyObject *__pyx_int_25655789;
  PyObject *__pyx_int_46892423;
  PyObject *__pyx_int_60953237;
  PyObject *__pyx_int_61202372;
  PyObject *__pyx_int_136632333;
  PyObject *__pyx_int_147687723;
  PyObject *__pyx_int_183956233;
  PyObject *__pyx_int_243238152;
  PyObject *__pyx_int_255004817;
  PyObject *__pyx_int_257203134;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_k_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_codeobj__17;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__59;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__72;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);
  Py_CLEAR(clear_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CachingGlycanCompositionVectorCo);
  Py_CLEAR(clear_module_state->__pyx_n_s_CachingGlycanCompositionVectorCo_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_CachingGlycanCompositionVectorCo_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_compute_addition_from_Non);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_compute_difference_from_N);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphBase___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphBase___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphBase_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphEdge);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphEdge___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphEdge__traverse);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphEdge_copy_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphEdge_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode_clone);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode_edge_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositionGraphNode_neighbors);
  Py_CLEAR(clear_module_state->__pyx_kp_s_CompositionGraphNode_s_d_0_2f);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder__build_initia);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder_find_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder_find_smallest);
  Py_CLEAR(clear_module_state->__pyx_n_s_DijkstraPathFinder_search);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dimensionality_does_not_match);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet_add_if_shorter);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet_edge_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_EdgeSet_remove);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Failed_to_select_a_destination_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenMonosaccharideResidue);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVector);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVectorContext);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVectorContext_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVectorContext_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVectorContext_e);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVector___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVector_addition);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVector_clone);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVector_differen);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlycanCompositionVector_distance);
  Py_CLEAR(clear_module_state->__pyx_n_s_HashableGlycanComposition);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_n_s_None);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object_in_invalid_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_The_unvisited_set_must_not_be_em);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_kp_u__12);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__4);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_n_s__73);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_n_s_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_if_shorter);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_node);
  Py_CLEAR(clear_module_state->__pyx_n_s_addition);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_build_initial_key_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_clone);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_components);
  Py_CLEAR(clear_module_state->__pyx_n_s_composition);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_deque);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_difference);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_distance);
  Py_CLEAR(clear_module_state->__pyx_n_s_dup);
  Py_CLEAR(clear_module_state->__pyx_n_s_edge);
  Py_CLEAR(clear_module_state->__pyx_n_s_edge_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_find_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_find_smallest_unvisited);
  Py_CLEAR(clear_module_state->__pyx_n_s_gc);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_gcvec);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycan_composition);
  Py_CLEAR(clear_module_state->__pyx_n_s_glycresoft__c_composition_networ);
  Py_CLEAR(clear_module_state->__pyx_n_s_glypy_structure_glycan_compositi);
  Py_CLEAR(clear_module_state->__pyx_n_s_graph);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_inf);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit);
  Py_CLEAR(clear_module_state->__pyx_kp_u_looking_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_marked);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_neighbors);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_node);
  Py_CLEAR(clear_module_state->__pyx_n_s_node1);
  Py_CLEAR(clear_module_state->__pyx_n_s_node2);
  Py_CLEAR(clear_module_state->__pyx_n_s_nodes);
  Py_CLEAR(clear_module_state->__pyx_n_s_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_other);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_popleft);
  Py_CLEAR(clear_module_state->__pyx_n_s_ptemp);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CompositionGraphB);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CompositionGraphN);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_DijkstraPathFinde);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_EdgeSet);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_reconstruct_GlycanCompositionVe);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_reindex_graph);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_result);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_s_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_score);
  Py_CLEAR(clear_module_state->__pyx_n_s_search);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_glycresoft__c_composition_ne);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_store);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_traverse);
  Py_CLEAR(clear_module_state->__pyx_n_s_unvisited);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_weight);
  Py_CLEAR(clear_module_state->__pyx_kp_u_with_smallest_distance);
  Py_CLEAR(clear_module_state->__pyx_float_0_);
  Py_CLEAR(clear_module_state->__pyx_float_1_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_19304315);
  Py_CLEAR(clear_module_state->__pyx_int_25300956);
  Py_CLEAR(clear_module_state->__pyx_int_25655789);
  Py_CLEAR(clear_module_state->__pyx_int_46892423);
  Py_CLEAR(clear_module_state->__pyx_int_60953237);
  Py_CLEAR(clear_module_state->__pyx_int_61202372);
  Py_CLEAR(clear_module_state->__pyx_int_136632333);
  Py_CLEAR(clear_module_state->__pyx_int_147687723);
  Py_CLEAR(clear_module_state->__pyx_int_183956233);
  Py_CLEAR(clear_module_state->__pyx_int_243238152);
  Py_CLEAR(clear_module_state->__pyx_int_255004817);
  Py_CLEAR(clear_module_state->__pyx_int_257203134);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_k_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__17);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__59);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);
  Py_VISIT(traverse_module_state->__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CachingGlycanCompositionVectorCo);
  Py_VISIT(traverse_module_state->__pyx_n_s_CachingGlycanCompositionVectorCo_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_CachingGlycanCompositionVectorCo_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_compute_addition_from_Non);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_compute_difference_from_N);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphBase___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphBase___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphBase_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphEdge);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphEdge___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphEdge__traverse);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphEdge_copy_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphEdge_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode_clone);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode_edge_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositionGraphNode_neighbors);
  Py_VISIT(traverse_module_state->__pyx_kp_s_CompositionGraphNode_s_d_0_2f);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder__build_initia);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder_find_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder_find_smallest);
  Py_VISIT(traverse_module_state->__pyx_n_s_DijkstraPathFinder_search);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dimensionality_does_not_match);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet_add_if_shorter);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet_edge_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_EdgeSet_remove);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Failed_to_select_a_destination_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenMonosaccharideResidue);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVector);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVectorContext);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVectorContext_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVectorContext_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVectorContext_e);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVector___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVector_addition);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVector_clone);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVector_differen);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlycanCompositionVector_distance);
  Py_VISIT(traverse_module_state->__pyx_n_s_HashableGlycanComposition);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_n_s_None);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object_in_invalid_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_The_unvisited_set_must_not_be_em);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_kp_u__12);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__4);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_n_s__73);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_n_s_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_if_shorter);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_node);
  Py_VISIT(traverse_module_state->__pyx_n_s_addition);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_build_initial_key_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_clone);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_components);
  Py_VISIT(traverse_module_state->__pyx_n_s_composition);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_deque);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_difference);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_distance);
  Py_VISIT(traverse_module_state->__pyx_n_s_dup);
  Py_VISIT(traverse_module_state->__pyx_n_s_edge);
  Py_VISIT(traverse_module_state->__pyx_n_s_edge_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_find_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_find_smallest_unvisited);
  Py_VISIT(traverse_module_state->__pyx_n_s_gc);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_gcvec);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycan_composition);
  Py_VISIT(traverse_module_state->__pyx_n_s_glycresoft__c_composition_networ);
  Py_VISIT(traverse_module_state->__pyx_n_s_glypy_structure_glycan_compositi);
  Py_VISIT(traverse_module_state->__pyx_n_s_graph);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_inf);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit);
  Py_VISIT(traverse_module_state->__pyx_kp_u_looking_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_marked);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_neighbors);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_node);
  Py_VISIT(traverse_module_state->__pyx_n_s_node1);
  Py_VISIT(traverse_module_state->__pyx_n_s_node2);
  Py_VISIT(traverse_module_state->__pyx_n_s_nodes);
  Py_VISIT(traverse_module_state->__pyx_n_s_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_other);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_popleft);
  Py_VISIT(traverse_module_state->__pyx_n_s_ptemp);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CompositionGraphB);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CompositionGraphN);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_DijkstraPathFinde);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_EdgeSet);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_reconstruct_GlycanCompositionVe);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_reindex_graph);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_result);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_s_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_score);
  Py_VISIT(traverse_module_state->__pyx_n_s_search);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_glycresoft__c_composition_ne);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_store);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_traverse);
  Py_VISIT(traverse_module_state->__pyx_n_s_unvisited);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_weight);
  Py_VISIT(traverse_module_state->__pyx_kp_u_with_smallest_distance);
  Py_VISIT(traverse_module_state->__pyx_float_0_);
  Py_VISIT(traverse_module_state->__pyx_float_1_0);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_19304315);
  Py_VISIT(traverse_module_state->__pyx_int_25300956);
  Py_VISIT(traverse_module_state->__pyx_int_25655789);
  Py_VISIT(traverse_module_state->__pyx_int_46892423);
  Py_VISIT(traverse_module_state->__pyx_int_60953237);
  Py_VISIT(traverse_module_state->__pyx_int_61202372);
  Py_VISIT(traverse_module_state->__pyx_int_136632333);
  Py_VISIT(traverse_module_state->__pyx_int_147687723);
  Py_VISIT(traverse_module_state->__pyx_int_183956233);
  Py_VISIT(traverse_module_state->__pyx_int_243238152);
  Py_VISIT(traverse_module_state->__pyx_int_255004817);
  Py_VISIT(traverse_module_state->__pyx_int_257203134);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_k_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__17);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__59);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext
#define __pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder __pyx_mstate_global->__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder
#endif
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext
#define __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder __pyx_mstate_global->__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_CachingGlycanCompositionVectorCo __pyx_mstate_global->__pyx_n_s_CachingGlycanCompositionVectorCo
#define __pyx_n_s_CachingGlycanCompositionVectorCo_2 __pyx_mstate_global->__pyx_n_s_CachingGlycanCompositionVectorCo_2
#define __pyx_n_s_CachingGlycanCompositionVectorCo_3 __pyx_mstate_global->__pyx_n_s_CachingGlycanCompositionVectorCo_3
#define __pyx_kp_s_Cannot_compute_addition_from_Non __pyx_mstate_global->__pyx_kp_s_Cannot_compute_addition_from_Non
#define __pyx_kp_s_Cannot_compute_difference_from_N __pyx_mstate_global->__pyx_kp_s_Cannot_compute_difference_from_N
#define __pyx_n_s_CompositionGraphBase __pyx_mstate_global->__pyx_n_s_CompositionGraphBase
#define __pyx_n_s_CompositionGraphBase___reduce_cy __pyx_mstate_global->__pyx_n_s_CompositionGraphBase___reduce_cy
#define __pyx_n_s_CompositionGraphBase___setstate __pyx_mstate_global->__pyx_n_s_CompositionGraphBase___setstate
#define __pyx_n_s_CompositionGraphBase_copy __pyx_mstate_global->__pyx_n_s_CompositionGraphBase_copy
#define __pyx_n_s_CompositionGraphEdge __pyx_mstate_global->__pyx_n_s_CompositionGraphEdge
#define __pyx_n_s_CompositionGraphEdge___reduce __pyx_mstate_global->__pyx_n_s_CompositionGraphEdge___reduce
#define __pyx_n_s_CompositionGraphEdge__traverse __pyx_mstate_global->__pyx_n_s_CompositionGraphEdge__traverse
#define __pyx_n_s_CompositionGraphEdge_copy_for __pyx_mstate_global->__pyx_n_s_CompositionGraphEdge_copy_for
#define __pyx_n_s_CompositionGraphEdge_remove __pyx_mstate_global->__pyx_n_s_CompositionGraphEdge_remove
#define __pyx_n_s_CompositionGraphNode __pyx_mstate_global->__pyx_n_s_CompositionGraphNode
#define __pyx_n_s_CompositionGraphNode___reduce_cy __pyx_mstate_global->__pyx_n_s_CompositionGraphNode___reduce_cy
#define __pyx_n_s_CompositionGraphNode___setstate __pyx_mstate_global->__pyx_n_s_CompositionGraphNode___setstate
#define __pyx_n_s_CompositionGraphNode_clone __pyx_mstate_global->__pyx_n_s_CompositionGraphNode_clone
#define __pyx_n_s_CompositionGraphNode_copy __pyx_mstate_global->__pyx_n_s_CompositionGraphNode_copy
#define __pyx_n_s_CompositionGraphNode_edge_to __pyx_mstate_global->__pyx_n_s_CompositionGraphNode_edge_to
#define __pyx_n_s_CompositionGraphNode_neighbors __pyx_mstate_global->__pyx_n_s_CompositionGraphNode_neighbors
#define __pyx_kp_s_CompositionGraphNode_s_d_0_2f __pyx_mstate_global->__pyx_kp_s_CompositionGraphNode_s_d_0_2f
#define __pyx_n_s_DijkstraPathFinder __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder
#define __pyx_n_s_DijkstraPathFinder___reduce_cyth __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder___reduce_cyth
#define __pyx_n_s_DijkstraPathFinder___setstate_cy __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder___setstate_cy
#define __pyx_n_s_DijkstraPathFinder__build_initia __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder__build_initia
#define __pyx_n_s_DijkstraPathFinder_find_path __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder_find_path
#define __pyx_n_s_DijkstraPathFinder_find_smallest __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder_find_smallest
#define __pyx_n_s_DijkstraPathFinder_search __pyx_mstate_global->__pyx_n_s_DijkstraPathFinder_search
#define __pyx_kp_s_Dimensionality_does_not_match __pyx_mstate_global->__pyx_kp_s_Dimensionality_does_not_match
#define __pyx_n_s_EdgeSet __pyx_mstate_global->__pyx_n_s_EdgeSet
#define __pyx_n_s_EdgeSet___reduce_cython __pyx_mstate_global->__pyx_n_s_EdgeSet___reduce_cython
#define __pyx_n_s_EdgeSet___setstate_cython __pyx_mstate_global->__pyx_n_s_EdgeSet___setstate_cython
#define __pyx_n_s_EdgeSet_add __pyx_mstate_global->__pyx_n_s_EdgeSet_add
#define __pyx_n_s_EdgeSet_add_if_shorter __pyx_mstate_global->__pyx_n_s_EdgeSet_add_if_shorter
#define __pyx_n_s_EdgeSet_edge_to __pyx_mstate_global->__pyx_n_s_EdgeSet_edge_to
#define __pyx_n_s_EdgeSet_remove __pyx_mstate_global->__pyx_n_s_EdgeSet_remove
#define __pyx_kp_u_Failed_to_select_a_destination_f __pyx_mstate_global->__pyx_kp_u_Failed_to_select_a_destination_f
#define __pyx_n_s_FrozenMonosaccharideResidue __pyx_mstate_global->__pyx_n_s_FrozenMonosaccharideResidue
#define __pyx_n_s_GlycanCompositionVector __pyx_mstate_global->__pyx_n_s_GlycanCompositionVector
#define __pyx_n_s_GlycanCompositionVectorContext __pyx_mstate_global->__pyx_n_s_GlycanCompositionVectorContext
#define __pyx_n_s_GlycanCompositionVectorContext_2 __pyx_mstate_global->__pyx_n_s_GlycanCompositionVectorContext_2
#define __pyx_n_s_GlycanCompositionVectorContext_d __pyx_mstate_global->__pyx_n_s_GlycanCompositionVectorContext_d
#define __pyx_n_s_GlycanCompositionVectorContext_e __pyx_mstate_global->__pyx_n_s_GlycanCompositionVectorContext_e
#define __pyx_n_s_GlycanCompositionVector___reduce __pyx_mstate_global->__pyx_n_s_GlycanCompositionVector___reduce
#define __pyx_n_s_GlycanCompositionVector_addition __pyx_mstate_global->__pyx_n_s_GlycanCompositionVector_addition
#define __pyx_n_s_GlycanCompositionVector_clone __pyx_mstate_global->__pyx_n_s_GlycanCompositionVector_clone
#define __pyx_n_s_GlycanCompositionVector_differen __pyx_mstate_global->__pyx_n_s_GlycanCompositionVector_differen
#define __pyx_n_s_GlycanCompositionVector_distance __pyx_mstate_global->__pyx_n_s_GlycanCompositionVector_distance
#define __pyx_n_s_HashableGlycanComposition __pyx_mstate_global->__pyx_n_s_HashableGlycanComposition
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_n_s_None __pyx_mstate_global->__pyx_n_s_None
#define __pyx_kp_s_Object_in_invalid_state __pyx_mstate_global->__pyx_kp_s_Object_in_invalid_state
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_kp_s_The_unvisited_set_must_not_be_em __pyx_mstate_global->__pyx_kp_s_The_unvisited_set_must_not_be_em
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_kp_u__12 __pyx_mstate_global->__pyx_kp_u__12
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_kp_u__4 __pyx_mstate_global->__pyx_kp_u__4
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_n_s__73 __pyx_mstate_global->__pyx_n_s__73
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_n_s_add __pyx_mstate_global->__pyx_n_s_add
#define __pyx_n_s_add_if_shorter __pyx_mstate_global->__pyx_n_s_add_if_shorter
#define __pyx_n_s_add_node __pyx_mstate_global->__pyx_n_s_add_node
#define __pyx_n_s_addition __pyx_mstate_global->__pyx_n_s_addition
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_build_initial_key_set __pyx_mstate_global->__pyx_n_s_build_initial_key_set
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_clone __pyx_mstate_global->__pyx_n_s_clone
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_components __pyx_mstate_global->__pyx_n_s_components
#define __pyx_n_s_composition __pyx_mstate_global->__pyx_n_s_composition
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_copy_for __pyx_mstate_global->__pyx_n_s_copy_for
#define __pyx_n_s_decode __pyx_mstate_global->__pyx_n_s_decode
#define __pyx_n_s_deque __pyx_mstate_global->__pyx_n_s_deque
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_n_s_difference __pyx_mstate_global->__pyx_n_s_difference
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_distance __pyx_mstate_global->__pyx_n_s_distance
#define __pyx_n_s_dup __pyx_mstate_global->__pyx_n_s_dup
#define __pyx_n_s_edge __pyx_mstate_global->__pyx_n_s_edge
#define __pyx_n_s_edge_to __pyx_mstate_global->__pyx_n_s_edge_to
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_end __pyx_mstate_global->__pyx_n_s_end
#define __pyx_n_s_find_path __pyx_mstate_global->__pyx_n_s_find_path
#define __pyx_n_s_find_smallest_unvisited __pyx_mstate_global->__pyx_n_s_find_smallest_unvisited
#define __pyx_n_s_gc __pyx_mstate_global->__pyx_n_s_gc
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_gcvec __pyx_mstate_global->__pyx_n_s_gcvec
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_glycan_composition __pyx_mstate_global->__pyx_n_s_glycan_composition
#define __pyx_n_s_glycresoft__c_composition_networ __pyx_mstate_global->__pyx_n_s_glycresoft__c_composition_networ
#define __pyx_n_s_glypy_structure_glycan_compositi __pyx_mstate_global->__pyx_n_s_glypy_structure_glycan_compositi
#define __pyx_n_s_graph __pyx_mstate_global->__pyx_n_s_graph
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_inf __pyx_mstate_global->__pyx_n_s_inf
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_limit __pyx_mstate_global->__pyx_n_s_limit
#define __pyx_kp_u_looking_for __pyx_mstate_global->__pyx_kp_u_looking_for
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_marked __pyx_mstate_global->__pyx_n_s_marked
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_neighbors __pyx_mstate_global->__pyx_n_s_neighbors
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_node __pyx_mstate_global->__pyx_n_s_node
#define __pyx_n_s_node1 __pyx_mstate_global->__pyx_n_s_node1
#define __pyx_n_s_node2 __pyx_mstate_global->__pyx_n_s_node2
#define __pyx_n_s_nodes __pyx_mstate_global->__pyx_n_s_nodes
#define __pyx_n_s_order __pyx_mstate_global->__pyx_n_s_order
#define __pyx_n_s_other __pyx_mstate_global->__pyx_n_s_other
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_popleft __pyx_mstate_global->__pyx_n_s_popleft
#define __pyx_n_s_ptemp __pyx_mstate_global->__pyx_n_s_ptemp
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_CompositionGraphB __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CompositionGraphB
#define __pyx_n_s_pyx_unpickle_CompositionGraphN __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CompositionGraphN
#define __pyx_n_s_pyx_unpickle_DijkstraPathFinde __pyx_mstate_global->__pyx_n_s_pyx_unpickle_DijkstraPathFinde
#define __pyx_n_s_pyx_unpickle_EdgeSet __pyx_mstate_global->__pyx_n_s_pyx_unpickle_EdgeSet
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_reconstruct_GlycanCompositionVe __pyx_mstate_global->__pyx_n_s_reconstruct_GlycanCompositionVe
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_reindex_graph __pyx_mstate_global->__pyx_n_s_reindex_graph
#define __pyx_n_s_remove __pyx_mstate_global->__pyx_n_s_remove
#define __pyx_n_s_result __pyx_mstate_global->__pyx_n_s_result
#define __pyx_kp_s_s_s_s __pyx_mstate_global->__pyx_kp_s_s_s_s
#define __pyx_n_s_score __pyx_mstate_global->__pyx_n_s_score
#define __pyx_n_s_search __pyx_mstate_global->__pyx_n_s_search
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_kp_s_src_glycresoft__c_composition_ne __pyx_mstate_global->__pyx_kp_s_src_glycresoft__c_composition_ne
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_store __pyx_mstate_global->__pyx_n_s_store
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_traverse __pyx_mstate_global->__pyx_n_s_traverse
#define __pyx_n_s_unvisited __pyx_mstate_global->__pyx_n_s_unvisited
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_n_s_weight __pyx_mstate_global->__pyx_n_s_weight
#define __pyx_kp_u_with_smallest_distance __pyx_mstate_global->__pyx_kp_u_with_smallest_distance
#define __pyx_float_0_ __pyx_mstate_global->__pyx_float_0_
#define __pyx_float_1_0 __pyx_mstate_global->__pyx_float_1_0
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_19304315 __pyx_mstate_global->__pyx_int_19304315
#define __pyx_int_25300956 __pyx_mstate_global->__pyx_int_25300956
#define __pyx_int_25655789 __pyx_mstate_global->__pyx_int_25655789
#define __pyx_int_46892423 __pyx_mstate_global->__pyx_int_46892423
#define __pyx_int_60953237 __pyx_mstate_global->__pyx_int_60953237
#define __pyx_int_61202372 __pyx_mstate_global->__pyx_int_61202372
#define __pyx_int_136632333 __pyx_mstate_global->__pyx_int_136632333
#define __pyx_int_147687723 __pyx_mstate_global->__pyx_int_147687723
#define __pyx_int_183956233 __pyx_mstate_global->__pyx_int_183956233
#define __pyx_int_243238152 __pyx_mstate_global->__pyx_int_243238152
#define __pyx_int_255004817 __pyx_mstate_global->__pyx_int_255004817
#define __pyx_int_257203134 __pyx_mstate_global->__pyx_int_257203134
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_k_ __pyx_mstate_global->__pyx_k_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_codeobj__17 __pyx_mstate_global->__pyx_codeobj__17
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__59 __pyx_mstate_global->__pyx_codeobj__59
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
/* #### Code section: module_code ### */

/* "glycresoft/_c/composition_network/graph.pyx":20
 * cdef class CompositionGraphNode(object):
 * 
 *     def __init__(self, composition, index, score=0., marked=False, **kwargs):             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         self.index = index
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_composition = 0;
  PyObject *__pyx_v_index = 0;
  PyObject *__pyx_v_score = 0;
  PyObject *__pyx_v_marked = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_index,&__pyx_n_s_score,&__pyx_n_s_marked,0};
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_0_));
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_composition)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 20, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_score);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_marked);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 20, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_composition = values[0];
    __pyx_v_index = values[1];
    __pyx_v_score = values[2];
    __pyx_v_marked = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 20, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode___init__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), __pyx_v_composition, __pyx_v_index, __pyx_v_score, __pyx_v_marked, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_composition, PyObject *__pyx_v_index, PyObject *__pyx_v_score, PyObject *__pyx_v_marked, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_hash_t __pyx_t_3;
  double __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":21
 * 
 *     def __init__(self, composition, index, score=0., marked=False, **kwargs):
 *         self.composition = composition             # <<<<<<<<<<<<<<
 *         self.index = index
 *         self.edges = EdgeSet()
 */
  __Pyx_INCREF(__pyx_v_composition);
  __Pyx_GIVEREF(__pyx_v_composition);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(__pyx_v_self->composition);
  __pyx_v_self->composition = __pyx_v_composition;

  /* "glycresoft/_c/composition_network/graph.pyx":22
 *     def __init__(self, composition, index, score=0., marked=False, **kwargs):
 *         self.composition = composition
 *         self.index = index             # <<<<<<<<<<<<<<
 *         self.edges = EdgeSet()
 *         self._str = str(self.composition)
 */
  __pyx_t_1 = PyInt_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":23
 *         self.composition = composition
 *         self.index = index
 *         self.edges = EdgeSet()             # <<<<<<<<<<<<<<
 *         self._str = str(self.composition)
 *         self._hash = hash(str(self._str))
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->edges);
  __Pyx_DECREF((PyObject *)__pyx_v_self->edges);
  __pyx_v_self->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":24
 *         self.index = index
 *         self.edges = EdgeSet()
 *         self._str = str(self.composition)             # <<<<<<<<<<<<<<
 *         self._hash = hash(str(self._str))
 *         self._score = score
 */
  __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_self->composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyString_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_str);
  __Pyx_DECREF(__pyx_v_self->_str);
  __pyx_v_self->_str = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":25
 *         self.edges = EdgeSet()
 *         self._str = str(self.composition)
 *         self._hash = hash(str(self._str))             # <<<<<<<<<<<<<<
 *         self._score = score
 *         self.internal_score = 0.0
 */
  __pyx_t_2 = __Pyx_PyStr_Str(__pyx_v_self->_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Hash(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_hash_t)-1))) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_hash = __pyx_t_3;

  /* "glycresoft/_c/composition_network/graph.pyx":26
 *         self._str = str(self.composition)
 *         self._hash = hash(str(self._str))
 *         self._score = score             # <<<<<<<<<<<<<<
 *         self.internal_score = 0.0
 *         self.marked = marked
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_score); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_v_self->_score = __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":27
 *         self._hash = hash(str(self._str))
 *         self._score = score
 *         self.internal_score = 0.0             # <<<<<<<<<<<<<<
 *         self.marked = marked
 * 
 */
  __pyx_v_self->internal_score = 0.0;

  /* "glycresoft/_c/composition_network/graph.pyx":28
 *         self._score = score
 *         self.internal_score = 0.0
 *         self.marked = marked             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_marked); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_v_self->marked = __pyx_t_5;

  /* "glycresoft/_c/composition_network/graph.pyx":20
 * cdef class CompositionGraphNode(object):
 * 
 *     def __init__(self, composition, index, score=0., marked=False, **kwargs):             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         self.index = index
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":30
 *         self.marked = marked
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glycan_composition(self):
 *         return self.composition
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18glycan_composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18glycan_composition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18glycan_composition___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18glycan_composition___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":32
 *     @property
 *     def glycan_composition(self):
 *         return self.composition             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->composition);
  __pyx_r = __pyx_v_self->composition;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":30
 *         self.marked = marked
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glycan_composition(self):
 *         return self.composition
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":34
 *         return self.composition
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order(self):
 *         return len(self.edges)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5order_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5order_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5order___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5order___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":36
 *     @property
 *     def order(self):
 *         return len(self.edges)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->edges);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":34
 *         return self.composition
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order(self):
 *         return len(self.edges)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.order.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":38
 *         return len(self.edges)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def score(self):
 *         if self._score == 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":40
 *     @property
 *     def score(self):
 *         if self._score == 0:             # <<<<<<<<<<<<<<
 *             return self.internal_score
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->_score == 0.0);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":41
 *     def score(self):
 *         if self._score == 0:
 *             return self.internal_score             # <<<<<<<<<<<<<<
 *         else:
 *             return self._score
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->internal_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":40
 *     @property
 *     def score(self):
 *         if self._score == 0:             # <<<<<<<<<<<<<<
 *             return self.internal_score
 *         else:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":43
 *             return self.internal_score
 *         else:
 *             return self._score             # <<<<<<<<<<<<<<
 * 
 *     @score.setter
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":38
 *         return len(self.edges)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def score(self):
 *         if self._score == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":45
 *             return self._score
 * 
 *     @score.setter             # <<<<<<<<<<<<<<
 *     def score(self, value):
 *         self._score = value
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":47
 *     @score.setter
 *     def score(self, value):
 *         self._score = value             # <<<<<<<<<<<<<<
 * 
 *     def edge_to(self, node):
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_v_self->_score = __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":45
 *             return self._score
 * 
 *     @score.setter             # <<<<<<<<<<<<<<
 *     def score(self, value):
 *         self._score = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":49
 *         self._score = value
 * 
 *     def edge_to(self, node):             # <<<<<<<<<<<<<<
 *         return self.edges.edge_to(self, node)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_3edge_to(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_2edge_to, "CompositionGraphNode.edge_to(self, node)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_3edge_to = {"edge_to", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_3edge_to, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_2edge_to};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_3edge_to(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_node = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("edge_to (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "edge_to") < 0)) __PYX_ERR(0, 49, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_node = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("edge_to", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 49, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.edge_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_2edge_to(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), __pyx_v_node);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_2edge_to(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("edge_to", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":50
 * 
 *     def edge_to(self, node):
 *         return self.edges.edge_to(self, node)             # <<<<<<<<<<<<<<
 * 
 *     def neighbors(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->edges), __pyx_n_s_edge_to); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_node};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":49
 *         self._score = value
 * 
 *     def edge_to(self, node):             # <<<<<<<<<<<<<<
 *         return self.edges.edge_to(self, node)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.edge_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":52
 *         return self.edges.edge_to(self, node)
 * 
 *     def neighbors(self):             # <<<<<<<<<<<<<<
 *         result = [edge._traverse(self) for edge in self.edges]
 *         return result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5neighbors(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4neighbors, "CompositionGraphNode.neighbors(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5neighbors = {"neighbors", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5neighbors, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4neighbors};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5neighbors(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("neighbors (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("neighbors", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "neighbors", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4neighbors(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4neighbors(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_7genexpr__pyx_v_edge = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("neighbors", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":53
 * 
 *     def neighbors(self):
 *         result = [edge._traverse(self) for edge in self.edges]             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self->edges))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self->edges))) {
      __pyx_t_2 = ((PyObject *)__pyx_v_self->edges); __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(((PyObject *)__pyx_v_self->edges)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 53, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 53, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 53, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 53, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 53, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 53, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 53, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_edge, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_7genexpr__pyx_v_edge, __pyx_n_s_traverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 53, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 53, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 53, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_edge); __pyx_7genexpr__pyx_v_edge = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_edge); __pyx_7genexpr__pyx_v_edge = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":54
 *     def neighbors(self):
 *         result = [edge._traverse(self) for edge in self.edges]
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":52
 *         return self.edges.edge_to(self, node)
 * 
 *     def neighbors(self):             # <<<<<<<<<<<<<<
 *         result = [edge._traverse(self) for edge in self.edges]
 *         return result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.neighbors", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_edge);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":56
 *         return result
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphNode other_typed
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_7__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_7__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6__eq__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_other_typed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":59
 *         cdef:
 *             CompositionGraphNode other_typed
 *         if isinstance(self, CompositionGraphNode):             # <<<<<<<<<<<<<<
 *             if isinstance(other, CompositionGraphNode):
 *                 other_typed = <CompositionGraphNode>other
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode); 
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":60
 *             CompositionGraphNode other_typed
 *         if isinstance(self, CompositionGraphNode):
 *             if isinstance(other, CompositionGraphNode):             # <<<<<<<<<<<<<<
 *                 other_typed = <CompositionGraphNode>other
 *                 return self._str == other_typed._str
 */
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode); 
    if (__pyx_t_1) {

      /* "glycresoft/_c/composition_network/graph.pyx":61
 *         if isinstance(self, CompositionGraphNode):
 *             if isinstance(other, CompositionGraphNode):
 *                 other_typed = <CompositionGraphNode>other             # <<<<<<<<<<<<<<
 *                 return self._str == other_typed._str
 *             else:
 */
      __pyx_t_2 = __pyx_v_other;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_other_typed = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":62
 *             if isinstance(other, CompositionGraphNode):
 *                 other_typed = <CompositionGraphNode>other
 *                 return self._str == other_typed._str             # <<<<<<<<<<<<<<
 *             else:
 *                 return self._str == str(other)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_self->_str, __pyx_v_other_typed->_str, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 62, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "glycresoft/_c/composition_network/graph.pyx":60
 *             CompositionGraphNode other_typed
 *         if isinstance(self, CompositionGraphNode):
 *             if isinstance(other, CompositionGraphNode):             # <<<<<<<<<<<<<<
 *                 other_typed = <CompositionGraphNode>other
 *                 return self._str == other_typed._str
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":64
 *                 return self._str == other_typed._str
 *             else:
 *                 return self._str == str(other)             # <<<<<<<<<<<<<<
 *         else:
 *             return str(self) == str(other)
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->_str, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;
    }

    /* "glycresoft/_c/composition_network/graph.pyx":59
 *         cdef:
 *             CompositionGraphNode other_typed
 *         if isinstance(self, CompositionGraphNode):             # <<<<<<<<<<<<<<
 *             if isinstance(other, CompositionGraphNode):
 *                 other_typed = <CompositionGraphNode>other
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":66
 *                 return self._str == str(other)
 *         else:
 *             return str(self) == str(other)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_Str(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":56
 *         return result
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphNode other_typed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_other_typed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":68
 *             return str(self) == str(other)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._str
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_9__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_9__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_8__str__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_8__str__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":69
 * 
 *     def __str__(self):
 *         return self._str             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_str);
  __pyx_r = __pyx_v_self->_str;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":68
 *             return str(self) == str(other)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._str
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":71
 *         return self._str
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_10__hash__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_10__hash__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  Py_hash_t __pyx_r;

  /* "glycresoft/_c/composition_network/graph.pyx":72
 * 
 *     def __hash__(self):
 *         return self._hash             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_hash;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":71
 *         return self._str
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

  /* function exit code */
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":74
 *         return self._hash
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (
 *             self._str, int(self.index) if self.index is not None else -1,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_12__repr__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_12__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":75
 * 
 *     def __repr__(self):
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (             # <<<<<<<<<<<<<<
 *             self._str, int(self.index) if self.index is not None else -1,
 *             self.score if self.score != 0 else self.internal_score)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycresoft/_c/composition_network/graph.pyx":76
 *     def __repr__(self):
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (
 *             self._str, int(self.index) if self.index is not None else -1,             # <<<<<<<<<<<<<<
 *             self.score if self.score != 0 else self.internal_score)
 * 
 */
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_int_neg_1);
    __pyx_t_1 = __pyx_int_neg_1;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":77
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (
 *             self._str, int(self.index) if self.index is not None else -1,
 *             self.score if self.score != 0 else self.internal_score)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->internal_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_t_2;
    __pyx_t_2 = 0;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":76
 *     def __repr__(self):
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (
 *             self._str, int(self.index) if self.index is not None else -1,             # <<<<<<<<<<<<<<
 *             self.score if self.score != 0 else self.internal_score)
 * 
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_str);
  __Pyx_GIVEREF(__pyx_v_self->_str);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_str)) __PYX_ERR(0, 76, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4)) __PYX_ERR(0, 76, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":75
 * 
 *     def __repr__(self):
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (             # <<<<<<<<<<<<<<
 *             self._str, int(self.index) if self.index is not None else -1,
 *             self.score if self.score != 0 else self.internal_score)
 */
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_CompositionGraphNode_s_d_0_2f, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":74
 *         return self._hash
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "CompositionGraphNode(%s, %d, %0.2f)" % (
 *             self._str, int(self.index) if self.index is not None else -1,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":79
 *             self.score if self.score != 0 else self.internal_score)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)
 *         dup.internal_score = self.internal_score
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_15copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14copy, "CompositionGraphNode.copy(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_15copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_15copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14copy};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_15copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14copy(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14copy(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_dup = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":80
 * 
 *     def copy(self):
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)             # <<<<<<<<<<<<<<
 *         dup.internal_score = self.internal_score
 *         return dup
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->composition);
  __Pyx_GIVEREF(__pyx_v_self->composition);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->composition)) __PYX_ERR(0, 80, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_dup = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":81
 *     def copy(self):
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)
 *         dup.internal_score = self.internal_score             # <<<<<<<<<<<<<<
 *         return dup
 * 
 */
  __pyx_t_4 = __pyx_v_self->internal_score;
  __pyx_v_dup->internal_score = __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":82
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)
 *         dup.internal_score = self.internal_score
 *         return dup             # <<<<<<<<<<<<<<
 * 
 *     def clone(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_dup);
  __pyx_r = ((PyObject *)__pyx_v_dup);
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":79
 *             self.score if self.score != 0 else self.internal_score)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)
 *         dup.internal_score = self.internal_score
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":84
 *         return dup
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_17clone(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_16clone, "CompositionGraphNode.clone(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_17clone = {"clone", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_17clone, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_16clone};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_17clone(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clone", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clone", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_16clone(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_16clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":85
 * 
 *     def clone(self):
 *         return self.copy()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":84
 *         return dup
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":3
 * cdef class CompositionGraphNode(object):
 *     cdef:
 *         public object composition             # <<<<<<<<<<<<<<
 *         public ssize_t index
 *         public str _str
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->composition);
  __pyx_r = __pyx_v_self->composition;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(__pyx_v_self->composition);
  __pyx_v_self->composition = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(__pyx_v_self->composition);
  __pyx_v_self->composition = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":4
 *     cdef:
 *         public object composition
 *         public ssize_t index             # <<<<<<<<<<<<<<
 *         public str _str
 *         public EdgeSet edges
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = PyInt_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":5
 *         public object composition
 *         public ssize_t index
 *         public str _str             # <<<<<<<<<<<<<<
 *         public EdgeSet edges
 *         public Py_hash_t _hash
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_str);
  __pyx_r = __pyx_v_self->_str;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_v_value))) __PYX_ERR(1, 5, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_str);
  __Pyx_DECREF(__pyx_v_self->_str);
  __pyx_v_self->_str = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode._str.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_str);
  __Pyx_DECREF(__pyx_v_self->_str);
  __pyx_v_self->_str = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":6
 *         public ssize_t index
 *         public str _str
 *         public EdgeSet edges             # <<<<<<<<<<<<<<
 *         public Py_hash_t _hash
 *         public double _score
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->edges);
  __pyx_r = ((PyObject *)__pyx_v_self->edges);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet))))) __PYX_ERR(1, 6, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->edges);
  __Pyx_DECREF((PyObject *)__pyx_v_self->edges);
  __pyx_v_self->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.edges.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->edges);
  __Pyx_DECREF((PyObject *)__pyx_v_self->edges);
  __pyx_v_self->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":7
 *         public str _str
 *         public EdgeSet edges
 *         public Py_hash_t _hash             # <<<<<<<<<<<<<<
 *         public double _score
 *         public double internal_score
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode._hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  Py_hash_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_AsHash_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_v_self->_hash = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode._hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":8
 *         public EdgeSet edges
 *         public Py_hash_t _hash
 *         public double _score             # <<<<<<<<<<<<<<
 *         public double internal_score
 *         public bint marked
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode._score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 8, __pyx_L1_error)
  __pyx_v_self->_score = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode._score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":9
 *         public Py_hash_t _hash
 *         public double _score
 *         public double internal_score             # <<<<<<<<<<<<<<
 *         public bint marked
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->internal_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.internal_score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 9, __pyx_L1_error)
  __pyx_v_self->internal_score = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.internal_score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":10
 *         public double _score
 *         public double internal_score
 *         public bint marked             # <<<<<<<<<<<<<<
 * 
 * cdef class EdgeSet(object):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->marked); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.marked.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v_self->marked = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.marked.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18__reduce_cython__, "CompositionGraphNode.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._hash, self._score, self._str, self.composition, self.edges, self.index, self.internal_score, self.marked)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->internal_score); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->marked); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(8); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_str);
  __Pyx_GIVEREF(__pyx_v_self->_str);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_self->_str)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->composition);
  __Pyx_GIVEREF(__pyx_v_self->composition);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_self->composition)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->edges);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->edges);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, ((PyObject *)__pyx_v_self->edges))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 7, __pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._hash, self._score, self._str, self.composition, self.edges, self.index, self.internal_score, self.marked)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_6 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v__dict = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "(tree fragment)":7
 *     state = (self._hash, self._score, self._str, self.composition, self.edges, self.index, self.internal_score, self.marked)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_7 = (__pyx_v__dict != Py_None);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._str is not None or self.composition is not None or self.edges is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._hash, self._score, self._str, self.composition, self.edges, self.index, self.internal_score, self.marked)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._str is not None or self.composition is not None or self.edges is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, None), state
 */
  /*else*/ {
    __pyx_t_8 = (__pyx_v_self->_str != ((PyObject*)Py_None));
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->composition != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->edges) != Py_None);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._str is not None or self.composition is not None or self.edges is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._str is not None or self.composition is not None or self.edges is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_CompositionGraphN); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_19304315);
    __Pyx_GIVEREF(__pyx_int_19304315);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_19304315)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._str is not None or self.composition is not None or self.edges is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, None), state
 *     else:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CompositionGraphNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_CompositionGraphN); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_19304315);
    __Pyx_GIVEREF(__pyx_int_19304315);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_19304315)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_20__setstate_cython__, "CompositionGraphNode.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_20__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_20__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_20__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CompositionGraphNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":90
 * cdef class EdgeSet(object):
 * 
 *     def __init__(self, store=None):             # <<<<<<<<<<<<<<
 *         if store is None:
 *             store = dict()
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_store = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_store,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_store);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 90, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 90, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_store = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 90, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet___init__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), __pyx_v_store);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v_store) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_store);

  /* "glycresoft/_c/composition_network/graph.pyx":91
 * 
 *     def __init__(self, store=None):
 *         if store is None:             # <<<<<<<<<<<<<<
 *             store = dict()
 *         self.store = dict(store)
 */
  __pyx_t_1 = (__pyx_v_store == Py_None);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":92
 *     def __init__(self, store=None):
 *         if store is None:
 *             store = dict()             # <<<<<<<<<<<<<<
 *         self.store = dict(store)
 * 
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_store, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":91
 * 
 *     def __init__(self, store=None):
 *         if store is None:             # <<<<<<<<<<<<<<
 *             store = dict()
 *         self.store = dict(store)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":93
 *         if store is None:
 *             store = dict()
 *         self.store = dict(store)             # <<<<<<<<<<<<<<
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_store); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->store);
  __Pyx_DECREF(__pyx_v_self->store);
  __pyx_v_self->store = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":90
 * cdef class EdgeSet(object):
 * 
 *     def __init__(self, store=None):             # <<<<<<<<<<<<<<
 *         if store is None:
 *             store = dict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_store);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":95
 *         self.store = dict(store)
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         if node2.index < node1.index:
 *             node1, node2 = node2, node1
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_3edge_to(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_2edge_to, "EdgeSet.edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_3edge_to = {"edge_to", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_3edge_to, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_2edge_to};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_3edge_to(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1 = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("edge_to (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node1,&__pyx_n_s_node2,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node1)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 95, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 95, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("edge_to", 1, 2, 2, 1); __PYX_ERR(0, 95, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "edge_to") < 0)) __PYX_ERR(0, 95, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)values[0]);
    __pyx_v_node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("edge_to", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 95, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.edge_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node1), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, 1, "node1", 0))) __PYX_ERR(0, 95, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node2), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, 1, "node2", 0))) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_2edge_to(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), __pyx_v_node1, __pyx_v_node2);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_2edge_to(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("edge_to", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_node1);
  __Pyx_INCREF((PyObject *)__pyx_v_node2);

  /* "glycresoft/_c/composition_network/graph.pyx":96
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):
 *         if node2.index < node1.index:             # <<<<<<<<<<<<<<
 *             node1, node2 = node2, node1
 *         return self.store[node1, node2]
 */
  __pyx_t_1 = (__pyx_v_node2->index < __pyx_v_node1->index);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":97
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):
 *         if node2.index < node1.index:
 *             node1, node2 = node2, node1             # <<<<<<<<<<<<<<
 *         return self.store[node1, node2]
 * 
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_node2);
    __pyx_t_3 = ((PyObject *)__pyx_v_node1);
    __pyx_v_node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":96
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):
 *         if node2.index < node1.index:             # <<<<<<<<<<<<<<
 *             node1, node2 = node2, node1
 *         return self.store[node1, node2]
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":98
 *         if node2.index < node1.index:
 *             node1, node2 = node2, node1
 *         return self.store[node1, node2]             # <<<<<<<<<<<<<<
 * 
 *     def add(self, CompositionGraphEdge edge):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->store == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 98, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_v_node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_node1))) __PYX_ERR(0, 98, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_node2))) __PYX_ERR(0, 98, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_self->store, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":95
 *         self.store = dict(store)
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         if node2.index < node1.index:
 *             node1, node2 = node2, node1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.edge_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node1);
  __Pyx_XDECREF((PyObject *)__pyx_v_node2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":100
 *         return self.store[node1, node2]
 * 
 *     def add(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store[edge.node1, edge.node2] = edge
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_4add, "EdgeSet.add(self, CompositionGraphEdge edge)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5add = {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5add, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_4add};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_edge,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_edge)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add") < 0)) __PYX_ERR(0, 100, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_edge = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 100, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_edge), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, 1, "edge", 0))) __PYX_ERR(0, 100, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_4add(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), __pyx_v_edge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_4add(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":101
 * 
 *     def add(self, CompositionGraphEdge edge):
 *         self.store[edge.node1, edge.node2] = edge             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint add_if_shorter(self, CompositionGraphEdge edge):
 */
  if (unlikely(__pyx_v_self->store == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 101, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_edge->node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_edge->node1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_edge->node1))) __PYX_ERR(0, 101, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_edge->node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_edge->node2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_edge->node2))) __PYX_ERR(0, 101, __pyx_L1_error);
  if (unlikely((PyDict_SetItem(__pyx_v_self->store, __pyx_t_1, ((PyObject *)__pyx_v_edge)) < 0))) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":100
 *         return self.store[node1, node2]
 * 
 *     def add(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store[edge.node1, edge.node2] = edge
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":103
 *         self.store[edge.node1, edge.node2] = edge
 * 
 *     cpdef bint add_if_shorter(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_add_if_shorter(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_prev_ptr;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_prev = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_if_shorter", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add_if_shorter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_edge)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 103, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":109
 *             CompositionGraphEdge prev
 * 
 *         key = (edge.node1, edge.node2)             # <<<<<<<<<<<<<<
 *         prev_ptr = PyDict_GetItem(self.store, key)
 *         if prev_ptr == NULL:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_edge->node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_edge->node1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_edge->node1))) __PYX_ERR(0, 109, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_edge->node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_edge->node2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_edge->node2))) __PYX_ERR(0, 109, __pyx_L1_error);
  __pyx_v_key = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":110
 * 
 *         key = (edge.node1, edge.node2)
 *         prev_ptr = PyDict_GetItem(self.store, key)             # <<<<<<<<<<<<<<
 *         if prev_ptr == NULL:
 *             PyDict_SetItem(self.store, key, edge)
 */
  __pyx_t_1 = __pyx_v_self->store;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_prev_ptr = PyDict_GetItem(__pyx_t_1, __pyx_v_key);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":111
 *         key = (edge.node1, edge.node2)
 *         prev_ptr = PyDict_GetItem(self.store, key)
 *         if prev_ptr == NULL:             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self.store, key, edge)
 *             return True
 */
  __pyx_t_6 = (__pyx_v_prev_ptr == NULL);
  if (__pyx_t_6) {

    /* "glycresoft/_c/composition_network/graph.pyx":112
 *         prev_ptr = PyDict_GetItem(self.store, key)
 *         if prev_ptr == NULL:
 *             PyDict_SetItem(self.store, key, edge)             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
    __pyx_t_1 = __pyx_v_self->store;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = PyDict_SetItem(__pyx_t_1, __pyx_v_key, ((PyObject *)__pyx_v_edge)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":113
 *         if prev_ptr == NULL:
 *             PyDict_SetItem(self.store, key, edge)
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             prev = <CompositionGraphEdge>prev_ptr
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":111
 *         key = (edge.node1, edge.node2)
 *         prev_ptr = PyDict_GetItem(self.store, key)
 *         if prev_ptr == NULL:             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self.store, key, edge)
 *             return True
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":115
 *             return True
 *         else:
 *             prev = <CompositionGraphEdge>prev_ptr             # <<<<<<<<<<<<<<
 *             if prev.order < edge.order:
 *                 return False
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)__pyx_v_prev_ptr);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_prev = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":116
 *         else:
 *             prev = <CompositionGraphEdge>prev_ptr
 *             if prev.order < edge.order:             # <<<<<<<<<<<<<<
 *                 return False
 *             else:
 */
    __pyx_t_6 = (__pyx_v_prev->order < __pyx_v_edge->order);
    if (__pyx_t_6) {

      /* "glycresoft/_c/composition_network/graph.pyx":117
 *             prev = <CompositionGraphEdge>prev_ptr
 *             if prev.order < edge.order:
 *                 return False             # <<<<<<<<<<<<<<
 *             else:
 *                 PyDict_SetItem(self.store, key, edge)
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "glycresoft/_c/composition_network/graph.pyx":116
 *         else:
 *             prev = <CompositionGraphEdge>prev_ptr
 *             if prev.order < edge.order:             # <<<<<<<<<<<<<<
 *                 return False
 *             else:
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":119
 *                 return False
 *             else:
 *                 PyDict_SetItem(self.store, key, edge)             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    /*else*/ {
      __pyx_t_1 = __pyx_v_self->store;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = PyDict_SetItem(__pyx_t_1, __pyx_v_key, ((PyObject *)__pyx_v_edge)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":120
 *             else:
 *                 PyDict_SetItem(self.store, key, edge)
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *     def remove(self, CompositionGraphEdge edge):
 */
      __pyx_r = 1;
      goto __pyx_L0;
    }
  }

  /* "glycresoft/_c/composition_network/graph.pyx":103
 *         self.store[edge.node1, edge.node2] = edge
 * 
 *     cpdef bint add_if_shorter(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.add_if_shorter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_prev);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_6add_if_shorter, "EdgeSet.add_if_shorter(self, CompositionGraphEdge edge) -> bool");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter = {"add_if_shorter", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_6add_if_shorter};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_if_shorter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_edge,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_edge)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 103, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_if_shorter") < 0)) __PYX_ERR(0, 103, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_edge = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_if_shorter", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 103, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.add_if_shorter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_edge), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, 1, "edge", 0))) __PYX_ERR(0, 103, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_6add_if_shorter(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), __pyx_v_edge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_6add_if_shorter(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_if_shorter", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_add_if_shorter(__pyx_v_self, __pyx_v_edge, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 103, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.add_if_shorter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":122
 *                 return True
 * 
 *     def remove(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store.pop((edge.node1, edge.node2))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_9remove(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_8remove, "EdgeSet.remove(self, CompositionGraphEdge edge)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_9remove = {"remove", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_9remove, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_8remove};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_9remove(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_edge,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_edge)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove") < 0)) __PYX_ERR(0, 122, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_edge = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 122, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_edge), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, 1, "edge", 0))) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_8remove(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), __pyx_v_edge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_8remove(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":123
 * 
 *     def remove(self, CompositionGraphEdge edge):
 *         self.store.pop((edge.node1, edge.node2))             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  if (unlikely(__pyx_v_self->store == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 123, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_edge->node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_edge->node1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_edge->node1))) __PYX_ERR(0, 123, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_edge->node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_edge->node2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_edge->node2))) __PYX_ERR(0, 123, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyDict_Pop(__pyx_v_self->store, __pyx_t_1, ((PyObject *)NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":122
 *                 return True
 * 
 *     def remove(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store.pop((edge.node1, edge.node2))
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":125
 *         self.store.pop((edge.node1, edge.node2))
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.store.values())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_11__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_10__iter__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_10__iter__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":126
 * 
 *     def __iter__(self):
 *         return iter(self.store.values())             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->store == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
    __PYX_ERR(0, 126, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Values(__pyx_v_self->store); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":125
 *         self.store.pop((edge.node1, edge.node2))
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.store.values())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":128
 *         return iter(self.store.values())
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.store)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_13__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_13__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_12__len__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_12__len__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":129
 * 
 *     def __len__(self):
 *         return len(self.store)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_1 = __pyx_v_self->store;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 129, __pyx_L1_error)
  }
  __pyx_t_2 = PyDict_Size(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":128
 *         return iter(self.store.values())
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.store)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":131
 *         return len(self.store)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(set(self.store.values()))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_14__repr__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_14__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":132
 * 
 *     def __repr__(self):
 *         return str(set(self.store.values()))             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->store == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
    __PYX_ERR(0, 132, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Values(__pyx_v_self->store); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":131
 *         return len(self.store)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(set(self.store.values()))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":134
 *         return str(set(self.store.values()))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self.store == other.store
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_17__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_17__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_16__eq__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_16__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":135
 * 
 *     def __eq__(self, other):
 *         return self.store == other.store             # <<<<<<<<<<<<<<
 * 
 *     cdef list get_edges(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_store); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->store, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":134
 *         return str(set(self.store.values()))
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self.store == other.store
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":137
 *         return self.store == other.store
 * 
 *     cdef list get_edges(self):             # <<<<<<<<<<<<<<
 *         return PyDict_Values(self.store)
 * 
 */

static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_get_edges(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_edges", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":138
 * 
 *     cdef list get_edges(self):
 *         return PyDict_Values(self.store)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->store;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Values(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":137
 *         return self.store == other.store
 * 
 *     cdef list get_edges(self):             # <<<<<<<<<<<<<<
 *         return PyDict_Values(self.store)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.get_edges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":14
 * cdef class EdgeSet(object):
 *     cdef:
 *         public dict store             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint add_if_shorter(self, CompositionGraphEdge edge)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->store);
  __pyx_r = __pyx_v_self->store;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(1, 14, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->store);
  __Pyx_DECREF(__pyx_v_self->store);
  __pyx_v_self->store = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.store.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->store);
  __Pyx_DECREF(__pyx_v_self->store);
  __pyx_v_self->store = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_18__reduce_cython__, "EdgeSet.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_18__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_18__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_18__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.store,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->store);
  __Pyx_GIVEREF(__pyx_v_self->store);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->store)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.store,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.store,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.store is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.store,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.store is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->store != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.store is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.store is not None
 *     if use_setstate:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_EdgeSet); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136632333);
    __Pyx_GIVEREF(__pyx_int_136632333);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136632333)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.store is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, None), state
 *     else:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EdgeSet__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_EdgeSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136632333);
    __Pyx_GIVEREF(__pyx_int_136632333);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136632333)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EdgeSet__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_20__setstate_cython__, "EdgeSet.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_20__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_20__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_20__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EdgeSet__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_EdgeSet__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EdgeSet__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.EdgeSet.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":145
 * cdef class CompositionGraphEdge(object):
 * 
 *     def __init__(self, node1, node2, order, weight=1.0):             # <<<<<<<<<<<<<<
 *         self.node1 = node1
 *         self.node2 = node2
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node1 = 0;
  PyObject *__pyx_v_node2 = 0;
  PyObject *__pyx_v_order = 0;
  PyObject *__pyx_v_weight = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node1,&__pyx_n_s_node2,&__pyx_n_s_order,&__pyx_n_s_weight,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_float_1_0));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node1)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node2)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(0, 145, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(0, 145, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_weight);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 145, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_node1 = values[0];
    __pyx_v_node2 = values[1];
    __pyx_v_order = values[2];
    __pyx_v_weight = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 145, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge___init__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), __pyx_v_node1, __pyx_v_node2, __pyx_v_order, __pyx_v_weight);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_node1, PyObject *__pyx_v_node2, PyObject *__pyx_v_order, PyObject *__pyx_v_weight) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  double __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":146
 * 
 *     def __init__(self, node1, node2, order, weight=1.0):
 *         self.node1 = node1             # <<<<<<<<<<<<<<
 *         self.node2 = node2
 *         self.order = order if order > 0 else 1
 */
  if (!(likely(((__pyx_v_node1) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_node1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 146, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_node1;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node1);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node1);
  __pyx_v_self->node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":147
 *     def __init__(self, node1, node2, order, weight=1.0):
 *         self.node1 = node1
 *         self.node2 = node2             # <<<<<<<<<<<<<<
 *         self.order = order if order > 0 else 1
 *         self.weight = weight
 */
  if (!(likely(((__pyx_v_node2) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_node2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 147, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_node2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node2);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node2);
  __pyx_v_self->node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":148
 *         self.node1 = node1
 *         self.node2 = node2
 *         self.order = order if order > 0 else 1             # <<<<<<<<<<<<<<
 *         self.weight = weight
 *         self._init()
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_order, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
    __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_v_order); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_4;
  } else {
    __pyx_t_2 = 1;
  }
  __pyx_v_self->order = __pyx_t_2;

  /* "glycresoft/_c/composition_network/graph.pyx":149
 *         self.node2 = node2
 *         self.order = order if order > 0 else 1
 *         self.weight = weight             # <<<<<<<<<<<<<<
 *         self._init()
 * 
 */
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_v_weight); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_v_self->weight = __pyx_t_5;

  /* "glycresoft/_c/composition_network/graph.pyx":150
 *         self.order = order if order > 0 else 1
 *         self.weight = weight
 *         self._init()             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self->__pyx_vtab)->_init(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":145
 * cdef class CompositionGraphEdge(object):
 * 
 *     def __init__(self, node1, node2, order, weight=1.0):             # <<<<<<<<<<<<<<
 *         self.node1 = node1
 *         self.node2 = node2
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":153
 * 
 *     @staticmethod
 *     cdef CompositionGraphEdge _create(CompositionGraphNode node1, CompositionGraphNode node2, long order, double weight):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphEdge self
 */

static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__create(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2, long __pyx_v_order, double __pyx_v_weight) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":157
 *             CompositionGraphEdge self
 * 
 *         self = CompositionGraphEdge.__new__(CompositionGraphEdge)             # <<<<<<<<<<<<<<
 *         self.node1 = node1
 *         self.node2 = node2
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":158
 * 
 *         self = CompositionGraphEdge.__new__(CompositionGraphEdge)
 *         self.node1 = node1             # <<<<<<<<<<<<<<
 *         self.node2 = node2
 *         self.order = order
 */
  __Pyx_INCREF((PyObject *)__pyx_v_node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node1);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node1);
  __pyx_v_self->node1 = __pyx_v_node1;

  /* "glycresoft/_c/composition_network/graph.pyx":159
 *         self = CompositionGraphEdge.__new__(CompositionGraphEdge)
 *         self.node1 = node1
 *         self.node2 = node2             # <<<<<<<<<<<<<<
 *         self.order = order
 *         self.weight = weight
 */
  __Pyx_INCREF((PyObject *)__pyx_v_node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node2);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node2);
  __pyx_v_self->node2 = __pyx_v_node2;

  /* "glycresoft/_c/composition_network/graph.pyx":160
 *         self.node1 = node1
 *         self.node2 = node2
 *         self.order = order             # <<<<<<<<<<<<<<
 *         self.weight = weight
 *         self._init()
 */
  __pyx_v_self->order = __pyx_v_order;

  /* "glycresoft/_c/composition_network/graph.pyx":161
 *         self.node2 = node2
 *         self.order = order
 *         self.weight = weight             # <<<<<<<<<<<<<<
 *         self._init()
 *         return self
 */
  __pyx_v_self->weight = __pyx_v_weight;

  /* "glycresoft/_c/composition_network/graph.pyx":162
 *         self.order = order
 *         self.weight = weight
 *         self._init()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self->__pyx_vtab)->_init(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":163
 *         self.weight = weight
 *         self._init()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef void _init(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":153
 * 
 *     @staticmethod
 *     cdef CompositionGraphEdge _create(CompositionGraphNode node1, CompositionGraphNode node2, long order, double weight):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphEdge self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":165
 *         return self
 * 
 *     cdef void _init(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             tuple key
 */

static void __pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__init(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_v_key = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_hash_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":168
 *         cdef:
 *             tuple key
 *         key = (self.node1, self.node2, self.order)             # <<<<<<<<<<<<<<
 *         self._hash = hash(key)
 *         self._str = "(%s, %s, %s)" % key
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->order); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->node1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->node1))) __PYX_ERR(0, 168, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->node2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->node2))) __PYX_ERR(0, 168, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_key = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":169
 *             tuple key
 *         key = (self.node1, self.node2, self.order)
 *         self._hash = hash(key)             # <<<<<<<<<<<<<<
 *         self._str = "(%s, %s, %s)" % key
 *         self.node1.edges.add_if_shorter(self)
 */
  __pyx_t_3 = PyObject_Hash(__pyx_v_key); if (unlikely(__pyx_t_3 == ((Py_hash_t)-1))) __PYX_ERR(0, 169, __pyx_L1_error)
  __pyx_v_self->_hash = __pyx_t_3;

  /* "glycresoft/_c/composition_network/graph.pyx":170
 *         key = (self.node1, self.node2, self.order)
 *         self._hash = hash(key)
 *         self._str = "(%s, %s, %s)" % key             # <<<<<<<<<<<<<<
 *         self.node1.edges.add_if_shorter(self)
 *         self.node2.edges.add_if_shorter(self)
 */
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_s_s_s, __pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyString_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_str);
  __Pyx_DECREF(__pyx_v_self->_str);
  __pyx_v_self->_str = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":171
 *         self._hash = hash(key)
 *         self._str = "(%s, %s, %s)" % key
 *         self.node1.edges.add_if_shorter(self)             # <<<<<<<<<<<<<<
 *         self.node2.edges.add_if_shorter(self)
 * 
 */
  ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self->node1->edges->__pyx_vtab)->add_if_shorter(__pyx_v_self->node1->edges, __pyx_v_self, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":172
 *         self._str = "(%s, %s, %s)" % key
 *         self.node1.edges.add_if_shorter(self)
 *         self.node2.edges.add_if_shorter(self)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self->node2->edges->__pyx_vtab)->add_if_shorter(__pyx_v_self->node2->edges, __pyx_v_self, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":165
 *         return self
 * 
 *     cdef void _init(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             tuple key
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_RefNannyFinishContext();
}

/* "glycresoft/_c/composition_network/graph.pyx":174
 *         self.node2.edges.add_if_shorter(self)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         str_key = str(key)
 *         if str_key == self.node1._str:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_2__getitem__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_2__getitem__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_str_key = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":175
 * 
 *     def __getitem__(self, key):
 *         str_key = str(key)             # <<<<<<<<<<<<<<
 *         if str_key == self.node1._str:
 *             return self.node2
 */
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_str_key = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":176
 *     def __getitem__(self, key):
 *         str_key = str(key)
 *         if str_key == self.node1._str:             # <<<<<<<<<<<<<<
 *             return self.node2
 *         elif str_key == self.node2._str:
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_str_key, __pyx_v_self->node1->_str, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":177
 *         str_key = str(key)
 *         if str_key == self.node1._str:
 *             return self.node2             # <<<<<<<<<<<<<<
 *         elif str_key == self.node2._str:
 *             return self.node1
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->node2);
    __pyx_r = ((PyObject *)__pyx_v_self->node2);
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":176
 *     def __getitem__(self, key):
 *         str_key = str(key)
 *         if str_key == self.node1._str:             # <<<<<<<<<<<<<<
 *             return self.node2
 *         elif str_key == self.node2._str:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":178
 *         if str_key == self.node1._str:
 *             return self.node2
 *         elif str_key == self.node2._str:             # <<<<<<<<<<<<<<
 *             return self.node1
 *         else:
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_str_key, __pyx_v_self->node2->_str, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 178, __pyx_L1_error)
  if (likely(__pyx_t_2)) {

    /* "glycresoft/_c/composition_network/graph.pyx":179
 *             return self.node2
 *         elif str_key == self.node2._str:
 *             return self.node1             # <<<<<<<<<<<<<<
 *         else:
 *             raise KeyError(key)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->node1);
    __pyx_r = ((PyObject *)__pyx_v_self->node1);
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":178
 *         if str_key == self.node1._str:
 *             return self.node2
 *         elif str_key == self.node2._str:             # <<<<<<<<<<<<<<
 *             return self.node1
 *         else:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":181
 *             return self.node1
 *         else:
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 * 
 *     def _traverse(self, node):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 181, __pyx_L1_error)
  }

  /* "glycresoft/_c/composition_network/graph.pyx":174
 *         self.node2.edges.add_if_shorter(self)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         str_key = str(key)
 *         if str_key == self.node1._str:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_str_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":183
 *             raise KeyError(key)
 * 
 *     def _traverse(self, node):             # <<<<<<<<<<<<<<
 *         return self.node1 if node is self.node2 else self.node2
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_traverse(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_traverse, "CompositionGraphEdge._traverse(self, node)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_traverse = {"_traverse", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_traverse, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_traverse};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_traverse(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_node = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_traverse (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_traverse") < 0)) __PYX_ERR(0, 183, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_node = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_traverse", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 183, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge._traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_traverse(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), __pyx_v_node);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_traverse(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_traverse", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":184
 * 
 *     def _traverse(self, node):
 *         return self.node1 if node is self.node2 else self.node2             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_node == ((PyObject *)__pyx_v_self->node2));
  if (__pyx_t_2) {
    __Pyx_INCREF((PyObject *)__pyx_v_self->node1);
    __pyx_t_1 = ((PyObject *)__pyx_v_self->node1);
  } else {
    __Pyx_INCREF((PyObject *)__pyx_v_self->node2);
    __pyx_t_1 = ((PyObject *)__pyx_v_self->node2);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":183
 *             raise KeyError(key)
 * 
 *     def _traverse(self, node):             # <<<<<<<<<<<<<<
 *         return self.node1 if node is self.node2 else self.node2
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":186
 *         return self.node1 if node is self.node2 else self.node2
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self._str == other._str
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_7__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_7__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6__eq__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":187
 * 
 *     def __eq__(self, other):
 *         return self._str == other._str             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->_str, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":186
 *         return self.node1 if node is self.node2 else self.node2
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self._str == other._str
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":189
 *         return self._str == other._str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._str
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_9__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_9__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_8__str__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_8__str__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":190
 * 
 *     def __str__(self):
 *         return self._str             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_str);
  __pyx_r = __pyx_v_self->_str;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":189
 *         return self._str == other._str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self._str
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":192
 *         return self._str
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self._str
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_10__repr__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_10__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":193
 * 
 *     def __repr__(self):
 *         return self._str             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_str);
  __pyx_r = __pyx_v_self->_str;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":192
 *         return self._str
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self._str
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":195
 *         return self._str
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_13__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_13__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_12__hash__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_12__hash__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  Py_hash_t __pyx_r;

  /* "glycresoft/_c/composition_network/graph.pyx":196
 * 
 *     def __hash__(self):
 *         return self._hash             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_r = __pyx_v_self->_hash;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":195
 *         return self._str
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

  /* function exit code */
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":198
 *         return self._hash
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_15__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_14__reduce__, "CompositionGraphEdge.__reduce__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_15__reduce__ = {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_15__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_14__reduce__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_15__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_14__reduce__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_14__reduce__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":199
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)             # <<<<<<<<<<<<<<
 * 
 *     cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v_self->order); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->weight); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->node1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self->node1))) __PYX_ERR(0, 199, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->node2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self->node2))) __PYX_ERR(0, 199, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":198
 *         return self._hash
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":201
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 *     cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_copy_for(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_for", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy_for); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_v_node1), ((PyObject *)__pyx_v_node2)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":202
 * 
 *     cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)             # <<<<<<<<<<<<<<
 * 
 *     def remove(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__create(__pyx_v_node1, __pyx_v_node2, __pyx_v_self->order, __pyx_v_self->weight)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":201
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 *     cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.copy_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_16copy_for, "CompositionGraphEdge.copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for = {"copy_for", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_16copy_for};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1 = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_for (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node1,&__pyx_n_s_node2,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node1)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 201, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 201, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("copy_for", 1, 2, 2, 1); __PYX_ERR(0, 201, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "copy_for") < 0)) __PYX_ERR(0, 201, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)values[0]);
    __pyx_v_node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copy_for", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 201, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.copy_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node1), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, 1, "node1", 0))) __PYX_ERR(0, 201, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node2), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, 1, "node2", 0))) __PYX_ERR(0, 201, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_16copy_for(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), __pyx_v_node1, __pyx_v_node2);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_16copy_for(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node1, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_for", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_copy_for(__pyx_v_self, __pyx_v_node1, __pyx_v_node2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.copy_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":204
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 *     def remove(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.node1.edges.remove(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_19remove(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_18remove, "CompositionGraphEdge.remove(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_19remove = {"remove", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_19remove, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_18remove};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_19remove(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("remove", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "remove", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_18remove(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_18remove(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":205
 * 
 *     def remove(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.node1.edges.remove(self)
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glycresoft/_c/composition_network/graph.pyx":206
 *     def remove(self):
 *         try:
 *             self.node1.edges.remove(self)             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             pass
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->node1->edges), __pyx_n_s_remove); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)__pyx_v_self)};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":205
 * 
 *     def remove(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.node1.edges.remove(self)
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":207
 *         try:
 *             self.node1.edges.remove(self)
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             pass
 *         try:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_8) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "glycresoft/_c/composition_network/graph.pyx":205
 * 
 *     def remove(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.node1.edges.remove(self)
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":209
 *         except KeyError:
 *             pass
 *         try:             # <<<<<<<<<<<<<<
 *             self.node2.edges.remove(self)
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "glycresoft/_c/composition_network/graph.pyx":210
 *             pass
 *         try:
 *             self.node2.edges.remove(self)             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             pass
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->node2->edges), __pyx_n_s_remove); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 210, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)__pyx_v_self)};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 210, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":209
 *         except KeyError:
 *             pass
 *         try:             # <<<<<<<<<<<<<<
 *             self.node2.edges.remove(self)
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":211
 *         try:
 *             self.node2.edges.remove(self)
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_8) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L10_exception_handled;
    }
    goto __pyx_L11_except_error;

    /* "glycresoft/_c/composition_network/graph.pyx":209
 *         except KeyError:
 *             pass
 *         try:             # <<<<<<<<<<<<<<
 *             self.node2.edges.remove(self)
 *         except KeyError:
 */
    __pyx_L11_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L10_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L14_try_end:;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":204
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 *     def remove(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.node1.edges.remove(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":22
 * cdef class CompositionGraphEdge(object):
 *     cdef:
 *         public CompositionGraphNode node1             # <<<<<<<<<<<<<<
 *         public CompositionGraphNode node2
 *         public long order
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node1);
  __pyx_r = ((PyObject *)__pyx_v_self->node1);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node1);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node1);
  __pyx_v_self->node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.node1.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node1);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node1);
  __pyx_v_self->node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":23
 *     cdef:
 *         public CompositionGraphNode node1
 *         public CompositionGraphNode node2             # <<<<<<<<<<<<<<
 *         public long order
 *         public double weight
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node2);
  __pyx_r = ((PyObject *)__pyx_v_self->node2);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(1, 23, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node2);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node2);
  __pyx_v_self->node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.node2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->node2);
  __Pyx_DECREF((PyObject *)__pyx_v_self->node2);
  __pyx_v_self->node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":24
 *         public CompositionGraphNode node1
 *         public CompositionGraphNode node2
 *         public long order             # <<<<<<<<<<<<<<
 *         public double weight
 *         public Py_hash_t _hash
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->order); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.order.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 24, __pyx_L1_error)
  __pyx_v_self->order = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.order.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":25
 *         public CompositionGraphNode node2
 *         public long order
 *         public double weight             # <<<<<<<<<<<<<<
 *         public Py_hash_t _hash
 *         public str _str
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L1_error)
  __pyx_v_self->weight = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge.weight.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":26
 *         public long order
 *         public double weight
 *         public Py_hash_t _hash             # <<<<<<<<<<<<<<
 *         public str _str
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge._hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  Py_hash_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_AsHash_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 26, __pyx_L1_error)
  __pyx_v_self->_hash = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge._hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":27
 *         public double weight
 *         public Py_hash_t _hash
 *         public str _str             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_str);
  __pyx_r = __pyx_v_self->_str;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_v_value))) __PYX_ERR(1, 27, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_str);
  __Pyx_DECREF(__pyx_v_self->_str);
  __pyx_v_self->_str = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphEdge._str.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_str);
  __Pyx_DECREF(__pyx_v_self->_str);
  __pyx_v_self->_str = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":216
 * 
 * @cython.binding(True)
 * cpdef reindex_graph(self):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_1reindex_graph(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_reindex_graph(PyObject *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node = 0;
  PyObject *__pyx_v_nodes = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reindex_graph", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":222
 *         list nodes
 * 
 *     nodes = self.nodes             # <<<<<<<<<<<<<<
 *     n = PyList_Size(nodes)
 *     for i in range(n):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 222, __pyx_L1_error)
  __pyx_v_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":223
 * 
 *     nodes = self.nodes
 *     n = PyList_Size(nodes)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 */
  __pyx_t_2 = PyList_Size(__pyx_v_nodes); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 223, __pyx_L1_error)
  __pyx_v_n = __pyx_t_2;

  /* "glycresoft/_c/composition_network/graph.pyx":224
 *     nodes = self.nodes
 *     n = PyList_Size(nodes)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 *         node.index = i
 */
  __pyx_t_3 = __pyx_v_n;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "glycresoft/_c/composition_network/graph.pyx":225
 *     n = PyList_Size(nodes)
 *     for i in range(n):
 *         node = <CompositionGraphNode>PyList_GetItem(nodes, i)             # <<<<<<<<<<<<<<
 *         node.index = i
 * 
 */
    __pyx_t_6 = PyList_GetItem(__pyx_v_nodes, __pyx_v_i); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 225, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_t_6);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":226
 *     for i in range(n):
 *         node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 *         node.index = i             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_node->index = __pyx_v_i;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":216
 * 
 * @cython.binding(True)
 * cpdef reindex_graph(self):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.reindex_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_1reindex_graph(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_reindex_graph, "reindex_graph(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_1reindex_graph = {"reindex_graph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_1reindex_graph, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_reindex_graph};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_1reindex_graph(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reindex_graph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "reindex_graph") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reindex_graph", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 216, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.reindex_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_reindex_graph(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_reindex_graph(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reindex_graph", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_reindex_graph(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.reindex_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":233
 * 
 * cdef class CompositionGraphBase(object):
 *     cdef list get_edges(self):             # <<<<<<<<<<<<<<
 *         return self.edges.get_edges()
 * 
 */

static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_get_edges(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_edges", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":234
 * cdef class CompositionGraphBase(object):
 *     cdef list get_edges(self):
 *         return self.edges.get_edges()             # <<<<<<<<<<<<<<
 * 
 *     cpdef CompositionGraphBase copy(CompositionGraphBase self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v_self->edges->__pyx_vtab)->get_edges(__pyx_v_self->edges); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":233
 * 
 * cdef class CompositionGraphBase(object):
 *     cdef list get_edges(self):             # <<<<<<<<<<<<<<
 *         return self.edges.get_edges()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.get_edges", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":236
 *         return self.edges.get_edges()
 * 
 *     cpdef CompositionGraphBase copy(CompositionGraphBase self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphNode n1, n2
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_n1 = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_n2 = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_e = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_graph = 0;
  PyObject *__pyx_v_edges = 0;
  PyObject *__pyx_v_nodes = 0;
  size_t __pyx_v_n;
  size_t __pyx_v_i;
  PyObject *__pyx_v_node = NULL;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  PyObject *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 236, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase))))) __PYX_ERR(0, 236, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":244
 *             size_t n, i
 * 
 *         graph = self.__class__([], self.distance_fn)             # <<<<<<<<<<<<<<
 *         graph._composition_normalizer = self._composition_normalizer.copy()
 *         for node in self.nodes:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_self->distance_fn};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase))))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_v_graph = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":245
 * 
 *         graph = self.__class__([], self.distance_fn)
 *         graph._composition_normalizer = self._composition_normalizer.copy()             # <<<<<<<<<<<<<<
 *         for node in self.nodes:
 *             graph.add_node(node.clone())
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_composition_normalizer, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_graph->_composition_normalizer);
  __Pyx_DECREF(__pyx_v_graph->_composition_normalizer);
  __pyx_v_graph->_composition_normalizer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":246
 *         graph = self.__class__([], self.distance_fn)
 *         graph._composition_normalizer = self._composition_normalizer.copy()
 *         for node in self.nodes:             # <<<<<<<<<<<<<<
 *             graph.add_node(node.clone())
 * 
 */
  if (unlikely(__pyx_v_self->nodes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 246, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->nodes; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 246, __pyx_L1_error)
      #endif
      if (__pyx_t_6 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 246, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":247
 *         graph._composition_normalizer = self._composition_normalizer.copy()
 *         for node in self.nodes:
 *             graph.add_node(node.clone())             # <<<<<<<<<<<<<<
 * 
 *         nodes = graph.nodes
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_graph), __pyx_n_s_add_node); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_clone); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_t_7 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":246
 *         graph = self.__class__([], self.distance_fn)
 *         graph._composition_normalizer = self._composition_normalizer.copy()
 *         for node in self.nodes:             # <<<<<<<<<<<<<<
 *             graph.add_node(node.clone())
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":249
 *             graph.add_node(node.clone())
 * 
 *         nodes = graph.nodes             # <<<<<<<<<<<<<<
 *         edges = self.get_edges()
 * 
 */
  __pyx_t_1 = __pyx_v_graph->nodes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":250
 * 
 *         nodes = graph.nodes
 *         edges = self.get_edges()             # <<<<<<<<<<<<<<
 * 
 *         n = PyList_Size(edges)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self->__pyx_vtab)->get_edges(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_edges = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":252
 *         edges = self.get_edges()
 * 
 *         n = PyList_Size(edges)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             edge = <CompositionGraphEdge>PyList_GET_ITEM(edges, i)
 */
  __pyx_t_6 = PyList_Size(__pyx_v_edges); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 252, __pyx_L1_error)
  __pyx_v_n = __pyx_t_6;

  /* "glycresoft/_c/composition_network/graph.pyx":253
 * 
 *         n = PyList_Size(edges)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             edge = <CompositionGraphEdge>PyList_GET_ITEM(edges, i)
 *             n1 = nodes[edge.node1.index]
 */
  __pyx_t_9 = __pyx_v_n;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "glycresoft/_c/composition_network/graph.pyx":254
 *         n = PyList_Size(edges)
 *         for i in range(n):
 *             edge = <CompositionGraphEdge>PyList_GET_ITEM(edges, i)             # <<<<<<<<<<<<<<
 *             n1 = nodes[edge.node1.index]
 *             n2 = nodes[edge.node2.index]
 */
    __pyx_t_12 = PyList_GET_ITEM(__pyx_v_edges, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_12);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_edge, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":255
 *         for i in range(n):
 *             edge = <CompositionGraphEdge>PyList_GET_ITEM(edges, i)
 *             n1 = nodes[edge.node1.index]             # <<<<<<<<<<<<<<
 *             n2 = nodes[edge.node2.index]
 *             e = edge.copy_for(n1, n2)
 */
    if (unlikely(__pyx_v_nodes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 255, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_nodes, __pyx_v_edge->node1->index, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_n1, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":256
 *             edge = <CompositionGraphEdge>PyList_GET_ITEM(edges, i)
 *             n1 = nodes[edge.node1.index]
 *             n2 = nodes[edge.node2.index]             # <<<<<<<<<<<<<<
 *             e = edge.copy_for(n1, n2)
 *             graph.edges.add(e)
 */
    if (unlikely(__pyx_v_nodes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 256, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_nodes, __pyx_v_edge->node2->index, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_n2, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":257
 *             n1 = nodes[edge.node1.index]
 *             n2 = nodes[edge.node2.index]
 *             e = edge.copy_for(n1, n2)             # <<<<<<<<<<<<<<
 *             graph.edges.add(e)
 *         graph.neighborhoods.update(self.neighborhoods.copy())
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_v_edge->__pyx_vtab)->copy_for(__pyx_v_edge, __pyx_v_n1, __pyx_v_n2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge))))) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_e, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":258
 *             n2 = nodes[edge.node2.index]
 *             e = edge.copy_for(n1, n2)
 *             graph.edges.add(e)             # <<<<<<<<<<<<<<
 *         graph.neighborhoods.update(self.neighborhoods.copy())
 *         return graph
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_graph->edges), __pyx_n_s_add); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_e)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":259
 *             e = edge.copy_for(n1, n2)
 *             graph.edges.add(e)
 *         graph.neighborhoods.update(self.neighborhoods.copy())             # <<<<<<<<<<<<<<
 *         return graph
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_graph->neighborhoods, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->neighborhoods, __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":260
 *             graph.edges.add(e)
 *         graph.neighborhoods.update(self.neighborhoods.copy())
 *         return graph             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_graph);
  __pyx_r = __pyx_v_graph;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":236
 *         return self.edges.get_edges()
 * 
 *     cpdef CompositionGraphBase copy(CompositionGraphBase self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphNode n1, n2
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_n1);
  __Pyx_XDECREF((PyObject *)__pyx_v_n2);
  __Pyx_XDECREF((PyObject *)__pyx_v_edge);
  __Pyx_XDECREF((PyObject *)__pyx_v_e);
  __Pyx_XDECREF((PyObject *)__pyx_v_graph);
  __Pyx_XDECREF(__pyx_v_edges);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_node);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy, "CompositionGraphBase.copy(self) -> CompositionGraphBase");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":40
 * cdef class CompositionGraphBase(object):
 *     cdef:
 *         public list nodes             # <<<<<<<<<<<<<<
 *         public dict node_map
 *         public object _composition_normalizer
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->nodes);
  __pyx_r = __pyx_v_self->nodes;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->nodes);
  __Pyx_DECREF(__pyx_v_self->nodes);
  __pyx_v_self->nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.nodes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->nodes);
  __Pyx_DECREF(__pyx_v_self->nodes);
  __pyx_v_self->nodes = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":41
 *     cdef:
 *         public list nodes
 *         public dict node_map             # <<<<<<<<<<<<<<
 *         public object _composition_normalizer
 *         public object distance_fn
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->node_map);
  __pyx_r = __pyx_v_self->node_map;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(1, 41, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->node_map);
  __Pyx_DECREF(__pyx_v_self->node_map);
  __pyx_v_self->node_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.node_map.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->node_map);
  __Pyx_DECREF(__pyx_v_self->node_map);
  __pyx_v_self->node_map = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":42
 *         public list nodes
 *         public dict node_map
 *         public object _composition_normalizer             # <<<<<<<<<<<<<<
 *         public object distance_fn
 *         public EdgeSet edges
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_composition_normalizer);
  __pyx_r = __pyx_v_self->_composition_normalizer;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_composition_normalizer);
  __Pyx_DECREF(__pyx_v_self->_composition_normalizer);
  __pyx_v_self->_composition_normalizer = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_composition_normalizer);
  __Pyx_DECREF(__pyx_v_self->_composition_normalizer);
  __pyx_v_self->_composition_normalizer = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":43
 *         public dict node_map
 *         public object _composition_normalizer
 *         public object distance_fn             # <<<<<<<<<<<<<<
 *         public EdgeSet edges
 *         public object neighborhoods
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->distance_fn);
  __pyx_r = __pyx_v_self->distance_fn;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->distance_fn);
  __Pyx_DECREF(__pyx_v_self->distance_fn);
  __pyx_v_self->distance_fn = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->distance_fn);
  __Pyx_DECREF(__pyx_v_self->distance_fn);
  __pyx_v_self->distance_fn = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":44
 *         public object _composition_normalizer
 *         public object distance_fn
 *         public EdgeSet edges             # <<<<<<<<<<<<<<
 *         public object neighborhoods
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->edges);
  __pyx_r = ((PyObject *)__pyx_v_self->edges);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet))))) __PYX_ERR(1, 44, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->edges);
  __Pyx_DECREF((PyObject *)__pyx_v_self->edges);
  __pyx_v_self->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.edges.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->edges);
  __Pyx_DECREF((PyObject *)__pyx_v_self->edges);
  __pyx_v_self->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":45
 *         public object distance_fn
 *         public EdgeSet edges
 *         public object neighborhoods             # <<<<<<<<<<<<<<
 * 
 *     cdef list get_edges(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->neighborhoods);
  __pyx_r = __pyx_v_self->neighborhoods;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->neighborhoods);
  __Pyx_DECREF(__pyx_v_self->neighborhoods);
  __pyx_v_self->neighborhoods = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->neighborhoods);
  __Pyx_DECREF(__pyx_v_self->neighborhoods);
  __pyx_v_self->neighborhoods = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_2__reduce_cython__, "CompositionGraphBase.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_2__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_2__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_2__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._composition_normalizer, self.distance_fn, self.edges, self.neighborhoods, self.node_map, self.nodes)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_composition_normalizer);
  __Pyx_GIVEREF(__pyx_v_self->_composition_normalizer);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_composition_normalizer)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->distance_fn);
  __Pyx_GIVEREF(__pyx_v_self->distance_fn);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->distance_fn)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->edges);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->edges);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->edges))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->neighborhoods);
  __Pyx_GIVEREF(__pyx_v_self->neighborhoods);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->neighborhoods)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->node_map);
  __Pyx_GIVEREF(__pyx_v_self->node_map);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->node_map)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->nodes);
  __Pyx_GIVEREF(__pyx_v_self->nodes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->nodes)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._composition_normalizer, self.distance_fn, self.edges, self.neighborhoods, self.node_map, self.nodes)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._composition_normalizer, self.distance_fn, self.edges, self.neighborhoods, self.node_map, self.nodes)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._composition_normalizer is not None or self.distance_fn is not None or self.edges is not None or self.neighborhoods is not None or self.node_map is not None or self.nodes is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._composition_normalizer, self.distance_fn, self.edges, self.neighborhoods, self.node_map, self.nodes)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._composition_normalizer is not None or self.distance_fn is not None or self.edges is not None or self.neighborhoods is not None or self.node_map is not None or self.nodes is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_composition_normalizer != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->distance_fn != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->edges) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->neighborhoods != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->node_map != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->nodes != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._composition_normalizer is not None or self.distance_fn is not None or self.edges is not None or self.neighborhoods is not None or self.node_map is not None or self.nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._composition_normalizer is not None or self.distance_fn is not None or self.edges is not None or self.neighborhoods is not None or self.node_map is not None or self.nodes is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_CompositionGraphB); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_183956233);
    __Pyx_GIVEREF(__pyx_int_183956233);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_183956233)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._composition_normalizer is not None or self.distance_fn is not None or self.edges is not None or self.neighborhoods is not None or self.node_map is not None or self.nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, None), state
 *     else:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CompositionGraphBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_CompositionGraphB); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_183956233);
    __Pyx_GIVEREF(__pyx_int_183956233);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_183956233)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_4__setstate_cython__, "CompositionGraphBase.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_4__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_4__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_4__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CompositionGraphBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CompositionGraphBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":273
 *         public double limit
 * 
 *     def __init__(self, graph, start, end, limit=INF):             # <<<<<<<<<<<<<<
 *         self.graph = graph
 *         self.start = start
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_graph = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_v_limit = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_graph,&__pyx_n_s_start,&__pyx_n_s_end,&__pyx_n_s_limit,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(__pyx_k_);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_graph)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 273, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_start)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 273, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(0, 273, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_end)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 273, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(0, 273, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_limit);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 273, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 273, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_graph = values[0];
    __pyx_v_start = values[1];
    __pyx_v_end = values[2];
    __pyx_v_limit = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 273, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder___init__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), __pyx_v_graph, __pyx_v_start, __pyx_v_end, __pyx_v_limit);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_graph, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_limit) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  double __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":274
 * 
 *     def __init__(self, graph, start, end, limit=INF):
 *         self.graph = graph             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */
  __Pyx_INCREF(__pyx_v_graph);
  __Pyx_GIVEREF(__pyx_v_graph);
  __Pyx_GOTREF(__pyx_v_self->graph);
  __Pyx_DECREF(__pyx_v_self->graph);
  __pyx_v_self->graph = __pyx_v_graph;

  /* "glycresoft/_c/composition_network/graph.pyx":275
 *     def __init__(self, graph, start, end, limit=INF):
 *         self.graph = graph
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         assert start._str is not None
 */
  if (!(likely(((__pyx_v_start) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_start, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 275, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_start;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->start);
  __Pyx_DECREF((PyObject *)__pyx_v_self->start);
  __pyx_v_self->start = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":276
 *         self.graph = graph
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         assert start._str is not None
 *         assert end._str is not None
 */
  if (!(likely(((__pyx_v_end) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_end, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 276, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_end;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->end);
  __Pyx_DECREF((PyObject *)__pyx_v_self->end);
  __pyx_v_self->end = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":277
 *         self.start = start
 *         self.end = end
 *         assert start._str is not None             # <<<<<<<<<<<<<<
 *         assert end._str is not None
 *         self.distance = dict()
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_start, __pyx_n_s_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 277, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 277, __pyx_L1_error)
  #endif

  /* "glycresoft/_c/composition_network/graph.pyx":278
 *         self.end = end
 *         assert start._str is not None
 *         assert end._str is not None             # <<<<<<<<<<<<<<
 *         self.distance = dict()
 *         self.distance[start._str] = 0
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_end, __pyx_n_s_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 278, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 278, __pyx_L1_error)
  #endif

  /* "glycresoft/_c/composition_network/graph.pyx":279
 *         assert start._str is not None
 *         assert end._str is not None
 *         self.distance = dict()             # <<<<<<<<<<<<<<
 *         self.distance[start._str] = 0
 *         self.unvisited_finite_distance = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->distance);
  __Pyx_DECREF(__pyx_v_self->distance);
  __pyx_v_self->distance = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":280
 *         assert end._str is not None
 *         self.distance = dict()
 *         self.distance[start._str] = 0             # <<<<<<<<<<<<<<
 *         self.unvisited_finite_distance = dict()
 *         self.limit = limit
 */
  if (unlikely(__pyx_v_self->distance == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 280, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_start, __pyx_n_s_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyDict_SetItem(__pyx_v_self->distance, __pyx_t_1, __pyx_int_0) < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":281
 *         self.distance = dict()
 *         self.distance[start._str] = 0
 *         self.unvisited_finite_distance = dict()             # <<<<<<<<<<<<<<
 *         self.limit = limit
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->unvisited_finite_distance);
  __Pyx_DECREF(__pyx_v_self->unvisited_finite_distance);
  __pyx_v_self->unvisited_finite_distance = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":282
 *         self.distance[start._str] = 0
 *         self.unvisited_finite_distance = dict()
 *         self.limit = limit             # <<<<<<<<<<<<<<
 * 
 *     cpdef set _build_initial_key_set(self):
 */
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_limit); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 282, __pyx_L1_error)
  __pyx_v_self->limit = __pyx_t_3;

  /* "glycresoft/_c/composition_network/graph.pyx":273
 *         public double limit
 * 
 *     def __init__(self, graph, start, end, limit=INF):             # <<<<<<<<<<<<<<
 *         self.graph = graph
 *         self.start = start
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":284
 *         self.limit = limit
 * 
 *     cpdef set _build_initial_key_set(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder__build_initial_key_set(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_node = 0;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_nodes = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_build_initial_key_set", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build_initial_key_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_2))) __PYX_ERR(0, 284, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":290
 *             set result
 *             list nodes
 *         result = set()             # <<<<<<<<<<<<<<
 *         nodes = self.graph.nodes
 *         n = PyList_Size(nodes)
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":291
 *             list nodes
 *         result = set()
 *         nodes = self.graph.nodes             # <<<<<<<<<<<<<<
 *         n = PyList_Size(nodes)
 *         for i in range(n):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->graph, __pyx_n_s_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 291, __pyx_L1_error)
  __pyx_v_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":292
 *         result = set()
 *         nodes = self.graph.nodes
 *         n = PyList_Size(nodes)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 */
  __pyx_t_6 = PyList_Size(__pyx_v_nodes); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 292, __pyx_L1_error)
  __pyx_v_n = __pyx_t_6;

  /* "glycresoft/_c/composition_network/graph.pyx":293
 *         nodes = self.graph.nodes
 *         n = PyList_Size(nodes)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 *             assert node._str is not None
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "glycresoft/_c/composition_network/graph.pyx":294
 *         n = PyList_Size(nodes)
 *         for i in range(n):
 *             node = <CompositionGraphNode>PyList_GetItem(nodes, i)             # <<<<<<<<<<<<<<
 *             assert node._str is not None
 *             result.add(node._str)
 */
    __pyx_t_10 = PyList_GetItem(__pyx_v_nodes, __pyx_v_i); if (unlikely(__pyx_t_10 == ((PyObject *)NULL))) __PYX_ERR(0, 294, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_t_10);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":295
 *         for i in range(n):
 *             node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 *             assert node._str is not None             # <<<<<<<<<<<<<<
 *             result.add(node._str)
 *         return result
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_11 = (__pyx_v_node->_str != ((PyObject*)Py_None));
      if (unlikely(!__pyx_t_11)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 295, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 295, __pyx_L1_error)
    #endif

    /* "glycresoft/_c/composition_network/graph.pyx":296
 *             node = <CompositionGraphNode>PyList_GetItem(nodes, i)
 *             assert node._str is not None
 *             result.add(node._str)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = __pyx_v_node->_str;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_12 = PySet_Add(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":297
 *             assert node._str is not None
 *             result.add(node._str)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef CompositionGraphNode find_smallest_unvisited(self, set unvisited):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":284
 *         self.limit = limit
 * 
 *     cpdef set _build_initial_key_set(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder._build_initial_key_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_2_build_initial_key_set, "DijkstraPathFinder._build_initial_key_set(self) -> set");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set = {"_build_initial_key_set", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_2_build_initial_key_set};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_build_initial_key_set (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_build_initial_key_set", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_build_initial_key_set", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_2_build_initial_key_set(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_2_build_initial_key_set(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_build_initial_key_set", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder__build_initial_key_set(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder._build_initial_key_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":299
 *         return result
 * 
 *     cpdef CompositionGraphNode find_smallest_unvisited(self, set unvisited):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double smallest_distance, distance
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_smallest_unvisited(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_unvisited, int __pyx_skip_dispatch) {
  double __pyx_v_smallest_distance;
  double __pyx_v_distance;
  PyObject *__pyx_v_smallest_node = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_iterable = 0;
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_node = NULL;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_UCS4 __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  double __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_smallest_unvisited", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_smallest_unvisited); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_unvisited};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 299, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":307
 *             PyObject* ptemp
 * 
 *         smallest_distance = INF             # <<<<<<<<<<<<<<
 *         smallest_node = None
 * 
 */
  __pyx_v_smallest_distance = __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;

  /* "glycresoft/_c/composition_network/graph.pyx":308
 * 
 *         smallest_distance = INF
 *         smallest_node = None             # <<<<<<<<<<<<<<
 * 
 *         if not unvisited:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_smallest_node = Py_None;

  /* "glycresoft/_c/composition_network/graph.pyx":310
 *         smallest_node = None
 * 
 *         if not unvisited:             # <<<<<<<<<<<<<<
 *             raise ValueError("The unvisited set must not be empty")
 * 
 */
  __pyx_t_6 = (__pyx_v_unvisited != Py_None)&&(PySet_GET_SIZE(__pyx_v_unvisited) != 0);
  __pyx_t_7 = (!__pyx_t_6);
  if (unlikely(__pyx_t_7)) {

    /* "glycresoft/_c/composition_network/graph.pyx":311
 * 
 *         if not unvisited:
 *             raise ValueError("The unvisited set must not be empty")             # <<<<<<<<<<<<<<
 * 
 *         if not self.unvisited_finite_distance:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 311, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":310
 *         smallest_node = None
 * 
 *         if not unvisited:             # <<<<<<<<<<<<<<
 *             raise ValueError("The unvisited set must not be empty")
 * 
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":313
 *             raise ValueError("The unvisited set must not be empty")
 * 
 *         if not self.unvisited_finite_distance:             # <<<<<<<<<<<<<<
 *             for key in unvisited:
 *                 ptemp = PyDict_GetItem(self.distance, key)
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->unvisited_finite_distance); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_7);
  if (__pyx_t_6) {

    /* "glycresoft/_c/composition_network/graph.pyx":314
 * 
 *         if not self.unvisited_finite_distance:
 *             for key in unvisited:             # <<<<<<<<<<<<<<
 *                 ptemp = PyDict_GetItem(self.distance, key)
 *                 if ptemp == NULL:
 */
    __pyx_t_8 = 0;
    __pyx_t_2 = __Pyx_set_iterator(__pyx_v_unvisited, 1, (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_9, &__pyx_t_8, &__pyx_t_2, __pyx_t_10);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 314, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":315
 *         if not self.unvisited_finite_distance:
 *             for key in unvisited:
 *                 ptemp = PyDict_GetItem(self.distance, key)             # <<<<<<<<<<<<<<
 *                 if ptemp == NULL:
 *                     distance = INF
 */
      __pyx_t_2 = __pyx_v_self->distance;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_ptemp = PyDict_GetItem(__pyx_t_2, __pyx_v_key);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":316
 *             for key in unvisited:
 *                 ptemp = PyDict_GetItem(self.distance, key)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     distance = INF
 *                 if distance <= smallest_distance:
 */
      __pyx_t_6 = (__pyx_v_ptemp == NULL);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":317
 *                 ptemp = PyDict_GetItem(self.distance, key)
 *                 if ptemp == NULL:
 *                     distance = INF             # <<<<<<<<<<<<<<
 *                 if distance <= smallest_distance:
 *                     smallest_node = key
 */
        __pyx_v_distance = __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;

        /* "glycresoft/_c/composition_network/graph.pyx":316
 *             for key in unvisited:
 *                 ptemp = PyDict_GetItem(self.distance, key)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     distance = INF
 *                 if distance <= smallest_distance:
 */
      }

      /* "glycresoft/_c/composition_network/graph.pyx":318
 *                 if ptemp == NULL:
 *                     distance = INF
 *                 if distance <= smallest_distance:             # <<<<<<<<<<<<<<
 *                     smallest_node = key
 *                     smallest_distance = distance
 */
      __pyx_t_6 = (__pyx_v_distance <= __pyx_v_smallest_distance);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":319
 *                     distance = INF
 *                 if distance <= smallest_distance:
 *                     smallest_node = key             # <<<<<<<<<<<<<<
 *                     smallest_distance = distance
 *             if smallest_node is None:
 */
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_DECREF_SET(__pyx_v_smallest_node, __pyx_v_key);

        /* "glycresoft/_c/composition_network/graph.pyx":320
 *                 if distance <= smallest_distance:
 *                     smallest_node = key
 *                     smallest_distance = distance             # <<<<<<<<<<<<<<
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 */
        __pyx_v_smallest_distance = __pyx_v_distance;

        /* "glycresoft/_c/composition_network/graph.pyx":318
 *                 if ptemp == NULL:
 *                     distance = INF
 *                 if distance <= smallest_distance:             # <<<<<<<<<<<<<<
 *                     smallest_node = key
 *                     smallest_distance = distance
 */
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":321
 *                     smallest_node = key
 *                     smallest_distance = distance
 *             if smallest_node is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 */
    __pyx_t_6 = (__pyx_v_smallest_node == Py_None);
    if (unlikely(__pyx_t_6)) {

      /* "glycresoft/_c/composition_network/graph.pyx":322
 *                     smallest_distance = distance
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "             # <<<<<<<<<<<<<<
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 *         else:
 */
      __pyx_t_1 = PyTuple_New(10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = 0;
      __pyx_t_12 = 127;
      __Pyx_INCREF(__pyx_kp_u_Failed_to_select_a_destination_f);
      __pyx_t_9 += 58;
      __Pyx_GIVEREF(__pyx_kp_u_Failed_to_select_a_destination_f);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Failed_to_select_a_destination_f);
      if (unlikely(__pyx_v_unvisited == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 322, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PySet_GET_SIZE(__pyx_v_unvisited); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 322, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_8, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_with_smallest_distance);
      __pyx_t_9 += 24;
      __Pyx_GIVEREF(__pyx_kp_u_with_smallest_distance);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_with_smallest_distance);

      /* "glycresoft/_c/composition_network/graph.pyx":323
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")             # <<<<<<<<<<<<<<
 *         else:
 *             iterable = PyDict_Items(self.unvisited_finite_distance)
 */
      __pyx_t_2 = PyFloat_FromDouble(__pyx_v_smallest_distance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u__3);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__3);
      PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__3);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_unvisited, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_looking_for);
      __pyx_t_9 += 13;
      __Pyx_GIVEREF(__pyx_kp_u_looking_for);
      PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u_looking_for);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_self->start->_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u__4);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__4);
      PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_kp_u__4);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_self->end->_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":322
 *                     smallest_distance = distance
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "             # <<<<<<<<<<<<<<
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 *         else:
 */
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 10, __pyx_t_9, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 322, __pyx_L1_error)

      /* "glycresoft/_c/composition_network/graph.pyx":321
 *                     smallest_node = key
 *                     smallest_distance = distance
 *             if smallest_node is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":313
 *             raise ValueError("The unvisited set must not be empty")
 * 
 *         if not self.unvisited_finite_distance:             # <<<<<<<<<<<<<<
 *             for key in unvisited:
 *                 ptemp = PyDict_GetItem(self.distance, key)
 */
    goto __pyx_L4;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":325
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 *         else:
 *             iterable = PyDict_Items(self.unvisited_finite_distance)             # <<<<<<<<<<<<<<
 *             for node, distance in iterable:
 *                 if distance <= smallest_distance:
 */
  /*else*/ {
    __pyx_t_1 = __pyx_v_self->unvisited_finite_distance;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyDict_Items(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_iterable = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":326
 *         else:
 *             iterable = PyDict_Items(self.unvisited_finite_distance)
 *             for node, distance in iterable:             # <<<<<<<<<<<<<<
 *                 if distance <= smallest_distance:
 *                     smallest_distance = distance
 */
    if (unlikely(__pyx_v_iterable == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 326, __pyx_L1_error)
    }
    __pyx_t_3 = __pyx_v_iterable; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_9 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 326, __pyx_L1_error)
        #endif
        if (__pyx_t_9 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 326, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 326, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_13 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_13);
        index = 0; __pyx_t_2 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_2)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_4 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_4)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_13), 2) < 0) __PYX_ERR(0, 326, __pyx_L1_error)
        __pyx_t_14 = NULL;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L13_unpacking_done;
        __pyx_L12_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_14 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 326, __pyx_L1_error)
        __pyx_L13_unpacking_done:;
      }
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_v_distance = __pyx_t_15;

      /* "glycresoft/_c/composition_network/graph.pyx":327
 *             iterable = PyDict_Items(self.unvisited_finite_distance)
 *             for node, distance in iterable:
 *                 if distance <= smallest_distance:             # <<<<<<<<<<<<<<
 *                     smallest_distance = distance
 *                     smallest_node = node
 */
      __pyx_t_6 = (__pyx_v_distance <= __pyx_v_smallest_distance);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":328
 *             for node, distance in iterable:
 *                 if distance <= smallest_distance:
 *                     smallest_distance = distance             # <<<<<<<<<<<<<<
 *                     smallest_node = node
 *             if smallest_node is None:
 */
        __pyx_v_smallest_distance = __pyx_v_distance;

        /* "glycresoft/_c/composition_network/graph.pyx":329
 *                 if distance <= smallest_distance:
 *                     smallest_distance = distance
 *                     smallest_node = node             # <<<<<<<<<<<<<<
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 */
        __Pyx_INCREF(__pyx_v_node);
        __Pyx_DECREF_SET(__pyx_v_smallest_node, __pyx_v_node);

        /* "glycresoft/_c/composition_network/graph.pyx":327
 *             iterable = PyDict_Items(self.unvisited_finite_distance)
 *             for node, distance in iterable:
 *                 if distance <= smallest_distance:             # <<<<<<<<<<<<<<
 *                     smallest_distance = distance
 *                     smallest_node = node
 */
      }

      /* "glycresoft/_c/composition_network/graph.pyx":326
 *         else:
 *             iterable = PyDict_Items(self.unvisited_finite_distance)
 *             for node, distance in iterable:             # <<<<<<<<<<<<<<
 *                 if distance <= smallest_distance:
 *                     smallest_distance = distance
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":330
 *                     smallest_distance = distance
 *                     smallest_node = node
 *             if smallest_node is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 */
    __pyx_t_6 = (__pyx_v_smallest_node == Py_None);
    if (unlikely(__pyx_t_6)) {

      /* "glycresoft/_c/composition_network/graph.pyx":331
 *                     smallest_node = node
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "             # <<<<<<<<<<<<<<
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 *         return self.graph[smallest_node]
 */
      __pyx_t_3 = PyTuple_New(10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = 0;
      __pyx_t_12 = 127;
      __Pyx_INCREF(__pyx_kp_u_Failed_to_select_a_destination_f);
      __pyx_t_9 += 58;
      __Pyx_GIVEREF(__pyx_kp_u_Failed_to_select_a_destination_f);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Failed_to_select_a_destination_f);
      if (unlikely(__pyx_v_unvisited == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 331, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PySet_GET_SIZE(__pyx_v_unvisited); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 331, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_8, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u_with_smallest_distance);
      __pyx_t_9 += 24;
      __Pyx_GIVEREF(__pyx_kp_u_with_smallest_distance);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_with_smallest_distance);

      /* "glycresoft/_c/composition_network/graph.pyx":332
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")             # <<<<<<<<<<<<<<
 *         return self.graph[smallest_node]
 * 
 */
      __pyx_t_1 = PyFloat_FromDouble(__pyx_v_smallest_distance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u__3);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__3);
      PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__3);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_unvisited, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u_looking_for);
      __pyx_t_9 += 13;
      __Pyx_GIVEREF(__pyx_kp_u_looking_for);
      PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_looking_for);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_self->start->_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u__4);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__4);
      PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u__4);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_self->end->_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_12;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":331
 *                     smallest_node = node
 *             if smallest_node is None:
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "             # <<<<<<<<<<<<<<
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 *         return self.graph[smallest_node]
 */
      __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 10, __pyx_t_9, __pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 331, __pyx_L1_error)

      /* "glycresoft/_c/composition_network/graph.pyx":330
 *                     smallest_distance = distance
 *                     smallest_node = node
 *             if smallest_node is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 */
    }
  }
  __pyx_L4:;

  /* "glycresoft/_c/composition_network/graph.pyx":333
 *                 raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
 *                                  f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
 *         return self.graph[smallest_node]             # <<<<<<<<<<<<<<
 * 
 *     cpdef find_path(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->graph, __pyx_v_smallest_node); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 333, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":299
 *         return result
 * 
 *     cpdef CompositionGraphNode find_smallest_unvisited(self, set unvisited):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double smallest_distance, distance
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.find_smallest_unvisited", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_smallest_node);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_iterable);
  __Pyx_XDECREF(__pyx_v_node);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_4find_smallest_unvisited, "DijkstraPathFinder.find_smallest_unvisited(self, set unvisited) -> CompositionGraphNode");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited = {"find_smallest_unvisited", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_4find_smallest_unvisited};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_unvisited = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_smallest_unvisited (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_unvisited,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_unvisited)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 299, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "find_smallest_unvisited") < 0)) __PYX_ERR(0, 299, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_unvisited = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_smallest_unvisited", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 299, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.find_smallest_unvisited", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_unvisited), (&PySet_Type), 1, "unvisited", 1))) __PYX_ERR(0, 299, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_4find_smallest_unvisited(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), __pyx_v_unvisited);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_4find_smallest_unvisited(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_unvisited) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_smallest_unvisited", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_smallest_unvisited(__pyx_v_self, __pyx_v_unvisited, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.find_smallest_unvisited", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":335
 *         return self.graph[smallest_node]
 * 
 *     cpdef find_path(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             set unvisited
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_path(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_unvisited = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_current_node = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v_terminal = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v_edges = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_v_edge = 0;
  PyObject *__pyx_v_ptemp;
  double __pyx_v_path_length;
  double __pyx_v_terminal_distance;
  long __pyx_v_n;
  int __pyx_v_tmp;
  PyObject *__pyx_v_visit_queue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  double __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_path", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":346
 *             int tmp
 * 
 *         unvisited = self._build_initial_key_set()             # <<<<<<<<<<<<<<
 * 
 *         visit_queue = deque([self.start])
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self->__pyx_vtab)->_build_initial_key_set(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_unvisited = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":348
 *         unvisited = self._build_initial_key_set()
 * 
 *         visit_queue = deque([self.start])             # <<<<<<<<<<<<<<
 *         n = 1
 *         while self.end._str in unvisited:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_self->start);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->start);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->start))) __PYX_ERR(0, 348, __pyx_L1_error);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_visit_queue = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":349
 * 
 *         visit_queue = deque([self.start])
 *         n = 1             # <<<<<<<<<<<<<<
 *         while self.end._str in unvisited:
 *             if n > 0:
 */
  __pyx_v_n = 1;

  /* "glycresoft/_c/composition_network/graph.pyx":350
 *         visit_queue = deque([self.start])
 *         n = 1
 *         while self.end._str in unvisited:             # <<<<<<<<<<<<<<
 *             if n > 0:
 *                 n -= 1
 */
  while (1) {
    if (unlikely(__pyx_v_unvisited == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 350, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PySet_ContainsTF(__pyx_v_self->end->_str, __pyx_v_unvisited, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 350, __pyx_L1_error)
    if (!__pyx_t_6) break;

    /* "glycresoft/_c/composition_network/graph.pyx":351
 *         n = 1
 *         while self.end._str in unvisited:
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 n -= 1
 *                 current_node = visit_queue.popleft()
 */
    __pyx_t_6 = (__pyx_v_n > 0);
    if (__pyx_t_6) {

      /* "glycresoft/_c/composition_network/graph.pyx":352
 *         while self.end._str in unvisited:
 *             if n > 0:
 *                 n -= 1             # <<<<<<<<<<<<<<
 *                 current_node = visit_queue.popleft()
 *             else:
 */
      __pyx_v_n = (__pyx_v_n - 1);

      /* "glycresoft/_c/composition_network/graph.pyx":353
 *             if n > 0:
 *                 n -= 1
 *                 current_node = visit_queue.popleft()             # <<<<<<<<<<<<<<
 *             else:
 *                 current_node = self.find_smallest_unvisited(unvisited)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_visit_queue, __pyx_n_s_popleft); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_current_node, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":351
 *         n = 1
 *         while self.end._str in unvisited:
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 n -= 1
 *                 current_node = visit_queue.popleft()
 */
      goto __pyx_L5;
    }

    /* "glycresoft/_c/composition_network/graph.pyx":355
 *                 current_node = visit_queue.popleft()
 *             else:
 *                 current_node = self.find_smallest_unvisited(unvisited)             # <<<<<<<<<<<<<<
 * 
 *             tmp = PySet_Discard(unvisited, current_node._str)
 */
    /*else*/ {
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self->__pyx_vtab)->find_smallest_unvisited(__pyx_v_self, __pyx_v_unvisited, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_current_node, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1));
      __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "glycresoft/_c/composition_network/graph.pyx":357
 *                 current_node = self.find_smallest_unvisited(unvisited)
 * 
 *             tmp = PySet_Discard(unvisited, current_node._str)             # <<<<<<<<<<<<<<
 *             if tmp == -1:
 *                 raise KeyError(current_node._str)
 */
    __pyx_t_1 = __pyx_v_current_node->_str;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = PySet_Discard(__pyx_v_unvisited, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_tmp = __pyx_t_6;

    /* "glycresoft/_c/composition_network/graph.pyx":358
 * 
 *             tmp = PySet_Discard(unvisited, current_node._str)
 *             if tmp == -1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(current_node._str)
 *             elif tmp == 0:
 */
    switch (__pyx_v_tmp) {
      case -1L:

      /* "glycresoft/_c/composition_network/graph.pyx":359
 *             tmp = PySet_Discard(unvisited, current_node._str)
 *             if tmp == -1:
 *                 raise KeyError(current_node._str)             # <<<<<<<<<<<<<<
 *             elif tmp == 0:
 *                 continue
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_v_current_node->_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 359, __pyx_L1_error)

      /* "glycresoft/_c/composition_network/graph.pyx":358
 * 
 *             tmp = PySet_Discard(unvisited, current_node._str)
 *             if tmp == -1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(current_node._str)
 *             elif tmp == 0:
 */
      break;
      case 0:

      /* "glycresoft/_c/composition_network/graph.pyx":361
 *                 raise KeyError(current_node._str)
 *             elif tmp == 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
 */
      goto __pyx_L3_continue;

      /* "glycresoft/_c/composition_network/graph.pyx":360
 *             if tmp == -1:
 *                 raise KeyError(current_node._str)
 *             elif tmp == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:
 */
      break;
      default: break;
    }

    /* "glycresoft/_c/composition_network/graph.pyx":362
 *             elif tmp == 0:
 *                 continue
 *             if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:             # <<<<<<<<<<<<<<
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
 *             edges = current_node.edges
 */
    __pyx_t_1 = __pyx_v_self->unvisited_finite_distance;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_current_node->_str;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = (PyDict_GetItem(__pyx_t_1, __pyx_t_2) != NULL);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "glycresoft/_c/composition_network/graph.pyx":363
 *                 continue
 *             if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)             # <<<<<<<<<<<<<<
 *             edges = current_node.edges
 *             for edge in edges:
 */
      __pyx_t_2 = __pyx_v_self->unvisited_finite_distance;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_1 = __pyx_v_current_node->_str;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = PyDict_DelItem(__pyx_t_2, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":362
 *             elif tmp == 0:
 *                 continue
 *             if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:             # <<<<<<<<<<<<<<
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
 *             edges = current_node.edges
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":364
 *             if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
 *             edges = current_node.edges             # <<<<<<<<<<<<<<
 *             for edge in edges:
 *                 terminal = edge._traverse(current_node)
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_current_node->edges);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_edges, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycresoft/_c/composition_network/graph.pyx":365
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
 *             edges = current_node.edges
 *             for edge in edges:             # <<<<<<<<<<<<<<
 *                 terminal = edge._traverse(current_node)
 *                 if terminal._str not in unvisited:
 */
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_edges))) || PyTuple_CheckExact(((PyObject *)__pyx_v_edges))) {
      __pyx_t_1 = ((PyObject *)__pyx_v_edges); __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_edges)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 365, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 365, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 365, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 365, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 365, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 365, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge))))) __PYX_ERR(0, 365, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_edge, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":366
 *             edges = current_node.edges
 *             for edge in edges:
 *                 terminal = edge._traverse(current_node)             # <<<<<<<<<<<<<<
 *                 if terminal._str not in unvisited:
 *                     continue
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_edge), __pyx_n_s_traverse); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_current_node)};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_terminal, ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":367
 *             for edge in edges:
 *                 terminal = edge._traverse(current_node)
 *                 if terminal._str not in unvisited:             # <<<<<<<<<<<<<<
 *                     continue
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)
 */
      if (unlikely(__pyx_v_unvisited == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 367, __pyx_L1_error)
      }
      __pyx_t_6 = (__Pyx_PySet_ContainsTF(__pyx_v_terminal->_str, __pyx_v_unvisited, Py_NE)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 367, __pyx_L1_error)
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":368
 *                 terminal = edge._traverse(current_node)
 *                 if terminal._str not in unvisited:
 *                     continue             # <<<<<<<<<<<<<<
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)
 *                 if ptemp == NULL:
 */
        goto __pyx_L7_continue;

        /* "glycresoft/_c/composition_network/graph.pyx":367
 *             for edge in edges:
 *                 terminal = edge._traverse(current_node)
 *                 if terminal._str not in unvisited:             # <<<<<<<<<<<<<<
 *                     continue
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)
 */
      }

      /* "glycresoft/_c/composition_network/graph.pyx":369
 *                 if terminal._str not in unvisited:
 *                     continue
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)             # <<<<<<<<<<<<<<
 *                 if ptemp == NULL:
 *                     path_length = INF
 */
      __pyx_t_2 = __pyx_v_self->distance;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = __pyx_v_current_node->_str;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_ptemp = PyDict_GetItem(__pyx_t_2, __pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":370
 *                     continue
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     path_length = INF
 *                 else:
 */
      __pyx_t_6 = (__pyx_v_ptemp == NULL);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":371
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)
 *                 if ptemp == NULL:
 *                     path_length = INF             # <<<<<<<<<<<<<<
 *                 else:
 *                     path_length = PyFloat_AsDouble(<object>ptemp)
 */
        __pyx_v_path_length = __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;

        /* "glycresoft/_c/composition_network/graph.pyx":370
 *                     continue
 *                 ptemp = PyDict_GetItem(self.distance, current_node._str)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     path_length = INF
 *                 else:
 */
        goto __pyx_L10;
      }

      /* "glycresoft/_c/composition_network/graph.pyx":373
 *                     path_length = INF
 *                 else:
 *                     path_length = PyFloat_AsDouble(<object>ptemp)             # <<<<<<<<<<<<<<
 *                 path_length += edge.order
 *                 ptemp = PyDict_GetItem(self.distance, terminal._str)
 */
      /*else*/ {
        __pyx_t_10 = PyFloat_AsDouble(((PyObject *)__pyx_v_ptemp)); if (unlikely(__pyx_t_10 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L1_error)
        __pyx_v_path_length = __pyx_t_10;
      }
      __pyx_L10:;

      /* "glycresoft/_c/composition_network/graph.pyx":374
 *                 else:
 *                     path_length = PyFloat_AsDouble(<object>ptemp)
 *                 path_length += edge.order             # <<<<<<<<<<<<<<
 *                 ptemp = PyDict_GetItem(self.distance, terminal._str)
 *                 if ptemp == NULL:
 */
      __pyx_v_path_length = (__pyx_v_path_length + __pyx_v_edge->order);

      /* "glycresoft/_c/composition_network/graph.pyx":375
 *                     path_length = PyFloat_AsDouble(<object>ptemp)
 *                 path_length += edge.order
 *                 ptemp = PyDict_GetItem(self.distance, terminal._str)             # <<<<<<<<<<<<<<
 *                 if ptemp == NULL:
 *                     terminal_distance = INF
 */
      __pyx_t_3 = __pyx_v_self->distance;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_2 = __pyx_v_terminal->_str;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_ptemp = PyDict_GetItem(__pyx_t_3, __pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":376
 *                 path_length += edge.order
 *                 ptemp = PyDict_GetItem(self.distance, terminal._str)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     terminal_distance = INF
 *                 else:
 */
      __pyx_t_6 = (__pyx_v_ptemp == NULL);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":377
 *                 ptemp = PyDict_GetItem(self.distance, terminal._str)
 *                 if ptemp == NULL:
 *                     terminal_distance = INF             # <<<<<<<<<<<<<<
 *                 else:
 *                     terminal_distance = PyFloat_AsDouble(<object>ptemp)
 */
        __pyx_v_terminal_distance = __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;

        /* "glycresoft/_c/composition_network/graph.pyx":376
 *                 path_length += edge.order
 *                 ptemp = PyDict_GetItem(self.distance, terminal._str)
 *                 if ptemp == NULL:             # <<<<<<<<<<<<<<
 *                     terminal_distance = INF
 *                 else:
 */
        goto __pyx_L11;
      }

      /* "glycresoft/_c/composition_network/graph.pyx":379
 *                     terminal_distance = INF
 *                 else:
 *                     terminal_distance = PyFloat_AsDouble(<object>ptemp)             # <<<<<<<<<<<<<<
 * 
 *                 if terminal_distance > path_length:
 */
      /*else*/ {
        __pyx_t_10 = PyFloat_AsDouble(((PyObject *)__pyx_v_ptemp)); if (unlikely(__pyx_t_10 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
        __pyx_v_terminal_distance = __pyx_t_10;
      }
      __pyx_L11:;

      /* "glycresoft/_c/composition_network/graph.pyx":381
 *                     terminal_distance = PyFloat_AsDouble(<object>ptemp)
 * 
 *                 if terminal_distance > path_length:             # <<<<<<<<<<<<<<
 *                     self.distance[terminal._str] = path_length
 *                     if terminal._str in unvisited and terminal_distance < self.limit:
 */
      __pyx_t_6 = (__pyx_v_terminal_distance > __pyx_v_path_length);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":382
 * 
 *                 if terminal_distance > path_length:
 *                     self.distance[terminal._str] = path_length             # <<<<<<<<<<<<<<
 *                     if terminal._str in unvisited and terminal_distance < self.limit:
 *                         self.unvisited_finite_distance[terminal._str] = path_length
 */
        __pyx_t_2 = PyFloat_FromDouble(__pyx_v_path_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__pyx_v_self->distance == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 382, __pyx_L1_error)
        }
        if (unlikely((PyDict_SetItem(__pyx_v_self->distance, __pyx_v_terminal->_str, __pyx_t_2) < 0))) __PYX_ERR(0, 382, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "glycresoft/_c/composition_network/graph.pyx":383
 *                 if terminal_distance > path_length:
 *                     self.distance[terminal._str] = path_length
 *                     if terminal._str in unvisited and terminal_distance < self.limit:             # <<<<<<<<<<<<<<
 *                         self.unvisited_finite_distance[terminal._str] = path_length
 *                 if terminal_distance < self.limit:
 */
        if (unlikely(__pyx_v_unvisited == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 383, __pyx_L1_error)
        }
        __pyx_t_11 = (__Pyx_PySet_ContainsTF(__pyx_v_terminal->_str, __pyx_v_unvisited, Py_EQ)); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 383, __pyx_L1_error)
        if (__pyx_t_11) {
        } else {
          __pyx_t_6 = __pyx_t_11;
          goto __pyx_L14_bool_binop_done;
        }
        __pyx_t_11 = (__pyx_v_terminal_distance < __pyx_v_self->limit);
        __pyx_t_6 = __pyx_t_11;
        __pyx_L14_bool_binop_done:;
        if (__pyx_t_6) {

          /* "glycresoft/_c/composition_network/graph.pyx":384
 *                     self.distance[terminal._str] = path_length
 *                     if terminal._str in unvisited and terminal_distance < self.limit:
 *                         self.unvisited_finite_distance[terminal._str] = path_length             # <<<<<<<<<<<<<<
 *                 if terminal_distance < self.limit:
 *                     visit_queue.append(terminal)
 */
          __pyx_t_2 = PyFloat_FromDouble(__pyx_v_path_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(__pyx_v_self->unvisited_finite_distance == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 384, __pyx_L1_error)
          }
          if (unlikely((PyDict_SetItem(__pyx_v_self->unvisited_finite_distance, __pyx_v_terminal->_str, __pyx_t_2) < 0))) __PYX_ERR(0, 384, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "glycresoft/_c/composition_network/graph.pyx":383
 *                 if terminal_distance > path_length:
 *                     self.distance[terminal._str] = path_length
 *                     if terminal._str in unvisited and terminal_distance < self.limit:             # <<<<<<<<<<<<<<
 *                         self.unvisited_finite_distance[terminal._str] = path_length
 *                 if terminal_distance < self.limit:
 */
        }

        /* "glycresoft/_c/composition_network/graph.pyx":381
 *                     terminal_distance = PyFloat_AsDouble(<object>ptemp)
 * 
 *                 if terminal_distance > path_length:             # <<<<<<<<<<<<<<
 *                     self.distance[terminal._str] = path_length
 *                     if terminal._str in unvisited and terminal_distance < self.limit:
 */
      }

      /* "glycresoft/_c/composition_network/graph.pyx":385
 *                     if terminal._str in unvisited and terminal_distance < self.limit:
 *                         self.unvisited_finite_distance[terminal._str] = path_length
 *                 if terminal_distance < self.limit:             # <<<<<<<<<<<<<<
 *                     visit_queue.append(terminal)
 *                     n += 1
 */
      __pyx_t_6 = (__pyx_v_terminal_distance < __pyx_v_self->limit);
      if (__pyx_t_6) {

        /* "glycresoft/_c/composition_network/graph.pyx":386
 *                         self.unvisited_finite_distance[terminal._str] = path_length
 *                 if terminal_distance < self.limit:
 *                     visit_queue.append(terminal)             # <<<<<<<<<<<<<<
 *                     n += 1
 * 
 */
        __pyx_t_12 = __Pyx_PyObject_Append(__pyx_v_visit_queue, ((PyObject *)__pyx_v_terminal)); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 386, __pyx_L1_error)

        /* "glycresoft/_c/composition_network/graph.pyx":387
 *                 if terminal_distance < self.limit:
 *                     visit_queue.append(terminal)
 *                     n += 1             # <<<<<<<<<<<<<<
 * 
 *     def search(self):
 */
        __pyx_v_n = (__pyx_v_n + 1);

        /* "glycresoft/_c/composition_network/graph.pyx":385
 *                     if terminal._str in unvisited and terminal_distance < self.limit:
 *                         self.unvisited_finite_distance[terminal._str] = path_length
 *                 if terminal_distance < self.limit:             # <<<<<<<<<<<<<<
 *                     visit_queue.append(terminal)
 *                     n += 1
 */
      }

      /* "glycresoft/_c/composition_network/graph.pyx":365
 *                 PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
 *             edges = current_node.edges
 *             for edge in edges:             # <<<<<<<<<<<<<<
 *                 terminal = edge._traverse(current_node)
 *                 if terminal._str not in unvisited:
 */
      __pyx_L7_continue:;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_L3_continue:;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":335
 *         return self.graph[smallest_node]
 * 
 *     cpdef find_path(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             set unvisited
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.find_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_unvisited);
  __Pyx_XDECREF((PyObject *)__pyx_v_current_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_terminal);
  __Pyx_XDECREF((PyObject *)__pyx_v_edges);
  __Pyx_XDECREF((PyObject *)__pyx_v_edge);
  __Pyx_XDECREF(__pyx_v_visit_queue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_6find_path, "DijkstraPathFinder.find_path(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path = {"find_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_6find_path};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_path (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("find_path", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "find_path", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_6find_path(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_6find_path(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_path", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_path(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.find_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":389
 *                     n += 1
 * 
 *     def search(self):             # <<<<<<<<<<<<<<
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_9search(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8search, "DijkstraPathFinder.search(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_9search = {"search", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_9search, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8search};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_9search(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("search (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("search", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "search", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8search(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8search(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("search", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":390
 * 
 *     def search(self):
 *         self.find_path()             # <<<<<<<<<<<<<<
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 *         if ptemp == NULL:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self->__pyx_vtab)->find_path(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":391
 *     def search(self):
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)             # <<<<<<<<<<<<<<
 *         if ptemp == NULL:
 *             return INF
 */
  __pyx_t_1 = __pyx_v_self->distance;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->end->_str;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_1, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":392
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             return INF
 *         else:
 */
  __pyx_t_3 = (__pyx_v_ptemp == NULL);
  if (__pyx_t_3) {

    /* "glycresoft/_c/composition_network/graph.pyx":393
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 *         if ptemp == NULL:
 *             return INF             # <<<<<<<<<<<<<<
 *         else:
 *             return <object>(ptemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_10glycresoft_2_c_19composition_network_5graph_INF); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":392
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 *         if ptemp == NULL:             # <<<<<<<<<<<<<<
 *             return INF
 *         else:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":395
 *             return INF
 *         else:
 *             return <object>(ptemp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_ptemp));
    __pyx_r = ((PyObject *)__pyx_v_ptemp);
    goto __pyx_L0;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":389
 *                     n += 1
 * 
 *     def search(self):             # <<<<<<<<<<<<<<
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.search", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":266
 * 
 *     cdef:
 *         public object graph             # <<<<<<<<<<<<<<
 *         public CompositionGraphNode start
 *         public CompositionGraphNode end
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->graph);
  __pyx_r = __pyx_v_self->graph;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->graph);
  __Pyx_DECREF(__pyx_v_self->graph);
  __pyx_v_self->graph = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->graph);
  __Pyx_DECREF(__pyx_v_self->graph);
  __pyx_v_self->graph = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":267
 *     cdef:
 *         public object graph
 *         public CompositionGraphNode start             # <<<<<<<<<<<<<<
 *         public CompositionGraphNode end
 *         public dict distance
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->start);
  __pyx_r = ((PyObject *)__pyx_v_self->start);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->start);
  __Pyx_DECREF((PyObject *)__pyx_v_self->start);
  __pyx_v_self->start = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->start);
  __Pyx_DECREF((PyObject *)__pyx_v_self->start);
  __pyx_v_self->start = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":268
 *         public object graph
 *         public CompositionGraphNode start
 *         public CompositionGraphNode end             # <<<<<<<<<<<<<<
 *         public dict distance
 *         public dict unvisited_finite_distance
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->end);
  __pyx_r = ((PyObject *)__pyx_v_self->end);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(0, 268, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->end);
  __Pyx_DECREF((PyObject *)__pyx_v_self->end);
  __pyx_v_self->end = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->end);
  __Pyx_DECREF((PyObject *)__pyx_v_self->end);
  __pyx_v_self->end = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":269
 *         public CompositionGraphNode start
 *         public CompositionGraphNode end
 *         public dict distance             # <<<<<<<<<<<<<<
 *         public dict unvisited_finite_distance
 *         public double limit
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->distance);
  __pyx_r = __pyx_v_self->distance;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 269, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->distance);
  __Pyx_DECREF(__pyx_v_self->distance);
  __pyx_v_self->distance = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.distance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->distance);
  __Pyx_DECREF(__pyx_v_self->distance);
  __pyx_v_self->distance = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":270
 *         public CompositionGraphNode end
 *         public dict distance
 *         public dict unvisited_finite_distance             # <<<<<<<<<<<<<<
 *         public double limit
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->unvisited_finite_distance);
  __pyx_r = __pyx_v_self->unvisited_finite_distance;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(0, 270, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->unvisited_finite_distance);
  __Pyx_DECREF(__pyx_v_self->unvisited_finite_distance);
  __pyx_v_self->unvisited_finite_distance = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.unvisited_finite_distance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->unvisited_finite_distance);
  __Pyx_DECREF(__pyx_v_self->unvisited_finite_distance);
  __pyx_v_self->unvisited_finite_distance = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":271
 *         public dict distance
 *         public dict unvisited_finite_distance
 *         public double limit             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, graph, start, end, limit=INF):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->limit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_v_self->limit = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.limit.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_10__reduce_cython__, "DijkstraPathFinder.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_10__reduce_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_10__reduce_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_10__reduce_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.distance, self.end, self.graph, self.limit, self.start, self.unvisited_finite_distance)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->limit); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->distance);
  __Pyx_GIVEREF(__pyx_v_self->distance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->distance)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->end);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->end);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->end))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->graph);
  __Pyx_GIVEREF(__pyx_v_self->graph);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->graph)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->start);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->start);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, ((PyObject *)__pyx_v_self->start))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->unvisited_finite_distance);
  __Pyx_GIVEREF(__pyx_v_self->unvisited_finite_distance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_v_self->unvisited_finite_distance)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.distance, self.end, self.graph, self.limit, self.start, self.unvisited_finite_distance)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.distance, self.end, self.graph, self.limit, self.start, self.unvisited_finite_distance)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.distance is not None or self.end is not None or self.graph is not None or self.start is not None or self.unvisited_finite_distance is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.distance, self.end, self.graph, self.limit, self.start, self.unvisited_finite_distance)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.distance is not None or self.end is not None or self.graph is not None or self.start is not None or self.unvisited_finite_distance is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->distance != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->end) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->graph != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->start) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->unvisited_finite_distance != ((PyObject*)Py_None));
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.distance is not None or self.end is not None or self.graph is not None or self.start is not None or self.unvisited_finite_distance is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.distance is not None or self.end is not None or self.graph is not None or self.start is not None or self.unvisited_finite_distance is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_DijkstraPathFinde); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_243238152);
    __Pyx_GIVEREF(__pyx_int_243238152);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_243238152)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.distance is not None or self.end is not None or self.graph is not None or self.start is not None or self.unvisited_finite_distance is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, None), state
 *     else:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DijkstraPathFinder__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_DijkstraPathFinde); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_243238152);
    __Pyx_GIVEREF(__pyx_int_243238152);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_243238152)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DijkstraPathFinder__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_12__setstate_cython__, "DijkstraPathFinder.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_12__setstate_cython__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_12__setstate_cython__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_12__setstate_cython__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DijkstraPathFinder__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_DijkstraPathFinder__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DijkstraPathFinder__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.DijkstraPathFinder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":398
 * 
 * 
 * cdef int initialize_glycan_composition_vector(size_t size, glycan_composition_vector* self) nogil:             # <<<<<<<<<<<<<<
 *     self.size = size
 *     self.counts = <int*>malloc(sizeof(int) * size)
 */

static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(size_t __pyx_v_size, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":399
 * 
 * cdef int initialize_glycan_composition_vector(size_t size, glycan_composition_vector* self) nogil:
 *     self.size = size             # <<<<<<<<<<<<<<
 *     self.counts = <int*>malloc(sizeof(int) * size)
 *     if self.counts == NULL:
 */
  __pyx_v_self->size = __pyx_v_size;

  /* "glycresoft/_c/composition_network/graph.pyx":400
 * cdef int initialize_glycan_composition_vector(size_t size, glycan_composition_vector* self) nogil:
 *     self.size = size
 *     self.counts = <int*>malloc(sizeof(int) * size)             # <<<<<<<<<<<<<<
 *     if self.counts == NULL:
 *         return 1
 */
  __pyx_v_self->counts = ((int *)malloc(((sizeof(int)) * __pyx_v_size)));

  /* "glycresoft/_c/composition_network/graph.pyx":401
 *     self.size = size
 *     self.counts = <int*>malloc(sizeof(int) * size)
 *     if self.counts == NULL:             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  __pyx_t_1 = (__pyx_v_self->counts == NULL);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":402
 *     self.counts = <int*>malloc(sizeof(int) * size)
 *     if self.counts == NULL:
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":401
 *     self.size = size
 *     self.counts = <int*>malloc(sizeof(int) * size)
 *     if self.counts == NULL:             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":403
 *     if self.counts == NULL:
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":398
 * 
 * 
 * cdef int initialize_glycan_composition_vector(size_t size, glycan_composition_vector* self) nogil:             # <<<<<<<<<<<<<<
 *     self.size = size
 *     self.counts = <int*>malloc(sizeof(int) * size)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":406
 * 
 * 
 * cdef double glycan_composition_vector_distance(glycan_composition_vector* self, glycan_composition_vector* other) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

static double __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_distance(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_self, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_other) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  double __pyx_v_distance;
  double __pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":410
 *         size_t i, n
 *         double distance
 *     distance = 0             # <<<<<<<<<<<<<<
 *     n = self.size
 *     if other.size != n:
 */
  __pyx_v_distance = 0.0;

  /* "glycresoft/_c/composition_network/graph.pyx":411
 *         double distance
 *     distance = 0
 *     n = self.size             # <<<<<<<<<<<<<<
 *     if other.size != n:
 *         return INF
 */
  __pyx_t_1 = __pyx_v_self->size;
  __pyx_v_n = __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":412
 *     distance = 0
 *     n = self.size
 *     if other.size != n:             # <<<<<<<<<<<<<<
 *         return INF
 *     for i in range(n):
 */
  __pyx_t_2 = (__pyx_v_other->size != __pyx_v_n);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":413
 *     n = self.size
 *     if other.size != n:
 *         return INF             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         distance += labs(self.counts[i] - other.counts[i])
 */
    __pyx_r = __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":412
 *     distance = 0
 *     n = self.size
 *     if other.size != n:             # <<<<<<<<<<<<<<
 *         return INF
 *     for i in range(n):
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":414
 *     if other.size != n:
 *         return INF
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         distance += labs(self.counts[i] - other.counts[i])
 *     return distance
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glycresoft/_c/composition_network/graph.pyx":415
 *         return INF
 *     for i in range(n):
 *         distance += labs(self.counts[i] - other.counts[i])             # <<<<<<<<<<<<<<
 *     return distance
 * 
 */
    __pyx_v_distance = (__pyx_v_distance + abs(((__pyx_v_self->counts[__pyx_v_i]) - (__pyx_v_other->counts[__pyx_v_i]))));
  }

  /* "glycresoft/_c/composition_network/graph.pyx":416
 *     for i in range(n):
 *         distance += labs(self.counts[i] - other.counts[i])
 *     return distance             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_distance;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":406
 * 
 * 
 * cdef double glycan_composition_vector_distance(glycan_composition_vector* self, glycan_composition_vector* other) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":419
 * 
 * 
 * cdef int glycan_composition_vector_difference(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_difference(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_self, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_other, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_into) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":423
 *         size_t i, n
 * 
 *     n = self.size             # <<<<<<<<<<<<<<
 *     if other.size != n:
 *         return 1
 */
  __pyx_t_1 = __pyx_v_self->size;
  __pyx_v_n = __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":424
 * 
 *     n = self.size
 *     if other.size != n:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_2 = (__pyx_v_other->size != __pyx_v_n);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":425
 *     n = self.size
 *     if other.size != n:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     if into.counts == NULL:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":424
 * 
 *     n = self.size
 *     if other.size != n:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":427
 *         return 1
 * 
 *     if into.counts == NULL:             # <<<<<<<<<<<<<<
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 */
  __pyx_t_2 = (__pyx_v_into->counts == NULL);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":428
 * 
 *     if into.counts == NULL:
 *         into.counts = <int*>malloc(sizeof(int) * n)             # <<<<<<<<<<<<<<
 *         into.size = n
 *         if into.counts == NULL:
 */
    __pyx_v_into->counts = ((int *)malloc(((sizeof(int)) * __pyx_v_n)));

    /* "glycresoft/_c/composition_network/graph.pyx":429
 *     if into.counts == NULL:
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n             # <<<<<<<<<<<<<<
 *         if into.counts == NULL:
 *             return 2
 */
    __pyx_v_into->size = __pyx_v_n;

    /* "glycresoft/_c/composition_network/graph.pyx":430
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 *         if into.counts == NULL:             # <<<<<<<<<<<<<<
 *             return 2
 * 
 */
    __pyx_t_2 = (__pyx_v_into->counts == NULL);
    if (__pyx_t_2) {

      /* "glycresoft/_c/composition_network/graph.pyx":431
 *         into.size = n
 *         if into.counts == NULL:
 *             return 2             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n):
 */
      __pyx_r = 2;
      goto __pyx_L0;

      /* "glycresoft/_c/composition_network/graph.pyx":430
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 *         if into.counts == NULL:             # <<<<<<<<<<<<<<
 *             return 2
 * 
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":427
 *         return 1
 * 
 *     if into.counts == NULL:             # <<<<<<<<<<<<<<
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":433
 *             return 2
 * 
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         into.counts[i] = self.counts[i] - other.counts[i]
 *     return 0
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glycresoft/_c/composition_network/graph.pyx":434
 * 
 *     for i in range(n):
 *         into.counts[i] = self.counts[i] - other.counts[i]             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    (__pyx_v_into->counts[__pyx_v_i]) = ((__pyx_v_self->counts[__pyx_v_i]) - (__pyx_v_other->counts[__pyx_v_i]));
  }

  /* "glycresoft/_c/composition_network/graph.pyx":435
 *     for i in range(n):
 *         into.counts[i] = self.counts[i] - other.counts[i]
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":419
 * 
 * 
 * cdef int glycan_composition_vector_difference(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":438
 * 
 * 
 * cdef int glycan_composition_vector_addition(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_addition(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_self, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_other, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_into) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":442
 *         size_t i, n
 * 
 *     n = self.size             # <<<<<<<<<<<<<<
 *     if other.size != n:
 *         return 1
 */
  __pyx_t_1 = __pyx_v_self->size;
  __pyx_v_n = __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":443
 * 
 *     n = self.size
 *     if other.size != n:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_2 = (__pyx_v_other->size != __pyx_v_n);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":444
 *     n = self.size
 *     if other.size != n:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     if into.counts == NULL:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":443
 * 
 *     n = self.size
 *     if other.size != n:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":446
 *         return 1
 * 
 *     if into.counts == NULL:             # <<<<<<<<<<<<<<
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 */
  __pyx_t_2 = (__pyx_v_into->counts == NULL);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":447
 * 
 *     if into.counts == NULL:
 *         into.counts = <int*>malloc(sizeof(int) * n)             # <<<<<<<<<<<<<<
 *         into.size = n
 *         if into.counts == NULL:
 */
    __pyx_v_into->counts = ((int *)malloc(((sizeof(int)) * __pyx_v_n)));

    /* "glycresoft/_c/composition_network/graph.pyx":448
 *     if into.counts == NULL:
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n             # <<<<<<<<<<<<<<
 *         if into.counts == NULL:
 *             return 2
 */
    __pyx_v_into->size = __pyx_v_n;

    /* "glycresoft/_c/composition_network/graph.pyx":449
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 *         if into.counts == NULL:             # <<<<<<<<<<<<<<
 *             return 2
 * 
 */
    __pyx_t_2 = (__pyx_v_into->counts == NULL);
    if (__pyx_t_2) {

      /* "glycresoft/_c/composition_network/graph.pyx":450
 *         into.size = n
 *         if into.counts == NULL:
 *             return 2             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n):
 */
      __pyx_r = 2;
      goto __pyx_L0;

      /* "glycresoft/_c/composition_network/graph.pyx":449
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 *         if into.counts == NULL:             # <<<<<<<<<<<<<<
 *             return 2
 * 
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":446
 *         return 1
 * 
 *     if into.counts == NULL:             # <<<<<<<<<<<<<<
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":452
 *             return 2
 * 
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         into.counts[i] = self.counts[i] + other.counts[i]
 *     return 0
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glycresoft/_c/composition_network/graph.pyx":453
 * 
 *     for i in range(n):
 *         into.counts[i] = self.counts[i] + other.counts[i]             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    (__pyx_v_into->counts[__pyx_v_i]) = ((__pyx_v_self->counts[__pyx_v_i]) + (__pyx_v_other->counts[__pyx_v_i]));
  }

  /* "glycresoft/_c/composition_network/graph.pyx":454
 *     for i in range(n):
 *         into.counts[i] = self.counts[i] + other.counts[i]
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":438
 * 
 * 
 * cdef int glycan_composition_vector_addition(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":457
 * 
 * 
 * cdef int destroy_glycan_composition_vector(glycan_composition_vector* self) nogil:             # <<<<<<<<<<<<<<
 *     free(self.counts)
 *     return 0
 */

static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_destroy_glycan_composition_vector(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_self) {
  int __pyx_r;

  /* "glycresoft/_c/composition_network/graph.pyx":458
 * 
 * cdef int destroy_glycan_composition_vector(glycan_composition_vector* self) nogil:
 *     free(self.counts)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_self->counts);

  /* "glycresoft/_c/composition_network/graph.pyx":459
 * cdef int destroy_glycan_composition_vector(glycan_composition_vector* self) nogil:
 *     free(self.counts)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":457
 * 
 * 
 * cdef int destroy_glycan_composition_vector(glycan_composition_vector* self) nogil:             # <<<<<<<<<<<<<<
 *     free(self.counts)
 *     return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":462
 * 
 * 
 * cdef int copy_glycan_composition_vector(glycan_composition_vector* self, glycan_composition_vector* into) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

static int __pyx_f_10glycresoft_2_c_19composition_network_5graph_copy_glycan_composition_vector(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_self, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_into) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":466
 *         size_t i, n
 * 
 *     n = self.size             # <<<<<<<<<<<<<<
 *     if into.size != self.size:
 *         free(into.counts)
 */
  __pyx_t_1 = __pyx_v_self->size;
  __pyx_v_n = __pyx_t_1;

  /* "glycresoft/_c/composition_network/graph.pyx":467
 * 
 *     n = self.size
 *     if into.size != self.size:             # <<<<<<<<<<<<<<
 *         free(into.counts)
 *         into.counts = NULL
 */
  __pyx_t_2 = (__pyx_v_into->size != __pyx_v_self->size);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":468
 *     n = self.size
 *     if into.size != self.size:
 *         free(into.counts)             # <<<<<<<<<<<<<<
 *         into.counts = NULL
 * 
 */
    free(__pyx_v_into->counts);

    /* "glycresoft/_c/composition_network/graph.pyx":469
 *     if into.size != self.size:
 *         free(into.counts)
 *         into.counts = NULL             # <<<<<<<<<<<<<<
 * 
 *     if into.counts == NULL:
 */
    __pyx_v_into->counts = NULL;

    /* "glycresoft/_c/composition_network/graph.pyx":467
 * 
 *     n = self.size
 *     if into.size != self.size:             # <<<<<<<<<<<<<<
 *         free(into.counts)
 *         into.counts = NULL
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":471
 *         into.counts = NULL
 * 
 *     if into.counts == NULL:             # <<<<<<<<<<<<<<
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 */
  __pyx_t_2 = (__pyx_v_into->counts == NULL);
  if (__pyx_t_2) {

    /* "glycresoft/_c/composition_network/graph.pyx":472
 * 
 *     if into.counts == NULL:
 *         into.counts = <int*>malloc(sizeof(int) * n)             # <<<<<<<<<<<<<<
 *         into.size = n
 *         if into.counts == NULL:
 */
    __pyx_v_into->counts = ((int *)malloc(((sizeof(int)) * __pyx_v_n)));

    /* "glycresoft/_c/composition_network/graph.pyx":473
 *     if into.counts == NULL:
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n             # <<<<<<<<<<<<<<
 *         if into.counts == NULL:
 *             return 1
 */
    __pyx_v_into->size = __pyx_v_n;

    /* "glycresoft/_c/composition_network/graph.pyx":474
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 *         if into.counts == NULL:             # <<<<<<<<<<<<<<
 *             return 1
 * 
 */
    __pyx_t_2 = (__pyx_v_into->counts == NULL);
    if (__pyx_t_2) {

      /* "glycresoft/_c/composition_network/graph.pyx":475
 *         into.size = n
 *         if into.counts == NULL:
 *             return 1             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n):
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "glycresoft/_c/composition_network/graph.pyx":474
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 *         if into.counts == NULL:             # <<<<<<<<<<<<<<
 *             return 1
 * 
 */
    }

    /* "glycresoft/_c/composition_network/graph.pyx":471
 *         into.counts = NULL
 * 
 *     if into.counts == NULL:             # <<<<<<<<<<<<<<
 *         into.counts = <int*>malloc(sizeof(int) * n)
 *         into.size = n
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":477
 *             return 1
 * 
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         into.counts[i] = self.counts[i]
 * 
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glycresoft/_c/composition_network/graph.pyx":478
 * 
 *     for i in range(n):
 *         into.counts[i] = self.counts[i]             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    (__pyx_v_into->counts[__pyx_v_i]) = (__pyx_v_self->counts[__pyx_v_i]);
  }

  /* "glycresoft/_c/composition_network/graph.pyx":480
 *         into.counts[i] = self.counts[i]
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":462
 * 
 * 
 * cdef int copy_glycan_composition_vector(glycan_composition_vector* self, glycan_composition_vector* into) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":485
 * cdef class GlycanCompositionVectorContext:
 * 
 *     def __init__(self, components):             # <<<<<<<<<<<<<<
 *         self.components = list(components)
 *         self.component_count = len(self.components)
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_components = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_components,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_components)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 485, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_components = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 485, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext___init__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_components);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_components) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":486
 * 
 *     def __init__(self, components):
 *         self.components = list(components)             # <<<<<<<<<<<<<<
 *         self.component_count = len(self.components)
 * 
 */
  __pyx_t_1 = PySequence_List(__pyx_v_components); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->components);
  __Pyx_DECREF(__pyx_v_self->components);
  __pyx_v_self->components = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":487
 *     def __init__(self, components):
 *         self.components = list(components)
 *         self.component_count = len(self.components)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_1 = __pyx_v_self->components;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 487, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->component_count = __pyx_t_2;

  /* "glycresoft/_c/composition_network/graph.pyx":485
 * cdef class GlycanCompositionVectorContext:
 * 
 *     def __init__(self, components):             # <<<<<<<<<<<<<<
 *         self.components = list(components)
 *         self.component_count = len(self.components)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":489
 *         self.component_count = len(self.components)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.components, )
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_3__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_2__reduce__, "GlycanCompositionVectorContext.__reduce__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_3__reduce__ = {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_2__reduce__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_3__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_2__reduce__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_2__reduce__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":490
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.components, )             # <<<<<<<<<<<<<<
 * 
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->components);
  __Pyx_GIVEREF(__pyx_v_self->components);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->components)) __PYX_ERR(0, 490, __pyx_L1_error);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":489
 *         self.component_count = len(self.components)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.components, )
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":492
 *         return self.__class__, (self.components, )
 * 
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:
 */

static struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode_raw(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_glycan_composition) {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_ptr;
  size_t __pyx_v_i;
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_raw", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":493
 * 
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition):
 *         cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))             # <<<<<<<<<<<<<<
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:
 *             raise MemoryError()
 */
  __pyx_v_ptr = ((struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)malloc((sizeof(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector))));

  /* "glycresoft/_c/composition_network/graph.pyx":494
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition):
 *         cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         for i in range(self.component_count):
 */
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(__pyx_v_self->component_count, __pyx_v_ptr); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 1);
  if (unlikely(__pyx_t_2)) {

    /* "glycresoft/_c/composition_network/graph.pyx":495
 *         cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         for i in range(self.component_count):
 *             ptr.counts[i] = glycan_composition[self.components[i]]
 */
    PyErr_NoMemory(); __PYX_ERR(0, 495, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":494
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition):
 *         cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         for i in range(self.component_count):
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":496
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:
 *             raise MemoryError()
 *         for i in range(self.component_count):             # <<<<<<<<<<<<<<
 *             ptr.counts[i] = glycan_composition[self.components[i]]
 *         return ptr
 */
  __pyx_t_3 = __pyx_v_self->component_count;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "glycresoft/_c/composition_network/graph.pyx":497
 *             raise MemoryError()
 *         for i in range(self.component_count):
 *             ptr.counts[i] = glycan_composition[self.components[i]]             # <<<<<<<<<<<<<<
 *         return ptr
 * 
 */
    if (unlikely(__pyx_v_self->components == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 497, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_self->components, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_glycan_composition, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    (__pyx_v_ptr->counts[__pyx_v_i]) = __pyx_t_1;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":498
 *         for i in range(self.component_count):
 *             ptr.counts[i] = glycan_composition[self.components[i]]
 *         return ptr             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycanCompositionVector encode(self, glycan_composition):
 */
  __pyx_r = __pyx_v_ptr;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":492
 *         return self.__class__, (self.components, )
 * 
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if initialize_glycan_composition_vector(self.component_count, ptr) == 1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.encode_raw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":500
 *         return ptr
 * 
 *     cpdef GlycanCompositionVector encode(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_glycan_composition, int __pyx_skip_dispatch) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_glycan_composition};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector))))) __PYX_ERR(0, 500, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":501
 * 
 *     cpdef GlycanCompositionVector encode(self, glycan_composition):
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))             # <<<<<<<<<<<<<<
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_6 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self->__pyx_vtab)->encode_raw(__pyx_v_self, __pyx_v_glycan_composition); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_t_6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":500
 *         return ptr
 * 
 *     cpdef GlycanCompositionVector encode(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_4encode, "GlycanCompositionVectorContext.encode(self, glycan_composition) -> GlycanCompositionVector");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode = {"encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_4encode};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_glycan_composition = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_glycan_composition,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glycan_composition)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "encode") < 0)) __PYX_ERR(0, 500, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_glycan_composition = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("encode", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 500, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_4encode(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_glycan_composition);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_4encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_glycan_composition) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode(__pyx_v_self, __pyx_v_glycan_composition, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":503
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  int __pyx_v_value;
  PyObject *__pyx_v_gc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_gcvec)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":507
 *             size_t i
 *             int value
 *         gc = HashableGlycanComposition()             # <<<<<<<<<<<<<<
 *         if self.component_count != gcvec.ptr.size:
 *             raise ValueError("Dimensionality does not match")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HashableGlycanComposition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_gc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":508
 *             int value
 *         gc = HashableGlycanComposition()
 *         if self.component_count != gcvec.ptr.size:             # <<<<<<<<<<<<<<
 *             raise ValueError("Dimensionality does not match")
 *         for i in range(self.component_count):
 */
  __pyx_t_6 = (__pyx_v_self->component_count != __pyx_v_gcvec->ptr->size);
  if (unlikely(__pyx_t_6)) {

    /* "glycresoft/_c/composition_network/graph.pyx":509
 *         gc = HashableGlycanComposition()
 *         if self.component_count != gcvec.ptr.size:
 *             raise ValueError("Dimensionality does not match")             # <<<<<<<<<<<<<<
 *         for i in range(self.component_count):
 *             value = gcvec.ptr.counts[i]
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 509, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":508
 *             int value
 *         gc = HashableGlycanComposition()
 *         if self.component_count != gcvec.ptr.size:             # <<<<<<<<<<<<<<
 *             raise ValueError("Dimensionality does not match")
 *         for i in range(self.component_count):
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":510
 *         if self.component_count != gcvec.ptr.size:
 *             raise ValueError("Dimensionality does not match")
 *         for i in range(self.component_count):             # <<<<<<<<<<<<<<
 *             value = gcvec.ptr.counts[i]
 *             if value != 0:
 */
  __pyx_t_7 = __pyx_v_self->component_count;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "glycresoft/_c/composition_network/graph.pyx":511
 *             raise ValueError("Dimensionality does not match")
 *         for i in range(self.component_count):
 *             value = gcvec.ptr.counts[i]             # <<<<<<<<<<<<<<
 *             if value != 0:
 *                 gc[self.components[i]] = value
 */
    __pyx_v_value = (__pyx_v_gcvec->ptr->counts[__pyx_v_i]);

    /* "glycresoft/_c/composition_network/graph.pyx":512
 *         for i in range(self.component_count):
 *             value = gcvec.ptr.counts[i]
 *             if value != 0:             # <<<<<<<<<<<<<<
 *                 gc[self.components[i]] = value
 *         return gc
 */
    __pyx_t_6 = (__pyx_v_value != 0);
    if (__pyx_t_6) {

      /* "glycresoft/_c/composition_network/graph.pyx":513
 *             value = gcvec.ptr.counts[i]
 *             if value != 0:
 *                 gc[self.components[i]] = value             # <<<<<<<<<<<<<<
 *         return gc
 * 
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(__pyx_v_self->components == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 513, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_self->components, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_gc, __pyx_t_2, __pyx_t_1) < 0))) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "glycresoft/_c/composition_network/graph.pyx":512
 *         for i in range(self.component_count):
 *             value = gcvec.ptr.counts[i]
 *             if value != 0:             # <<<<<<<<<<<<<<
 *                 gc[self.components[i]] = value
 *         return gc
 */
    }
  }

  /* "glycresoft/_c/composition_network/graph.pyx":514
 *             if value != 0:
 *                 gc[self.components[i]] = value
 *         return gc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_gc);
  __pyx_r = __pyx_v_gc;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":503
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_gc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_6decode, "GlycanCompositionVectorContext.decode(self, GlycanCompositionVector gcvec)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode = {"decode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_6decode};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gcvec,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gcvec)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 503, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "decode") < 0)) __PYX_ERR(0, 503, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gcvec = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("decode", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 503, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gcvec), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "gcvec", 0))) __PYX_ERR(0, 503, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_6decode(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_gcvec);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_6decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_decode(__pyx_v_self, __pyx_v_gcvec, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":70
 * cdef class GlycanCompositionVectorContext:
 *     cdef:
 *         public list components             # <<<<<<<<<<<<<<
 *         public size_t component_count
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->components);
  __pyx_r = __pyx_v_self->components;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_value))) __PYX_ERR(1, 70, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->components);
  __Pyx_DECREF(__pyx_v_self->components);
  __pyx_v_self->components = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.components.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->components);
  __Pyx_DECREF(__pyx_v_self->components);
  __pyx_v_self->components = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":71
 *     cdef:
 *         public list components
 *         public size_t component_count             # <<<<<<<<<<<<<<
 * 
 *     cdef glycan_composition_vector* encode_raw(self, glycan_composition)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->component_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.component_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 71, __pyx_L1_error)
  __pyx_v_self->component_count = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVectorContext.component_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":517
 * 
 * 
 * cpdef _reconstruct_GlycanCompositionVector(state):             # <<<<<<<<<<<<<<
 *     cdef:
 *         glycan_composition_vector* ptr
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_3_reconstruct_GlycanCompositionVector(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph__reconstruct_GlycanCompositionVector(PyObject *__pyx_v_state, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_ptr;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_v_counts = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  long __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_reconstruct_GlycanCompositionVector", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":522
 *         size_t i, n
 *         list counts
 *     counts = state[0]             # <<<<<<<<<<<<<<
 *     n = len(counts)
 *     ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 522, __pyx_L1_error)
  __pyx_v_counts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":523
 *         list counts
 *     counts = state[0]
 *     n = len(counts)             # <<<<<<<<<<<<<<
 *     ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *     initialize_glycan_composition_vector(n, ptr)
 */
  if (unlikely(__pyx_v_counts == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 523, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_counts); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 523, __pyx_L1_error)
  __pyx_v_n = __pyx_t_2;

  /* "glycresoft/_c/composition_network/graph.pyx":524
 *     counts = state[0]
 *     n = len(counts)
 *     ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))             # <<<<<<<<<<<<<<
 *     initialize_glycan_composition_vector(n, ptr)
 *     for i in range(n):
 */
  __pyx_v_ptr = ((struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)malloc((sizeof(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector))));

  /* "glycresoft/_c/composition_network/graph.pyx":525
 *     n = len(counts)
 *     ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *     initialize_glycan_composition_vector(n, ptr)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         ptr.counts[i] = PyInt_AsLong(counts[i])
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(__pyx_v_n, __pyx_v_ptr); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 525, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":526
 *     ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *     initialize_glycan_composition_vector(n, ptr)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ptr.counts[i] = PyInt_AsLong(counts[i])
 *     return GlycanCompositionVector._create(ptr)
 */
  __pyx_t_4 = __pyx_v_n;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "glycresoft/_c/composition_network/graph.pyx":527
 *     initialize_glycan_composition_vector(n, ptr)
 *     for i in range(n):
 *         ptr.counts[i] = PyInt_AsLong(counts[i])             # <<<<<<<<<<<<<<
 *     return GlycanCompositionVector._create(ptr)
 * 
 */
    if (unlikely(__pyx_v_counts == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 527, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_counts, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyInt_AsLong(__pyx_t_1); if (unlikely(__pyx_t_7 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_ptr->counts[__pyx_v_i]) = __pyx_t_7;
  }

  /* "glycresoft/_c/composition_network/graph.pyx":528
 *     for i in range(n):
 *         ptr.counts[i] = PyInt_AsLong(counts[i])
 *     return GlycanCompositionVector._create(ptr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_v_ptr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":517
 * 
 * 
 * cpdef _reconstruct_GlycanCompositionVector(state):             # <<<<<<<<<<<<<<
 *     cdef:
 *         glycan_composition_vector* ptr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph._reconstruct_GlycanCompositionVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_counts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_3_reconstruct_GlycanCompositionVector(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_2_reconstruct_GlycanCompositionVector, "_reconstruct_GlycanCompositionVector(state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_3_reconstruct_GlycanCompositionVector = {"_reconstruct_GlycanCompositionVector", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_3_reconstruct_GlycanCompositionVector, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_2_reconstruct_GlycanCompositionVector};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_3_reconstruct_GlycanCompositionVector(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_reconstruct_GlycanCompositionVector (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_reconstruct_GlycanCompositionVector") < 0)) __PYX_ERR(0, 517, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_reconstruct_GlycanCompositionVector", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 517, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph._reconstruct_GlycanCompositionVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_2_reconstruct_GlycanCompositionVector(__pyx_self, __pyx_v_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_2_reconstruct_GlycanCompositionVector(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_reconstruct_GlycanCompositionVector", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph__reconstruct_GlycanCompositionVector(__pyx_v_state, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph._reconstruct_GlycanCompositionVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":538
 * 
 *     @staticmethod
 *     cdef GlycanCompositionVector _create(glycan_composition_vector* ptr):             # <<<<<<<<<<<<<<
 *         cdef GlycanCompositionVector self = GlycanCompositionVector.__new__(GlycanCompositionVector)
 *         self.ptr = ptr
 */

static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_ptr) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":539
 *     @staticmethod
 *     cdef GlycanCompositionVector _create(glycan_composition_vector* ptr):
 *         cdef GlycanCompositionVector self = GlycanCompositionVector.__new__(GlycanCompositionVector)             # <<<<<<<<<<<<<<
 *         self.ptr = ptr
 *         return self
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(((PyTypeObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":540
 *     cdef GlycanCompositionVector _create(glycan_composition_vector* ptr):
 *         cdef GlycanCompositionVector self = GlycanCompositionVector.__new__(GlycanCompositionVector)
 *         self.ptr = ptr             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_v_self->ptr = __pyx_v_ptr;

  /* "glycresoft/_c/composition_network/graph.pyx":541
 *         cdef GlycanCompositionVector self = GlycanCompositionVector.__new__(GlycanCompositionVector)
 *         self.ptr = ptr
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":538
 * 
 *     @staticmethod
 *     cdef GlycanCompositionVector _create(glycan_composition_vector* ptr):             # <<<<<<<<<<<<<<
 *         cdef GlycanCompositionVector self = GlycanCompositionVector.__new__(GlycanCompositionVector)
 *         self.ptr = ptr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":543
 *         return self
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return _reconstruct_GlycanCompositionVector, (list(self), )
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_1__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector___reduce__, "GlycanCompositionVector.__reduce__(self)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_1__reduce__ = {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_1__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector___reduce__};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_1__reduce__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce__", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector___reduce__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector___reduce__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":544
 * 
 *     def __reduce__(self):
 *         return _reconstruct_GlycanCompositionVector, (list(self), )             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_reconstruct_GlycanCompositionVe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":543
 *         return self
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return _reconstruct_GlycanCompositionVector, (list(self), )
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":546
 *         return _reconstruct_GlycanCompositionVector, (list(self), )
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.ptr != NULL:
 *             destroy_glycan_composition_vector(self.ptr)
 */

/* Python wrapper */
static void __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_2__dealloc__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_2__dealloc__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":547
 * 
 *     def __dealloc__(self):
 *         if self.ptr != NULL:             # <<<<<<<<<<<<<<
 *             destroy_glycan_composition_vector(self.ptr)
 *             free(self.ptr)
 */
  __pyx_t_1 = (__pyx_v_self->ptr != NULL);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":548
 *     def __dealloc__(self):
 *         if self.ptr != NULL:
 *             destroy_glycan_composition_vector(self.ptr)             # <<<<<<<<<<<<<<
 *             free(self.ptr)
 *             self.ptr = NULL
 */
    __pyx_t_2 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_destroy_glycan_composition_vector(__pyx_v_self->ptr); if (unlikely(__pyx_t_2 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":549
 *         if self.ptr != NULL:
 *             destroy_glycan_composition_vector(self.ptr)
 *             free(self.ptr)             # <<<<<<<<<<<<<<
 *             self.ptr = NULL
 * 
 */
    free(__pyx_v_self->ptr);

    /* "glycresoft/_c/composition_network/graph.pyx":550
 *             destroy_glycan_composition_vector(self.ptr)
 *             free(self.ptr)
 *             self.ptr = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_v_self->ptr = NULL;

    /* "glycresoft/_c/composition_network/graph.pyx":547
 * 
 *     def __dealloc__(self):
 *         if self.ptr != NULL:             # <<<<<<<<<<<<<<
 *             destroy_glycan_composition_vector(self.ptr)
 *             free(self.ptr)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":546
 *         return _reconstruct_GlycanCompositionVector, (list(self), )
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.ptr != NULL:
 *             destroy_glycan_composition_vector(self.ptr)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glycresoft._c.composition_network.graph.GlycanCompositionVector.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
}

/* "glycresoft/_c/composition_network/graph.pyx":552
 *             self.ptr = NULL
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_4__getitem__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_4__getitem__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":553
 * 
 *     def __getitem__(self, int i):
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":554
 *     def __getitem__(self, int i):
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 554, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":553
 * 
 *     def __getitem__(self, int i):
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":555
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(i)
 *         return self.ptr.counts[i]
 */
  __pyx_t_3 = (__pyx_v_i >= __pyx_v_self->ptr->size);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_i < 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":556
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)             # <<<<<<<<<<<<<<
 *         return self.ptr.counts[i]
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 556, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":555
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(i)
 *         return self.ptr.counts[i]
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":557
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)
 *         return self.ptr.counts[i]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_self->ptr->counts[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":552
 *             self.ptr = NULL
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":559
 *         return self.ptr.counts[i]
 * 
 *     def __setitem__(self, int i, int value):             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_value) {
  int __pyx_v_i;
  int __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L3_error)
  }
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_6__setitem__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), ((int)__pyx_v_i), ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_6__setitem__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, int __pyx_v_i, int __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":560
 * 
 *     def __setitem__(self, int i, int value):
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":561
 *     def __setitem__(self, int i, int value):
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 561, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 561, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":560
 * 
 *     def __setitem__(self, int i, int value):
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":562
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(i)
 *         self.ptr.counts[i] = value
 */
  __pyx_t_3 = (__pyx_v_i >= __pyx_v_self->ptr->size);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_i < 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":563
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)             # <<<<<<<<<<<<<<
 *         self.ptr.counts[i] = value
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 563, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":562
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if i >= self.ptr.size or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(i)
 *         self.ptr.counts[i] = value
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":564
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)
 *         self.ptr.counts[i] = value             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  (__pyx_v_self->ptr->counts[__pyx_v_i]) = __pyx_v_value;

  /* "glycresoft/_c/composition_network/graph.pyx":559
 *         return self.ptr.counts[i]
 * 
 *     def __setitem__(self, int i, int value):             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":566
 *         self.ptr.counts[i] = value
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_9__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_9__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_8__len__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_8__len__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":567
 * 
 *     def __len__(self):
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         return self.ptr.size
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":568
 *     def __len__(self):
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         return self.ptr.size
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 568, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":567
 * 
 *     def __len__(self):
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         return self.ptr.size
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":569
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         return self.ptr.size             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->ptr->size;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":566
 *         self.ptr.counts[i] = value
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":571
 *         return self.ptr.size
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_10__repr__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_10__repr__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr1__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":572
 * 
 *     def __repr__(self):
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"             # <<<<<<<<<<<<<<
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__7);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__7);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__7);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 572, __pyx_L1_error)
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_8genexpr1__pyx_v_i = __pyx_t_8;
      __pyx_t_5 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self), __pyx_8genexpr1__pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__8);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__8);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__8);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":571
 *         return self.ptr.size
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":574
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return INF
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_13distance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_distance(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("distance", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":575
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return INF
 *         if self.ptr == NULL:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_other) == Py_None);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":576
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:
 *         if other is None:
 *             return INF             # <<<<<<<<<<<<<<
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 */
    __pyx_r = __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":575
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return INF
 *         if self.ptr == NULL:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":577
 *         if other is None:
 *             return INF
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         return glycan_composition_vector_distance(self.ptr, other.ptr)
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":578
 *             return INF
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         return glycan_composition_vector_distance(self.ptr, other.ptr)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 578, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":577
 *         if other is None:
 *             return INF
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         return glycan_composition_vector_distance(self.ptr, other.ptr)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":579
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         return glycan_composition_vector_distance(self.ptr, other.ptr)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, GlycanCompositionVector other):
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_distance(__pyx_v_self->ptr, __pyx_v_other->ptr); if (unlikely(__pyx_t_3 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 579, __pyx_L1_error)
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":574
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return INF
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1.0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_13distance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_12distance, "GlycanCompositionVector.distance(self, GlycanCompositionVector other) -> double");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_13distance = {"distance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_13distance, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_12distance};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_13distance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("distance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 574, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "distance") < 0)) __PYX_ERR(0, 574, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("distance", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 574, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_12distance(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_12distance(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("distance", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_distance(__pyx_v_self, __pyx_v_other, 1); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":581
 *         return glycan_composition_vector_distance(self.ptr, other.ptr)
 * 
 *     def __eq__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_15__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_15__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_14__eq__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_14__eq__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  double __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":582
 * 
 *     def __eq__(self, GlycanCompositionVector other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         return self.distance(other) == 0
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_other) == Py_None);
  if (__pyx_t_1) {

    /* "glycresoft/_c/composition_network/graph.pyx":583
 *     def __eq__(self, GlycanCompositionVector other):
 *         if other is None:
 *             return False             # <<<<<<<<<<<<<<
 *         return self.distance(other) == 0
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":582
 * 
 *     def __eq__(self, GlycanCompositionVector other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         return self.distance(other) == 0
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":584
 *         if other is None:
 *             return False
 *         return self.distance(other) == 0             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, GlycanCompositionVector other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_distance(__pyx_v_self, __pyx_v_other, 0); if (unlikely(__pyx_t_2 == ((double)-1.0))) __PYX_ERR(0, 584, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_t_2 == 0.0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":581
 *         return glycan_composition_vector_distance(self.ptr, other.ptr)
 * 
 *     def __eq__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":586
 *         return self.distance(other) == 0
 * 
 *     def __ne__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_17__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_17__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 586, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_16__ne__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_16__ne__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":587
 * 
 *     def __ne__(self, GlycanCompositionVector other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":586
 *         return self.distance(other) == 0
 * 
 *     def __ne__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":589
 *         return not self == other
 * 
 *     cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_19difference(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_difference(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_new;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("difference", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":592
 *         cdef:
 *              glycan_composition_vector* new
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":593
 *              glycan_composition_vector* new
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         if other is None:
 *             raise TypeError("Cannot compute difference from `None`")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 593, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":592
 *         cdef:
 *              glycan_composition_vector* new
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":594
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if other is None:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot compute difference from `None`")
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_other) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":595
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 *             raise TypeError("Cannot compute difference from `None`")             # <<<<<<<<<<<<<<
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 595, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":594
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if other is None:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot compute difference from `None`")
 * 
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":597
 *             raise TypeError("Cannot compute difference from `None`")
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))             # <<<<<<<<<<<<<<
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:
 */
  __pyx_v_new = ((struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)malloc((sizeof(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector))));

  /* "glycresoft/_c/composition_network/graph.pyx":598
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         initialize_glycan_composition_vector(self.ptr.size, new)             # <<<<<<<<<<<<<<
 *         if new == NULL:
 *             raise MemoryError()
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(__pyx_v_self->ptr->size, __pyx_v_new); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 598, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":599
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:
 */
  __pyx_t_1 = (__pyx_v_new == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":600
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:
 *             raise MemoryError()
 */
    PyErr_NoMemory(); __PYX_ERR(0, 600, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":599
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":601
 *         if new == NULL:
 *             raise MemoryError()
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_difference(__pyx_v_self->ptr, __pyx_v_other->ptr, __pyx_v_new); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 601, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_3 == 2);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":602
 *             raise MemoryError()
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(new)
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 602, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":601
 *         if new == NULL:
 *             raise MemoryError()
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":603
 *         if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_v_new)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":589
 *         return not self == other
 * 
 *     cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.difference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_19difference(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_18difference, "GlycanCompositionVector.difference(self, GlycanCompositionVector other) -> GlycanCompositionVector");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_19difference = {"difference", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_19difference, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_18difference};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_19difference(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("difference (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 589, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "difference") < 0)) __PYX_ERR(0, 589, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("difference", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 589, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.difference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 589, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_18difference(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_18difference(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("difference", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_difference(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.difference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":605
 *         return GlycanCompositionVector._create(new)
 * 
 *     cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_21addition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_addition(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_new;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addition", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":608
 *         cdef:
 *              glycan_composition_vector* new
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":609
 *              glycan_composition_vector* new
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         if other is None:
 *             raise TypeError("Cannot compute addition from `None`")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 609, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":608
 *         cdef:
 *              glycan_composition_vector* new
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":610
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if other is None:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot compute addition from `None`")
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_other) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":611
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 *             raise TypeError("Cannot compute addition from `None`")             # <<<<<<<<<<<<<<
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 611, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":610
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         if other is None:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot compute addition from `None`")
 * 
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":613
 *             raise TypeError("Cannot compute addition from `None`")
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))             # <<<<<<<<<<<<<<
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:
 */
  __pyx_v_new = ((struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)malloc((sizeof(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector))));

  /* "glycresoft/_c/composition_network/graph.pyx":614
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         initialize_glycan_composition_vector(self.ptr.size, new)             # <<<<<<<<<<<<<<
 *         if new == NULL:
 *             raise MemoryError()
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(__pyx_v_self->ptr->size, __pyx_v_new); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":615
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:
 */
  __pyx_t_1 = (__pyx_v_new == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":616
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:
 *             raise MemoryError()
 */
    PyErr_NoMemory(); __PYX_ERR(0, 616, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":615
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":617
 *         if new == NULL:
 *             raise MemoryError()
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_addition(__pyx_v_self->ptr, __pyx_v_other->ptr, __pyx_v_new); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 617, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_3 == 2);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":618
 *             raise MemoryError()
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(new)
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 618, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":617
 *         if new == NULL:
 *             raise MemoryError()
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":619
 *         if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, GlycanCompositionVector other):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_v_new)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":605
 *         return GlycanCompositionVector._create(new)
 * 
 *     cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.addition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_21addition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_20addition, "GlycanCompositionVector.addition(self, GlycanCompositionVector other) -> GlycanCompositionVector");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_21addition = {"addition", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_21addition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_20addition};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_21addition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addition (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "addition") < 0)) __PYX_ERR(0, 605, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addition", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 605, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.addition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 605, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_20addition(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_20addition(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addition", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_addition(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.addition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":621
 *         return GlycanCompositionVector._create(new)
 * 
 *     def __sub__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         return self.difference(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_23__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_23__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 621, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_22__sub__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_22__sub__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sub__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":622
 * 
 *     def __sub__(self, GlycanCompositionVector other):
 *         return self.difference(other)             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, GlycanCompositionVector other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_difference(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":621
 *         return GlycanCompositionVector._create(new)
 * 
 *     def __sub__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         return self.difference(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":624
 *         return self.difference(other)
 * 
 *     def __add__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         return self.addition(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_25__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_25__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "other", 0))) __PYX_ERR(0, 624, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_24__add__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self), ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_24__add__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":625
 * 
 *     def __add__(self, GlycanCompositionVector other):
 *         return self.addition(other)             # <<<<<<<<<<<<<<
 * 
 *     def __neg__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_addition(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":624
 *         return self.difference(other)
 * 
 *     def __add__(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         return self.addition(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":627
 *         return self.addition(other)
 * 
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_27__neg__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_27__neg__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neg__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_26__neg__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_26__neg__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_new;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__neg__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":631
 *              glycan_composition_vector* new
 * 
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":632
 * 
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 632, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":631
 *              glycan_composition_vector* new
 * 
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":633
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))             # <<<<<<<<<<<<<<
 *         if new == NULL:
 *             raise MemoryError()
 */
  __pyx_v_new = ((struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)malloc((sizeof(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector))));

  /* "glycresoft/_c/composition_network/graph.pyx":634
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 */
  __pyx_t_1 = (__pyx_v_new == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":635
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 635, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":634
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":636
 *         if new == NULL:
 *             raise MemoryError()
 *         initialize_glycan_composition_vector(self.ptr.size, new)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.ptr.size):
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector(__pyx_v_self->ptr->size, __pyx_v_new); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":638
 *         initialize_glycan_composition_vector(self.ptr.size, new)
 * 
 *         for i in range(self.ptr.size):             # <<<<<<<<<<<<<<
 *             new.counts[i] = -self.ptr.counts[i]
 *         return GlycanCompositionVector._create(new)
 */
  __pyx_t_4 = __pyx_v_self->ptr->size;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "glycresoft/_c/composition_network/graph.pyx":639
 * 
 *         for i in range(self.ptr.size):
 *             new.counts[i] = -self.ptr.counts[i]             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(new)
 * 
 */
    (__pyx_v_new->counts[__pyx_v_i]) = (-(__pyx_v_self->ptr->counts[__pyx_v_i]));
  }

  /* "glycresoft/_c/composition_network/graph.pyx":640
 *         for i in range(self.ptr.size):
 *             new.counts[i] = -self.ptr.counts[i]
 *         return GlycanCompositionVector._create(new)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_v_new)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":627
 *         return self.addition(other)
 * 
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__neg__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":642
 *         return GlycanCompositionVector._create(new)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_29__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_29__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_28__hash__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_28__hash__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  Py_hash_t __pyx_v_hash_val;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":647
 *             Py_hash_t hash_val
 * 
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         hash_val = 0
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":648
 * 
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         hash_val = 0
 *         n = self.ptr.size
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 648, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":647
 *             Py_hash_t hash_val
 * 
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         hash_val = 0
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":649
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         hash_val = 0             # <<<<<<<<<<<<<<
 *         n = self.ptr.size
 *         for i in range(n):
 */
  __pyx_v_hash_val = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":650
 *             raise ValueError("Object in invalid state")
 *         hash_val = 0
 *         n = self.ptr.size             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             hash_val = (hash_val ^ self.ptr.counts[i]) * 82520UL
 */
  __pyx_t_3 = __pyx_v_self->ptr->size;
  __pyx_v_n = __pyx_t_3;

  /* "glycresoft/_c/composition_network/graph.pyx":651
 *         hash_val = 0
 *         n = self.ptr.size
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             hash_val = (hash_val ^ self.ptr.counts[i]) * 82520UL
 *         return hash_val
 */
  __pyx_t_3 = __pyx_v_n;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "glycresoft/_c/composition_network/graph.pyx":652
 *         n = self.ptr.size
 *         for i in range(n):
 *             hash_val = (hash_val ^ self.ptr.counts[i]) * 82520UL             # <<<<<<<<<<<<<<
 *         return hash_val
 * 
 */
    __pyx_v_hash_val = ((__pyx_v_hash_val ^ (__pyx_v_self->ptr->counts[__pyx_v_i])) * 82520UL);
  }

  /* "glycresoft/_c/composition_network/graph.pyx":653
 *         for i in range(n):
 *             hash_val = (hash_val ^ self.ptr.counts[i]) * 82520UL
 *         return hash_val             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycanCompositionVector clone(self):
 */
  __pyx_r = __pyx_v_hash_val;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":642
 *         return GlycanCompositionVector._create(new)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":655
 *         return hash_val
 * 
 *     cpdef GlycanCompositionVector clone(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             glycan_composition_vector* new
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_31clone(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_v_new;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":658
 *         cdef:
 *             glycan_composition_vector* new
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  __pyx_t_1 = (__pyx_v_self->ptr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":659
 *             glycan_composition_vector* new
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 659, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":658
 *         cdef:
 *             glycan_composition_vector* new
 *         if self.ptr == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":660
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))             # <<<<<<<<<<<<<<
 *         if new == NULL:
 *             raise MemoryError()
 */
  __pyx_v_new = ((struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)malloc((sizeof(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector))));

  /* "glycresoft/_c/composition_network/graph.pyx":661
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         new.counts = NULL
 */
  __pyx_t_1 = (__pyx_v_new == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":662
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         new.counts = NULL
 *         new.size = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 662, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":661
 *             raise ValueError("Object in invalid state")
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 *         if new == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         new.counts = NULL
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":663
 *         if new == NULL:
 *             raise MemoryError()
 *         new.counts = NULL             # <<<<<<<<<<<<<<
 *         new.size = 0
 *         if copy_glycan_composition_vector(self.ptr, new) != 0:
 */
  __pyx_v_new->counts = NULL;

  /* "glycresoft/_c/composition_network/graph.pyx":664
 *             raise MemoryError()
 *         new.counts = NULL
 *         new.size = 0             # <<<<<<<<<<<<<<
 *         if copy_glycan_composition_vector(self.ptr, new) != 0:
 *             raise MemoryError()
 */
  __pyx_v_new->size = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":665
 *         new.counts = NULL
 *         new.size = 0
 *         if copy_glycan_composition_vector(self.ptr, new) != 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)
 */
  __pyx_t_3 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_copy_glycan_composition_vector(__pyx_v_self->ptr, __pyx_v_new); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 665, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_1)) {

    /* "glycresoft/_c/composition_network/graph.pyx":666
 *         new.size = 0
 *         if copy_glycan_composition_vector(self.ptr, new) != 0:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(new)
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 666, __pyx_L1_error)

    /* "glycresoft/_c/composition_network/graph.pyx":665
 *         new.counts = NULL
 *         new.size = 0
 *         if copy_glycan_composition_vector(self.ptr, new) != 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":667
 *         if copy_glycan_composition_vector(self.ptr, new) != 0:
 *             raise MemoryError()
 *         return GlycanCompositionVector._create(new)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_v_new)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":655
 *         return hash_val
 * 
 *     cpdef GlycanCompositionVector clone(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             glycan_composition_vector* new
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_31clone(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_30clone, "GlycanCompositionVector.clone(self) -> GlycanCompositionVector");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_31clone = {"clone", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_31clone, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_30clone};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_31clone(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clone", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clone", 0))) return NULL;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_30clone(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_30clone(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.GlycanCompositionVector.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":672
 * cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):
 * 
 *     def __init__(self, components):             # <<<<<<<<<<<<<<
 *         super().__init__(components)
 *         self.encode_cache = {}
 */

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_components = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_components,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_components)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 672, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_components = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 672, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext___init__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_components);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext___init__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_components) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "glycresoft/_c/composition_network/graph.pyx":673
 * 
 *     def __init__(self, components):
 *         super().__init__(components)             # <<<<<<<<<<<<<<
 *         self.encode_cache = {}
 *         self.decode_cache = {}
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext))) __PYX_ERR(0, 673, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 673, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_components};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":674
 *     def __init__(self, components):
 *         super().__init__(components)
 *         self.encode_cache = {}             # <<<<<<<<<<<<<<
 *         self.decode_cache = {}
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->encode_cache);
  __Pyx_DECREF(__pyx_v_self->encode_cache);
  __pyx_v_self->encode_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":675
 *         super().__init__(components)
 *         self.encode_cache = {}
 *         self.decode_cache = {}             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycanCompositionVector encode(self, gc):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->decode_cache);
  __Pyx_DECREF(__pyx_v_self->decode_cache);
  __pyx_v_self->decode_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":672
 * cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):
 * 
 *     def __init__(self, components):             # <<<<<<<<<<<<<<
 *         super().__init__(components)
 *         self.encode_cache = {}
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":677
 *         self.decode_cache = {}
 * 
 *     cpdef GlycanCompositionVector encode(self, gc):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_gc, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ptemp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_value = 0;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_gc};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector))))) __PYX_ERR(0, 677, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":682
 *             GlycanCompositionVector value
 * 
 *         ptemp = PyDict_GetItem(self.encode_cache, gc)             # <<<<<<<<<<<<<<
 *         if ptemp != NULL:
 *             return <GlycanCompositionVector>ptemp
 */
  __pyx_t_1 = __pyx_v_self->encode_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_1, __pyx_v_gc);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":683
 * 
 *         ptemp = PyDict_GetItem(self.encode_cache, gc)
 *         if ptemp != NULL:             # <<<<<<<<<<<<<<
 *             return <GlycanCompositionVector>ptemp
 *         value = GlycanCompositionVector._create(self.encode_raw(gc))
 */
  __pyx_t_6 = (__pyx_v_ptemp != NULL);
  if (__pyx_t_6) {

    /* "glycresoft/_c/composition_network/graph.pyx":684
 *         ptemp = PyDict_GetItem(self.encode_cache, gc)
 *         if ptemp != NULL:
 *             return <GlycanCompositionVector>ptemp             # <<<<<<<<<<<<<<
 *         value = GlycanCompositionVector._create(self.encode_raw(gc))
 *         PyDict_SetItem(self.encode_cache, gc, value)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_ptemp));
    __pyx_r = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_v_ptemp);
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":683
 * 
 *         ptemp = PyDict_GetItem(self.encode_cache, gc)
 *         if ptemp != NULL:             # <<<<<<<<<<<<<<
 *             return <GlycanCompositionVector>ptemp
 *         value = GlycanCompositionVector._create(self.encode_raw(gc))
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":685
 *         if ptemp != NULL:
 *             return <GlycanCompositionVector>ptemp
 *         value = GlycanCompositionVector._create(self.encode_raw(gc))             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(self.encode_cache, gc, value)
 *         PyDict_SetItem(self.decode_cache, value, gc)
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.encode_raw(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_gc); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 685, __pyx_L1_error)
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create(__pyx_t_7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":686
 *             return <GlycanCompositionVector>ptemp
 *         value = GlycanCompositionVector._create(self.encode_raw(gc))
 *         PyDict_SetItem(self.encode_cache, gc, value)             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(self.decode_cache, value, gc)
 *         return value
 */
  __pyx_t_1 = __pyx_v_self->encode_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_8 = PyDict_SetItem(__pyx_t_1, __pyx_v_gc, ((PyObject *)__pyx_v_value)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":687
 *         value = GlycanCompositionVector._create(self.encode_raw(gc))
 *         PyDict_SetItem(self.encode_cache, gc, value)
 *         PyDict_SetItem(self.decode_cache, value, gc)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
  __pyx_t_1 = __pyx_v_self->decode_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_8 = PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_v_value), __pyx_v_gc); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":688
 *         PyDict_SetItem(self.encode_cache, gc, value)
 *         PyDict_SetItem(self.decode_cache, value, gc)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":677
 *         self.decode_cache = {}
 * 
 *     cpdef GlycanCompositionVector encode(self, gc):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_2encode, "CachingGlycanCompositionVectorContext.encode(self, gc) -> GlycanCompositionVector");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode = {"encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_2encode};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gc,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "encode") < 0)) __PYX_ERR(0, 677, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gc = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("encode", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 677, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_2encode(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_gc);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_2encode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_gc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode(__pyx_v_self, __pyx_v_gc, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pyx":690
 *         return value
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_gcvec)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycresoft/_c/composition_network/graph.pyx":693
 *         cdef:
 *             PyObject* ptemp
 *         ptemp = PyDict_GetItem(self.decode_cache, gcvec)             # <<<<<<<<<<<<<<
 *         if ptemp != NULL:
 *             return <object>ptemp
 */
  __pyx_t_1 = __pyx_v_self->decode_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_1, ((PyObject *)__pyx_v_gcvec));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":694
 *             PyObject* ptemp
 *         ptemp = PyDict_GetItem(self.decode_cache, gcvec)
 *         if ptemp != NULL:             # <<<<<<<<<<<<<<
 *             return <object>ptemp
 * 
 */
  __pyx_t_6 = (__pyx_v_ptemp != NULL);
  if (__pyx_t_6) {

    /* "glycresoft/_c/composition_network/graph.pyx":695
 *         ptemp = PyDict_GetItem(self.decode_cache, gcvec)
 *         if ptemp != NULL:
 *             return <object>ptemp             # <<<<<<<<<<<<<<
 * 
 *         value = super(CachingGlycanCompositionVectorContext, self).decode(gcvec)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_ptemp));
    __pyx_r = ((PyObject *)__pyx_v_ptemp);
    goto __pyx_L0;

    /* "glycresoft/_c/composition_network/graph.pyx":694
 *             PyObject* ptemp
 *         ptemp = PyDict_GetItem(self.decode_cache, gcvec)
 *         if ptemp != NULL:             # <<<<<<<<<<<<<<
 *             return <object>ptemp
 * 
 */
  }

  /* "glycresoft/_c/composition_network/graph.pyx":697
 *             return <object>ptemp
 * 
 *         value = super(CachingGlycanCompositionVectorContext, self).decode(gcvec)             # <<<<<<<<<<<<<<
 *         self.encode_cache[value] = gcvec
 *         self.decode_cache[gcvec] = value
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext))) __PYX_ERR(0, 697, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 697, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_gcvec)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":698
 * 
 *         value = super(CachingGlycanCompositionVectorContext, self).decode(gcvec)
 *         self.encode_cache[value] = gcvec             # <<<<<<<<<<<<<<
 *         self.decode_cache[gcvec] = value
 *         return value
 */
  if (unlikely(__pyx_v_self->encode_cache == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 698, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->encode_cache, __pyx_v_value, ((PyObject *)__pyx_v_gcvec)) < 0))) __PYX_ERR(0, 698, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":699
 *         value = super(CachingGlycanCompositionVectorContext, self).decode(gcvec)
 *         self.encode_cache[value] = gcvec
 *         self.decode_cache[gcvec] = value             # <<<<<<<<<<<<<<
 *         return value
 */
  if (unlikely(__pyx_v_self->decode_cache == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 699, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->decode_cache, ((PyObject *)__pyx_v_gcvec), __pyx_v_value) < 0))) __PYX_ERR(0, 699, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":700
 *         self.encode_cache[value] = gcvec
 *         self.decode_cache[gcvec] = value
 *         return value             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "glycresoft/_c/composition_network/graph.pyx":690
 *         return value
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_4decode, "CachingGlycanCompositionVectorContext.decode(self, GlycanCompositionVector gcvec)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode = {"decode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_4decode};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gcvec,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gcvec)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "decode") < 0)) __PYX_ERR(0, 690, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gcvec = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("decode", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 690, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gcvec), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, 1, "gcvec", 0))) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_4decode(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self), __pyx_v_gcvec);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_4decode(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_v_gcvec) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode(__pyx_v_self, __pyx_v_gcvec, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":95
 * cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):
 *     cdef:
 *         public dict encode_cache             # <<<<<<<<<<<<<<
 *         public dict decode_cache
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->encode_cache);
  __pyx_r = __pyx_v_self->encode_cache;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(1, 95, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->encode_cache);
  __Pyx_DECREF(__pyx_v_self->encode_cache);
  __pyx_v_self->encode_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.encode_cache.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->encode_cache);
  __Pyx_DECREF(__pyx_v_self->encode_cache);
  __pyx_v_self->encode_cache = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycresoft/_c/composition_network/graph.pxd":96
 *     cdef:
 *         public dict encode_cache
 *         public dict decode_cache             # <<<<<<<<<<<<<<
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache___get__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache___get__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->decode_cache);
  __pyx_r = __pyx_v_self->decode_cache;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_2__set__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_2__set__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(1, 96, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->decode_cache);
  __Pyx_DECREF(__pyx_v_self->decode_cache);
  __pyx_v_self->decode_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext.decode_cache.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_4__del__(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_4__del__(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->decode_cache);
  __Pyx_DECREF(__pyx_v_self->decode_cache);
  __pyx_v_self->decode_cache = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_5__pyx_unpickle_CompositionGraphNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_4__pyx_unpickle_CompositionGraphNode, "__pyx_unpickle_CompositionGraphNode(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_5__pyx_unpickle_CompositionGraphNode = {"__pyx_unpickle_CompositionGraphNode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_5__pyx_unpickle_CompositionGraphNode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_4__pyx_unpickle_CompositionGraphNode};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_5__pyx_unpickle_CompositionGraphNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CompositionGraphNode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CompositionGraphNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CompositionGraphNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CompositionGraphNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CompositionGraphNode", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_CompositionGraphNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_4__pyx_unpickle_CompositionGraphNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_4__pyx_unpickle_CompositionGraphNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CompositionGraphNode", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1268f7b, 0x1820fdc, 0x3a5dfc4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__11, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1268f7b, 0x1820fdc, 0x3a5dfc4):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 *     __pyx_result = CompositionGraphNode.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1268f7b, 0x1820fdc, 0x3a5dfc4):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CompositionGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1268f7b, 0x1820fdc, 0x3a5dfc4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 *     __pyx_result = CompositionGraphNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 *     __pyx_result = CompositionGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CompositionGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphNode__set_state(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 *     __pyx_result = CompositionGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_CompositionGraphNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphNode__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  double __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CompositionGraphNode__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_hash = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_score = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_str);
  __Pyx_DECREF(__pyx_v___pyx_result->_str);
  __pyx_v___pyx_result->_str = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->composition);
  __Pyx_DECREF(__pyx_v___pyx_result->composition);
  __pyx_v___pyx_result->composition = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->edges);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->edges);
  __pyx_v___pyx_result->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyInt_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_4 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->index = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->internal_score = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->marked = __pyx_t_5;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 8);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "(tree fragment)":14
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_CompositionGraphNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_EdgeSet(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7__pyx_unpickle_EdgeSet(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_6__pyx_unpickle_EdgeSet, "__pyx_unpickle_EdgeSet(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7__pyx_unpickle_EdgeSet = {"__pyx_unpickle_EdgeSet", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7__pyx_unpickle_EdgeSet, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_6__pyx_unpickle_EdgeSet};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7__pyx_unpickle_EdgeSet(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_EdgeSet (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EdgeSet", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EdgeSet", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_EdgeSet") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EdgeSet", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_EdgeSet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_6__pyx_unpickle_EdgeSet(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_6__pyx_unpickle_EdgeSet(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EdgeSet", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x824d80d, 0x3a21295, 0x8cd892b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x824d80d, 0x3a21295, 0x8cd892b):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum
 *     __pyx_result = EdgeSet.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x824d80d, 0x3a21295, 0x8cd892b):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = EdgeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x824d80d, 0x3a21295, 0x8cd892b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum
 *     __pyx_result = EdgeSet.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum
 *     __pyx_result = EdgeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = EdgeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_EdgeSet__set_state(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x824d80d, 0x3a21295, 0x8cd892b) = (store))" % __pyx_checksum
 *     __pyx_result = EdgeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.store = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EdgeSet(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_EdgeSet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.store = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_EdgeSet__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EdgeSet__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.store = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->store);
  __Pyx_DECREF(__pyx_v___pyx_result->store);
  __pyx_v___pyx_result->store = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.store = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.store = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.store = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_EdgeSet__set_state(<EdgeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EdgeSet__set_state(EdgeSet __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.store = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_EdgeSet__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_9__pyx_unpickle_CompositionGraphBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_8__pyx_unpickle_CompositionGraphBase, "__pyx_unpickle_CompositionGraphBase(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_9__pyx_unpickle_CompositionGraphBase = {"__pyx_unpickle_CompositionGraphBase", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_9__pyx_unpickle_CompositionGraphBase, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_8__pyx_unpickle_CompositionGraphBase};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_9__pyx_unpickle_CompositionGraphBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CompositionGraphBase (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CompositionGraphBase", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CompositionGraphBase", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CompositionGraphBase") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CompositionGraphBase", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_CompositionGraphBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_8__pyx_unpickle_CompositionGraphBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_8__pyx_unpickle_CompositionGraphBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CompositionGraphBase", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaf6f309, 0xf549bbe, 0x2cb8587):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__14, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaf6f309, 0xf549bbe, 0x2cb8587):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum
 *     __pyx_result = CompositionGraphBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaf6f309, 0xf549bbe, 0x2cb8587):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CompositionGraphBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaf6f309, 0xf549bbe, 0x2cb8587):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum
 *     __pyx_result = CompositionGraphBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum
 *     __pyx_result = CompositionGraphBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CompositionGraphBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphBase__set_state(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaf6f309, 0xf549bbe, 0x2cb8587) = (_composition_normalizer, distance_fn, edges, neighborhoods, node_map, nodes))" % __pyx_checksum
 *     __pyx_result = CompositionGraphBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_CompositionGraphBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_CompositionGraphBase__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CompositionGraphBase__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_composition_normalizer);
  __Pyx_DECREF(__pyx_v___pyx_result->_composition_normalizer);
  __pyx_v___pyx_result->_composition_normalizer = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->distance_fn);
  __Pyx_DECREF(__pyx_v___pyx_result->distance_fn);
  __pyx_v___pyx_result->distance_fn = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->edges);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->edges);
  __pyx_v___pyx_result->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->neighborhoods);
  __Pyx_DECREF(__pyx_v___pyx_result->neighborhoods);
  __pyx_v___pyx_result->neighborhoods = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->node_map);
  __Pyx_DECREF(__pyx_v___pyx_result->node_map);
  __pyx_v___pyx_result->node_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->nodes);
  __Pyx_DECREF(__pyx_v___pyx_result->nodes);
  __pyx_v___pyx_result->nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 6);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_CompositionGraphBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DijkstraPathFinder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_11__pyx_unpickle_DijkstraPathFinder(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10glycresoft_2_c_19composition_network_5graph_10__pyx_unpickle_DijkstraPathFinder, "__pyx_unpickle_DijkstraPathFinder(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_10glycresoft_2_c_19composition_network_5graph_11__pyx_unpickle_DijkstraPathFinder = {"__pyx_unpickle_DijkstraPathFinder", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_11__pyx_unpickle_DijkstraPathFinder, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_10__pyx_unpickle_DijkstraPathFinder};
static PyObject *__pyx_pw_10glycresoft_2_c_19composition_network_5graph_11__pyx_unpickle_DijkstraPathFinder(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DijkstraPathFinder (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DijkstraPathFinder", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DijkstraPathFinder", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_DijkstraPathFinder") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DijkstraPathFinder", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_DijkstraPathFinder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10glycresoft_2_c_19composition_network_5graph_10__pyx_unpickle_DijkstraPathFinder(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10glycresoft_2_c_19composition_network_5graph_10__pyx_unpickle_DijkstraPathFinder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DijkstraPathFinder", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe7f8508, 0xf331091, 0x18779ed):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__15, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe7f8508, 0xf331091, 0x18779ed):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum
 *     __pyx_result = DijkstraPathFinder.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe7f8508, 0xf331091, 0x18779ed):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = DijkstraPathFinder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe7f8508, 0xf331091, 0x18779ed):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum
 *     __pyx_result = DijkstraPathFinder.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum
 *     __pyx_result = DijkstraPathFinder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = DijkstraPathFinder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_DijkstraPathFinder__set_state(((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe7f8508, 0xf331091, 0x18779ed) = (distance, end, graph, limit, start, unvisited_finite_distance))" % __pyx_checksum
 *     __pyx_result = DijkstraPathFinder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DijkstraPathFinder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_DijkstraPathFinder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10glycresoft_2_c_19composition_network_5graph___pyx_unpickle_DijkstraPathFinder__set_state(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DijkstraPathFinder__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->distance);
  __Pyx_DECREF(__pyx_v___pyx_result->distance);
  __pyx_v___pyx_result->distance = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->end);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->end);
  __pyx_v___pyx_result->end = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->graph);
  __Pyx_DECREF(__pyx_v___pyx_result->graph);
  __pyx_v___pyx_result->graph = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->limit = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->start);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->start);
  __pyx_v___pyx_result->start = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->unvisited_finite_distance);
  __Pyx_DECREF(__pyx_v___pyx_result->unvisited_finite_distance);
  __pyx_v___pyx_result->unvisited_finite_distance = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 6);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DijkstraPathFinder__set_state(<DijkstraPathFinder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DijkstraPathFinder__set_state(DijkstraPathFinder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.distance = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.graph = __pyx_state[2]; __pyx_result.limit = __pyx_state[3]; __pyx_result.start = __pyx_state[4]; __pyx_result.unvisited_finite_distance = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycresoft._c.composition_network.graph.__pyx_unpickle_DijkstraPathFinder__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *__pyx_freelist_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode[5000];
static int __pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode[--__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)o);
  p->composition = Py_None; Py_INCREF(Py_None);
  p->_str = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->composition);
  Py_CLEAR(p->_str);
  Py_CLEAR(p->edges);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode < 5000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode[__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode++] = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)o;
  if (p->composition) {
    e = (*v)(p->composition, a); if (e) return e;
  }
  if (p->edges) {
    e = (*v)(((PyObject *)p->edges), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)o;
  tmp = ((PyObject*)p->composition);
  p->composition = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->edges);
  p->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_7__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_7__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_glycan_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18glycan_composition_1__get__(o);
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_order(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5order_1__get__(o);
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_composition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11composition_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__str(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__str(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4_str_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_edges(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_edges(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5edges_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5_hash_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6_score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_internal_score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_internal_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14internal_score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_marked(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_marked(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_6marked_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode[] = {
  {"edge_to", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_3edge_to, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_2edge_to},
  {"neighbors", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5neighbors, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_4neighbors},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_15copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_14copy},
  {"clone", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_17clone, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_16clone},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_20__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode[] = {
  {(char *)"glycan_composition", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_glycan_composition, 0, (char *)0, 0},
  {(char *)"order", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_order, 0, (char *)0, 0},
  {(char *)"score", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_score, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_score, (char *)0, 0},
  {(char *)"composition", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_composition, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_composition, (char *)PyDoc_STR("composition: object"), 0},
  {(char *)"index", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_index, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_index, (char *)PyDoc_STR("index: 'Py_ssize_t'"), 0},
  {(char *)"_str", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__str, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__str, (char *)PyDoc_STR("_str: str"), 0},
  {(char *)"edges", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_edges, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_edges, (char *)PyDoc_STR("edges: glycresoft._c.composition_network.graph.EdgeSet"), 0},
  {(char *)"_hash", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__hash, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__hash, (char *)PyDoc_STR("_hash: 'Py_hash_t'"), 0},
  {(char *)"_score", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__score, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode__score, (char *)PyDoc_STR("_score: 'double'"), 0},
  {(char *)"internal_score", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_internal_score, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_internal_score, (char *)PyDoc_STR("internal_score: 'double'"), 0},
  {(char *)"marked", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_marked, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_marked, (char *)PyDoc_STR("marked: 'bool'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__},
  {Py_tp_hash, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11__hash__},
  {Py_tp_str, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_9__str__},
  {Py_tp_doc, (void *)PyDoc_STR("CompositionGraphNode(composition, index, score=0., marked=False, **kwargs)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode_spec = {
  "glycresoft._c.composition_network.graph.CompositionGraphNode",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""CompositionGraphNode", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_13__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_11__hash__, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_9__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("CompositionGraphNode(composition, index, score=0., marked=False, **kwargs)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_clear*/
  __pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_EdgeSet __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_EdgeSet;

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_EdgeSet(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_EdgeSet;
  p->store = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_EdgeSet(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_EdgeSet) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->store);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_EdgeSet(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)o;
  if (p->store) {
    e = (*v)(p->store, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_EdgeSet(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)o;
  tmp = ((PyObject*)p->store);
  p->store = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_EdgeSet(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_17__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_17__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_store(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_store(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5store_5__del__(o);
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_EdgeSet[] = {
  {"edge_to", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_3edge_to, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_2edge_to},
  {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5add, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_4add},
  {"remove", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_9remove, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_8remove},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_20__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_EdgeSet[] = {
  {(char *)"store", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_store, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_store, (char *)PyDoc_STR("store: dict"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__},
  {Py_sq_length, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_13__len__},
  {Py_mp_length, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_13__len__},
  {Py_tp_doc, (void *)PyDoc_STR("EdgeSet(store=None)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {Py_tp_iter, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_11__iter__},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_EdgeSet},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet_spec = {
  "glycresoft._c.composition_network.graph.EdgeSet",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_EdgeSet = {
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_13__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_EdgeSet = {
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_13__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""EdgeSet", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_15__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_EdgeSet, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_EdgeSet, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("EdgeSet(store=None)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_clear*/
  __pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_EdgeSet, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *__pyx_freelist_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge[15000];
static int __pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge[--__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;
  p->node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  p->node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  p->_str = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->node1);
  Py_CLEAR(p->node2);
  Py_CLEAR(p->_str);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge < 15000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge[__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge++] = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)o;
  if (p->node1) {
    e = (*v)(((PyObject *)p->node1), a); if (e) return e;
  }
  if (p->node2) {
    e = (*v)(((PyObject *)p->node2), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *)o;
  tmp = ((PyObject*)p->node1);
  p->node1 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->node2);
  p->node2 = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_7__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_7__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node1(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node1_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5node2_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_order(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_order(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5order_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_weight(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_6weight_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_hash_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__str(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__str(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_str_5__del__(o);
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge[] = {
  {"_traverse", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_traverse, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_4_traverse},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_15__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_14__reduce__},
  {"remove", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_19remove, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_18remove},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge[] = {
  {(char *)"node1", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node1, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node1, (char *)PyDoc_STR("node1: glycresoft._c.composition_network.graph.CompositionGraphNode"), 0},
  {(char *)"node2", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node2, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_node2, (char *)PyDoc_STR("node2: glycresoft._c.composition_network.graph.CompositionGraphNode"), 0},
  {(char *)"order", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_order, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_order, (char *)PyDoc_STR("order: 'long'"), 0},
  {(char *)"weight", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_weight, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_weight, (char *)PyDoc_STR("weight: 'double'"), 0},
  {(char *)"_hash", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__hash, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__hash, (char *)PyDoc_STR("_hash: 'Py_hash_t'"), 0},
  {(char *)"_str", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__str, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__str, (char *)PyDoc_STR("_str: str"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__},
  {Py_sq_item, (void *)__pyx_sq_item_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_mp_subscript, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_3__getitem__},
  {Py_tp_hash, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_13__hash__},
  {Py_tp_str, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_9__str__},
  {Py_tp_doc, (void *)PyDoc_STR("CompositionGraphEdge(node1, node2, order, weight=1.0)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge_spec = {
  "glycresoft._c.composition_network.graph.CompositionGraphEdge",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_CompositionGraphEdge = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_CompositionGraphEdge = {
  0, /*mp_length*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_3__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""CompositionGraphEdge", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_CompositionGraphEdge, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_CompositionGraphEdge, /*tp_as_mapping*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_13__hash__, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_9__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("CompositionGraphEdge(node1, node2, order, weight=1.0)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_clear*/
  __pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;
  p->nodes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->node_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_composition_normalizer = Py_None; Py_INCREF(Py_None);
  p->distance_fn = Py_None; Py_INCREF(Py_None);
  p->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)Py_None); Py_INCREF(Py_None);
  p->neighborhoods = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->nodes);
  Py_CLEAR(p->node_map);
  Py_CLEAR(p->_composition_normalizer);
  Py_CLEAR(p->distance_fn);
  Py_CLEAR(p->edges);
  Py_CLEAR(p->neighborhoods);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)o;
  if (p->nodes) {
    e = (*v)(p->nodes, a); if (e) return e;
  }
  if (p->node_map) {
    e = (*v)(p->node_map, a); if (e) return e;
  }
  if (p->_composition_normalizer) {
    e = (*v)(p->_composition_normalizer, a); if (e) return e;
  }
  if (p->distance_fn) {
    e = (*v)(p->distance_fn, a); if (e) return e;
  }
  if (p->edges) {
    e = (*v)(((PyObject *)p->edges), a); if (e) return e;
  }
  if (p->neighborhoods) {
    e = (*v)(p->neighborhoods, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *)o;
  tmp = ((PyObject*)p->nodes);
  p->nodes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->node_map);
  p->node_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_composition_normalizer);
  p->_composition_normalizer = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->distance_fn);
  p->distance_fn = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->edges);
  p->edges = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->neighborhoods);
  p->neighborhoods = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_nodes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_nodes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5nodes_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_node_map(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_node_map(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_8node_map_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase__composition_normalizer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase__composition_normalizer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_23_composition_normalizer_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_distance_fn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_distance_fn(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_11distance_fn_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_edges(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_edges(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5edges_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_neighborhoods(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_neighborhoods(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_13neighborhoods_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_4__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase[] = {
  {(char *)"nodes", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_nodes, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_nodes, (char *)PyDoc_STR("nodes: list"), 0},
  {(char *)"node_map", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_node_map, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_node_map, (char *)PyDoc_STR("node_map: dict"), 0},
  {(char *)"_composition_normalizer", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase__composition_normalizer, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase__composition_normalizer, (char *)PyDoc_STR("_composition_normalizer: object"), 0},
  {(char *)"distance_fn", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_distance_fn, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_distance_fn, (char *)PyDoc_STR("distance_fn: object"), 0},
  {(char *)"edges", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_edges, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_edges, (char *)PyDoc_STR("edges: glycresoft._c.composition_network.graph.EdgeSet"), 0},
  {(char *)"neighborhoods", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_neighborhoods, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_neighborhoods, (char *)PyDoc_STR("neighborhoods: object"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase_spec = {
  "glycresoft._c.composition_network.graph.CompositionGraphBase",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""CompositionGraphBase", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  p->components = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->components);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)o;
  if (p->components) {
    e = (*v)(p->components, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *)o;
  tmp = ((PyObject*)p->components);
  p->components = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_components(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_components(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_10components_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_component_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_component_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_15component_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext[] = {
  {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_2__reduce__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext[] = {
  {(char *)"components", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_components, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_components, (char *)PyDoc_STR("components: list"), 0},
  {(char *)"component_count", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_component_count, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_component_count, (char *)PyDoc_STR("component_count: 'size_t'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext},
  {Py_tp_doc, (void *)PyDoc_STR("GlycanCompositionVectorContext(components)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext_spec = {
  "glycresoft._c.composition_network.graph.GlycanCompositionVectorContext",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""GlycanCompositionVectorContext", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("GlycanCompositionVectorContext(components)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *__pyx_freelist_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector[2000];
static int __pyx_freecount_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector = 0;
#endif

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)))) {
    o = (PyObject*)__pyx_freelist_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector[--__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector];
    memset(o, 0, sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector < 2000) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)))) {
    __pyx_freelist_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector[__pyx_freecount_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector++] = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}
static PyObject *__pyx_sq_item_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_7__setitem__(o, i, v);
  }
  else {
    __Pyx_TypeName o_type_name;
    o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by " __Pyx_FMT_TYPENAME, o_type_name);
    __Pyx_DECREF_TypeName(o_type_name);
    return -1;
  }
}

static PyObject *__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_15__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_17__ne__(o1, o2);
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static CYTHON_INLINE PyObject *__pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_add : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_add);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_add == &__pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
    if (maybe_self_is_left) {
        PyObject *res;
        res = __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_25__add__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_add == &__pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)
#endif
            || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
    if (maybe_self_is_right) {
        return __pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_maybe_call_slot(__Pyx_PyType_GetSlot(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, tp_base, PyTypeObject*), left, right );
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_subtract : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_subtract);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_subtract == &__pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
    if (maybe_self_is_left) {
        PyObject *res;
        res = __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_23__sub__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_subtract == &__pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)
#endif
            || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);
    if (maybe_self_is_right) {
        return __pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_maybe_call_slot(__Pyx_PyType_GetSlot(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, tp_base, PyTypeObject*), left, right );
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static PyObject *__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__(self);
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector[] = {
  {"__reduce__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_1__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector___reduce__},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__, METH_NOARGS|METH_COEXIST, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_tp_repr, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__},
  {Py_nb_add, (void *)__pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_nb_subtract, (void *)__pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_nb_negative, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_27__neg__},
  {Py_sq_length, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_9__len__},
  {Py_sq_item, (void *)__pyx_sq_item_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_mp_length, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_9__len__},
  {Py_mp_subscript, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_5__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_tp_hash, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_29__hash__},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_spec = {
  "glycresoft._c.composition_network.graph.GlycanCompositionVector",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_GlycanCompositionVector = {
  __pyx_nb_add_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*nb_add*/
  __pyx_nb_subtract_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_27__neg__, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_bool*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_GlycanCompositionVector = {
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_9__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GlycanCompositionVector = {
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_9__len__, /*mp_length*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_5__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""GlycanCompositionVector", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_11__repr__, /*tp_repr*/
  &__pyx_tp_as_number_GlycanCompositionVector, /*tp_as_number*/
  &__pyx_tp_as_sequence_GlycanCompositionVector, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GlycanCompositionVector, /*tp_as_mapping*/
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_29__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *p;
  PyObject *o = __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext*)__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;
  p->encode_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->decode_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->encode_cache);
  Py_CLEAR(p->decode_cache);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(o);
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)o;
  e = __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(o, v, a); if (e) return e;
  if (p->encode_cache) {
    e = (*v)(p->encode_cache, a); if (e) return e;
  }
  if (p->decode_cache) {
    e = (*v)(p->decode_cache, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext *)o;
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext(o);
  tmp = ((PyObject*)p->encode_cache);
  p->encode_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->decode_cache);
  p->decode_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode_cache(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode_cache(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12encode_cache_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode_cache(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode_cache(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_12decode_cache_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext[] = {
  {(char *)"encode_cache", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode_cache, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode_cache, (char *)PyDoc_STR("encode_cache: dict"), 0},
  {(char *)"decode_cache", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode_cache, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode_cache, (char *)PyDoc_STR("decode_cache: dict"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext},
  {Py_tp_doc, (void *)PyDoc_STR("CachingGlycanCompositionVectorContext(components)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext_spec = {
  "glycresoft._c.composition_network.graph.CachingGlycanCompositionVectorContext",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""CachingGlycanCompositionVectorContext", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("CachingGlycanCompositionVectorContext(components)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;

static PyObject *__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)o);
  p->__pyx_vtab = __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
  p->graph = Py_None; Py_INCREF(Py_None);
  p->start = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  p->end = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  p->distance = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->unvisited_finite_distance = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder(PyObject *o) {
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->graph);
  Py_CLEAR(p->start);
  Py_CLEAR(p->end);
  Py_CLEAR(p->distance);
  Py_CLEAR(p->unvisited_finite_distance);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)o;
  if (p->graph) {
    e = (*v)(p->graph, a); if (e) return e;
  }
  if (p->start) {
    e = (*v)(((PyObject *)p->start), a); if (e) return e;
  }
  if (p->end) {
    e = (*v)(((PyObject *)p->end), a); if (e) return e;
  }
  if (p->distance) {
    e = (*v)(p->distance, a); if (e) return e;
  }
  if (p->unvisited_finite_distance) {
    e = (*v)(p->unvisited_finite_distance, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *p = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *)o;
  tmp = ((PyObject*)p->graph);
  p->graph = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->start);
  p->start = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->end);
  p->end = ((struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->distance);
  p->distance = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->unvisited_finite_distance);
  p->unvisited_finite_distance = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_graph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_graph(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5graph_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_start(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5start_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_end(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3end_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_distance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_distance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8distance_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_unvisited_finite_distance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_unvisited_finite_distance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_3__set__(o, v);
  }
  else {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_25unvisited_finite_distance_5__del__(o);
  }
}

static PyObject *__pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_1__get__(o);
}

static int __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_limit(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5limit_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder[] = {
  {"search", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_9search, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_8search},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_10__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_12__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder[] = {
  {(char *)"graph", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_graph, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_graph, (char *)PyDoc_STR("graph: object"), 0},
  {(char *)"start", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_start, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_start, (char *)PyDoc_STR("start: glycresoft._c.composition_network.graph.CompositionGraphNode"), 0},
  {(char *)"end", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_end, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_end, (char *)PyDoc_STR("end: glycresoft._c.composition_network.graph.CompositionGraphNode"), 0},
  {(char *)"distance", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_distance, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_distance, (char *)PyDoc_STR("distance: dict"), 0},
  {(char *)"unvisited_finite_distance", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_unvisited_finite_distance, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_unvisited_finite_distance, (char *)PyDoc_STR("unvisited_finite_distance: dict"), 0},
  {(char *)"limit", __pyx_getprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_limit, __pyx_setprop_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_limit, (char *)PyDoc_STR("limit: 'double'"), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder},
  {Py_tp_doc, (void *)PyDoc_STR("DijkstraPathFinder(graph, start, end, limit=INF)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder},
  {Py_tp_clear, (void *)__pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder},
  {Py_tp_methods, (void *)__pyx_methods_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder},
  {Py_tp_getset, (void *)__pyx_getsets_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder},
  {Py_tp_init, (void *)__pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder},
  {0, 0},
};
static PyType_Spec __pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder_spec = {
  "glycresoft._c.composition_network.graph.DijkstraPathFinder",
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder_slots,
};
#else

static PyTypeObject __pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycresoft._c.composition_network.graph.""DijkstraPathFinder", /*tp_name*/
  sizeof(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("DijkstraPathFinder(graph, start, end, limit=INF)"), /*tp_doc*/
  __pyx_tp_traverse_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, /*tp_traverse*/
  __pyx_tp_clear_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_CachingGlycanCompositionVectorCo, __pyx_k_CachingGlycanCompositionVectorCo, sizeof(__pyx_k_CachingGlycanCompositionVectorCo), 0, 0, 1, 1},
    {&__pyx_n_s_CachingGlycanCompositionVectorCo_2, __pyx_k_CachingGlycanCompositionVectorCo_2, sizeof(__pyx_k_CachingGlycanCompositionVectorCo_2), 0, 0, 1, 1},
    {&__pyx_n_s_CachingGlycanCompositionVectorCo_3, __pyx_k_CachingGlycanCompositionVectorCo_3, sizeof(__pyx_k_CachingGlycanCompositionVectorCo_3), 0, 0, 1, 1},
    {&__pyx_kp_s_Cannot_compute_addition_from_Non, __pyx_k_Cannot_compute_addition_from_Non, sizeof(__pyx_k_Cannot_compute_addition_from_Non), 0, 0, 1, 0},
    {&__pyx_kp_s_Cannot_compute_difference_from_N, __pyx_k_Cannot_compute_difference_from_N, sizeof(__pyx_k_Cannot_compute_difference_from_N), 0, 0, 1, 0},
    {&__pyx_n_s_CompositionGraphBase, __pyx_k_CompositionGraphBase, sizeof(__pyx_k_CompositionGraphBase), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphBase___reduce_cy, __pyx_k_CompositionGraphBase___reduce_cy, sizeof(__pyx_k_CompositionGraphBase___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphBase___setstate, __pyx_k_CompositionGraphBase___setstate, sizeof(__pyx_k_CompositionGraphBase___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphBase_copy, __pyx_k_CompositionGraphBase_copy, sizeof(__pyx_k_CompositionGraphBase_copy), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphEdge, __pyx_k_CompositionGraphEdge, sizeof(__pyx_k_CompositionGraphEdge), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphEdge___reduce, __pyx_k_CompositionGraphEdge___reduce, sizeof(__pyx_k_CompositionGraphEdge___reduce), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphEdge__traverse, __pyx_k_CompositionGraphEdge__traverse, sizeof(__pyx_k_CompositionGraphEdge__traverse), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphEdge_copy_for, __pyx_k_CompositionGraphEdge_copy_for, sizeof(__pyx_k_CompositionGraphEdge_copy_for), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphEdge_remove, __pyx_k_CompositionGraphEdge_remove, sizeof(__pyx_k_CompositionGraphEdge_remove), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode, __pyx_k_CompositionGraphNode, sizeof(__pyx_k_CompositionGraphNode), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode___reduce_cy, __pyx_k_CompositionGraphNode___reduce_cy, sizeof(__pyx_k_CompositionGraphNode___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode___setstate, __pyx_k_CompositionGraphNode___setstate, sizeof(__pyx_k_CompositionGraphNode___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode_clone, __pyx_k_CompositionGraphNode_clone, sizeof(__pyx_k_CompositionGraphNode_clone), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode_copy, __pyx_k_CompositionGraphNode_copy, sizeof(__pyx_k_CompositionGraphNode_copy), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode_edge_to, __pyx_k_CompositionGraphNode_edge_to, sizeof(__pyx_k_CompositionGraphNode_edge_to), 0, 0, 1, 1},
    {&__pyx_n_s_CompositionGraphNode_neighbors, __pyx_k_CompositionGraphNode_neighbors, sizeof(__pyx_k_CompositionGraphNode_neighbors), 0, 0, 1, 1},
    {&__pyx_kp_s_CompositionGraphNode_s_d_0_2f, __pyx_k_CompositionGraphNode_s_d_0_2f, sizeof(__pyx_k_CompositionGraphNode_s_d_0_2f), 0, 0, 1, 0},
    {&__pyx_n_s_DijkstraPathFinder, __pyx_k_DijkstraPathFinder, sizeof(__pyx_k_DijkstraPathFinder), 0, 0, 1, 1},
    {&__pyx_n_s_DijkstraPathFinder___reduce_cyth, __pyx_k_DijkstraPathFinder___reduce_cyth, sizeof(__pyx_k_DijkstraPathFinder___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_DijkstraPathFinder___setstate_cy, __pyx_k_DijkstraPathFinder___setstate_cy, sizeof(__pyx_k_DijkstraPathFinder___setstate_cy), 0, 0, 1, 1},
    {&__pyx_n_s_DijkstraPathFinder__build_initia, __pyx_k_DijkstraPathFinder__build_initia, sizeof(__pyx_k_DijkstraPathFinder__build_initia), 0, 0, 1, 1},
    {&__pyx_n_s_DijkstraPathFinder_find_path, __pyx_k_DijkstraPathFinder_find_path, sizeof(__pyx_k_DijkstraPathFinder_find_path), 0, 0, 1, 1},
    {&__pyx_n_s_DijkstraPathFinder_find_smallest, __pyx_k_DijkstraPathFinder_find_smallest, sizeof(__pyx_k_DijkstraPathFinder_find_smallest), 0, 0, 1, 1},
    {&__pyx_n_s_DijkstraPathFinder_search, __pyx_k_DijkstraPathFinder_search, sizeof(__pyx_k_DijkstraPathFinder_search), 0, 0, 1, 1},
    {&__pyx_kp_s_Dimensionality_does_not_match, __pyx_k_Dimensionality_does_not_match, sizeof(__pyx_k_Dimensionality_does_not_match), 0, 0, 1, 0},
    {&__pyx_n_s_EdgeSet, __pyx_k_EdgeSet, sizeof(__pyx_k_EdgeSet), 0, 0, 1, 1},
    {&__pyx_n_s_EdgeSet___reduce_cython, __pyx_k_EdgeSet___reduce_cython, sizeof(__pyx_k_EdgeSet___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EdgeSet___setstate_cython, __pyx_k_EdgeSet___setstate_cython, sizeof(__pyx_k_EdgeSet___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EdgeSet_add, __pyx_k_EdgeSet_add, sizeof(__pyx_k_EdgeSet_add), 0, 0, 1, 1},
    {&__pyx_n_s_EdgeSet_add_if_shorter, __pyx_k_EdgeSet_add_if_shorter, sizeof(__pyx_k_EdgeSet_add_if_shorter), 0, 0, 1, 1},
    {&__pyx_n_s_EdgeSet_edge_to, __pyx_k_EdgeSet_edge_to, sizeof(__pyx_k_EdgeSet_edge_to), 0, 0, 1, 1},
    {&__pyx_n_s_EdgeSet_remove, __pyx_k_EdgeSet_remove, sizeof(__pyx_k_EdgeSet_remove), 0, 0, 1, 1},
    {&__pyx_kp_u_Failed_to_select_a_destination_f, __pyx_k_Failed_to_select_a_destination_f, sizeof(__pyx_k_Failed_to_select_a_destination_f), 0, 1, 0, 0},
    {&__pyx_n_s_FrozenMonosaccharideResidue, __pyx_k_FrozenMonosaccharideResidue, sizeof(__pyx_k_FrozenMonosaccharideResidue), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVector, __pyx_k_GlycanCompositionVector, sizeof(__pyx_k_GlycanCompositionVector), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVectorContext, __pyx_k_GlycanCompositionVectorContext, sizeof(__pyx_k_GlycanCompositionVectorContext), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVectorContext_2, __pyx_k_GlycanCompositionVectorContext_2, sizeof(__pyx_k_GlycanCompositionVectorContext_2), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVectorContext_d, __pyx_k_GlycanCompositionVectorContext_d, sizeof(__pyx_k_GlycanCompositionVectorContext_d), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVectorContext_e, __pyx_k_GlycanCompositionVectorContext_e, sizeof(__pyx_k_GlycanCompositionVectorContext_e), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVector___reduce, __pyx_k_GlycanCompositionVector___reduce, sizeof(__pyx_k_GlycanCompositionVector___reduce), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVector_addition, __pyx_k_GlycanCompositionVector_addition, sizeof(__pyx_k_GlycanCompositionVector_addition), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVector_clone, __pyx_k_GlycanCompositionVector_clone, sizeof(__pyx_k_GlycanCompositionVector_clone), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVector_differen, __pyx_k_GlycanCompositionVector_differen, sizeof(__pyx_k_GlycanCompositionVector_differen), 0, 0, 1, 1},
    {&__pyx_n_s_GlycanCompositionVector_distance, __pyx_k_GlycanCompositionVector_distance, sizeof(__pyx_k_GlycanCompositionVector_distance), 0, 0, 1, 1},
    {&__pyx_n_s_HashableGlycanComposition, __pyx_k_HashableGlycanComposition, sizeof(__pyx_k_HashableGlycanComposition), 0, 0, 1, 1},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
    {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
    {&__pyx_kp_s_Object_in_invalid_state, __pyx_k_Object_in_invalid_state, sizeof(__pyx_k_Object_in_invalid_state), 0, 0, 1, 0},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_kp_s_The_unvisited_set_must_not_be_em, __pyx_k_The_unvisited_set_must_not_be_em, sizeof(__pyx_k_The_unvisited_set_must_not_be_em), 0, 0, 1, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_n_s__73, __pyx_k__73, sizeof(__pyx_k__73), 0, 0, 1, 1},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
    {&__pyx_n_s_add_if_shorter, __pyx_k_add_if_shorter, sizeof(__pyx_k_add_if_shorter), 0, 0, 1, 1},
    {&__pyx_n_s_add_node, __pyx_k_add_node, sizeof(__pyx_k_add_node), 0, 0, 1, 1},
    {&__pyx_n_s_addition, __pyx_k_addition, sizeof(__pyx_k_addition), 0, 0, 1, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_build_initial_key_set, __pyx_k_build_initial_key_set, sizeof(__pyx_k_build_initial_key_set), 0, 0, 1, 1},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_n_s_components, __pyx_k_components, sizeof(__pyx_k_components), 0, 0, 1, 1},
    {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_copy_for, __pyx_k_copy_for, sizeof(__pyx_k_copy_for), 0, 0, 1, 1},
    {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
    {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_n_s_difference, __pyx_k_difference, sizeof(__pyx_k_difference), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_distance, __pyx_k_distance, sizeof(__pyx_k_distance), 0, 0, 1, 1},
    {&__pyx_n_s_dup, __pyx_k_dup, sizeof(__pyx_k_dup), 0, 0, 1, 1},
    {&__pyx_n_s_edge, __pyx_k_edge, sizeof(__pyx_k_edge), 0, 0, 1, 1},
    {&__pyx_n_s_edge_to, __pyx_k_edge_to, sizeof(__pyx_k_edge_to), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
    {&__pyx_n_s_find_path, __pyx_k_find_path, sizeof(__pyx_k_find_path), 0, 0, 1, 1},
    {&__pyx_n_s_find_smallest_unvisited, __pyx_k_find_smallest_unvisited, sizeof(__pyx_k_find_smallest_unvisited), 0, 0, 1, 1},
    {&__pyx_n_s_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_gcvec, __pyx_k_gcvec, sizeof(__pyx_k_gcvec), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_glycan_composition, __pyx_k_glycan_composition, sizeof(__pyx_k_glycan_composition), 0, 0, 1, 1},
    {&__pyx_n_s_glycresoft__c_composition_networ, __pyx_k_glycresoft__c_composition_networ, sizeof(__pyx_k_glycresoft__c_composition_networ), 0, 0, 1, 1},
    {&__pyx_n_s_glypy_structure_glycan_compositi, __pyx_k_glypy_structure_glycan_compositi, sizeof(__pyx_k_glypy_structure_glycan_compositi), 0, 0, 1, 1},
    {&__pyx_n_s_graph, __pyx_k_graph, sizeof(__pyx_k_graph), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_limit, __pyx_k_limit, sizeof(__pyx_k_limit), 0, 0, 1, 1},
    {&__pyx_kp_u_looking_for, __pyx_k_looking_for, sizeof(__pyx_k_looking_for), 0, 1, 0, 0},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_marked, __pyx_k_marked, sizeof(__pyx_k_marked), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_neighbors, __pyx_k_neighbors, sizeof(__pyx_k_neighbors), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
    {&__pyx_n_s_node1, __pyx_k_node1, sizeof(__pyx_k_node1), 0, 0, 1, 1},
    {&__pyx_n_s_node2, __pyx_k_node2, sizeof(__pyx_k_node2), 0, 0, 1, 1},
    {&__pyx_n_s_nodes, __pyx_k_nodes, sizeof(__pyx_k_nodes), 0, 0, 1, 1},
    {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
    {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
    {&__pyx_n_s_popleft, __pyx_k_popleft, sizeof(__pyx_k_popleft), 0, 0, 1, 1},
    {&__pyx_n_s_ptemp, __pyx_k_ptemp, sizeof(__pyx_k_ptemp), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CompositionGraphB, __pyx_k_pyx_unpickle_CompositionGraphB, sizeof(__pyx_k_pyx_unpickle_CompositionGraphB), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CompositionGraphN, __pyx_k_pyx_unpickle_CompositionGraphN, sizeof(__pyx_k_pyx_unpickle_CompositionGraphN), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_DijkstraPathFinde, __pyx_k_pyx_unpickle_DijkstraPathFinde, sizeof(__pyx_k_pyx_unpickle_DijkstraPathFinde), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_EdgeSet, __pyx_k_pyx_unpickle_EdgeSet, sizeof(__pyx_k_pyx_unpickle_EdgeSet), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_reconstruct_GlycanCompositionVe, __pyx_k_reconstruct_GlycanCompositionVe, sizeof(__pyx_k_reconstruct_GlycanCompositionVe), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_reindex_graph, __pyx_k_reindex_graph, sizeof(__pyx_k_reindex_graph), 0, 0, 1, 1},
    {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
    {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
    {&__pyx_kp_s_s_s_s, __pyx_k_s_s_s, sizeof(__pyx_k_s_s_s), 0, 0, 1, 0},
    {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
    {&__pyx_n_s_search, __pyx_k_search, sizeof(__pyx_k_search), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_k_src_glycresoft__c_composition_ne, sizeof(__pyx_k_src_glycresoft__c_composition_ne), 0, 0, 1, 0},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_store, __pyx_k_store, sizeof(__pyx_k_store), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_traverse, __pyx_k_traverse, sizeof(__pyx_k_traverse), 0, 0, 1, 1},
    {&__pyx_n_s_unvisited, __pyx_k_unvisited, sizeof(__pyx_k_unvisited), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {&__pyx_n_s_weight, __pyx_k_weight, sizeof(__pyx_k_weight), 0, 0, 1, 1},
    {&__pyx_kp_u_with_smallest_distance, __pyx_k_with_smallest_distance, sizeof(__pyx_k_with_smallest_distance), 0, 1, 0, 0},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 181, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 224, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 277, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 495, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 556, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 595, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 673, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "glycresoft/_c/composition_network/graph.pyx":311
 * 
 *         if not unvisited:
 *             raise ValueError("The unvisited set must not be empty")             # <<<<<<<<<<<<<<
 * 
 *         if not self.unvisited_finite_distance:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_The_unvisited_set_must_not_be_em); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "glycresoft/_c/composition_network/graph.pyx":509
 *         gc = HashableGlycanComposition()
 *         if self.component_count != gcvec.ptr.size:
 *             raise ValueError("Dimensionality does not match")             # <<<<<<<<<<<<<<
 *         for i in range(self.component_count):
 *             value = gcvec.ptr.counts[i]
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_Dimensionality_does_not_match); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "glycresoft/_c/composition_network/graph.pyx":554
 *     def __getitem__(self, int i):
 *         if self.ptr == NULL:
 *             raise ValueError("Object in invalid state")             # <<<<<<<<<<<<<<
 *         if i >= self.ptr.size or i < 0:
 *             raise IndexError(i)
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Object_in_invalid_state); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "glycresoft/_c/composition_network/graph.pyx":595
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 *             raise TypeError("Cannot compute difference from `None`")             # <<<<<<<<<<<<<<
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_Cannot_compute_difference_from_N); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "glycresoft/_c/composition_network/graph.pyx":611
 *             raise ValueError("Object in invalid state")
 *         if other is None:
 *             raise TypeError("Cannot compute addition from `None`")             # <<<<<<<<<<<<<<
 * 
 *         new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_Cannot_compute_addition_from_Non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1268f7b, 0x1820fdc, 0x3a5dfc4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1268f7b, 0x1820fdc, 0x3a5dfc4) = (_hash, _score, _str, composition, edges, index, internal_score, marked))" % __pyx_checksum
 */
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_int_19304315, __pyx_int_25300956, __pyx_int_61202372); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_int_136632333, __pyx_int_60953237, __pyx_int_147687723); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_int_183956233, __pyx_int_257203134, __pyx_int_46892423); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_int_243238152, __pyx_int_255004817, __pyx_int_25655789); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "glycresoft/_c/composition_network/graph.pyx":49
 *         self._score = value
 * 
 *     def edge_to(self, node):             # <<<<<<<<<<<<<<
 *         return self.edges.edge_to(self, node)
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_node); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_edge_to, 49, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 49, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":52
 *         return self.edges.edge_to(self, node)
 * 
 *     def neighbors(self):             # <<<<<<<<<<<<<<
 *         result = [edge._traverse(self) for edge in self.edges]
 *         return result
 */
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_edge); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_neighbors, 52, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 52, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":79
 *             self.score if self.score != 0 else self.internal_score)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)
 *         dup.internal_score = self.internal_score
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_dup); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_copy, 79, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 79, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":84
 *         return dup
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_clone, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__24 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphNode__set_state(self, __pyx_state)
 */
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":95
 *         self.store = dict(store)
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         if node2.index < node1.index:
 *             node1, node2 = node2, node1
 */
  __pyx_tuple__28 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_node1, __pyx_n_s_node2); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_edge_to, 95, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 95, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":100
 *         return self.store[node1, node2]
 * 
 *     def add(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store[edge.node1, edge.node2] = edge
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_edge); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_add, 100, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 100, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":103
 *         self.store[edge.node1, edge.node2] = edge
 * 
 *     cpdef bint add_if_shorter(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_add_if_shorter, 103, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 103, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":122
 *                 return True
 * 
 *     def remove(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store.pop((edge.node1, edge.node2))
 * 
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_remove, 122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 122, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EdgeSet__set_state(self, __pyx_state)
 */
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":183
 *             raise KeyError(key)
 * 
 *     def _traverse(self, node):             # <<<<<<<<<<<<<<
 *         return self.node1 if node is self.node2 else self.node2
 * 
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_traverse, 183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 183, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":198
 *         return self._hash
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 */
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_reduce, 198, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 198, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":201
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 *     cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_copy_for, 201, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 201, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":204
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 *     def remove(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.node1.edges.remove(self)
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_remove, 204, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 204, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":216
 * 
 * @cython.binding(True)
 * cpdef reindex_graph(self):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_reindex_graph, 216, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 216, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":236
 *         return self.edges.get_edges()
 * 
 *     cpdef CompositionGraphBase copy(CompositionGraphBase self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphNode n1, n2
 */
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_copy, 236, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 236, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphBase__set_state(self, __pyx_state)
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":284
 *         self.limit = limit
 * 
 *     cpdef set _build_initial_key_set(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_build_initial_key_set, 284, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 284, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":299
 *         return result
 * 
 *     cpdef CompositionGraphNode find_smallest_unvisited(self, set unvisited):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double smallest_distance, distance
 */
  __pyx_tuple__45 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_unvisited); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_find_smallest_unvisited, 299, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 299, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":335
 *         return self.graph[smallest_node]
 * 
 *     cpdef find_path(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             set unvisited
 */
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_find_path, 335, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 335, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":389
 *                     n += 1
 * 
 *     def search(self):             # <<<<<<<<<<<<<<
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 */
  __pyx_tuple__48 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ptemp); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_search, 389, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 389, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DijkstraPathFinder__set_state(self, __pyx_state)
 */
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":489
 *         self.component_count = len(self.components)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.components, )
 * 
 */
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_reduce, 489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 489, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":500
 *         return ptr
 * 
 *     cpdef GlycanCompositionVector encode(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 */
  __pyx_tuple__53 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_glycan_composition); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_encode, 500, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 500, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":503
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */
  __pyx_tuple__55 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_gcvec); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_decode, 503, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 503, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":517
 * 
 * 
 * cpdef _reconstruct_GlycanCompositionVector(state):             # <<<<<<<<<<<<<<
 *     cdef:
 *         glycan_composition_vector* ptr
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_n_s_state); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_reconstruct_GlycanCompositionVe, 517, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 517, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":543
 *         return self
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return _reconstruct_GlycanCompositionVector, (list(self), )
 * 
 */
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_reduce, 543, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 543, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":574
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return INF
 */
  __pyx_tuple__60 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_distance, 574, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 574, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":589
 *         return not self == other
 * 
 *     cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_difference, 589, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 589, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":605
 *         return GlycanCompositionVector._create(new)
 * 
 *     cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_addition, 605, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 605, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":655
 *         return hash_val
 * 
 *     cpdef GlycanCompositionVector clone(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             glycan_composition_vector* new
 */
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_clone, 655, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 655, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":677
 *         self.decode_cache = {}
 * 
 *     cpdef GlycanCompositionVector encode(self, gc):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_gc); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_encode, 677, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 677, __pyx_L1_error)

  /* "glycresoft/_c/composition_network/graph.pyx":690
 *         return value
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycresoft__c_composition_ne, __pyx_n_s_decode, 690, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 690, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__68 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CompositionGraphN, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_EdgeSet, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CompositionGraphB, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DijkstraPathFinde, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.method_name = &__pyx_n_s_pop;
  __pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_values.method_name = &__pyx_n_s_values;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_ = PyFloat_FromDouble(0.); if (unlikely(!__pyx_float_0_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_19304315 = PyInt_FromLong(19304315L); if (unlikely(!__pyx_int_19304315)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_25300956 = PyInt_FromLong(25300956L); if (unlikely(!__pyx_int_25300956)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_25655789 = PyInt_FromLong(25655789L); if (unlikely(!__pyx_int_25655789)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_46892423 = PyInt_FromLong(46892423L); if (unlikely(!__pyx_int_46892423)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_60953237 = PyInt_FromLong(60953237L); if (unlikely(!__pyx_int_60953237)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_61202372 = PyInt_FromLong(61202372L); if (unlikely(!__pyx_int_61202372)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136632333 = PyInt_FromLong(136632333L); if (unlikely(!__pyx_int_136632333)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_147687723 = PyInt_FromLong(147687723L); if (unlikely(!__pyx_int_147687723)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_183956233 = PyInt_FromLong(183956233L); if (unlikely(!__pyx_int_183956233)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_243238152 = PyInt_FromLong(243238152L); if (unlikely(!__pyx_int_243238152)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_255004817 = PyInt_FromLong(255004817L); if (unlikely(!__pyx_int_255004817)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_257203134 = PyInt_FromLong(257203134L); if (unlikely(!__pyx_int_257203134)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("reindex_graph", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_reindex_graph, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("initialize_glycan_composition_vector", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_initialize_glycan_composition_vector, "int (size_t, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroy_glycan_composition_vector", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_destroy_glycan_composition_vector, "int (struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("copy_glycan_composition_vector", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_copy_glycan_composition_vector, "int (struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("glycan_composition_vector_distance", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_distance, "double (struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("glycan_composition_vector_difference", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_difference, "int (struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("glycan_composition_vector_addition", (void (*)(void))__pyx_f_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector_addition, "int (struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *, struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode)) __PYX_ERR(0, 18, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CompositionGraphNode, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_EdgeSet = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_EdgeSet;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_EdgeSet.add_if_shorter = (int (*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_add_if_shorter;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_EdgeSet.get_edges = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_EdgeSet *))__pyx_f_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_get_edges;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet)) __PYX_ERR(0, 88, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_EdgeSet;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_EdgeSet) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EdgeSet, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge._create = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, long, double))__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__create;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge._init = (void (*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *))__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge__init;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge.copy_for = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_copy_for;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge)) __PYX_ERR(0, 143, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CompositionGraphEdge, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase.get_edges = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *))__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_get_edges;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase.copy = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_copy;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase)) __PYX_ERR(0, 232, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CompositionGraphBase, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext.encode_raw = (struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, PyObject *))__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode_raw;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext.encode = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_encode;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext.decode = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_decode;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext)) __PYX_ERR(0, 483, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext) < 0) __PYX_ERR(0, 483, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext) < 0) __PYX_ERR(0, 483, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext) < 0) __PYX_ERR(0, 483, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext) < 0) __PYX_ERR(0, 483, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanCompositionVectorContext, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext) < 0) __PYX_ERR(0, 483, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector._create = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*)(struct __pyx_t_10glycresoft_2_c_19composition_network_5graph_glycan_composition_vector *))__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector__create;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector.distance = (double (*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_distance;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector.difference = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_difference;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector.addition = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_addition;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector.clone = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_clone;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector)) __PYX_ERR(0, 535, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanCompositionVector, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext.__pyx_base = *__pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext.__pyx_base.encode = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_encode;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext.__pyx_base.decode = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext *, struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_decode;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext->tp_base = __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CachingGlycanCompositionVectorCo, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
  __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder = &__pyx_vtable_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder._build_initial_key_set = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder__build_initial_key_set;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder.find_smallest_unvisited = (struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_smallest_unvisited;
  __pyx_vtable_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder.find_path = (PyObject *(*)(struct __pyx_obj_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder *, int __pyx_skip_dispatch))__pyx_f_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_find_path;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder_spec, NULL); if (unlikely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder)) __PYX_ERR(0, 263, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder_spec, __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #else
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder = &__pyx_type_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder->tp_dictoffset && __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_vtabptr_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DijkstraPathFinder, (PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_graph(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_graph},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "graph",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initgraph(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initgraph(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_graph(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_graph(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_graph(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'graph' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("graph", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "graph" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_graph(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_glycresoft___c__composition_network__graph) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "glycresoft._c.composition_network.graph")) {
      if (unlikely((PyDict_SetItemString(modules, "glycresoft._c.composition_network.graph", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely((__Pyx_modinit_function_export_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "glycresoft/_c/composition_network/graph.pyx":5
 * from libc.stdlib cimport malloc, free, abs as labs
 * 
 * from collections import deque             # <<<<<<<<<<<<<<
 * 
 * from cpython.object cimport PyObject
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_deque);
  __Pyx_GIVEREF(__pyx_n_s_deque);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_deque)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_deque); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_deque, __pyx_t_2) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":14
 * from cpython.int cimport PyInt_AsLong
 * 
 * from glypy.structure.glycan_composition import HashableGlycanComposition, FrozenMonosaccharideResidue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_HashableGlycanComposition);
  __Pyx_GIVEREF(__pyx_n_s_HashableGlycanComposition);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_HashableGlycanComposition)) __PYX_ERR(0, 14, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_FrozenMonosaccharideResidue);
  __Pyx_GIVEREF(__pyx_n_s_FrozenMonosaccharideResidue);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_FrozenMonosaccharideResidue)) __PYX_ERR(0, 14, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_glypy_structure_glycan_compositi, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_HashableGlycanComposition); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HashableGlycanComposition, __pyx_t_3) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FrozenMonosaccharideResidue); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FrozenMonosaccharideResidue, __pyx_t_3) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":49
 *         self._score = value
 * 
 *     def edge_to(self, node):             # <<<<<<<<<<<<<<
 *         return self.edges.edge_to(self, node)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_3edge_to, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphNode_edge_to, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, __pyx_n_s_edge_to, __pyx_t_2) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);

  /* "glycresoft/_c/composition_network/graph.pyx":52
 *         return self.edges.edge_to(self, node)
 * 
 *     def neighbors(self):             # <<<<<<<<<<<<<<
 *         result = [edge._traverse(self) for edge in self.edges]
 *         return result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_5neighbors, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphNode_neighbors, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, __pyx_n_s_neighbors, __pyx_t_2) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);

  /* "glycresoft/_c/composition_network/graph.pyx":79
 *             self.score if self.score != 0 else self.internal_score)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         dup = CompositionGraphNode(self.composition, self.index, self.score)
 *         dup.internal_score = self.internal_score
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_15copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphNode_copy, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, __pyx_n_s_copy, __pyx_t_2) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);

  /* "glycresoft/_c/composition_network/graph.pyx":84
 *         return dup
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_17clone, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphNode_clone, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, __pyx_n_s_clone, __pyx_t_2) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphNode___reduce_cy, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphNode, (type(self), 0x1268f7b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphNode__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphNode_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphNode___setstate, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphNode);

  /* "glycresoft/_c/composition_network/graph.pyx":95
 *         self.store = dict(store)
 * 
 *     def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         if node2.index < node1.index:
 *             node1, node2 = node2, node1
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_3edge_to, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EdgeSet_edge_to, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_n_s_edge_to, __pyx_t_2) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);

  /* "glycresoft/_c/composition_network/graph.pyx":100
 *         return self.store[node1, node2]
 * 
 *     def add(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store[edge.node1, edge.node2] = edge
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_5add, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EdgeSet_add, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_n_s_add, __pyx_t_2) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);

  /* "glycresoft/_c/composition_network/graph.pyx":103
 *         self.store[edge.node1, edge.node2] = edge
 * 
 *     cpdef bint add_if_shorter(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object key
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_7add_if_shorter, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EdgeSet_add_if_shorter, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_n_s_add_if_shorter, __pyx_t_2) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);

  /* "glycresoft/_c/composition_network/graph.pyx":122
 *                 return True
 * 
 *     def remove(self, CompositionGraphEdge edge):             # <<<<<<<<<<<<<<
 *         self.store.pop((edge.node1, edge.node2))
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_9remove, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EdgeSet_remove, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_n_s_remove, __pyx_t_2) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EdgeSet___reduce_cython, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EdgeSet, (type(self), 0x824d80d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EdgeSet__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7EdgeSet_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EdgeSet___setstate_cython, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_EdgeSet);

  /* "glycresoft/_c/composition_network/graph.pyx":183
 *             raise KeyError(key)
 * 
 *     def _traverse(self, node):             # <<<<<<<<<<<<<<
 *         return self.node1 if node is self.node2 else self.node2
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_5_traverse, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphEdge__traverse, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, __pyx_n_s_traverse, __pyx_t_2) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);

  /* "glycresoft/_c/composition_network/graph.pyx":198
 *         return self._hash
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_15__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphEdge___reduce, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, __pyx_n_s_reduce, __pyx_t_2) < 0) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);

  /* "glycresoft/_c/composition_network/graph.pyx":201
 *         return self.__class__, (self.node1, self.node2, self.order, self.weight)
 * 
 *     cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):             # <<<<<<<<<<<<<<
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_17copy_for, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphEdge_copy_for, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, __pyx_n_s_copy_for, __pyx_t_2) < 0) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);

  /* "glycresoft/_c/composition_network/graph.pyx":204
 *         return CompositionGraphEdge._create(node1, node2, self.order, self.weight)
 * 
 *     def remove(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.node1.edges.remove(self)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphEdge_19remove, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphEdge_remove, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge, __pyx_n_s_remove, __pyx_t_2) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphEdge);

  /* "glycresoft/_c/composition_network/graph.pyx":216
 * 
 * @cython.binding(True)
 * cpdef reindex_graph(self):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_1reindex_graph, 0, __pyx_n_s_reindex_graph, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reindex_graph, __pyx_t_2) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":229
 * 
 * 
 * cdef double INF = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyString_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 229, __pyx_L1_error)
  __pyx_v_10glycresoft_2_c_19composition_network_5graph_INF = __pyx_t_4;

  /* "glycresoft/_c/composition_network/graph.pyx":236
 *         return self.edges.get_edges()
 * 
 *     cpdef CompositionGraphBase copy(CompositionGraphBase self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CompositionGraphNode n1, n2
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphBase_copy, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, __pyx_n_s_copy, __pyx_t_2) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphBase___reduce_cy, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CompositionGraphBase, (type(self), 0xaf6f309, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CompositionGraphBase__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_20CompositionGraphBase_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CompositionGraphBase___setstate, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CompositionGraphBase);

  /* "glycresoft/_c/composition_network/graph.pyx":273
 *         public double limit
 * 
 *     def __init__(self, graph, start, end, limit=INF):             # <<<<<<<<<<<<<<
 *         self.graph = graph
 *         self.start = start
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_10glycresoft_2_c_19composition_network_5graph_INF); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_ = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":284
 *         self.limit = limit
 * 
 *     cpdef set _build_initial_key_set(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_3_build_initial_key_set, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DijkstraPathFinder__build_initia, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_n_s_build_initial_key_set, __pyx_t_2) < 0) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);

  /* "glycresoft/_c/composition_network/graph.pyx":299
 *         return result
 * 
 *     cpdef CompositionGraphNode find_smallest_unvisited(self, set unvisited):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double smallest_distance, distance
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_5find_smallest_unvisited, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DijkstraPathFinder_find_smallest, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_n_s_find_smallest_unvisited, __pyx_t_2) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);

  /* "glycresoft/_c/composition_network/graph.pyx":335
 *         return self.graph[smallest_node]
 * 
 *     cpdef find_path(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             set unvisited
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_7find_path, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DijkstraPathFinder_find_path, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_n_s_find_path, __pyx_t_2) < 0) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);

  /* "glycresoft/_c/composition_network/graph.pyx":389
 *                     n += 1
 * 
 *     def search(self):             # <<<<<<<<<<<<<<
 *         self.find_path()
 *         ptemp = PyDict_GetItem(self.distance, self.end._str)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_9search, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DijkstraPathFinder_search, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_n_s_search, __pyx_t_2) < 0) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DijkstraPathFinder___reduce_cyth, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DijkstraPathFinder, (type(self), 0xe7f8508, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DijkstraPathFinder__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_18DijkstraPathFinder_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DijkstraPathFinder___setstate_cy, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_DijkstraPathFinder);

  /* "glycresoft/_c/composition_network/graph.pyx":489
 *         self.component_count = len(self.components)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.components, )
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_3__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVectorContext_2, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, __pyx_n_s_reduce, __pyx_t_2) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);

  /* "glycresoft/_c/composition_network/graph.pyx":500
 *         return ptr
 * 
 *     cpdef GlycanCompositionVector encode(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_5encode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVectorContext_e, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, __pyx_n_s_encode, __pyx_t_2) < 0) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);

  /* "glycresoft/_c/composition_network/graph.pyx":503
 *         return GlycanCompositionVector._create(self.encode_raw(glycan_composition))
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_30GlycanCompositionVectorContext_7decode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVectorContext_d, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext, __pyx_n_s_decode, __pyx_t_2) < 0) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVectorContext);

  /* "glycresoft/_c/composition_network/graph.pyx":517
 * 
 * 
 * cpdef _reconstruct_GlycanCompositionVector(state):             # <<<<<<<<<<<<<<
 *     cdef:
 *         glycan_composition_vector* ptr
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_3_reconstruct_GlycanCompositionVector, 0, __pyx_n_s_reconstruct_GlycanCompositionVe, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reconstruct_GlycanCompositionVe, __pyx_t_2) < 0) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":543
 *         return self
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return _reconstruct_GlycanCompositionVector, (list(self), )
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_1__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVector___reduce, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, __pyx_n_s_reduce, __pyx_t_2) < 0) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);

  /* "glycresoft/_c/composition_network/graph.pyx":574
 *         return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"
 * 
 *     cpdef double distance(self, GlycanCompositionVector other) except -1:             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return INF
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_13distance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVector_distance, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, __pyx_n_s_distance, __pyx_t_2) < 0) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);

  /* "glycresoft/_c/composition_network/graph.pyx":589
 *         return not self == other
 * 
 *     cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_19difference, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVector_differen, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, __pyx_n_s_difference, __pyx_t_2) < 0) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);

  /* "glycresoft/_c/composition_network/graph.pyx":605
 *         return GlycanCompositionVector._create(new)
 * 
 *     cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other):             # <<<<<<<<<<<<<<
 *         cdef:
 *              glycan_composition_vector* new
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_21addition, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVector_addition, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, __pyx_n_s_addition, __pyx_t_2) < 0) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);

  /* "glycresoft/_c/composition_network/graph.pyx":655
 *         return hash_val
 * 
 *     cpdef GlycanCompositionVector clone(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             glycan_composition_vector* new
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_23GlycanCompositionVector_31clone, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlycanCompositionVector_clone, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector, __pyx_n_s_clone, __pyx_t_2) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_GlycanCompositionVector);

  /* "glycresoft/_c/composition_network/graph.pyx":677
 *         self.decode_cache = {}
 * 
 *     cpdef GlycanCompositionVector encode(self, gc):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_3encode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CachingGlycanCompositionVectorCo_2, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, __pyx_n_s_encode, __pyx_t_2) < 0) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);

  /* "glycresoft/_c/composition_network/graph.pyx":690
 *         return value
 * 
 *     cpdef decode(self, GlycanCompositionVector gcvec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* ptemp
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_37CachingGlycanCompositionVectorContext_5decode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CachingGlycanCompositionVectorCo_3, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext, __pyx_n_s_decode, __pyx_t_2) < 0) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10glycresoft_2_c_19composition_network_5graph_CachingGlycanCompositionVectorContext);

  /* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_5__pyx_unpickle_CompositionGraphNode, 0, __pyx_n_s_pyx_unpickle_CompositionGraphN, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CompositionGraphN, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_CompositionGraphNode__set_state(<CompositionGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphNode__set_state(CompositionGraphNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result._score = __pyx_state[1]; __pyx_result._str = __pyx_state[2]; __pyx_result.composition = __pyx_state[3]; __pyx_result.edges = __pyx_state[4]; __pyx_result.index = __pyx_state[5]; __pyx_result.internal_score = __pyx_state[6]; __pyx_result.marked = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_7__pyx_unpickle_EdgeSet, 0, __pyx_n_s_pyx_unpickle_EdgeSet, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_EdgeSet, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CompositionGraphBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_9__pyx_unpickle_CompositionGraphBase, 0, __pyx_n_s_pyx_unpickle_CompositionGraphB, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CompositionGraphB, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_CompositionGraphBase__set_state(<CompositionGraphBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CompositionGraphBase__set_state(CompositionGraphBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._composition_normalizer = __pyx_state[0]; __pyx_result.distance_fn = __pyx_state[1]; __pyx_result.edges = __pyx_state[2]; __pyx_result.neighborhoods = __pyx_state[3]; __pyx_result.node_map = __pyx_state[4]; __pyx_result.nodes = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10glycresoft_2_c_19composition_network_5graph_11__pyx_unpickle_DijkstraPathFinder, 0, __pyx_n_s_pyx_unpickle_DijkstraPathFinde, NULL, __pyx_n_s_glycresoft__c_composition_networ, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DijkstraPathFinde, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycresoft/_c/composition_network/graph.pyx":1
 * cimport cython             # <<<<<<<<<<<<<<
 * 
 * from libc.stdlib cimport malloc, free, abs as labs
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init glycresoft._c.composition_network.graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init glycresoft._c.composition_network.graph");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyStr_Str */
static CYTHON_INLINE PyObject* __Pyx_PyStr_Str(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_n_s_None;
    return __Pyx_NewRef(obj);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a != b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3 & PY_VERSION_HEX < 0x030d0000
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_values */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_values, d);
    else
        return PyDict_Values(d);
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__12);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_11
#define __PYX_HAVE_RT_ImportType_3_0_11
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_11 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_11 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* pybytes_as_double */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
#if CYTHON_ASSUME_SAFE_MACROS
        double value = PyFloat_AS_DOUBLE(float_value);
#else
        double value = PyFloat_AsDouble(float_value);
#endif
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__73);
    }
    return name;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
